/* Automatically @generated by tree-sitter v0.25.8 */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 1442
#define LARGE_STATE_COUNT 29
#define SYMBOL_COUNT 214
#define ALIAS_COUNT 5
#define TOKEN_COUNT 95
#define EXTERNAL_TOKEN_COUNT 0
#define FIELD_COUNT 35
#define MAX_ALIAS_SEQUENCE_LENGTH 9
#define MAX_RESERVED_WORD_SET_SIZE 25
#define PRODUCTION_ID_COUNT 111
#define SUPERTYPE_COUNT 4

enum ts_symbol_identifiers {
  sym_identifier = 1,
  aux_sym_source_file_token1 = 2,
  anon_sym_SEMI = 3,
  anon_sym_NULL = 4,
  anon_sym_package = 5,
  anon_sym_import = 6,
  anon_sym_DOT = 7,
  sym_blank_identifier = 8,
  anon_sym_LPAREN = 9,
  anon_sym_RPAREN = 10,
  anon_sym_const = 11,
  anon_sym_COMMA = 12,
  anon_sym_EQ = 13,
  anon_sym_var = 14,
  anon_sym_func = 15,
  anon_sym_LBRACK = 16,
  anon_sym_RBRACK = 17,
  anon_sym_DOT_DOT_DOT = 18,
  anon_sym_type = 19,
  anon_sym_STAR = 20,
  anon_sym_struct = 21,
  anon_sym_TILDE = 22,
  anon_sym_LBRACE = 23,
  anon_sym_RBRACE = 24,
  anon_sym_interface = 25,
  anon_sym_PIPE = 26,
  anon_sym_map = 27,
  anon_sym_chan = 28,
  anon_sym_LT_DASH = 29,
  anon_sym_COLON_EQ = 30,
  anon_sym_PLUS_PLUS = 31,
  anon_sym_DASH_DASH = 32,
  anon_sym_STAR_EQ = 33,
  anon_sym_SLASH_EQ = 34,
  anon_sym_PERCENT_EQ = 35,
  anon_sym_LT_LT_EQ = 36,
  anon_sym_GT_GT_EQ = 37,
  anon_sym_AMP_EQ = 38,
  anon_sym_AMP_CARET_EQ = 39,
  anon_sym_PLUS_EQ = 40,
  anon_sym_DASH_EQ = 41,
  anon_sym_PIPE_EQ = 42,
  anon_sym_CARET_EQ = 43,
  anon_sym_COLON = 44,
  anon_sym_fallthrough = 45,
  anon_sym_break = 46,
  anon_sym_continue = 47,
  anon_sym_goto = 48,
  anon_sym_return = 49,
  anon_sym_go = 50,
  anon_sym_defer = 51,
  anon_sym_if = 52,
  anon_sym_else = 53,
  anon_sym_for = 54,
  anon_sym_range = 55,
  anon_sym_switch = 56,
  anon_sym_case = 57,
  anon_sym_default = 58,
  anon_sym_select = 59,
  anon_sym_new = 60,
  anon_sym_make = 61,
  anon_sym_PLUS = 62,
  anon_sym_DASH = 63,
  anon_sym_BANG = 64,
  anon_sym_CARET = 65,
  anon_sym_AMP = 66,
  anon_sym_SLASH = 67,
  anon_sym_PERCENT = 68,
  anon_sym_LT_LT = 69,
  anon_sym_GT_GT = 70,
  anon_sym_AMP_CARET = 71,
  anon_sym_EQ_EQ = 72,
  anon_sym_BANG_EQ = 73,
  anon_sym_LT = 74,
  anon_sym_LT_EQ = 75,
  anon_sym_GT = 76,
  anon_sym_GT_EQ = 77,
  anon_sym_AMP_AMP = 78,
  anon_sym_PIPE_PIPE = 79,
  anon_sym_BQUOTE = 80,
  aux_sym_raw_string_literal_token1 = 81,
  anon_sym_DQUOTE = 82,
  aux_sym_interpreted_string_literal_token1 = 83,
  anon_sym_DQUOTE2 = 84,
  sym_escape_sequence = 85,
  sym_int_literal = 86,
  sym_float_literal = 87,
  sym_imaginary_literal = 88,
  sym_rune_literal = 89,
  sym_nil = 90,
  sym_true = 91,
  sym_false = 92,
  sym_iota = 93,
  sym_comment = 94,
  sym_source_file = 95,
  sym_package_clause = 96,
  sym_import_declaration = 97,
  sym_import_spec = 98,
  sym_dot = 99,
  sym_import_spec_list = 100,
  sym__declaration = 101,
  sym_const_declaration = 102,
  sym_const_spec = 103,
  sym_var_declaration = 104,
  sym_var_spec = 105,
  sym_var_spec_list = 106,
  sym_function_declaration = 107,
  sym_method_declaration = 108,
  sym_type_parameter_list = 109,
  sym_type_parameter_declaration = 110,
  sym_parameter_list = 111,
  sym_parameter_declaration = 112,
  sym_variadic_parameter_declaration = 113,
  sym_type_alias = 114,
  sym_type_declaration = 115,
  sym_type_spec = 116,
  sym_expression_list = 117,
  sym_parenthesized_type = 118,
  sym__simple_type = 119,
  sym_generic_type = 120,
  sym_type_arguments = 121,
  sym_pointer_type = 122,
  sym_array_type = 123,
  sym_implicit_length_array_type = 124,
  sym_slice_type = 125,
  sym_struct_type = 126,
  sym_negated_type = 127,
  sym_field_declaration_list = 128,
  sym_field_declaration = 129,
  sym_interface_type = 130,
  sym_method_elem = 131,
  sym_type_elem = 132,
  sym_map_type = 133,
  sym_channel_type = 134,
  sym_function_type = 135,
  sym_block = 136,
  sym_statement_list = 137,
  sym__statement = 138,
  sym_empty_statement = 139,
  sym__simple_statement = 140,
  sym_expression_statement = 141,
  sym_send_statement = 142,
  sym_receive_statement = 143,
  sym_inc_statement = 144,
  sym_dec_statement = 145,
  sym_assignment_statement = 146,
  sym_short_var_declaration = 147,
  sym_labeled_statement = 148,
  sym_empty_labeled_statement = 149,
  sym_fallthrough_statement = 150,
  sym_break_statement = 151,
  sym_continue_statement = 152,
  sym_goto_statement = 153,
  sym_return_statement = 154,
  sym_go_statement = 155,
  sym_defer_statement = 156,
  sym_if_statement = 157,
  sym_for_statement = 158,
  sym_for_clause = 159,
  sym_range_clause = 160,
  sym_expression_switch_statement = 161,
  sym_expression_case = 162,
  sym_default_case = 163,
  sym_type_switch_statement = 164,
  sym__type_switch_header = 165,
  sym_type_case = 166,
  sym_select_statement = 167,
  sym_communication_case = 168,
  sym__expression = 169,
  sym_parenthesized_expression = 170,
  sym_call_expression = 171,
  sym_variadic_argument = 172,
  sym_special_argument_list = 173,
  sym_argument_list = 174,
  sym_selector_expression = 175,
  sym_index_expression = 176,
  sym_slice_expression = 177,
  sym_type_assertion_expression = 178,
  sym_type_conversion_expression = 179,
  sym_type_instantiation_expression = 180,
  sym_composite_literal = 181,
  sym_literal_value = 182,
  sym_literal_element = 183,
  sym_keyed_element = 184,
  sym_func_literal = 185,
  sym_unary_expression = 186,
  sym_binary_expression = 187,
  sym_qualified_type = 188,
  sym_raw_string_literal = 189,
  sym_interpreted_string_literal = 190,
  aux_sym_source_file_repeat1 = 191,
  aux_sym_import_spec_list_repeat1 = 192,
  aux_sym_const_declaration_repeat1 = 193,
  aux_sym_const_spec_repeat1 = 194,
  aux_sym_var_spec_repeat1 = 195,
  aux_sym_var_spec_list_repeat1 = 196,
  aux_sym_type_parameter_list_repeat1 = 197,
  aux_sym_parameter_list_repeat1 = 198,
  aux_sym_type_declaration_repeat1 = 199,
  aux_sym_expression_list_repeat1 = 200,
  aux_sym_type_arguments_repeat1 = 201,
  aux_sym_field_declaration_list_repeat1 = 202,
  aux_sym_field_declaration_repeat1 = 203,
  aux_sym_interface_type_repeat1 = 204,
  aux_sym_type_elem_repeat1 = 205,
  aux_sym_statement_list_repeat1 = 206,
  aux_sym_expression_switch_statement_repeat1 = 207,
  aux_sym_type_switch_statement_repeat1 = 208,
  aux_sym_type_case_repeat1 = 209,
  aux_sym_select_statement_repeat1 = 210,
  aux_sym_argument_list_repeat1 = 211,
  aux_sym_literal_value_repeat1 = 212,
  aux_sym_interpreted_string_literal_repeat1 = 213,
  alias_sym_field_identifier = 214,
  alias_sym_label_name = 215,
  alias_sym_package_identifier = 216,
  alias_sym_type_constraint = 217,
  alias_sym_type_identifier = 218,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [aux_sym_source_file_token1] = "source_file_token1",
  [anon_sym_SEMI] = ";",
  [anon_sym_NULL] = "\0",
  [anon_sym_package] = "package",
  [anon_sym_import] = "import",
  [anon_sym_DOT] = ".",
  [sym_blank_identifier] = "blank_identifier",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_const] = "const",
  [anon_sym_COMMA] = ",",
  [anon_sym_EQ] = "=",
  [anon_sym_var] = "var",
  [anon_sym_func] = "func",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_DOT_DOT_DOT] = "...",
  [anon_sym_type] = "type",
  [anon_sym_STAR] = "*",
  [anon_sym_struct] = "struct",
  [anon_sym_TILDE] = "~",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_interface] = "interface",
  [anon_sym_PIPE] = "|",
  [anon_sym_map] = "map",
  [anon_sym_chan] = "chan",
  [anon_sym_LT_DASH] = "<-",
  [anon_sym_COLON_EQ] = ":=",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_AMP_CARET_EQ] = "&^=",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_COLON] = ":",
  [anon_sym_fallthrough] = "fallthrough",
  [anon_sym_break] = "break",
  [anon_sym_continue] = "continue",
  [anon_sym_goto] = "goto",
  [anon_sym_return] = "return",
  [anon_sym_go] = "go",
  [anon_sym_defer] = "defer",
  [anon_sym_if] = "if",
  [anon_sym_else] = "else",
  [anon_sym_for] = "for",
  [anon_sym_range] = "range",
  [anon_sym_switch] = "switch",
  [anon_sym_case] = "case",
  [anon_sym_default] = "default",
  [anon_sym_select] = "select",
  [anon_sym_new] = "identifier",
  [anon_sym_make] = "identifier",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_BANG] = "!",
  [anon_sym_CARET] = "^",
  [anon_sym_AMP] = "&",
  [anon_sym_SLASH] = "/",
  [anon_sym_PERCENT] = "%",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_AMP_CARET] = "&^",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_LT] = "<",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_GT] = ">",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_BQUOTE] = "`",
  [aux_sym_raw_string_literal_token1] = "raw_string_literal_content",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym_interpreted_string_literal_token1] = "interpreted_string_literal_content",
  [anon_sym_DQUOTE2] = "\"",
  [sym_escape_sequence] = "escape_sequence",
  [sym_int_literal] = "int_literal",
  [sym_float_literal] = "float_literal",
  [sym_imaginary_literal] = "imaginary_literal",
  [sym_rune_literal] = "rune_literal",
  [sym_nil] = "nil",
  [sym_true] = "true",
  [sym_false] = "false",
  [sym_iota] = "iota",
  [sym_comment] = "comment",
  [sym_source_file] = "source_file",
  [sym_package_clause] = "package_clause",
  [sym_import_declaration] = "import_declaration",
  [sym_import_spec] = "import_spec",
  [sym_dot] = "dot",
  [sym_import_spec_list] = "import_spec_list",
  [sym__declaration] = "_declaration",
  [sym_const_declaration] = "const_declaration",
  [sym_const_spec] = "const_spec",
  [sym_var_declaration] = "var_declaration",
  [sym_var_spec] = "var_spec",
  [sym_var_spec_list] = "var_spec_list",
  [sym_function_declaration] = "function_declaration",
  [sym_method_declaration] = "method_declaration",
  [sym_type_parameter_list] = "type_parameter_list",
  [sym_type_parameter_declaration] = "type_parameter_declaration",
  [sym_parameter_list] = "parameter_list",
  [sym_parameter_declaration] = "parameter_declaration",
  [sym_variadic_parameter_declaration] = "variadic_parameter_declaration",
  [sym_type_alias] = "type_alias",
  [sym_type_declaration] = "type_declaration",
  [sym_type_spec] = "type_spec",
  [sym_expression_list] = "expression_list",
  [sym_parenthesized_type] = "parenthesized_type",
  [sym__simple_type] = "_simple_type",
  [sym_generic_type] = "generic_type",
  [sym_type_arguments] = "type_arguments",
  [sym_pointer_type] = "pointer_type",
  [sym_array_type] = "array_type",
  [sym_implicit_length_array_type] = "implicit_length_array_type",
  [sym_slice_type] = "slice_type",
  [sym_struct_type] = "struct_type",
  [sym_negated_type] = "negated_type",
  [sym_field_declaration_list] = "field_declaration_list",
  [sym_field_declaration] = "field_declaration",
  [sym_interface_type] = "interface_type",
  [sym_method_elem] = "method_elem",
  [sym_type_elem] = "type_elem",
  [sym_map_type] = "map_type",
  [sym_channel_type] = "channel_type",
  [sym_function_type] = "function_type",
  [sym_block] = "block",
  [sym_statement_list] = "statement_list",
  [sym__statement] = "_statement",
  [sym_empty_statement] = "empty_statement",
  [sym__simple_statement] = "_simple_statement",
  [sym_expression_statement] = "expression_statement",
  [sym_send_statement] = "send_statement",
  [sym_receive_statement] = "receive_statement",
  [sym_inc_statement] = "inc_statement",
  [sym_dec_statement] = "dec_statement",
  [sym_assignment_statement] = "assignment_statement",
  [sym_short_var_declaration] = "short_var_declaration",
  [sym_labeled_statement] = "labeled_statement",
  [sym_empty_labeled_statement] = "labeled_statement",
  [sym_fallthrough_statement] = "fallthrough_statement",
  [sym_break_statement] = "break_statement",
  [sym_continue_statement] = "continue_statement",
  [sym_goto_statement] = "goto_statement",
  [sym_return_statement] = "return_statement",
  [sym_go_statement] = "go_statement",
  [sym_defer_statement] = "defer_statement",
  [sym_if_statement] = "if_statement",
  [sym_for_statement] = "for_statement",
  [sym_for_clause] = "for_clause",
  [sym_range_clause] = "range_clause",
  [sym_expression_switch_statement] = "expression_switch_statement",
  [sym_expression_case] = "expression_case",
  [sym_default_case] = "default_case",
  [sym_type_switch_statement] = "type_switch_statement",
  [sym__type_switch_header] = "_type_switch_header",
  [sym_type_case] = "type_case",
  [sym_select_statement] = "select_statement",
  [sym_communication_case] = "communication_case",
  [sym__expression] = "_expression",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym_call_expression] = "call_expression",
  [sym_variadic_argument] = "variadic_argument",
  [sym_special_argument_list] = "argument_list",
  [sym_argument_list] = "argument_list",
  [sym_selector_expression] = "selector_expression",
  [sym_index_expression] = "index_expression",
  [sym_slice_expression] = "slice_expression",
  [sym_type_assertion_expression] = "type_assertion_expression",
  [sym_type_conversion_expression] = "type_conversion_expression",
  [sym_type_instantiation_expression] = "type_instantiation_expression",
  [sym_composite_literal] = "composite_literal",
  [sym_literal_value] = "literal_value",
  [sym_literal_element] = "literal_element",
  [sym_keyed_element] = "keyed_element",
  [sym_func_literal] = "func_literal",
  [sym_unary_expression] = "unary_expression",
  [sym_binary_expression] = "binary_expression",
  [sym_qualified_type] = "qualified_type",
  [sym_raw_string_literal] = "raw_string_literal",
  [sym_interpreted_string_literal] = "interpreted_string_literal",
  [aux_sym_source_file_repeat1] = "source_file_repeat1",
  [aux_sym_import_spec_list_repeat1] = "import_spec_list_repeat1",
  [aux_sym_const_declaration_repeat1] = "const_declaration_repeat1",
  [aux_sym_const_spec_repeat1] = "const_spec_repeat1",
  [aux_sym_var_spec_repeat1] = "var_spec_repeat1",
  [aux_sym_var_spec_list_repeat1] = "var_spec_list_repeat1",
  [aux_sym_type_parameter_list_repeat1] = "type_parameter_list_repeat1",
  [aux_sym_parameter_list_repeat1] = "parameter_list_repeat1",
  [aux_sym_type_declaration_repeat1] = "type_declaration_repeat1",
  [aux_sym_expression_list_repeat1] = "expression_list_repeat1",
  [aux_sym_type_arguments_repeat1] = "type_arguments_repeat1",
  [aux_sym_field_declaration_list_repeat1] = "field_declaration_list_repeat1",
  [aux_sym_field_declaration_repeat1] = "field_declaration_repeat1",
  [aux_sym_interface_type_repeat1] = "interface_type_repeat1",
  [aux_sym_type_elem_repeat1] = "type_elem_repeat1",
  [aux_sym_statement_list_repeat1] = "statement_list_repeat1",
  [aux_sym_expression_switch_statement_repeat1] = "expression_switch_statement_repeat1",
  [aux_sym_type_switch_statement_repeat1] = "type_switch_statement_repeat1",
  [aux_sym_type_case_repeat1] = "type_case_repeat1",
  [aux_sym_select_statement_repeat1] = "select_statement_repeat1",
  [aux_sym_argument_list_repeat1] = "argument_list_repeat1",
  [aux_sym_literal_value_repeat1] = "literal_value_repeat1",
  [aux_sym_interpreted_string_literal_repeat1] = "interpreted_string_literal_repeat1",
  [alias_sym_field_identifier] = "field_identifier",
  [alias_sym_label_name] = "label_name",
  [alias_sym_package_identifier] = "package_identifier",
  [alias_sym_type_constraint] = "type_constraint",
  [alias_sym_type_identifier] = "type_identifier",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [aux_sym_source_file_token1] = aux_sym_source_file_token1,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_NULL] = anon_sym_NULL,
  [anon_sym_package] = anon_sym_package,
  [anon_sym_import] = anon_sym_import,
  [anon_sym_DOT] = anon_sym_DOT,
  [sym_blank_identifier] = sym_blank_identifier,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_const] = anon_sym_const,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_var] = anon_sym_var,
  [anon_sym_func] = anon_sym_func,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_DOT_DOT_DOT] = anon_sym_DOT_DOT_DOT,
  [anon_sym_type] = anon_sym_type,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_struct] = anon_sym_struct,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_interface] = anon_sym_interface,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_map] = anon_sym_map,
  [anon_sym_chan] = anon_sym_chan,
  [anon_sym_LT_DASH] = anon_sym_LT_DASH,
  [anon_sym_COLON_EQ] = anon_sym_COLON_EQ,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_AMP_CARET_EQ] = anon_sym_AMP_CARET_EQ,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_fallthrough] = anon_sym_fallthrough,
  [anon_sym_break] = anon_sym_break,
  [anon_sym_continue] = anon_sym_continue,
  [anon_sym_goto] = anon_sym_goto,
  [anon_sym_return] = anon_sym_return,
  [anon_sym_go] = anon_sym_go,
  [anon_sym_defer] = anon_sym_defer,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_range] = anon_sym_range,
  [anon_sym_switch] = anon_sym_switch,
  [anon_sym_case] = anon_sym_case,
  [anon_sym_default] = anon_sym_default,
  [anon_sym_select] = anon_sym_select,
  [anon_sym_new] = sym_identifier,
  [anon_sym_make] = sym_identifier,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_AMP_CARET] = anon_sym_AMP_CARET,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_BQUOTE] = anon_sym_BQUOTE,
  [aux_sym_raw_string_literal_token1] = aux_sym_raw_string_literal_token1,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym_interpreted_string_literal_token1] = aux_sym_interpreted_string_literal_token1,
  [anon_sym_DQUOTE2] = anon_sym_DQUOTE,
  [sym_escape_sequence] = sym_escape_sequence,
  [sym_int_literal] = sym_int_literal,
  [sym_float_literal] = sym_float_literal,
  [sym_imaginary_literal] = sym_imaginary_literal,
  [sym_rune_literal] = sym_rune_literal,
  [sym_nil] = sym_nil,
  [sym_true] = sym_true,
  [sym_false] = sym_false,
  [sym_iota] = sym_iota,
  [sym_comment] = sym_comment,
  [sym_source_file] = sym_source_file,
  [sym_package_clause] = sym_package_clause,
  [sym_import_declaration] = sym_import_declaration,
  [sym_import_spec] = sym_import_spec,
  [sym_dot] = sym_dot,
  [sym_import_spec_list] = sym_import_spec_list,
  [sym__declaration] = sym__declaration,
  [sym_const_declaration] = sym_const_declaration,
  [sym_const_spec] = sym_const_spec,
  [sym_var_declaration] = sym_var_declaration,
  [sym_var_spec] = sym_var_spec,
  [sym_var_spec_list] = sym_var_spec_list,
  [sym_function_declaration] = sym_function_declaration,
  [sym_method_declaration] = sym_method_declaration,
  [sym_type_parameter_list] = sym_type_parameter_list,
  [sym_type_parameter_declaration] = sym_type_parameter_declaration,
  [sym_parameter_list] = sym_parameter_list,
  [sym_parameter_declaration] = sym_parameter_declaration,
  [sym_variadic_parameter_declaration] = sym_variadic_parameter_declaration,
  [sym_type_alias] = sym_type_alias,
  [sym_type_declaration] = sym_type_declaration,
  [sym_type_spec] = sym_type_spec,
  [sym_expression_list] = sym_expression_list,
  [sym_parenthesized_type] = sym_parenthesized_type,
  [sym__simple_type] = sym__simple_type,
  [sym_generic_type] = sym_generic_type,
  [sym_type_arguments] = sym_type_arguments,
  [sym_pointer_type] = sym_pointer_type,
  [sym_array_type] = sym_array_type,
  [sym_implicit_length_array_type] = sym_implicit_length_array_type,
  [sym_slice_type] = sym_slice_type,
  [sym_struct_type] = sym_struct_type,
  [sym_negated_type] = sym_negated_type,
  [sym_field_declaration_list] = sym_field_declaration_list,
  [sym_field_declaration] = sym_field_declaration,
  [sym_interface_type] = sym_interface_type,
  [sym_method_elem] = sym_method_elem,
  [sym_type_elem] = sym_type_elem,
  [sym_map_type] = sym_map_type,
  [sym_channel_type] = sym_channel_type,
  [sym_function_type] = sym_function_type,
  [sym_block] = sym_block,
  [sym_statement_list] = sym_statement_list,
  [sym__statement] = sym__statement,
  [sym_empty_statement] = sym_empty_statement,
  [sym__simple_statement] = sym__simple_statement,
  [sym_expression_statement] = sym_expression_statement,
  [sym_send_statement] = sym_send_statement,
  [sym_receive_statement] = sym_receive_statement,
  [sym_inc_statement] = sym_inc_statement,
  [sym_dec_statement] = sym_dec_statement,
  [sym_assignment_statement] = sym_assignment_statement,
  [sym_short_var_declaration] = sym_short_var_declaration,
  [sym_labeled_statement] = sym_labeled_statement,
  [sym_empty_labeled_statement] = sym_labeled_statement,
  [sym_fallthrough_statement] = sym_fallthrough_statement,
  [sym_break_statement] = sym_break_statement,
  [sym_continue_statement] = sym_continue_statement,
  [sym_goto_statement] = sym_goto_statement,
  [sym_return_statement] = sym_return_statement,
  [sym_go_statement] = sym_go_statement,
  [sym_defer_statement] = sym_defer_statement,
  [sym_if_statement] = sym_if_statement,
  [sym_for_statement] = sym_for_statement,
  [sym_for_clause] = sym_for_clause,
  [sym_range_clause] = sym_range_clause,
  [sym_expression_switch_statement] = sym_expression_switch_statement,
  [sym_expression_case] = sym_expression_case,
  [sym_default_case] = sym_default_case,
  [sym_type_switch_statement] = sym_type_switch_statement,
  [sym__type_switch_header] = sym__type_switch_header,
  [sym_type_case] = sym_type_case,
  [sym_select_statement] = sym_select_statement,
  [sym_communication_case] = sym_communication_case,
  [sym__expression] = sym__expression,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym_call_expression] = sym_call_expression,
  [sym_variadic_argument] = sym_variadic_argument,
  [sym_special_argument_list] = sym_argument_list,
  [sym_argument_list] = sym_argument_list,
  [sym_selector_expression] = sym_selector_expression,
  [sym_index_expression] = sym_index_expression,
  [sym_slice_expression] = sym_slice_expression,
  [sym_type_assertion_expression] = sym_type_assertion_expression,
  [sym_type_conversion_expression] = sym_type_conversion_expression,
  [sym_type_instantiation_expression] = sym_type_instantiation_expression,
  [sym_composite_literal] = sym_composite_literal,
  [sym_literal_value] = sym_literal_value,
  [sym_literal_element] = sym_literal_element,
  [sym_keyed_element] = sym_keyed_element,
  [sym_func_literal] = sym_func_literal,
  [sym_unary_expression] = sym_unary_expression,
  [sym_binary_expression] = sym_binary_expression,
  [sym_qualified_type] = sym_qualified_type,
  [sym_raw_string_literal] = sym_raw_string_literal,
  [sym_interpreted_string_literal] = sym_interpreted_string_literal,
  [aux_sym_source_file_repeat1] = aux_sym_source_file_repeat1,
  [aux_sym_import_spec_list_repeat1] = aux_sym_import_spec_list_repeat1,
  [aux_sym_const_declaration_repeat1] = aux_sym_const_declaration_repeat1,
  [aux_sym_const_spec_repeat1] = aux_sym_const_spec_repeat1,
  [aux_sym_var_spec_repeat1] = aux_sym_var_spec_repeat1,
  [aux_sym_var_spec_list_repeat1] = aux_sym_var_spec_list_repeat1,
  [aux_sym_type_parameter_list_repeat1] = aux_sym_type_parameter_list_repeat1,
  [aux_sym_parameter_list_repeat1] = aux_sym_parameter_list_repeat1,
  [aux_sym_type_declaration_repeat1] = aux_sym_type_declaration_repeat1,
  [aux_sym_expression_list_repeat1] = aux_sym_expression_list_repeat1,
  [aux_sym_type_arguments_repeat1] = aux_sym_type_arguments_repeat1,
  [aux_sym_field_declaration_list_repeat1] = aux_sym_field_declaration_list_repeat1,
  [aux_sym_field_declaration_repeat1] = aux_sym_field_declaration_repeat1,
  [aux_sym_interface_type_repeat1] = aux_sym_interface_type_repeat1,
  [aux_sym_type_elem_repeat1] = aux_sym_type_elem_repeat1,
  [aux_sym_statement_list_repeat1] = aux_sym_statement_list_repeat1,
  [aux_sym_expression_switch_statement_repeat1] = aux_sym_expression_switch_statement_repeat1,
  [aux_sym_type_switch_statement_repeat1] = aux_sym_type_switch_statement_repeat1,
  [aux_sym_type_case_repeat1] = aux_sym_type_case_repeat1,
  [aux_sym_select_statement_repeat1] = aux_sym_select_statement_repeat1,
  [aux_sym_argument_list_repeat1] = aux_sym_argument_list_repeat1,
  [aux_sym_literal_value_repeat1] = aux_sym_literal_value_repeat1,
  [aux_sym_interpreted_string_literal_repeat1] = aux_sym_interpreted_string_literal_repeat1,
  [alias_sym_field_identifier] = alias_sym_field_identifier,
  [alias_sym_label_name] = alias_sym_label_name,
  [alias_sym_package_identifier] = alias_sym_package_identifier,
  [alias_sym_type_constraint] = alias_sym_type_constraint,
  [alias_sym_type_identifier] = alias_sym_type_identifier,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_source_file_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_NULL] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_package] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_import] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [sym_blank_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_const] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_var] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_func] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_type] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_struct] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_interface] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_map] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_chan] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fallthrough] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_break] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_continue] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_goto] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_return] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_go] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_defer] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_range] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_switch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_case] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_default] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_select] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_new] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_make] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_raw_string_literal_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_interpreted_string_literal_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_int_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_float_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_imaginary_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_rune_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_nil] = {
    .visible = true,
    .named = true,
  },
  [sym_true] = {
    .visible = true,
    .named = true,
  },
  [sym_false] = {
    .visible = true,
    .named = true,
  },
  [sym_iota] = {
    .visible = true,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_source_file] = {
    .visible = true,
    .named = true,
  },
  [sym_package_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_import_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_import_spec] = {
    .visible = true,
    .named = true,
  },
  [sym_dot] = {
    .visible = true,
    .named = true,
  },
  [sym_import_spec_list] = {
    .visible = true,
    .named = true,
  },
  [sym__declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_const_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_const_spec] = {
    .visible = true,
    .named = true,
  },
  [sym_var_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_var_spec] = {
    .visible = true,
    .named = true,
  },
  [sym_var_spec_list] = {
    .visible = true,
    .named = true,
  },
  [sym_function_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_method_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_type_parameter_list] = {
    .visible = true,
    .named = true,
  },
  [sym_type_parameter_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter_list] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_parameter_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_type_alias] = {
    .visible = true,
    .named = true,
  },
  [sym_type_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_type_spec] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_list] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_type] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_generic_type] = {
    .visible = true,
    .named = true,
  },
  [sym_type_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_pointer_type] = {
    .visible = true,
    .named = true,
  },
  [sym_array_type] = {
    .visible = true,
    .named = true,
  },
  [sym_implicit_length_array_type] = {
    .visible = true,
    .named = true,
  },
  [sym_slice_type] = {
    .visible = true,
    .named = true,
  },
  [sym_struct_type] = {
    .visible = true,
    .named = true,
  },
  [sym_negated_type] = {
    .visible = true,
    .named = true,
  },
  [sym_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_field_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_type] = {
    .visible = true,
    .named = true,
  },
  [sym_method_elem] = {
    .visible = true,
    .named = true,
  },
  [sym_type_elem] = {
    .visible = true,
    .named = true,
  },
  [sym_map_type] = {
    .visible = true,
    .named = true,
  },
  [sym_channel_type] = {
    .visible = true,
    .named = true,
  },
  [sym_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym_block] = {
    .visible = true,
    .named = true,
  },
  [sym_statement_list] = {
    .visible = true,
    .named = true,
  },
  [sym__statement] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_empty_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_statement] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_expression_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_send_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_receive_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_inc_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_dec_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_short_var_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_empty_labeled_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_fallthrough_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_break_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_continue_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_goto_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_return_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_go_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_defer_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_for_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_range_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_switch_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_case] = {
    .visible = true,
    .named = true,
  },
  [sym_default_case] = {
    .visible = true,
    .named = true,
  },
  [sym_type_switch_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__type_switch_header] = {
    .visible = false,
    .named = true,
  },
  [sym_type_case] = {
    .visible = true,
    .named = true,
  },
  [sym_select_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_communication_case] = {
    .visible = true,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_special_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym_selector_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_index_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_slice_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_type_assertion_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_type_conversion_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_type_instantiation_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_composite_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_literal_value] = {
    .visible = true,
    .named = true,
  },
  [sym_literal_element] = {
    .visible = true,
    .named = true,
  },
  [sym_keyed_element] = {
    .visible = true,
    .named = true,
  },
  [sym_func_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_qualified_type] = {
    .visible = true,
    .named = true,
  },
  [sym_raw_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_interpreted_string_literal] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_source_file_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_import_spec_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_const_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_const_spec_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_var_spec_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_var_spec_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_parameter_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_parameter_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expression_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_field_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_field_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_interface_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_elem_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_statement_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expression_switch_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_switch_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_case_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_select_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_argument_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_literal_value_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_interpreted_string_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_field_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_label_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_package_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_type_constraint] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_type_identifier] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alias = 1,
  field_alternative = 2,
  field_arguments = 3,
  field_body = 4,
  field_capacity = 5,
  field_channel = 6,
  field_communication = 7,
  field_condition = 8,
  field_consequence = 9,
  field_element = 10,
  field_end = 11,
  field_field = 12,
  field_function = 13,
  field_index = 14,
  field_initializer = 15,
  field_key = 16,
  field_label = 17,
  field_left = 18,
  field_length = 19,
  field_name = 20,
  field_operand = 21,
  field_operator = 22,
  field_package = 23,
  field_parameters = 24,
  field_path = 25,
  field_receiver = 26,
  field_result = 27,
  field_right = 28,
  field_start = 29,
  field_tag = 30,
  field_type = 31,
  field_type_arguments = 32,
  field_type_parameters = 33,
  field_update = 34,
  field_value = 35,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_alternative] = "alternative",
  [field_arguments] = "arguments",
  [field_body] = "body",
  [field_capacity] = "capacity",
  [field_channel] = "channel",
  [field_communication] = "communication",
  [field_condition] = "condition",
  [field_consequence] = "consequence",
  [field_element] = "element",
  [field_end] = "end",
  [field_field] = "field",
  [field_function] = "function",
  [field_index] = "index",
  [field_initializer] = "initializer",
  [field_key] = "key",
  [field_label] = "label",
  [field_left] = "left",
  [field_length] = "length",
  [field_name] = "name",
  [field_operand] = "operand",
  [field_operator] = "operator",
  [field_package] = "package",
  [field_parameters] = "parameters",
  [field_path] = "path",
  [field_receiver] = "receiver",
  [field_result] = "result",
  [field_right] = "right",
  [field_start] = "start",
  [field_tag] = "tag",
  [field_type] = "type",
  [field_type_arguments] = "type_arguments",
  [field_type_parameters] = "type_parameters",
  [field_update] = "update",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [3] = {.index = 0, .length = 1},
  [4] = {.index = 1, .length = 1},
  [5] = {.index = 2, .length = 1},
  [6] = {.index = 3, .length = 2},
  [7] = {.index = 5, .length = 1},
  [9] = {.index = 6, .length = 1},
  [10] = {.index = 7, .length = 2},
  [11] = {.index = 9, .length = 2},
  [12] = {.index = 11, .length = 2},
  [13] = {.index = 11, .length = 2},
  [14] = {.index = 9, .length = 2},
  [15] = {.index = 13, .length = 2},
  [16] = {.index = 13, .length = 2},
  [17] = {.index = 15, .length = 2},
  [18] = {.index = 17, .length = 2},
  [19] = {.index = 19, .length = 1},
  [20] = {.index = 20, .length = 2},
  [21] = {.index = 22, .length = 2},
  [22] = {.index = 24, .length = 2},
  [23] = {.index = 26, .length = 1},
  [24] = {.index = 17, .length = 2},
  [25] = {.index = 19, .length = 1},
  [26] = {.index = 27, .length = 1},
  [27] = {.index = 28, .length = 1},
  [28] = {.index = 29, .length = 2},
  [29] = {.index = 31, .length = 1},
  [30] = {.index = 32, .length = 1},
  [31] = {.index = 33, .length = 2},
  [32] = {.index = 35, .length = 3},
  [33] = {.index = 38, .length = 2},
  [34] = {.index = 40, .length = 2},
  [35] = {.index = 42, .length = 2},
  [36] = {.index = 44, .length = 3},
  [37] = {.index = 47, .length = 2},
  [38] = {.index = 49, .length = 1},
  [39] = {.index = 50, .length = 3},
  [40] = {.index = 53, .length = 2},
  [41] = {.index = 55, .length = 1},
  [42] = {.index = 56, .length = 3},
  [43] = {.index = 59, .length = 3},
  [44] = {.index = 62, .length = 3},
  [45] = {.index = 65, .length = 3},
  [46] = {.index = 68, .length = 3},
  [47] = {.index = 71, .length = 1},
  [48] = {.index = 72, .length = 2},
  [49] = {.index = 74, .length = 2},
  [50] = {.index = 76, .length = 3},
  [51] = {.index = 55, .length = 1},
  [52] = {.index = 17, .length = 2},
  [53] = {.index = 79, .length = 2},
  [54] = {.index = 79, .length = 2},
  [55] = {.index = 81, .length = 2},
  [56] = {.index = 83, .length = 1},
  [57] = {.index = 84, .length = 1},
  [58] = {.index = 85, .length = 1},
  [59] = {.index = 86, .length = 3},
  [60] = {.index = 89, .length = 1},
  [61] = {.index = 90, .length = 2},
  [62] = {.index = 92, .length = 1},
  [63] = {.index = 93, .length = 2},
  [64] = {.index = 95, .length = 3},
  [65] = {.index = 98, .length = 3},
  [66] = {.index = 101, .length = 3},
  [67] = {.index = 74, .length = 2},
  [68] = {.index = 17, .length = 2},
  [69] = {.index = 104, .length = 4},
  [70] = {.index = 108, .length = 4},
  [71] = {.index = 112, .length = 4},
  [72] = {.index = 116, .length = 4},
  [73] = {.index = 120, .length = 4},
  [74] = {.index = 124, .length = 3},
  [75] = {.index = 127, .length = 2},
  [76] = {.index = 127, .length = 2},
  [77] = {.index = 49, .length = 1},
  [78] = {.index = 129, .length = 3},
  [79] = {.index = 50, .length = 3},
  [80] = {.index = 132, .length = 3},
  [81] = {.index = 135, .length = 2},
  [82] = {.index = 137, .length = 3},
  [83] = {.index = 140, .length = 3},
  [84] = {.index = 143, .length = 2},
  [85] = {.index = 145, .length = 2},
  [86] = {.index = 147, .length = 2},
  [87] = {.index = 149, .length = 2},
  [88] = {.index = 151, .length = 1},
  [89] = {.index = 152, .length = 1},
  [90] = {.index = 153, .length = 2},
  [91] = {.index = 155, .length = 2},
  [92] = {.index = 157, .length = 2},
  [93] = {.index = 159, .length = 2},
  [94] = {.index = 161, .length = 4},
  [95] = {.index = 165, .length = 4},
  [96] = {.index = 50, .length = 3},
  [97] = {.index = 169, .length = 5},
  [98] = {.index = 174, .length = 5},
  [99] = {.index = 179, .length = 4},
  [100] = {.index = 183, .length = 3},
  [101] = {.index = 186, .length = 2},
  [102] = {.index = 188, .length = 1},
  [103] = {.index = 189, .length = 3},
  [104] = {.index = 192, .length = 4},
  [105] = {.index = 196, .length = 2},
  [106] = {.index = 198, .length = 3},
  [107] = {.index = 201, .length = 2},
  [108] = {.index = 203, .length = 2},
  [109] = {.index = 205, .length = 4},
  [110] = {.index = 209, .length = 3},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_path, 0},
  [1] =
    {field_name, 0},
  [2] =
    {field_parameters, 1},
  [3] =
    {field_operand, 1},
    {field_operator, 0},
  [5] =
    {field_value, 1},
  [6] =
    {field_body, 1},
  [7] =
    {field_arguments, 1},
    {field_function, 0},
  [9] =
    {field_type, 0},
    {field_type_arguments, 1},
  [11] =
    {field_body, 1},
    {field_type, 0},
  [13] =
    {field_name, 0},
    {field_path, 1},
  [15] =
    {field_name, 0},
    {field_name, 1},
  [17] =
    {field_name, 0},
    {field_type, 1},
  [19] =
    {field_type, 0},
  [20] =
    {field_name, 1},
    {field_parameters, 2},
  [22] =
    {field_parameters, 1},
    {field_result, 2},
  [24] =
    {field_body, 2},
    {field_parameters, 1},
  [26] =
    {field_element, 2},
  [27] =
    {field_label, 0},
  [28] =
    {field_value, 2},
  [29] =
    {field_condition, 1},
    {field_consequence, 2},
  [31] =
    {field_right, 1},
  [32] =
    {field_body, 2},
  [33] =
    {field_name, 2},
    {field_package, 0},
  [35] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [38] =
    {field_left, 0},
    {field_right, 2},
  [40] =
    {field_field, 2},
    {field_operand, 0},
  [42] =
    {field_channel, 0},
    {field_value, 2},
  [44] =
    {field_arguments, 2},
    {field_function, 0},
    {field_type_arguments, 1},
  [47] =
    {field_name, 0},
    {field_value, 2},
  [49] =
    {field_name, 1},
  [50] =
    {field_name, 0},
    {field_name, 1, .inherited = true},
    {field_type, 2},
  [53] =
    {field_name, 0, .inherited = true},
    {field_name, 1, .inherited = true},
  [55] =
    {field_type, 1},
  [56] =
    {field_name, 1},
    {field_parameters, 3},
    {field_type_parameters, 2},
  [59] =
    {field_name, 1},
    {field_parameters, 2},
    {field_result, 3},
  [62] =
    {field_body, 3},
    {field_name, 1},
    {field_parameters, 2},
  [65] =
    {field_name, 2},
    {field_parameters, 3},
    {field_receiver, 1},
  [68] =
    {field_body, 3},
    {field_parameters, 1},
    {field_result, 2},
  [71] =
    {field_element, 3},
  [72] =
    {field_element, 3},
    {field_length, 1},
  [74] =
    {field_name, 0},
    {field_type, 2},
  [76] =
    {field_name, 0},
    {field_type, 2},
    {field_type_parameters, 1},
  [79] =
    {field_tag, 1},
    {field_type, 0},
  [81] =
    {field_name, 0},
    {field_parameters, 1},
  [83] =
    {field_update, 2},
  [84] =
    {field_condition, 1},
  [85] =
    {field_initializer, 0},
  [86] =
    {field_alias, 1, .inherited = true},
    {field_initializer, 1, .inherited = true},
    {field_value, 1, .inherited = true},
  [89] =
    {field_right, 0},
  [90] =
    {field_operand, 2},
    {field_type, 0},
  [92] =
    {field_operand, 0},
  [93] =
    {field_index, 2},
    {field_operand, 0},
  [95] =
    {field_name, 0},
    {field_type, 1},
    {field_value, 3},
  [98] =
    {field_name, 0},
    {field_name, 1},
    {field_value, 3},
  [101] =
    {field_name, 0},
    {field_name, 1, .inherited = true},
    {field_value, 3},
  [104] =
    {field_name, 1},
    {field_parameters, 3},
    {field_result, 4},
    {field_type_parameters, 2},
  [108] =
    {field_body, 4},
    {field_name, 1},
    {field_parameters, 3},
    {field_type_parameters, 2},
  [112] =
    {field_body, 4},
    {field_name, 1},
    {field_parameters, 2},
    {field_result, 3},
  [116] =
    {field_name, 2},
    {field_parameters, 3},
    {field_receiver, 1},
    {field_result, 4},
  [120] =
    {field_body, 4},
    {field_name, 2},
    {field_parameters, 3},
    {field_receiver, 1},
  [124] =
    {field_name, 0},
    {field_type, 3},
    {field_type_parameters, 1},
  [127] =
    {field_tag, 2},
    {field_type, 1},
  [129] =
    {field_name, 0},
    {field_tag, 2},
    {field_type, 1},
  [132] =
    {field_name, 0},
    {field_parameters, 1},
    {field_result, 2},
  [135] =
    {field_key, 2},
    {field_value, 4},
  [137] =
    {field_condition, 3},
    {field_consequence, 4},
    {field_initializer, 1},
  [140] =
    {field_alternative, 4},
    {field_condition, 1},
    {field_consequence, 2},
  [143] =
    {field_condition, 1},
    {field_update, 3},
  [145] =
    {field_left, 0},
    {field_right, 3},
  [147] =
    {field_initializer, 0},
    {field_update, 3},
  [149] =
    {field_condition, 2},
    {field_initializer, 0},
  [151] =
    {field_initializer, 1},
  [152] =
    {field_communication, 1},
  [153] =
    {field_key, 0},
    {field_value, 2},
  [155] =
    {field_operand, 0},
    {field_type, 3},
  [157] =
    {field_end, 3},
    {field_operand, 0},
  [159] =
    {field_operand, 0},
    {field_start, 2},
  [161] =
    {field_name, 0},
    {field_name, 1},
    {field_type, 2},
    {field_value, 4},
  [165] =
    {field_name, 0},
    {field_name, 1, .inherited = true},
    {field_type, 2},
    {field_value, 4},
  [169] =
    {field_body, 5},
    {field_name, 1},
    {field_parameters, 3},
    {field_result, 4},
    {field_type_parameters, 2},
  [174] =
    {field_body, 5},
    {field_name, 2},
    {field_parameters, 3},
    {field_receiver, 1},
    {field_result, 4},
  [179] =
    {field_name, 0},
    {field_name, 1, .inherited = true},
    {field_tag, 3},
    {field_type, 2},
  [183] =
    {field_condition, 2},
    {field_initializer, 0},
    {field_update, 4},
  [186] =
    {field_initializer, 1},
    {field_value, 3},
  [188] =
    {field_value, 0},
  [189] =
    {field_end, 4},
    {field_operand, 0},
    {field_start, 2},
  [192] =
    {field_alternative, 6},
    {field_condition, 3},
    {field_consequence, 4},
    {field_initializer, 1},
  [196] =
    {field_type, 1},
    {field_type, 2},
  [198] =
    {field_capacity, 5},
    {field_end, 3},
    {field_operand, 0},
  [201] =
    {field_alias, 0},
    {field_value, 2},
  [203] =
    {field_initializer, 0},
    {field_value, 2},
  [205] =
    {field_capacity, 6},
    {field_end, 4},
    {field_operand, 0},
    {field_start, 2},
  [209] =
    {field_alias, 2},
    {field_initializer, 0},
    {field_value, 4},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = alias_sym_type_identifier,
  },
  [2] = {
    [1] = alias_sym_package_identifier,
  },
  [8] = {
    [1] = alias_sym_label_name,
  },
  [11] = {
    [0] = alias_sym_type_identifier,
  },
  [12] = {
    [0] = alias_sym_type_identifier,
  },
  [16] = {
    [0] = alias_sym_package_identifier,
  },
  [24] = {
    [0] = alias_sym_type_identifier,
  },
  [25] = {
    [0] = alias_sym_type_identifier,
  },
  [26] = {
    [0] = alias_sym_label_name,
  },
  [31] = {
    [0] = alias_sym_package_identifier,
    [2] = alias_sym_type_identifier,
  },
  [34] = {
    [2] = alias_sym_field_identifier,
  },
  [45] = {
    [2] = alias_sym_field_identifier,
  },
  [49] = {
    [0] = alias_sym_type_identifier,
  },
  [50] = {
    [0] = alias_sym_type_identifier,
  },
  [51] = {
    [1] = alias_sym_type_identifier,
  },
  [52] = {
    [0] = alias_sym_field_identifier,
  },
  [53] = {
    [0] = alias_sym_type_identifier,
  },
  [55] = {
    [0] = alias_sym_field_identifier,
  },
  [68] = {
    [1] = alias_sym_type_constraint,
  },
  [72] = {
    [2] = alias_sym_field_identifier,
  },
  [73] = {
    [2] = alias_sym_field_identifier,
  },
  [74] = {
    [0] = alias_sym_type_identifier,
  },
  [75] = {
    [1] = alias_sym_type_identifier,
  },
  [77] = {
    [1] = alias_sym_field_identifier,
  },
  [78] = {
    [0] = alias_sym_field_identifier,
  },
  [79] = {
    [0] = alias_sym_field_identifier,
  },
  [80] = {
    [0] = alias_sym_field_identifier,
  },
  [96] = {
    [2] = alias_sym_type_constraint,
  },
  [98] = {
    [2] = alias_sym_field_identifier,
  },
  [99] = {
    [0] = alias_sym_field_identifier,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym_type_elem, 2,
    sym_type_elem,
    alias_sym_type_constraint,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 11,
  [14] = 11,
  [15] = 11,
  [16] = 11,
  [17] = 11,
  [18] = 11,
  [19] = 19,
  [20] = 20,
  [21] = 21,
  [22] = 22,
  [23] = 23,
  [24] = 24,
  [25] = 25,
  [26] = 26,
  [27] = 27,
  [28] = 28,
  [29] = 29,
  [30] = 29,
  [31] = 29,
  [32] = 29,
  [33] = 29,
  [34] = 29,
  [35] = 35,
  [36] = 36,
  [37] = 35,
  [38] = 35,
  [39] = 36,
  [40] = 35,
  [41] = 36,
  [42] = 35,
  [43] = 36,
  [44] = 36,
  [45] = 36,
  [46] = 35,
  [47] = 47,
  [48] = 48,
  [49] = 49,
  [50] = 50,
  [51] = 51,
  [52] = 52,
  [53] = 53,
  [54] = 51,
  [55] = 55,
  [56] = 56,
  [57] = 52,
  [58] = 58,
  [59] = 51,
  [60] = 60,
  [61] = 53,
  [62] = 52,
  [63] = 60,
  [64] = 53,
  [65] = 65,
  [66] = 60,
  [67] = 52,
  [68] = 58,
  [69] = 60,
  [70] = 53,
  [71] = 52,
  [72] = 60,
  [73] = 53,
  [74] = 52,
  [75] = 60,
  [76] = 76,
  [77] = 53,
  [78] = 51,
  [79] = 51,
  [80] = 51,
  [81] = 81,
  [82] = 81,
  [83] = 83,
  [84] = 83,
  [85] = 85,
  [86] = 86,
  [87] = 87,
  [88] = 88,
  [89] = 89,
  [90] = 83,
  [91] = 91,
  [92] = 89,
  [93] = 93,
  [94] = 93,
  [95] = 85,
  [96] = 96,
  [97] = 85,
  [98] = 86,
  [99] = 99,
  [100] = 81,
  [101] = 93,
  [102] = 83,
  [103] = 103,
  [104] = 86,
  [105] = 91,
  [106] = 81,
  [107] = 107,
  [108] = 89,
  [109] = 85,
  [110] = 86,
  [111] = 111,
  [112] = 85,
  [113] = 113,
  [114] = 83,
  [115] = 115,
  [116] = 86,
  [117] = 93,
  [118] = 81,
  [119] = 119,
  [120] = 81,
  [121] = 85,
  [122] = 86,
  [123] = 91,
  [124] = 124,
  [125] = 125,
  [126] = 83,
  [127] = 127,
  [128] = 128,
  [129] = 93,
  [130] = 93,
  [131] = 131,
  [132] = 132,
  [133] = 133,
  [134] = 134,
  [135] = 135,
  [136] = 136,
  [137] = 137,
  [138] = 131,
  [139] = 139,
  [140] = 140,
  [141] = 141,
  [142] = 134,
  [143] = 136,
  [144] = 137,
  [145] = 139,
  [146] = 140,
  [147] = 141,
  [148] = 134,
  [149] = 136,
  [150] = 137,
  [151] = 139,
  [152] = 140,
  [153] = 141,
  [154] = 134,
  [155] = 136,
  [156] = 137,
  [157] = 139,
  [158] = 140,
  [159] = 141,
  [160] = 134,
  [161] = 136,
  [162] = 162,
  [163] = 139,
  [164] = 140,
  [165] = 141,
  [166] = 134,
  [167] = 136,
  [168] = 137,
  [169] = 139,
  [170] = 140,
  [171] = 141,
  [172] = 172,
  [173] = 173,
  [174] = 135,
  [175] = 131,
  [176] = 134,
  [177] = 134,
  [178] = 136,
  [179] = 137,
  [180] = 139,
  [181] = 140,
  [182] = 141,
  [183] = 172,
  [184] = 173,
  [185] = 185,
  [186] = 135,
  [187] = 135,
  [188] = 136,
  [189] = 189,
  [190] = 137,
  [191] = 172,
  [192] = 192,
  [193] = 173,
  [194] = 139,
  [195] = 135,
  [196] = 140,
  [197] = 141,
  [198] = 172,
  [199] = 173,
  [200] = 173,
  [201] = 201,
  [202] = 202,
  [203] = 172,
  [204] = 172,
  [205] = 173,
  [206] = 206,
  [207] = 207,
  [208] = 208,
  [209] = 172,
  [210] = 173,
  [211] = 172,
  [212] = 173,
  [213] = 172,
  [214] = 172,
  [215] = 162,
  [216] = 185,
  [217] = 208,
  [218] = 132,
  [219] = 162,
  [220] = 185,
  [221] = 208,
  [222] = 132,
  [223] = 162,
  [224] = 185,
  [225] = 208,
  [226] = 132,
  [227] = 162,
  [228] = 185,
  [229] = 208,
  [230] = 132,
  [231] = 162,
  [232] = 185,
  [233] = 208,
  [234] = 132,
  [235] = 137,
  [236] = 236,
  [237] = 237,
  [238] = 238,
  [239] = 239,
  [240] = 240,
  [241] = 241,
  [242] = 242,
  [243] = 243,
  [244] = 244,
  [245] = 245,
  [246] = 246,
  [247] = 247,
  [248] = 248,
  [249] = 249,
  [250] = 250,
  [251] = 251,
  [252] = 252,
  [253] = 253,
  [254] = 254,
  [255] = 255,
  [256] = 256,
  [257] = 257,
  [258] = 258,
  [259] = 259,
  [260] = 260,
  [261] = 261,
  [262] = 262,
  [263] = 263,
  [264] = 264,
  [265] = 265,
  [266] = 266,
  [267] = 267,
  [268] = 268,
  [269] = 269,
  [270] = 270,
  [271] = 271,
  [272] = 272,
  [273] = 273,
  [274] = 274,
  [275] = 275,
  [276] = 276,
  [277] = 277,
  [278] = 278,
  [279] = 279,
  [280] = 280,
  [281] = 281,
  [282] = 282,
  [283] = 283,
  [284] = 284,
  [285] = 285,
  [286] = 286,
  [287] = 287,
  [288] = 288,
  [289] = 289,
  [290] = 290,
  [291] = 243,
  [292] = 247,
  [293] = 244,
  [294] = 239,
  [295] = 246,
  [296] = 240,
  [297] = 242,
  [298] = 245,
  [299] = 248,
  [300] = 264,
  [301] = 261,
  [302] = 262,
  [303] = 263,
  [304] = 265,
  [305] = 252,
  [306] = 266,
  [307] = 272,
  [308] = 274,
  [309] = 275,
  [310] = 276,
  [311] = 277,
  [312] = 288,
  [313] = 278,
  [314] = 285,
  [315] = 279,
  [316] = 280,
  [317] = 282,
  [318] = 284,
  [319] = 283,
  [320] = 286,
  [321] = 287,
  [322] = 259,
  [323] = 257,
  [324] = 270,
  [325] = 268,
  [326] = 267,
  [327] = 289,
  [328] = 253,
  [329] = 250,
  [330] = 258,
  [331] = 255,
  [332] = 251,
  [333] = 249,
  [334] = 273,
  [335] = 269,
  [336] = 256,
  [337] = 254,
  [338] = 260,
  [339] = 239,
  [340] = 340,
  [341] = 243,
  [342] = 245,
  [343] = 246,
  [344] = 244,
  [345] = 247,
  [346] = 242,
  [347] = 248,
  [348] = 255,
  [349] = 251,
  [350] = 266,
  [351] = 351,
  [352] = 272,
  [353] = 274,
  [354] = 252,
  [355] = 275,
  [356] = 276,
  [357] = 277,
  [358] = 288,
  [359] = 249,
  [360] = 285,
  [361] = 262,
  [362] = 284,
  [363] = 263,
  [364] = 351,
  [365] = 286,
  [366] = 278,
  [367] = 283,
  [368] = 368,
  [369] = 267,
  [370] = 289,
  [371] = 253,
  [372] = 287,
  [373] = 258,
  [374] = 273,
  [375] = 259,
  [376] = 250,
  [377] = 261,
  [378] = 264,
  [379] = 268,
  [380] = 265,
  [381] = 260,
  [382] = 280,
  [383] = 270,
  [384] = 257,
  [385] = 269,
  [386] = 282,
  [387] = 256,
  [388] = 254,
  [389] = 279,
  [390] = 244,
  [391] = 246,
  [392] = 392,
  [393] = 393,
  [394] = 248,
  [395] = 239,
  [396] = 252,
  [397] = 288,
  [398] = 285,
  [399] = 284,
  [400] = 400,
  [401] = 289,
  [402] = 253,
  [403] = 258,
  [404] = 273,
  [405] = 264,
  [406] = 268,
  [407] = 269,
  [408] = 256,
  [409] = 254,
  [410] = 260,
  [411] = 261,
  [412] = 262,
  [413] = 263,
  [414] = 265,
  [415] = 266,
  [416] = 272,
  [417] = 274,
  [418] = 275,
  [419] = 276,
  [420] = 277,
  [421] = 278,
  [422] = 279,
  [423] = 280,
  [424] = 282,
  [425] = 283,
  [426] = 286,
  [427] = 287,
  [428] = 259,
  [429] = 257,
  [430] = 270,
  [431] = 239,
  [432] = 250,
  [433] = 251,
  [434] = 249,
  [435] = 255,
  [436] = 267,
  [437] = 340,
  [438] = 247,
  [439] = 351,
  [440] = 244,
  [441] = 243,
  [442] = 242,
  [443] = 245,
  [444] = 243,
  [445] = 242,
  [446] = 245,
  [447] = 247,
  [448] = 246,
  [449] = 351,
  [450] = 248,
  [451] = 451,
  [452] = 452,
  [453] = 453,
  [454] = 454,
  [455] = 455,
  [456] = 456,
  [457] = 457,
  [458] = 239,
  [459] = 459,
  [460] = 460,
  [461] = 266,
  [462] = 254,
  [463] = 260,
  [464] = 261,
  [465] = 262,
  [466] = 263,
  [467] = 265,
  [468] = 272,
  [469] = 274,
  [470] = 275,
  [471] = 276,
  [472] = 277,
  [473] = 278,
  [474] = 279,
  [475] = 280,
  [476] = 282,
  [477] = 283,
  [478] = 286,
  [479] = 287,
  [480] = 259,
  [481] = 257,
  [482] = 270,
  [483] = 284,
  [484] = 250,
  [485] = 251,
  [486] = 249,
  [487] = 267,
  [488] = 340,
  [489] = 289,
  [490] = 258,
  [491] = 273,
  [492] = 454,
  [493] = 459,
  [494] = 264,
  [495] = 256,
  [496] = 351,
  [497] = 255,
  [498] = 498,
  [499] = 246,
  [500] = 351,
  [501] = 244,
  [502] = 460,
  [503] = 252,
  [504] = 253,
  [505] = 243,
  [506] = 242,
  [507] = 245,
  [508] = 247,
  [509] = 288,
  [510] = 498,
  [511] = 498,
  [512] = 285,
  [513] = 269,
  [514] = 268,
  [515] = 239,
  [516] = 239,
  [517] = 248,
  [518] = 518,
  [519] = 519,
  [520] = 520,
  [521] = 519,
  [522] = 400,
  [523] = 520,
  [524] = 518,
  [525] = 518,
  [526] = 519,
  [527] = 520,
  [528] = 518,
  [529] = 529,
  [530] = 518,
  [531] = 518,
  [532] = 532,
  [533] = 286,
  [534] = 534,
  [535] = 535,
  [536] = 243,
  [537] = 242,
  [538] = 245,
  [539] = 247,
  [540] = 540,
  [541] = 541,
  [542] = 542,
  [543] = 543,
  [544] = 255,
  [545] = 239,
  [546] = 252,
  [547] = 288,
  [548] = 285,
  [549] = 284,
  [550] = 267,
  [551] = 289,
  [552] = 253,
  [553] = 258,
  [554] = 273,
  [555] = 264,
  [556] = 268,
  [557] = 269,
  [558] = 256,
  [559] = 254,
  [560] = 260,
  [561] = 261,
  [562] = 262,
  [563] = 263,
  [564] = 265,
  [565] = 266,
  [566] = 272,
  [567] = 274,
  [568] = 275,
  [569] = 569,
  [570] = 277,
  [571] = 278,
  [572] = 279,
  [573] = 280,
  [574] = 282,
  [575] = 283,
  [576] = 287,
  [577] = 259,
  [578] = 257,
  [579] = 270,
  [580] = 250,
  [581] = 251,
  [582] = 249,
  [583] = 569,
  [584] = 584,
  [585] = 585,
  [586] = 243,
  [587] = 242,
  [588] = 245,
  [589] = 247,
  [590] = 585,
  [591] = 591,
  [592] = 569,
  [593] = 400,
  [594] = 584,
  [595] = 585,
  [596] = 584,
  [597] = 597,
  [598] = 276,
  [599] = 599,
  [600] = 600,
  [601] = 601,
  [602] = 456,
  [603] = 452,
  [604] = 604,
  [605] = 600,
  [606] = 606,
  [607] = 607,
  [608] = 608,
  [609] = 604,
  [610] = 610,
  [611] = 608,
  [612] = 612,
  [613] = 599,
  [614] = 614,
  [615] = 340,
  [616] = 452,
  [617] = 604,
  [618] = 600,
  [619] = 607,
  [620] = 608,
  [621] = 610,
  [622] = 599,
  [623] = 452,
  [624] = 604,
  [625] = 600,
  [626] = 626,
  [627] = 608,
  [628] = 599,
  [629] = 452,
  [630] = 604,
  [631] = 600,
  [632] = 608,
  [633] = 633,
  [634] = 599,
  [635] = 452,
  [636] = 604,
  [637] = 600,
  [638] = 608,
  [639] = 599,
  [640] = 610,
  [641] = 606,
  [642] = 607,
  [643] = 340,
  [644] = 644,
  [645] = 645,
  [646] = 646,
  [647] = 647,
  [648] = 648,
  [649] = 648,
  [650] = 650,
  [651] = 651,
  [652] = 652,
  [653] = 653,
  [654] = 650,
  [655] = 650,
  [656] = 647,
  [657] = 657,
  [658] = 644,
  [659] = 653,
  [660] = 647,
  [661] = 661,
  [662] = 648,
  [663] = 650,
  [664] = 651,
  [665] = 652,
  [666] = 653,
  [667] = 651,
  [668] = 668,
  [669] = 456,
  [670] = 648,
  [671] = 647,
  [672] = 672,
  [673] = 652,
  [674] = 647,
  [675] = 675,
  [676] = 648,
  [677] = 650,
  [678] = 651,
  [679] = 652,
  [680] = 653,
  [681] = 681,
  [682] = 653,
  [683] = 644,
  [684] = 652,
  [685] = 644,
  [686] = 686,
  [687] = 651,
  [688] = 647,
  [689] = 652,
  [690] = 648,
  [691] = 650,
  [692] = 692,
  [693] = 693,
  [694] = 653,
  [695] = 657,
  [696] = 696,
  [697] = 657,
  [698] = 644,
  [699] = 644,
  [700] = 651,
  [701] = 701,
  [702] = 702,
  [703] = 703,
  [704] = 704,
  [705] = 705,
  [706] = 706,
  [707] = 707,
  [708] = 706,
  [709] = 702,
  [710] = 710,
  [711] = 706,
  [712] = 712,
  [713] = 713,
  [714] = 714,
  [715] = 715,
  [716] = 716,
  [717] = 717,
  [718] = 718,
  [719] = 703,
  [720] = 702,
  [721] = 718,
  [722] = 714,
  [723] = 717,
  [724] = 704,
  [725] = 706,
  [726] = 710,
  [727] = 727,
  [728] = 716,
  [729] = 714,
  [730] = 730,
  [731] = 716,
  [732] = 718,
  [733] = 717,
  [734] = 702,
  [735] = 704,
  [736] = 704,
  [737] = 706,
  [738] = 738,
  [739] = 739,
  [740] = 715,
  [741] = 710,
  [742] = 716,
  [743] = 718,
  [744] = 702,
  [745] = 710,
  [746] = 707,
  [747] = 747,
  [748] = 704,
  [749] = 710,
  [750] = 750,
  [751] = 704,
  [752] = 716,
  [753] = 718,
  [754] = 702,
  [755] = 706,
  [756] = 714,
  [757] = 715,
  [758] = 758,
  [759] = 716,
  [760] = 714,
  [761] = 705,
  [762] = 713,
  [763] = 717,
  [764] = 701,
  [765] = 718,
  [766] = 717,
  [767] = 705,
  [768] = 713,
  [769] = 701,
  [770] = 714,
  [771] = 701,
  [772] = 705,
  [773] = 713,
  [774] = 717,
  [775] = 703,
  [776] = 701,
  [777] = 705,
  [778] = 713,
  [779] = 701,
  [780] = 705,
  [781] = 713,
  [782] = 710,
  [783] = 783,
  [784] = 784,
  [785] = 785,
  [786] = 786,
  [787] = 787,
  [788] = 788,
  [789] = 783,
  [790] = 785,
  [791] = 786,
  [792] = 784,
  [793] = 785,
  [794] = 784,
  [795] = 783,
  [796] = 786,
  [797] = 797,
  [798] = 798,
  [799] = 797,
  [800] = 800,
  [801] = 801,
  [802] = 802,
  [803] = 801,
  [804] = 804,
  [805] = 805,
  [806] = 806,
  [807] = 807,
  [808] = 808,
  [809] = 809,
  [810] = 810,
  [811] = 811,
  [812] = 812,
  [813] = 813,
  [814] = 814,
  [815] = 815,
  [816] = 816,
  [817] = 817,
  [818] = 818,
  [819] = 819,
  [820] = 820,
  [821] = 821,
  [822] = 822,
  [823] = 822,
  [824] = 824,
  [825] = 825,
  [826] = 826,
  [827] = 827,
  [828] = 828,
  [829] = 829,
  [830] = 830,
  [831] = 831,
  [832] = 832,
  [833] = 833,
  [834] = 834,
  [835] = 835,
  [836] = 836,
  [837] = 837,
  [838] = 822,
  [839] = 839,
  [840] = 840,
  [841] = 841,
  [842] = 842,
  [843] = 843,
  [844] = 797,
  [845] = 845,
  [846] = 843,
  [847] = 801,
  [848] = 848,
  [849] = 849,
  [850] = 850,
  [851] = 850,
  [852] = 852,
  [853] = 849,
  [854] = 854,
  [855] = 797,
  [856] = 856,
  [857] = 854,
  [858] = 832,
  [859] = 836,
  [860] = 834,
  [861] = 809,
  [862] = 810,
  [863] = 807,
  [864] = 830,
  [865] = 865,
  [866] = 806,
  [867] = 867,
  [868] = 829,
  [869] = 821,
  [870] = 812,
  [871] = 818,
  [872] = 831,
  [873] = 815,
  [874] = 814,
  [875] = 801,
  [876] = 805,
  [877] = 811,
  [878] = 813,
  [879] = 816,
  [880] = 817,
  [881] = 804,
  [882] = 824,
  [883] = 826,
  [884] = 808,
  [885] = 837,
  [886] = 827,
  [887] = 833,
  [888] = 819,
  [889] = 798,
  [890] = 890,
  [891] = 891,
  [892] = 892,
  [893] = 893,
  [894] = 894,
  [895] = 895,
  [896] = 896,
  [897] = 897,
  [898] = 797,
  [899] = 890,
  [900] = 797,
  [901] = 901,
  [902] = 890,
  [903] = 891,
  [904] = 891,
  [905] = 800,
  [906] = 906,
  [907] = 906,
  [908] = 908,
  [909] = 906,
  [910] = 910,
  [911] = 911,
  [912] = 912,
  [913] = 913,
  [914] = 914,
  [915] = 915,
  [916] = 916,
  [917] = 917,
  [918] = 801,
  [919] = 919,
  [920] = 920,
  [921] = 921,
  [922] = 801,
  [923] = 923,
  [924] = 924,
  [925] = 925,
  [926] = 926,
  [927] = 927,
  [928] = 928,
  [929] = 929,
  [930] = 819,
  [931] = 931,
  [932] = 932,
  [933] = 933,
  [934] = 821,
  [935] = 935,
  [936] = 936,
  [937] = 937,
  [938] = 814,
  [939] = 939,
  [940] = 940,
  [941] = 941,
  [942] = 942,
  [943] = 798,
  [944] = 804,
  [945] = 827,
  [946] = 946,
  [947] = 836,
  [948] = 818,
  [949] = 949,
  [950] = 834,
  [951] = 809,
  [952] = 810,
  [953] = 807,
  [954] = 954,
  [955] = 955,
  [956] = 956,
  [957] = 957,
  [958] = 958,
  [959] = 959,
  [960] = 800,
  [961] = 961,
  [962] = 962,
  [963] = 831,
  [964] = 805,
  [965] = 965,
  [966] = 811,
  [967] = 813,
  [968] = 824,
  [969] = 969,
  [970] = 970,
  [971] = 833,
  [972] = 972,
  [973] = 973,
  [974] = 817,
  [975] = 826,
  [976] = 808,
  [977] = 977,
  [978] = 830,
  [979] = 979,
  [980] = 980,
  [981] = 981,
  [982] = 982,
  [983] = 983,
  [984] = 829,
  [985] = 832,
  [986] = 986,
  [987] = 987,
  [988] = 812,
  [989] = 989,
  [990] = 990,
  [991] = 991,
  [992] = 992,
  [993] = 993,
  [994] = 994,
  [995] = 995,
  [996] = 996,
  [997] = 997,
  [998] = 998,
  [999] = 999,
  [1000] = 1000,
  [1001] = 1001,
  [1002] = 1002,
  [1003] = 1003,
  [1004] = 1004,
  [1005] = 1005,
  [1006] = 1006,
  [1007] = 1007,
  [1008] = 1008,
  [1009] = 1009,
  [1010] = 1010,
  [1011] = 1011,
  [1012] = 1012,
  [1013] = 1013,
  [1014] = 1014,
  [1015] = 1015,
  [1016] = 1016,
  [1017] = 1017,
  [1018] = 1018,
  [1019] = 983,
  [1020] = 837,
  [1021] = 806,
  [1022] = 1022,
  [1023] = 1023,
  [1024] = 1024,
  [1025] = 815,
  [1026] = 1026,
  [1027] = 816,
  [1028] = 1028,
  [1029] = 1029,
  [1030] = 1030,
  [1031] = 1031,
  [1032] = 1032,
  [1033] = 1033,
  [1034] = 867,
  [1035] = 1035,
  [1036] = 1036,
  [1037] = 867,
  [1038] = 1038,
  [1039] = 867,
  [1040] = 1035,
  [1041] = 1041,
  [1042] = 1042,
  [1043] = 1043,
  [1044] = 1033,
  [1045] = 1041,
  [1046] = 1046,
  [1047] = 1047,
  [1048] = 1010,
  [1049] = 931,
  [1050] = 1035,
  [1051] = 1051,
  [1052] = 1052,
  [1053] = 867,
  [1054] = 1038,
  [1055] = 1055,
  [1056] = 1056,
  [1057] = 1042,
  [1058] = 867,
  [1059] = 956,
  [1060] = 1055,
  [1061] = 1038,
  [1062] = 1055,
  [1063] = 1041,
  [1064] = 901,
  [1065] = 1065,
  [1066] = 1066,
  [1067] = 1067,
  [1068] = 1068,
  [1069] = 1069,
  [1070] = 1070,
  [1071] = 1071,
  [1072] = 995,
  [1073] = 1073,
  [1074] = 1074,
  [1075] = 955,
  [1076] = 1076,
  [1077] = 1077,
  [1078] = 253,
  [1079] = 1079,
  [1080] = 1080,
  [1081] = 892,
  [1082] = 895,
  [1083] = 1083,
  [1084] = 1080,
  [1085] = 1083,
  [1086] = 1086,
  [1087] = 1087,
  [1088] = 998,
  [1089] = 1089,
  [1090] = 1090,
  [1091] = 1091,
  [1092] = 1087,
  [1093] = 1093,
  [1094] = 1094,
  [1095] = 1095,
  [1096] = 1096,
  [1097] = 1080,
  [1098] = 892,
  [1099] = 895,
  [1100] = 1100,
  [1101] = 1083,
  [1102] = 1102,
  [1103] = 901,
  [1104] = 1087,
  [1105] = 1105,
  [1106] = 1106,
  [1107] = 1107,
  [1108] = 1108,
  [1109] = 1109,
  [1110] = 1110,
  [1111] = 1111,
  [1112] = 1112,
  [1113] = 1113,
  [1114] = 1080,
  [1115] = 1115,
  [1116] = 1076,
  [1117] = 892,
  [1118] = 895,
  [1119] = 1119,
  [1120] = 1120,
  [1121] = 1121,
  [1122] = 1122,
  [1123] = 1083,
  [1124] = 1124,
  [1125] = 1125,
  [1126] = 901,
  [1127] = 1087,
  [1128] = 1128,
  [1129] = 1111,
  [1130] = 1130,
  [1131] = 251,
  [1132] = 249,
  [1133] = 1133,
  [1134] = 1134,
  [1135] = 1135,
  [1136] = 1080,
  [1137] = 892,
  [1138] = 895,
  [1139] = 1139,
  [1140] = 1140,
  [1141] = 1141,
  [1142] = 1083,
  [1143] = 901,
  [1144] = 1087,
  [1145] = 1145,
  [1146] = 1124,
  [1147] = 1083,
  [1148] = 250,
  [1149] = 892,
  [1150] = 1080,
  [1151] = 1083,
  [1152] = 1152,
  [1153] = 901,
  [1154] = 895,
  [1155] = 1119,
  [1156] = 1121,
  [1157] = 1070,
  [1158] = 1087,
  [1159] = 252,
  [1160] = 1160,
  [1161] = 1080,
  [1162] = 1162,
  [1163] = 1051,
  [1164] = 1164,
  [1165] = 1165,
  [1166] = 1166,
  [1167] = 1167,
  [1168] = 1168,
  [1169] = 1169,
  [1170] = 1170,
  [1171] = 1095,
  [1172] = 1172,
  [1173] = 1173,
  [1174] = 1174,
  [1175] = 1175,
  [1176] = 1176,
  [1177] = 1177,
  [1178] = 1178,
  [1179] = 1179,
  [1180] = 1180,
  [1181] = 1181,
  [1182] = 1182,
  [1183] = 1089,
  [1184] = 1184,
  [1185] = 1102,
  [1186] = 1106,
  [1187] = 1164,
  [1188] = 1165,
  [1189] = 1178,
  [1190] = 1190,
  [1191] = 1191,
  [1192] = 1166,
  [1193] = 1193,
  [1194] = 1164,
  [1195] = 1195,
  [1196] = 1167,
  [1197] = 1168,
  [1198] = 1170,
  [1199] = 1167,
  [1200] = 1200,
  [1201] = 1179,
  [1202] = 1141,
  [1203] = 1203,
  [1204] = 1168,
  [1205] = 1128,
  [1206] = 1206,
  [1207] = 1207,
  [1208] = 1086,
  [1209] = 1165,
  [1210] = 1178,
  [1211] = 1211,
  [1212] = 1190,
  [1213] = 1213,
  [1214] = 1191,
  [1215] = 1215,
  [1216] = 1190,
  [1217] = 1164,
  [1218] = 1167,
  [1219] = 1219,
  [1220] = 1168,
  [1221] = 1170,
  [1222] = 1222,
  [1223] = 1074,
  [1224] = 1224,
  [1225] = 1225,
  [1226] = 1226,
  [1227] = 1227,
  [1228] = 1191,
  [1229] = 1229,
  [1230] = 1165,
  [1231] = 1231,
  [1232] = 1178,
  [1233] = 1190,
  [1234] = 1139,
  [1235] = 1191,
  [1236] = 1160,
  [1237] = 1164,
  [1238] = 1167,
  [1239] = 1168,
  [1240] = 1170,
  [1241] = 1165,
  [1242] = 1231,
  [1243] = 1169,
  [1244] = 798,
  [1245] = 1222,
  [1246] = 1231,
  [1247] = 1247,
  [1248] = 800,
  [1249] = 1164,
  [1250] = 1178,
  [1251] = 1190,
  [1252] = 1252,
  [1253] = 1166,
  [1254] = 1165,
  [1255] = 1169,
  [1256] = 1067,
  [1257] = 1167,
  [1258] = 1191,
  [1259] = 1178,
  [1260] = 1190,
  [1261] = 1071,
  [1262] = 1168,
  [1263] = 1191,
  [1264] = 1170,
  [1265] = 1179,
  [1266] = 800,
  [1267] = 1170,
  [1268] = 1268,
  [1269] = 1269,
  [1270] = 1270,
  [1271] = 1271,
  [1272] = 1272,
  [1273] = 1270,
  [1274] = 1271,
  [1275] = 1275,
  [1276] = 1276,
  [1277] = 1277,
  [1278] = 1272,
  [1279] = 1279,
  [1280] = 1280,
  [1281] = 1281,
  [1282] = 1270,
  [1283] = 1272,
  [1284] = 1271,
  [1285] = 1270,
  [1286] = 1271,
  [1287] = 942,
  [1288] = 1288,
  [1289] = 1289,
  [1290] = 1290,
  [1291] = 946,
  [1292] = 1281,
  [1293] = 1272,
  [1294] = 1294,
  [1295] = 1270,
  [1296] = 1296,
  [1297] = 1297,
  [1298] = 1298,
  [1299] = 1271,
  [1300] = 1281,
  [1301] = 1301,
  [1302] = 939,
  [1303] = 1303,
  [1304] = 1304,
  [1305] = 1296,
  [1306] = 1289,
  [1307] = 1014,
  [1308] = 1308,
  [1309] = 1309,
  [1310] = 1270,
  [1311] = 1015,
  [1312] = 1309,
  [1313] = 1271,
  [1314] = 1314,
  [1315] = 1289,
  [1316] = 1316,
  [1317] = 1317,
  [1318] = 1318,
  [1319] = 1272,
  [1320] = 1289,
  [1321] = 1281,
  [1322] = 1281,
  [1323] = 1281,
  [1324] = 1289,
  [1325] = 1325,
  [1326] = 1272,
  [1327] = 1296,
  [1328] = 1328,
  [1329] = 1329,
  [1330] = 1330,
  [1331] = 1289,
  [1332] = 1332,
  [1333] = 1333,
  [1334] = 1334,
  [1335] = 1335,
  [1336] = 1269,
  [1337] = 1269,
  [1338] = 1269,
  [1339] = 1269,
  [1340] = 1269,
  [1341] = 1341,
  [1342] = 1341,
  [1343] = 1343,
  [1344] = 1344,
  [1345] = 1345,
  [1346] = 1341,
  [1347] = 1344,
  [1348] = 1348,
  [1349] = 1343,
  [1350] = 1350,
  [1351] = 1351,
  [1352] = 1352,
  [1353] = 1343,
  [1354] = 1354,
  [1355] = 1355,
  [1356] = 1356,
  [1357] = 1357,
  [1358] = 1358,
  [1359] = 1359,
  [1360] = 1356,
  [1361] = 1361,
  [1362] = 1351,
  [1363] = 1363,
  [1364] = 1364,
  [1365] = 1355,
  [1366] = 1352,
  [1367] = 1356,
  [1368] = 1368,
  [1369] = 1344,
  [1370] = 1370,
  [1371] = 1371,
  [1372] = 1372,
  [1373] = 1373,
  [1374] = 1363,
  [1375] = 1375,
  [1376] = 1350,
  [1377] = 1377,
  [1378] = 1378,
  [1379] = 1379,
  [1380] = 1352,
  [1381] = 1352,
  [1382] = 1377,
  [1383] = 1377,
  [1384] = 1384,
  [1385] = 1385,
  [1386] = 1355,
  [1387] = 1387,
  [1388] = 1388,
  [1389] = 1363,
  [1390] = 1375,
  [1391] = 1391,
  [1392] = 1352,
  [1393] = 1352,
  [1394] = 1394,
  [1395] = 1395,
  [1396] = 1388,
  [1397] = 1397,
  [1398] = 1355,
  [1399] = 1344,
  [1400] = 1400,
  [1401] = 1375,
  [1402] = 1377,
  [1403] = 1355,
  [1404] = 1356,
  [1405] = 1405,
  [1406] = 1406,
  [1407] = 1375,
  [1408] = 1408,
  [1409] = 1341,
  [1410] = 1341,
  [1411] = 1375,
  [1412] = 1352,
  [1413] = 1413,
  [1414] = 1343,
  [1415] = 1415,
  [1416] = 1416,
  [1417] = 1355,
  [1418] = 1343,
  [1419] = 1419,
  [1420] = 1350,
  [1421] = 1375,
  [1422] = 1344,
  [1423] = 1377,
  [1424] = 1356,
  [1425] = 1425,
  [1426] = 1377,
  [1427] = 1355,
  [1428] = 1344,
  [1429] = 1388,
  [1430] = 1375,
  [1431] = 1364,
  [1432] = 1351,
  [1433] = 1364,
  [1434] = 1356,
  [1435] = 1364,
  [1436] = 1341,
  [1437] = 1364,
  [1438] = 1345,
  [1439] = 1364,
  [1440] = 1440,
  [1441] = 1343,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym__expression,
  sym__simple_statement,
  sym__simple_type,
  sym__statement,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym__expression] = {.index = 0, .length = 25},
  [sym__simple_statement] = {.index = 25, .length = 6},
  [sym__simple_type] = {.index = 31, .length = 12},
  [sym__statement] = {.index = 43, .length = 19},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    anon_sym_make,
    anon_sym_new,
    sym_binary_expression,
    sym_call_expression,
    sym_composite_literal,
    sym_false,
    sym_float_literal,
    sym_func_literal,
    sym_identifier,
    sym_imaginary_literal,
    sym_index_expression,
    sym_int_literal,
    sym_interpreted_string_literal,
    sym_iota,
    sym_nil,
    sym_parenthesized_expression,
    sym_raw_string_literal,
    sym_rune_literal,
    sym_selector_expression,
    sym_slice_expression,
    sym_true,
    sym_type_assertion_expression,
    sym_type_conversion_expression,
    sym_type_instantiation_expression,
    sym_unary_expression,
  [25] =
    sym_assignment_statement,
    sym_dec_statement,
    sym_expression_statement,
    sym_inc_statement,
    sym_send_statement,
    sym_short_var_declaration,
  [31] =
    alias_sym_type_identifier,
    sym_array_type,
    sym_channel_type,
    sym_function_type,
    sym_generic_type,
    sym_interface_type,
    sym_map_type,
    sym_negated_type,
    sym_pointer_type,
    sym_qualified_type,
    sym_slice_type,
    sym_struct_type,
  [43] =
    sym__simple_statement,
    sym_block,
    sym_break_statement,
    sym_const_declaration,
    sym_continue_statement,
    sym_defer_statement,
    sym_empty_statement,
    sym_expression_switch_statement,
    sym_fallthrough_statement,
    sym_for_statement,
    sym_go_statement,
    sym_goto_statement,
    sym_if_statement,
    sym_labeled_statement,
    sym_return_statement,
    sym_select_statement,
    sym_type_declaration,
    sym_type_switch_statement,
    sym_var_declaration,
};

static const TSCharacterRange sym_identifier_character_set_1[] = {
  {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6},
  {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x370, 0x374}, {0x376, 0x377}, {0x37b, 0x37d},
  {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5}, {0x3f7, 0x481}, {0x48a, 0x52f},
  {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x620, 0x64a}, {0x66e, 0x66f}, {0x671, 0x6d3},
  {0x6d5, 0x6d5}, {0x6e5, 0x6e6}, {0x6ee, 0x6ef}, {0x6fa, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x710}, {0x712, 0x72f}, {0x74d, 0x7a5},
  {0x7b1, 0x7b1}, {0x7ca, 0x7ea}, {0x7f4, 0x7f5}, {0x7fa, 0x7fa}, {0x800, 0x815}, {0x81a, 0x81a}, {0x824, 0x824}, {0x828, 0x828},
  {0x840, 0x858}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x8a0, 0x8c9}, {0x904, 0x939}, {0x93d, 0x93d}, {0x950, 0x950},
  {0x958, 0x961}, {0x971, 0x980}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9},
  {0x9bd, 0x9bd}, {0x9ce, 0x9ce}, {0x9dc, 0x9dd}, {0x9df, 0x9e1}, {0x9f0, 0x9f1}, {0x9fc, 0x9fc}, {0xa05, 0xa0a}, {0xa0f, 0xa10},
  {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39}, {0xa59, 0xa5c}, {0xa5e, 0xa5e}, {0xa72, 0xa74},
  {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9}, {0xabd, 0xabd}, {0xad0, 0xad0},
  {0xae0, 0xae1}, {0xaf9, 0xaf9}, {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39},
  {0xb3d, 0xb3d}, {0xb5c, 0xb5d}, {0xb5f, 0xb61}, {0xb71, 0xb71}, {0xb83, 0xb83}, {0xb85, 0xb8a}, {0xb8e, 0xb90}, {0xb92, 0xb95},
  {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9}, {0xbd0, 0xbd0}, {0xc05, 0xc0c},
  {0xc0e, 0xc10}, {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3d, 0xc3d}, {0xc58, 0xc5a}, {0xc5d, 0xc5d}, {0xc60, 0xc61}, {0xc80, 0xc80},
  {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9}, {0xcbd, 0xcbd}, {0xcdd, 0xcde}, {0xce0, 0xce1},
  {0xcf1, 0xcf2}, {0xd04, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd3a}, {0xd3d, 0xd3d}, {0xd4e, 0xd4e}, {0xd54, 0xd56}, {0xd5f, 0xd61},
  {0xd7a, 0xd7f}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xe01, 0xe30}, {0xe32, 0xe32},
  {0xe40, 0xe46}, {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xeb0}, {0xeb2, 0xeb2},
  {0xebd, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf40, 0xf47}, {0xf49, 0xf6c}, {0xf88, 0xf8c},
  {0x1000, 0x102a}, {0x103f, 0x103f}, {0x1050, 0x1055}, {0x105a, 0x105d}, {0x1061, 0x1061}, {0x1065, 0x1066}, {0x106e, 0x1070}, {0x1075, 0x1081},
  {0x108e, 0x108e}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256},
  {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0},
  {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x1380, 0x138f}, {0x13a0, 0x13f5}, {0x13f8, 0x13fd},
  {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1711}, {0x171f, 0x1731}, {0x1740, 0x1751},
  {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1780, 0x17b3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dc}, {0x1820, 0x1878}, {0x1880, 0x18a8}, {0x18aa, 0x18aa},
  {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1950, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x1a00, 0x1a16}, {0x1a20, 0x1a54},
  {0x1aa7, 0x1aa7}, {0x1b05, 0x1b33}, {0x1b45, 0x1b4c}, {0x1b83, 0x1ba0}, {0x1bae, 0x1baf}, {0x1bba, 0x1be5}, {0x1c00, 0x1c23}, {0x1c4d, 0x1c4f},
  {0x1c5a, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1ce9, 0x1cec}, {0x1cee, 0x1cf3}, {0x1cf5, 0x1cf6}, {0x1cfa, 0x1cfa},
  {0x1d00, 0x1dbf}, {0x1e00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b},
  {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3},
  {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x2102, 0x2102},
  {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139},
  {0x213c, 0x213f}, {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cee}, {0x2cf2, 0x2cf3}, {0x2d00, 0x2d25},
  {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d80, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6},
  {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x3005, 0x3007}, {0x3021, 0x3029}, {0x3031, 0x3035},
  {0x3038, 0x303c}, {0x3041, 0x3096}, {0x309d, 0x309f}, {0x30a1, 0x30fa}, {0x30fc, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf},
  {0x31f0, 0x31ff}, {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa61f}, {0xa62a, 0xa62b}, {0xa640, 0xa66e},
  {0xa67f, 0xa69d}, {0xa6a0, 0xa6ef}, {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc},
  {0xa7f2, 0xa801}, {0xa803, 0xa805}, {0xa807, 0xa80a}, {0xa80c, 0xa822}, {0xa840, 0xa873}, {0xa882, 0xa8b3}, {0xa8f2, 0xa8f7}, {0xa8fb, 0xa8fb},
  {0xa8fd, 0xa8fe}, {0xa90a, 0xa925}, {0xa930, 0xa946}, {0xa960, 0xa97c}, {0xa984, 0xa9b2}, {0xa9cf, 0xa9cf}, {0xa9e0, 0xa9e4}, {0xa9e6, 0xa9ef},
  {0xa9fa, 0xa9fe}, {0xaa00, 0xaa28}, {0xaa40, 0xaa42}, {0xaa44, 0xaa4b}, {0xaa60, 0xaa76}, {0xaa7a, 0xaa7a}, {0xaa7e, 0xaaaf}, {0xaab1, 0xaab1},
  {0xaab5, 0xaab6}, {0xaab9, 0xaabd}, {0xaac0, 0xaac0}, {0xaac2, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaea}, {0xaaf2, 0xaaf4}, {0xab01, 0xab06},
  {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabe2}, {0xac00, 0xd7a3},
  {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb1d}, {0xfb1f, 0xfb28},
  {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d},
  {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77}, {0xfe79, 0xfe79}, {0xfe7b, 0xfe7b},
  {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff21, 0xff3a}, {0xff41, 0xff5a}, {0xff66, 0xff9d}, {0xffa0, 0xffbe}, {0xffc2, 0xffc7}, {0xffca, 0xffcf},
  {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d}, {0x1003f, 0x1004d}, {0x10050, 0x1005d},
  {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x10375}, {0x10380, 0x1039d},
  {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563},
  {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc},
  {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805},
  {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2},
  {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a00}, {0x10a10, 0x10a13}, {0x10a15, 0x10a17},
  {0x10a19, 0x10a35}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae4}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72},
  {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d23}, {0x10d4a, 0x10d65}, {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9},
  {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10f00, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f45}, {0x10f70, 0x10f81}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6},
  {0x11003, 0x11037}, {0x11071, 0x11072}, {0x11075, 0x11075}, {0x11083, 0x110af}, {0x110d0, 0x110e8}, {0x11103, 0x11126}, {0x11144, 0x11144}, {0x11147, 0x11147},
  {0x11150, 0x11172}, {0x11176, 0x11176}, {0x11183, 0x111b2}, {0x111c1, 0x111c4}, {0x111da, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x1122b},
  {0x1123f, 0x11240}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112de}, {0x11305, 0x1130c},
  {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133d, 0x1133d}, {0x11350, 0x11350}, {0x1135d, 0x11361},
  {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113b7}, {0x113d1, 0x113d1}, {0x113d3, 0x113d3}, {0x11400, 0x11434},
  {0x11447, 0x1144a}, {0x1145f, 0x11461}, {0x11480, 0x114af}, {0x114c4, 0x114c5}, {0x114c7, 0x114c7}, {0x11580, 0x115ae}, {0x115d8, 0x115db}, {0x11600, 0x1162f},
  {0x11644, 0x11644}, {0x11680, 0x116aa}, {0x116b8, 0x116b8}, {0x11700, 0x1171a}, {0x11740, 0x11746}, {0x11800, 0x1182b}, {0x118a0, 0x118df}, {0x118ff, 0x11906},
  {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x1192f}, {0x1193f, 0x1193f}, {0x11941, 0x11941}, {0x119a0, 0x119a7}, {0x119aa, 0x119d0},
  {0x119e1, 0x119e1}, {0x119e3, 0x119e3}, {0x11a00, 0x11a00}, {0x11a0b, 0x11a32}, {0x11a3a, 0x11a3a}, {0x11a50, 0x11a50}, {0x11a5c, 0x11a89}, {0x11a9d, 0x11a9d},
  {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c2e}, {0x11c40, 0x11c40}, {0x11c72, 0x11c8f}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09},
  {0x11d0b, 0x11d30}, {0x11d46, 0x11d46}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d89}, {0x11d98, 0x11d98}, {0x11ee0, 0x11ef2}, {0x11f02, 0x11f02},
  {0x11f04, 0x11f10}, {0x11f12, 0x11f33}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f},
  {0x13441, 0x13446}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x1611d}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a70, 0x16abe}, {0x16ad0, 0x16aed},
  {0x16b00, 0x16b2f}, {0x16b40, 0x16b43}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a}, {0x16f50, 0x16f50},
  {0x16f93, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe3}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb},
  {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a},
  {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6},
  {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c},
  {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da},
  {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2},
  {0x1d7c4, 0x1d7cb}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e030, 0x1e06d}, {0x1e100, 0x1e12c}, {0x1e137, 0x1e13d}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ad},
  {0x1e2c0, 0x1e2eb}, {0x1e4d0, 0x1e4eb}, {0x1e5d0, 0x1e5ed}, {0x1e5f0, 0x1e5f0}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe},
  {0x1e800, 0x1e8c4}, {0x1e900, 0x1e943}, {0x1e94b, 0x1e94b}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27},
  {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b},
  {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f},
  {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89},
  {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1},
  {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
};

static const TSCharacterRange sym_identifier_character_set_2[] = {
  {'0', '9'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xb7, 0xb7}, {0xba, 0xba},
  {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x300, 0x374},
  {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5}, {0x3f7, 0x481},
  {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf}, {0x5c1, 0x5c2},
  {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3}, {0x6d5, 0x6dc},
  {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa}, {0x7fd, 0x7fd},
  {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963}, {0x966, 0x96f},
  {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9}, {0x9bc, 0x9c4},
  {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9fc, 0x9fc}, {0x9fe, 0x9fe},
  {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39},
  {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e}, {0xa66, 0xa75},
  {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9}, {0xabc, 0xac5},
  {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03}, {0xb05, 0xb0c},
  {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48}, {0xb4b, 0xb4d},
  {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83}, {0xb85, 0xb8a}, {0xb8e, 0xb90},
  {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9}, {0xbbe, 0xbc2},
  {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c}, {0xc0e, 0xc10}, {0xc12, 0xc28},
  {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d}, {0xc60, 0xc63},
  {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9}, {0xcbc, 0xcc4},
  {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef}, {0xcf1, 0xcf3}, {0xd00, 0xd0c},
  {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63}, {0xd66, 0xd6f}, {0xd7a, 0xd7f},
  {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca}, {0xdcf, 0xdd4},
  {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59}, {0xe81, 0xe82},
  {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6}, {0xec8, 0xece},
  {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35}, {0xf37, 0xf37}, {0xf39, 0xf39},
  {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049}, {0x1050, 0x109d},
  {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256}, {0x1258, 0x1258},
  {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0}, {0x12c2, 0x12c5},
  {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371}, {0x1380, 0x138f}, {0x13a0, 0x13f5},
  {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715}, {0x171f, 0x1734},
  {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd}, {0x17e0, 0x17e9},
  {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1920, 0x192b}, {0x1930, 0x193b},
  {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b}, {0x1a20, 0x1a5e}, {0x1a60, 0x1a7c},
  {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59}, {0x1b6b, 0x1b73},
  {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1cd0, 0x1cd2},
  {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b},
  {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3},
  {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040}, {0x2054, 0x2054}, {0x2071, 0x2071},
  {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210a, 0x2113},
  {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f}, {0x2145, 0x2149},
  {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67},
  {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce},
  {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035}, {0x3038, 0x303c}, {0x3041, 0x3096},
  {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff}, {0x3400, 0x4dbf},
  {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1}, {0xa717, 0xa71f},
  {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c}, {0xa840, 0xa873},
  {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953}, {0xa960, 0xa97c}, {0xa980, 0xa9c0},
  {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76}, {0xaa7a, 0xaac2}, {0xaadb, 0xaadd},
  {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a},
  {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d},
  {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41},
  {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9}, {0xfe00, 0xfe0f},
  {0xfe20, 0xfe2f}, {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77}, {0xfe79, 0xfe79}, {0xfe7b, 0xfe7b},
  {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f}, {0xff41, 0xff5a}, {0xff65, 0xffbe}, {0xffc2, 0xffc7},
  {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d}, {0x1003f, 0x1004d},
  {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x102e0, 0x102e0}, {0x10300, 0x1031f},
  {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104a0, 0x104a9},
  {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595},
  {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767},
  {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c},
  {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939}, {0x10980, 0x109b7},
  {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a38, 0x10a3a}, {0x10a3f, 0x10a3f},
  {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72}, {0x10b80, 0x10b91},
  {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65}, {0x10d69, 0x10d6d}, {0x10d6f, 0x10d85},
  {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f50}, {0x10f70, 0x10f85},
  {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2}, {0x110d0, 0x110e8}, {0x110f0, 0x110f9},
  {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4}, {0x111c9, 0x111cc}, {0x111ce, 0x111da},
  {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d},
  {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330},
  {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350}, {0x11357, 0x11357}, {0x1135d, 0x11363},
  {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113c0}, {0x113c2, 0x113c2},
  {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459}, {0x1145e, 0x11461}, {0x11480, 0x114c5},
  {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640}, {0x11644, 0x11644}, {0x11650, 0x11659},
  {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x11739}, {0x11740, 0x11746}, {0x11800, 0x1183a},
  {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935}, {0x11937, 0x11938}, {0x1193b, 0x11943},
  {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e}, {0x11a47, 0x11a47}, {0x11a50, 0x11a99},
  {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36}, {0x11c38, 0x11c40}, {0x11c50, 0x11c59},
  {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36}, {0x11d3a, 0x11d3a}, {0x11d3c, 0x11d3d},
  {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91}, {0x11d93, 0x11d98}, {0x11da0, 0x11da9},
  {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399}, {0x12400, 0x1246e},
  {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x16139}, {0x16800, 0x16a38},
  {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4}, {0x16b00, 0x16b36}, {0x16b40, 0x16b43},
  {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a}, {0x16f4f, 0x16f87},
  {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3},
  {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb},
  {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9}, {0x1cf00, 0x1cf2d}, {0x1cf30, 0x1cf46},
  {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c},
  {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505},
  {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550},
  {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e},
  {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36}, {0x1da3b, 0x1da6c}, {0x1da75, 0x1da75},
  {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006}, {0x1e008, 0x1e018}, {0x1e01b, 0x1e021},
  {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d}, {0x1e140, 0x1e149}, {0x1e14e, 0x1e14e},
  {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe},
  {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24},
  {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49},
  {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d},
  {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e},
  {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1fbf0, 0x1fbf9}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739},
  {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af}, {0xe0100, 0xe01ef},
};

static const TSCharacterRange sym_rune_literal_character_set_1[] = {
  {'"', '"'}, {'\'', '\''}, {'0', '7'}, {'U', 'U'}, {'\\', '\\'}, {'a', 'b'}, {'f', 'f'}, {'n', 'n'},
  {'r', 'r'}, {'t', 'v'}, {'x', 'x'},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(59);
      ADVANCE_MAP(
        0, 62,
        '!', 106,
        '"', 148,
        '%', 115,
        '&', 110,
        '\'', 22,
        '(', 66,
        ')', 67,
        '*', 75,
        '+', 100,
        ',', 68,
        '-', 103,
        '.', 65,
        '/', 113,
        '0', 152,
        ':', 98,
        ';', 61,
        '<', 124,
        '=', 70,
        '>', 129,
        '[', 71,
        '\\', 20,
        ']', 72,
        '^', 108,
        '`', 135,
        '{', 77,
        '|', 80,
        '}', 78,
        '~', 76,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(57);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(154);
      if (set_contains(sym_identifier_character_set_1, 685, lookahead)) ADVANCE(134);
      END_STATE();
    case 1:
      ADVANCE_MAP(
        0, 62,
        '\n', 60,
        '!', 17,
        '%', 114,
        '&', 111,
        '(', 66,
        '*', 74,
        '+', 99,
        ',', 68,
        '-', 102,
        '.', 63,
        '/', 112,
        ';', 61,
        '<', 127,
        '=', 19,
        '>', 130,
        '[', 71,
        '^', 107,
        '{', 77,
        '|', 81,
        '}', 78,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(1);
      if (set_contains(sym_identifier_character_set_1, 685, lookahead)) ADVANCE(134);
      END_STATE();
    case 2:
      if (lookahead == '\n') SKIP(16);
      if (lookahead == '"') ADVANCE(148);
      if (lookahead == '/') ADVANCE(143);
      if (lookahead == '\\') ADVANCE(20);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(146);
      if (lookahead != 0) ADVANCE(147);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '!', 17,
        '"', 142,
        '%', 114,
        '&', 111,
        '(', 66,
        ')', 67,
        '*', 74,
        '+', 99,
        ',', 68,
        '-', 102,
        '.', 64,
        '/', 112,
        ':', 98,
        '<', 125,
        '=', 70,
        '>', 130,
        '[', 71,
        ']', 72,
        '^', 107,
        '`', 135,
        '{', 77,
        '|', 81,
        '~', 76,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(3);
      if (set_contains(sym_identifier_character_set_1, 685, lookahead)) ADVANCE(134);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        '!', 17,
        '%', 115,
        '&', 110,
        '(', 66,
        '*', 75,
        '+', 100,
        ',', 68,
        '-', 103,
        '.', 63,
        '/', 113,
        ':', 18,
        ';', 61,
        '<', 124,
        '=', 70,
        '>', 129,
        '[', 71,
        '^', 108,
        '{', 77,
        '|', 80,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(4);
      END_STATE();
    case 5:
      ADVANCE_MAP(
        '!', 17,
        '%', 115,
        '&', 110,
        '(', 66,
        '*', 75,
        '+', 101,
        ',', 68,
        '-', 104,
        '.', 63,
        '/', 113,
        ':', 18,
        '<', 126,
        '=', 70,
        '>', 129,
        '[', 71,
        '^', 108,
        '{', 77,
        '|', 80,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(5);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '!', 17,
        '%', 114,
        '&', 111,
        '(', 66,
        ')', 67,
        '*', 74,
        '+', 99,
        ',', 68,
        '-', 102,
        '.', 64,
        '/', 112,
        ':', 98,
        ';', 61,
        '<', 127,
        '=', 70,
        '>', 130,
        '[', 71,
        ']', 72,
        '^', 107,
        '{', 77,
        '|', 81,
        '}', 78,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(6);
      END_STATE();
    case 7:
      if (lookahead == '\'') ADVANCE(162);
      END_STATE();
    case 8:
      if (lookahead == '*') ADVANCE(10);
      if (lookahead == '/') ADVANCE(164);
      END_STATE();
    case 9:
      if (lookahead == '*') ADVANCE(9);
      if (lookahead == '/') ADVANCE(163);
      if (lookahead != 0) ADVANCE(10);
      END_STATE();
    case 10:
      if (lookahead == '*') ADVANCE(9);
      if (lookahead != 0) ADVANCE(10);
      END_STATE();
    case 11:
      if (lookahead == '-') ADVANCE(82);
      END_STATE();
    case 12:
      if (lookahead == '.') ADVANCE(14);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(158);
      END_STATE();
    case 13:
      if (lookahead == '.') ADVANCE(160);
      if (lookahead == '_') ADVANCE(35);
      if (lookahead == 'i') ADVANCE(161);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(26);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(13);
      END_STATE();
    case 14:
      if (lookahead == '.') ADVANCE(73);
      END_STATE();
    case 15:
      if (lookahead == '.') ADVANCE(40);
      if (lookahead == '_') ADVANCE(39);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(155);
      END_STATE();
    case 16:
      if (lookahead == '/') ADVANCE(8);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(16);
      END_STATE();
    case 17:
      if (lookahead == '=') ADVANCE(123);
      END_STATE();
    case 18:
      if (lookahead == '=') ADVANCE(83);
      END_STATE();
    case 19:
      if (lookahead == '=') ADVANCE(122);
      END_STATE();
    case 20:
      if (lookahead == 'U') ADVANCE(53);
      if (lookahead == 'u') ADVANCE(45);
      if (lookahead == 'x') ADVANCE(41);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(151);
      if (lookahead != 0) ADVANCE(149);
      END_STATE();
    case 21:
      if (lookahead == 'U') ADVANCE(54);
      if (lookahead == 'u') ADVANCE(46);
      if (lookahead == 'x') ADVANCE(42);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(32);
      if (set_contains(sym_rune_literal_character_set_1, 11, lookahead)) ADVANCE(7);
      END_STATE();
    case 22:
      if (lookahead == '\\') ADVANCE(21);
      if (lookahead != 0 &&
          lookahead != '\'') ADVANCE(7);
      END_STATE();
    case 23:
      if (lookahead == '_') ADVANCE(28);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(156);
      END_STATE();
    case 24:
      if (lookahead == '_') ADVANCE(40);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(26);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(24);
      END_STATE();
    case 25:
      if (lookahead == '_') ADVANCE(31);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(157);
      END_STATE();
    case 26:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(36);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(159);
      END_STATE();
    case 27:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(26);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(24);
      END_STATE();
    case 28:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(156);
      END_STATE();
    case 29:
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(13);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(153);
      END_STATE();
    case 30:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(7);
      END_STATE();
    case 31:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(157);
      END_STATE();
    case 32:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(30);
      END_STATE();
    case 33:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(154);
      END_STATE();
    case 34:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(158);
      END_STATE();
    case 35:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(13);
      END_STATE();
    case 36:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(159);
      END_STATE();
    case 37:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(7);
      END_STATE();
    case 38:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(149);
      END_STATE();
    case 39:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(155);
      END_STATE();
    case 40:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(24);
      END_STATE();
    case 41:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(38);
      END_STATE();
    case 42:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(37);
      END_STATE();
    case 43:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(41);
      END_STATE();
    case 44:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(42);
      END_STATE();
    case 45:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(43);
      END_STATE();
    case 46:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(44);
      END_STATE();
    case 47:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(45);
      END_STATE();
    case 48:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(46);
      END_STATE();
    case 49:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(47);
      END_STATE();
    case 50:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(48);
      END_STATE();
    case 51:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(49);
      END_STATE();
    case 52:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(50);
      END_STATE();
    case 53:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(51);
      END_STATE();
    case 54:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(52);
      END_STATE();
    case 55:
      if (eof) ADVANCE(59);
      ADVANCE_MAP(
        0, 62,
        '\n', 60,
        '!', 105,
        '"', 142,
        '&', 109,
        '\'', 22,
        '(', 66,
        ')', 67,
        '*', 74,
        '+', 99,
        ',', 68,
        '-', 102,
        '.', 34,
        '/', 8,
        '0', 152,
        ';', 61,
        '<', 11,
        '=', 69,
        '[', 71,
        '^', 107,
        '`', 135,
        '{', 77,
        '|', 79,
        '}', 78,
        '~', 76,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(55);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(154);
      if (set_contains(sym_identifier_character_set_1, 685, lookahead)) ADVANCE(134);
      END_STATE();
    case 56:
      if (eof) ADVANCE(59);
      ADVANCE_MAP(
        0, 62,
        '\n', 60,
        '!', 17,
        '"', 142,
        '%', 115,
        '&', 110,
        '(', 66,
        ')', 67,
        '*', 75,
        '+', 100,
        ',', 68,
        '-', 103,
        '.', 63,
        '/', 113,
        ':', 98,
        ';', 61,
        '<', 124,
        '=', 70,
        '>', 129,
        '[', 71,
        '^', 108,
        '`', 135,
        '{', 77,
        '|', 80,
        '}', 78,
        '~', 76,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(56);
      if (set_contains(sym_identifier_character_set_1, 685, lookahead)) ADVANCE(134);
      END_STATE();
    case 57:
      if (eof) ADVANCE(59);
      ADVANCE_MAP(
        0, 62,
        '!', 106,
        '"', 142,
        '%', 115,
        '&', 110,
        '\'', 22,
        '(', 66,
        ')', 67,
        '*', 75,
        '+', 100,
        ',', 68,
        '-', 103,
        '.', 65,
        '/', 113,
        '0', 152,
        ':', 98,
        ';', 61,
        '<', 124,
        '=', 70,
        '>', 129,
        '[', 71,
        ']', 72,
        '^', 108,
        '`', 135,
        '{', 77,
        '|', 80,
        '}', 78,
        '~', 76,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(57);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(154);
      if (set_contains(sym_identifier_character_set_1, 685, lookahead)) ADVANCE(134);
      END_STATE();
    case 58:
      if (eof) ADVANCE(59);
      ADVANCE_MAP(
        '!', 105,
        '"', 142,
        '&', 109,
        '\'', 22,
        '(', 66,
        ')', 67,
        '*', 74,
        '+', 99,
        ',', 68,
        '-', 102,
        '.', 12,
        '/', 8,
        '0', 152,
        ':', 97,
        ';', 61,
        '<', 11,
        '[', 71,
        ']', 72,
        '^', 107,
        '`', 135,
        '{', 77,
        '}', 78,
        '~', 76,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(58);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(154);
      if (set_contains(sym_identifier_character_set_1, 685, lookahead)) ADVANCE(134);
      END_STATE();
    case 59:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 60:
      ACCEPT_TOKEN(aux_sym_source_file_token1);
      if (lookahead == '\n') ADVANCE(60);
      END_STATE();
    case 61:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 62:
      ACCEPT_TOKEN(anon_sym_NULL);
      END_STATE();
    case 63:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 64:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(14);
      END_STATE();
    case 65:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(14);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(158);
      END_STATE();
    case 66:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 67:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 68:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 69:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 70:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(122);
      END_STATE();
    case 71:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 72:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 73:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_DOT);
      END_STATE();
    case 74:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 75:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '=') ADVANCE(86);
      END_STATE();
    case 76:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 77:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 78:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 80:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(95);
      if (lookahead == '|') ADVANCE(133);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(133);
      END_STATE();
    case 82:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      END_STATE();
    case 83:
      ACCEPT_TOKEN(anon_sym_COLON_EQ);
      END_STATE();
    case 84:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 86:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 88:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 89:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 90:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 91:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_AMP_CARET_EQ);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 94:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 95:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 96:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 97:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 98:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '=') ADVANCE(83);
      END_STATE();
    case 99:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 100:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(84);
      if (lookahead == '=') ADVANCE(93);
      END_STATE();
    case 101:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '=') ADVANCE(93);
      END_STATE();
    case 102:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 103:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(85);
      if (lookahead == '=') ADVANCE(94);
      END_STATE();
    case 104:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '=') ADVANCE(94);
      END_STATE();
    case 105:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 106:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(123);
      END_STATE();
    case 107:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(96);
      END_STATE();
    case 109:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 110:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(132);
      if (lookahead == '=') ADVANCE(91);
      if (lookahead == '^') ADVANCE(121);
      END_STATE();
    case 111:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(132);
      if (lookahead == '^') ADVANCE(120);
      END_STATE();
    case 112:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(10);
      if (lookahead == '/') ADVANCE(164);
      END_STATE();
    case 113:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(10);
      if (lookahead == '/') ADVANCE(164);
      if (lookahead == '=') ADVANCE(87);
      END_STATE();
    case 114:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 115:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(88);
      END_STATE();
    case 116:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 117:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(89);
      END_STATE();
    case 118:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 119:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(90);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(anon_sym_AMP_CARET);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(anon_sym_AMP_CARET);
      if (lookahead == '=') ADVANCE(92);
      END_STATE();
    case 122:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '-') ADVANCE(82);
      if (lookahead == '<') ADVANCE(117);
      if (lookahead == '=') ADVANCE(128);
      END_STATE();
    case 125:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '-') ADVANCE(82);
      if (lookahead == '<') ADVANCE(116);
      if (lookahead == '=') ADVANCE(128);
      END_STATE();
    case 126:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(117);
      if (lookahead == '=') ADVANCE(128);
      END_STATE();
    case 127:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(116);
      if (lookahead == '=') ADVANCE(128);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 129:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(131);
      if (lookahead == '>') ADVANCE(119);
      END_STATE();
    case 130:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(131);
      if (lookahead == '>') ADVANCE(118);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 134:
      ACCEPT_TOKEN(sym_identifier);
      if (set_contains(sym_identifier_character_set_2, 800, lookahead)) ADVANCE(134);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(aux_sym_raw_string_literal_token1);
      if (lookahead == '\n') ADVANCE(141);
      if (lookahead != 0 &&
          lookahead != '`') ADVANCE(136);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(aux_sym_raw_string_literal_token1);
      if (lookahead == '*') ADVANCE(139);
      if (lookahead == '/') ADVANCE(136);
      if (lookahead != 0 &&
          lookahead != '`') ADVANCE(141);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(aux_sym_raw_string_literal_token1);
      if (lookahead == '*') ADVANCE(138);
      if (lookahead == '/') ADVANCE(141);
      if (lookahead != 0 &&
          lookahead != '`') ADVANCE(139);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(aux_sym_raw_string_literal_token1);
      if (lookahead == '*') ADVANCE(138);
      if (lookahead != 0 &&
          lookahead != '`') ADVANCE(139);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(aux_sym_raw_string_literal_token1);
      if (lookahead == '/') ADVANCE(137);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(140);
      if (lookahead != 0 &&
          lookahead != '`') ADVANCE(141);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(aux_sym_raw_string_literal_token1);
      if (lookahead != 0 &&
          lookahead != '`') ADVANCE(141);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(aux_sym_interpreted_string_literal_token1);
      if (lookahead == '*') ADVANCE(145);
      if (lookahead == '/') ADVANCE(147);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(147);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(aux_sym_interpreted_string_literal_token1);
      if (lookahead == '*') ADVANCE(144);
      if (lookahead == '/') ADVANCE(147);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(145);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(aux_sym_interpreted_string_literal_token1);
      if (lookahead == '*') ADVANCE(144);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(145);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(aux_sym_interpreted_string_literal_token1);
      if (lookahead == '/') ADVANCE(143);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(146);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(147);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(aux_sym_interpreted_string_literal_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(147);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(anon_sym_DQUOTE2);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(149);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(150);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(sym_int_literal);
      ADVANCE_MAP(
        '.', 160,
        '_', 29,
        'i', 161,
        'B', 23,
        'b', 23,
        'E', 26,
        'e', 26,
        'O', 25,
        'o', 25,
        'X', 15,
        'x', 15,
        '8', 13,
        '9', 13,
      );
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(153);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(sym_int_literal);
      if (lookahead == '.') ADVANCE(160);
      if (lookahead == '_') ADVANCE(29);
      if (lookahead == 'i') ADVANCE(161);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(26);
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(13);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(153);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(sym_int_literal);
      if (lookahead == '.') ADVANCE(160);
      if (lookahead == '_') ADVANCE(33);
      if (lookahead == 'i') ADVANCE(161);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(26);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(154);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(sym_int_literal);
      if (lookahead == '.') ADVANCE(27);
      if (lookahead == '_') ADVANCE(39);
      if (lookahead == 'i') ADVANCE(161);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(26);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(155);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(sym_int_literal);
      if (lookahead == '_') ADVANCE(28);
      if (lookahead == 'i') ADVANCE(161);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(156);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(sym_int_literal);
      if (lookahead == '_') ADVANCE(31);
      if (lookahead == 'i') ADVANCE(161);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(157);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(sym_float_literal);
      if (lookahead == '_') ADVANCE(34);
      if (lookahead == 'i') ADVANCE(161);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(26);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(158);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(sym_float_literal);
      if (lookahead == '_') ADVANCE(36);
      if (lookahead == 'i') ADVANCE(161);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(159);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(sym_float_literal);
      if (lookahead == 'i') ADVANCE(161);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(26);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(158);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(sym_imaginary_literal);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(sym_rune_literal);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(sym_comment);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(164);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        '_', 1,
        'b', 2,
        'c', 3,
        'd', 4,
        'e', 5,
        'f', 6,
        'g', 7,
        'i', 8,
        'm', 9,
        'n', 10,
        'p', 11,
        'r', 12,
        's', 13,
        't', 14,
        'v', 15,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(0);
      END_STATE();
    case 1:
      ACCEPT_TOKEN(sym_blank_identifier);
      END_STATE();
    case 2:
      if (lookahead == 'r') ADVANCE(16);
      END_STATE();
    case 3:
      if (lookahead == 'a') ADVANCE(17);
      if (lookahead == 'h') ADVANCE(18);
      if (lookahead == 'o') ADVANCE(19);
      END_STATE();
    case 4:
      if (lookahead == 'e') ADVANCE(20);
      END_STATE();
    case 5:
      if (lookahead == 'l') ADVANCE(21);
      END_STATE();
    case 6:
      if (lookahead == 'a') ADVANCE(22);
      if (lookahead == 'o') ADVANCE(23);
      if (lookahead == 'u') ADVANCE(24);
      END_STATE();
    case 7:
      if (lookahead == 'o') ADVANCE(25);
      END_STATE();
    case 8:
      if (lookahead == 'f') ADVANCE(26);
      if (lookahead == 'm') ADVANCE(27);
      if (lookahead == 'n') ADVANCE(28);
      if (lookahead == 'o') ADVANCE(29);
      END_STATE();
    case 9:
      if (lookahead == 'a') ADVANCE(30);
      END_STATE();
    case 10:
      if (lookahead == 'e') ADVANCE(31);
      if (lookahead == 'i') ADVANCE(32);
      END_STATE();
    case 11:
      if (lookahead == 'a') ADVANCE(33);
      END_STATE();
    case 12:
      if (lookahead == 'a') ADVANCE(34);
      if (lookahead == 'e') ADVANCE(35);
      END_STATE();
    case 13:
      if (lookahead == 'e') ADVANCE(36);
      if (lookahead == 't') ADVANCE(37);
      if (lookahead == 'w') ADVANCE(38);
      END_STATE();
    case 14:
      if (lookahead == 'r') ADVANCE(39);
      if (lookahead == 'y') ADVANCE(40);
      END_STATE();
    case 15:
      if (lookahead == 'a') ADVANCE(41);
      END_STATE();
    case 16:
      if (lookahead == 'e') ADVANCE(42);
      END_STATE();
    case 17:
      if (lookahead == 's') ADVANCE(43);
      END_STATE();
    case 18:
      if (lookahead == 'a') ADVANCE(44);
      END_STATE();
    case 19:
      if (lookahead == 'n') ADVANCE(45);
      END_STATE();
    case 20:
      if (lookahead == 'f') ADVANCE(46);
      END_STATE();
    case 21:
      if (lookahead == 's') ADVANCE(47);
      END_STATE();
    case 22:
      if (lookahead == 'l') ADVANCE(48);
      END_STATE();
    case 23:
      if (lookahead == 'r') ADVANCE(49);
      END_STATE();
    case 24:
      if (lookahead == 'n') ADVANCE(50);
      END_STATE();
    case 25:
      ACCEPT_TOKEN(anon_sym_go);
      if (lookahead == 't') ADVANCE(51);
      END_STATE();
    case 26:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 27:
      if (lookahead == 'p') ADVANCE(52);
      END_STATE();
    case 28:
      if (lookahead == 't') ADVANCE(53);
      END_STATE();
    case 29:
      if (lookahead == 't') ADVANCE(54);
      END_STATE();
    case 30:
      if (lookahead == 'k') ADVANCE(55);
      if (lookahead == 'p') ADVANCE(56);
      END_STATE();
    case 31:
      if (lookahead == 'w') ADVANCE(57);
      END_STATE();
    case 32:
      if (lookahead == 'l') ADVANCE(58);
      END_STATE();
    case 33:
      if (lookahead == 'c') ADVANCE(59);
      END_STATE();
    case 34:
      if (lookahead == 'n') ADVANCE(60);
      END_STATE();
    case 35:
      if (lookahead == 't') ADVANCE(61);
      END_STATE();
    case 36:
      if (lookahead == 'l') ADVANCE(62);
      END_STATE();
    case 37:
      if (lookahead == 'r') ADVANCE(63);
      END_STATE();
    case 38:
      if (lookahead == 'i') ADVANCE(64);
      END_STATE();
    case 39:
      if (lookahead == 'u') ADVANCE(65);
      END_STATE();
    case 40:
      if (lookahead == 'p') ADVANCE(66);
      END_STATE();
    case 41:
      if (lookahead == 'r') ADVANCE(67);
      END_STATE();
    case 42:
      if (lookahead == 'a') ADVANCE(68);
      END_STATE();
    case 43:
      if (lookahead == 'e') ADVANCE(69);
      END_STATE();
    case 44:
      if (lookahead == 'n') ADVANCE(70);
      END_STATE();
    case 45:
      if (lookahead == 's') ADVANCE(71);
      if (lookahead == 't') ADVANCE(72);
      END_STATE();
    case 46:
      if (lookahead == 'a') ADVANCE(73);
      if (lookahead == 'e') ADVANCE(74);
      END_STATE();
    case 47:
      if (lookahead == 'e') ADVANCE(75);
      END_STATE();
    case 48:
      if (lookahead == 'l') ADVANCE(76);
      if (lookahead == 's') ADVANCE(77);
      END_STATE();
    case 49:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 50:
      if (lookahead == 'c') ADVANCE(78);
      END_STATE();
    case 51:
      if (lookahead == 'o') ADVANCE(79);
      END_STATE();
    case 52:
      if (lookahead == 'o') ADVANCE(80);
      END_STATE();
    case 53:
      if (lookahead == 'e') ADVANCE(81);
      END_STATE();
    case 54:
      if (lookahead == 'a') ADVANCE(82);
      END_STATE();
    case 55:
      if (lookahead == 'e') ADVANCE(83);
      END_STATE();
    case 56:
      ACCEPT_TOKEN(anon_sym_map);
      END_STATE();
    case 57:
      ACCEPT_TOKEN(anon_sym_new);
      END_STATE();
    case 58:
      ACCEPT_TOKEN(sym_nil);
      END_STATE();
    case 59:
      if (lookahead == 'k') ADVANCE(84);
      END_STATE();
    case 60:
      if (lookahead == 'g') ADVANCE(85);
      END_STATE();
    case 61:
      if (lookahead == 'u') ADVANCE(86);
      END_STATE();
    case 62:
      if (lookahead == 'e') ADVANCE(87);
      END_STATE();
    case 63:
      if (lookahead == 'u') ADVANCE(88);
      END_STATE();
    case 64:
      if (lookahead == 't') ADVANCE(89);
      END_STATE();
    case 65:
      if (lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 66:
      if (lookahead == 'e') ADVANCE(91);
      END_STATE();
    case 67:
      ACCEPT_TOKEN(anon_sym_var);
      END_STATE();
    case 68:
      if (lookahead == 'k') ADVANCE(92);
      END_STATE();
    case 69:
      ACCEPT_TOKEN(anon_sym_case);
      END_STATE();
    case 70:
      ACCEPT_TOKEN(anon_sym_chan);
      END_STATE();
    case 71:
      if (lookahead == 't') ADVANCE(93);
      END_STATE();
    case 72:
      if (lookahead == 'i') ADVANCE(94);
      END_STATE();
    case 73:
      if (lookahead == 'u') ADVANCE(95);
      END_STATE();
    case 74:
      if (lookahead == 'r') ADVANCE(96);
      END_STATE();
    case 75:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 76:
      if (lookahead == 't') ADVANCE(97);
      END_STATE();
    case 77:
      if (lookahead == 'e') ADVANCE(98);
      END_STATE();
    case 78:
      ACCEPT_TOKEN(anon_sym_func);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_goto);
      END_STATE();
    case 80:
      if (lookahead == 'r') ADVANCE(99);
      END_STATE();
    case 81:
      if (lookahead == 'r') ADVANCE(100);
      END_STATE();
    case 82:
      ACCEPT_TOKEN(sym_iota);
      END_STATE();
    case 83:
      ACCEPT_TOKEN(anon_sym_make);
      END_STATE();
    case 84:
      if (lookahead == 'a') ADVANCE(101);
      END_STATE();
    case 85:
      if (lookahead == 'e') ADVANCE(102);
      END_STATE();
    case 86:
      if (lookahead == 'r') ADVANCE(103);
      END_STATE();
    case 87:
      if (lookahead == 'c') ADVANCE(104);
      END_STATE();
    case 88:
      if (lookahead == 'c') ADVANCE(105);
      END_STATE();
    case 89:
      if (lookahead == 'c') ADVANCE(106);
      END_STATE();
    case 90:
      ACCEPT_TOKEN(sym_true);
      END_STATE();
    case 91:
      ACCEPT_TOKEN(anon_sym_type);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_break);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_const);
      END_STATE();
    case 94:
      if (lookahead == 'n') ADVANCE(107);
      END_STATE();
    case 95:
      if (lookahead == 'l') ADVANCE(108);
      END_STATE();
    case 96:
      ACCEPT_TOKEN(anon_sym_defer);
      END_STATE();
    case 97:
      if (lookahead == 'h') ADVANCE(109);
      END_STATE();
    case 98:
      ACCEPT_TOKEN(sym_false);
      END_STATE();
    case 99:
      if (lookahead == 't') ADVANCE(110);
      END_STATE();
    case 100:
      if (lookahead == 'f') ADVANCE(111);
      END_STATE();
    case 101:
      if (lookahead == 'g') ADVANCE(112);
      END_STATE();
    case 102:
      ACCEPT_TOKEN(anon_sym_range);
      END_STATE();
    case 103:
      if (lookahead == 'n') ADVANCE(113);
      END_STATE();
    case 104:
      if (lookahead == 't') ADVANCE(114);
      END_STATE();
    case 105:
      if (lookahead == 't') ADVANCE(115);
      END_STATE();
    case 106:
      if (lookahead == 'h') ADVANCE(116);
      END_STATE();
    case 107:
      if (lookahead == 'u') ADVANCE(117);
      END_STATE();
    case 108:
      if (lookahead == 't') ADVANCE(118);
      END_STATE();
    case 109:
      if (lookahead == 'r') ADVANCE(119);
      END_STATE();
    case 110:
      ACCEPT_TOKEN(anon_sym_import);
      END_STATE();
    case 111:
      if (lookahead == 'a') ADVANCE(120);
      END_STATE();
    case 112:
      if (lookahead == 'e') ADVANCE(121);
      END_STATE();
    case 113:
      ACCEPT_TOKEN(anon_sym_return);
      END_STATE();
    case 114:
      ACCEPT_TOKEN(anon_sym_select);
      END_STATE();
    case 115:
      ACCEPT_TOKEN(anon_sym_struct);
      END_STATE();
    case 116:
      ACCEPT_TOKEN(anon_sym_switch);
      END_STATE();
    case 117:
      if (lookahead == 'e') ADVANCE(122);
      END_STATE();
    case 118:
      ACCEPT_TOKEN(anon_sym_default);
      END_STATE();
    case 119:
      if (lookahead == 'o') ADVANCE(123);
      END_STATE();
    case 120:
      if (lookahead == 'c') ADVANCE(124);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(anon_sym_package);
      END_STATE();
    case 122:
      ACCEPT_TOKEN(anon_sym_continue);
      END_STATE();
    case 123:
      if (lookahead == 'u') ADVANCE(125);
      END_STATE();
    case 124:
      if (lookahead == 'e') ADVANCE(126);
      END_STATE();
    case 125:
      if (lookahead == 'g') ADVANCE(127);
      END_STATE();
    case 126:
      ACCEPT_TOKEN(anon_sym_interface);
      END_STATE();
    case 127:
      if (lookahead == 'h') ADVANCE(128);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_fallthrough);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0},
  [1] = {.lex_state = 58, .reserved_word_set_id = 1},
  [2] = {.lex_state = 58, .reserved_word_set_id = 1},
  [3] = {.lex_state = 58, .reserved_word_set_id = 1},
  [4] = {.lex_state = 58, .reserved_word_set_id = 1},
  [5] = {.lex_state = 58, .reserved_word_set_id = 1},
  [6] = {.lex_state = 58, .reserved_word_set_id = 1},
  [7] = {.lex_state = 58, .reserved_word_set_id = 1},
  [8] = {.lex_state = 58, .reserved_word_set_id = 1},
  [9] = {.lex_state = 58, .reserved_word_set_id = 2},
  [10] = {.lex_state = 58, .reserved_word_set_id = 2},
  [11] = {.lex_state = 58, .reserved_word_set_id = 3},
  [12] = {.lex_state = 58, .reserved_word_set_id = 2},
  [13] = {.lex_state = 58, .reserved_word_set_id = 1},
  [14] = {.lex_state = 58, .reserved_word_set_id = 1},
  [15] = {.lex_state = 58, .reserved_word_set_id = 3},
  [16] = {.lex_state = 58, .reserved_word_set_id = 3},
  [17] = {.lex_state = 58, .reserved_word_set_id = 3},
  [18] = {.lex_state = 58, .reserved_word_set_id = 1},
  [19] = {.lex_state = 58, .reserved_word_set_id = 3},
  [20] = {.lex_state = 58, .reserved_word_set_id = 3},
  [21] = {.lex_state = 58, .reserved_word_set_id = 4},
  [22] = {.lex_state = 58, .reserved_word_set_id = 5},
  [23] = {.lex_state = 58, .reserved_word_set_id = 1},
  [24] = {.lex_state = 58, .reserved_word_set_id = 1},
  [25] = {.lex_state = 58, .reserved_word_set_id = 1},
  [26] = {.lex_state = 58, .reserved_word_set_id = 1},
  [27] = {.lex_state = 58, .reserved_word_set_id = 5},
  [28] = {.lex_state = 55, .reserved_word_set_id = 6},
  [29] = {.lex_state = 58, .reserved_word_set_id = 5},
  [30] = {.lex_state = 58, .reserved_word_set_id = 5},
  [31] = {.lex_state = 58, .reserved_word_set_id = 5},
  [32] = {.lex_state = 58, .reserved_word_set_id = 5},
  [33] = {.lex_state = 58, .reserved_word_set_id = 1},
  [34] = {.lex_state = 58, .reserved_word_set_id = 1},
  [35] = {.lex_state = 58, .reserved_word_set_id = 5},
  [36] = {.lex_state = 58, .reserved_word_set_id = 1},
  [37] = {.lex_state = 58, .reserved_word_set_id = 5},
  [38] = {.lex_state = 58, .reserved_word_set_id = 5},
  [39] = {.lex_state = 58, .reserved_word_set_id = 5},
  [40] = {.lex_state = 58, .reserved_word_set_id = 1},
  [41] = {.lex_state = 58, .reserved_word_set_id = 5},
  [42] = {.lex_state = 58, .reserved_word_set_id = 5},
  [43] = {.lex_state = 58, .reserved_word_set_id = 5},
  [44] = {.lex_state = 58, .reserved_word_set_id = 5},
  [45] = {.lex_state = 58, .reserved_word_set_id = 1},
  [46] = {.lex_state = 58, .reserved_word_set_id = 1},
  [47] = {.lex_state = 58, .reserved_word_set_id = 1},
  [48] = {.lex_state = 58, .reserved_word_set_id = 1},
  [49] = {.lex_state = 58, .reserved_word_set_id = 1},
  [50] = {.lex_state = 58, .reserved_word_set_id = 5},
  [51] = {.lex_state = 58, .reserved_word_set_id = 1},
  [52] = {.lex_state = 58, .reserved_word_set_id = 5},
  [53] = {.lex_state = 58, .reserved_word_set_id = 5},
  [54] = {.lex_state = 58, .reserved_word_set_id = 5},
  [55] = {.lex_state = 58, .reserved_word_set_id = 1},
  [56] = {.lex_state = 58, .reserved_word_set_id = 1},
  [57] = {.lex_state = 58, .reserved_word_set_id = 5},
  [58] = {.lex_state = 58, .reserved_word_set_id = 1},
  [59] = {.lex_state = 58, .reserved_word_set_id = 1},
  [60] = {.lex_state = 58, .reserved_word_set_id = 5},
  [61] = {.lex_state = 58, .reserved_word_set_id = 5},
  [62] = {.lex_state = 58, .reserved_word_set_id = 5},
  [63] = {.lex_state = 58, .reserved_word_set_id = 5},
  [64] = {.lex_state = 58, .reserved_word_set_id = 5},
  [65] = {.lex_state = 58, .reserved_word_set_id = 5},
  [66] = {.lex_state = 58, .reserved_word_set_id = 5},
  [67] = {.lex_state = 58, .reserved_word_set_id = 5},
  [68] = {.lex_state = 58, .reserved_word_set_id = 5},
  [69] = {.lex_state = 58, .reserved_word_set_id = 5},
  [70] = {.lex_state = 58, .reserved_word_set_id = 5},
  [71] = {.lex_state = 58, .reserved_word_set_id = 1},
  [72] = {.lex_state = 58, .reserved_word_set_id = 1},
  [73] = {.lex_state = 58, .reserved_word_set_id = 1},
  [74] = {.lex_state = 58, .reserved_word_set_id = 1},
  [75] = {.lex_state = 58, .reserved_word_set_id = 1},
  [76] = {.lex_state = 58, .reserved_word_set_id = 5},
  [77] = {.lex_state = 58, .reserved_word_set_id = 1},
  [78] = {.lex_state = 58, .reserved_word_set_id = 5},
  [79] = {.lex_state = 58, .reserved_word_set_id = 5},
  [80] = {.lex_state = 58, .reserved_word_set_id = 5},
  [81] = {.lex_state = 58, .reserved_word_set_id = 5},
  [82] = {.lex_state = 58, .reserved_word_set_id = 5},
  [83] = {.lex_state = 58, .reserved_word_set_id = 5},
  [84] = {.lex_state = 58, .reserved_word_set_id = 5},
  [85] = {.lex_state = 58, .reserved_word_set_id = 5},
  [86] = {.lex_state = 58, .reserved_word_set_id = 5},
  [87] = {.lex_state = 58, .reserved_word_set_id = 1},
  [88] = {.lex_state = 58, .reserved_word_set_id = 1},
  [89] = {.lex_state = 58, .reserved_word_set_id = 1},
  [90] = {.lex_state = 58, .reserved_word_set_id = 5},
  [91] = {.lex_state = 58, .reserved_word_set_id = 1},
  [92] = {.lex_state = 58, .reserved_word_set_id = 5},
  [93] = {.lex_state = 58, .reserved_word_set_id = 5},
  [94] = {.lex_state = 58, .reserved_word_set_id = 5},
  [95] = {.lex_state = 58, .reserved_word_set_id = 5},
  [96] = {.lex_state = 58, .reserved_word_set_id = 1},
  [97] = {.lex_state = 58, .reserved_word_set_id = 5},
  [98] = {.lex_state = 58, .reserved_word_set_id = 5},
  [99] = {.lex_state = 58, .reserved_word_set_id = 5},
  [100] = {.lex_state = 58, .reserved_word_set_id = 5},
  [101] = {.lex_state = 58, .reserved_word_set_id = 5},
  [102] = {.lex_state = 58, .reserved_word_set_id = 1},
  [103] = {.lex_state = 58, .reserved_word_set_id = 5},
  [104] = {.lex_state = 58, .reserved_word_set_id = 5},
  [105] = {.lex_state = 58, .reserved_word_set_id = 5},
  [106] = {.lex_state = 58, .reserved_word_set_id = 1},
  [107] = {.lex_state = 58, .reserved_word_set_id = 5},
  [108] = {.lex_state = 58, .reserved_word_set_id = 1},
  [109] = {.lex_state = 58, .reserved_word_set_id = 1},
  [110] = {.lex_state = 58, .reserved_word_set_id = 1},
  [111] = {.lex_state = 58, .reserved_word_set_id = 5},
  [112] = {.lex_state = 58, .reserved_word_set_id = 5},
  [113] = {.lex_state = 58, .reserved_word_set_id = 5},
  [114] = {.lex_state = 58, .reserved_word_set_id = 1},
  [115] = {.lex_state = 58, .reserved_word_set_id = 5},
  [116] = {.lex_state = 58, .reserved_word_set_id = 5},
  [117] = {.lex_state = 58, .reserved_word_set_id = 1},
  [118] = {.lex_state = 58, .reserved_word_set_id = 1},
  [119] = {.lex_state = 58, .reserved_word_set_id = 5},
  [120] = {.lex_state = 58, .reserved_word_set_id = 5},
  [121] = {.lex_state = 58, .reserved_word_set_id = 1},
  [122] = {.lex_state = 58, .reserved_word_set_id = 1},
  [123] = {.lex_state = 58, .reserved_word_set_id = 1},
  [124] = {.lex_state = 58, .reserved_word_set_id = 1},
  [125] = {.lex_state = 58, .reserved_word_set_id = 5},
  [126] = {.lex_state = 58, .reserved_word_set_id = 5},
  [127] = {.lex_state = 58, .reserved_word_set_id = 1},
  [128] = {.lex_state = 58, .reserved_word_set_id = 5},
  [129] = {.lex_state = 58, .reserved_word_set_id = 5},
  [130] = {.lex_state = 58, .reserved_word_set_id = 1},
  [131] = {.lex_state = 58, .reserved_word_set_id = 5},
  [132] = {.lex_state = 58, .reserved_word_set_id = 5},
  [133] = {.lex_state = 58, .reserved_word_set_id = 5},
  [134] = {.lex_state = 58, .reserved_word_set_id = 5},
  [135] = {.lex_state = 58, .reserved_word_set_id = 5},
  [136] = {.lex_state = 58, .reserved_word_set_id = 5},
  [137] = {.lex_state = 58, .reserved_word_set_id = 5},
  [138] = {.lex_state = 58, .reserved_word_set_id = 5},
  [139] = {.lex_state = 58, .reserved_word_set_id = 5},
  [140] = {.lex_state = 58, .reserved_word_set_id = 5},
  [141] = {.lex_state = 58, .reserved_word_set_id = 5},
  [142] = {.lex_state = 58, .reserved_word_set_id = 5},
  [143] = {.lex_state = 58, .reserved_word_set_id = 5},
  [144] = {.lex_state = 58, .reserved_word_set_id = 5},
  [145] = {.lex_state = 58, .reserved_word_set_id = 5},
  [146] = {.lex_state = 58, .reserved_word_set_id = 5},
  [147] = {.lex_state = 58, .reserved_word_set_id = 5},
  [148] = {.lex_state = 58, .reserved_word_set_id = 5},
  [149] = {.lex_state = 58, .reserved_word_set_id = 5},
  [150] = {.lex_state = 58, .reserved_word_set_id = 5},
  [151] = {.lex_state = 58, .reserved_word_set_id = 5},
  [152] = {.lex_state = 58, .reserved_word_set_id = 5},
  [153] = {.lex_state = 58, .reserved_word_set_id = 5},
  [154] = {.lex_state = 58, .reserved_word_set_id = 5},
  [155] = {.lex_state = 58, .reserved_word_set_id = 5},
  [156] = {.lex_state = 58, .reserved_word_set_id = 5},
  [157] = {.lex_state = 58, .reserved_word_set_id = 5},
  [158] = {.lex_state = 58, .reserved_word_set_id = 5},
  [159] = {.lex_state = 58, .reserved_word_set_id = 5},
  [160] = {.lex_state = 58, .reserved_word_set_id = 1},
  [161] = {.lex_state = 58, .reserved_word_set_id = 1},
  [162] = {.lex_state = 58, .reserved_word_set_id = 5},
  [163] = {.lex_state = 58, .reserved_word_set_id = 1},
  [164] = {.lex_state = 58, .reserved_word_set_id = 1},
  [165] = {.lex_state = 58, .reserved_word_set_id = 1},
  [166] = {.lex_state = 58, .reserved_word_set_id = 1},
  [167] = {.lex_state = 58, .reserved_word_set_id = 1},
  [168] = {.lex_state = 58, .reserved_word_set_id = 1},
  [169] = {.lex_state = 58, .reserved_word_set_id = 1},
  [170] = {.lex_state = 58, .reserved_word_set_id = 1},
  [171] = {.lex_state = 58, .reserved_word_set_id = 1},
  [172] = {.lex_state = 58, .reserved_word_set_id = 1},
  [173] = {.lex_state = 58, .reserved_word_set_id = 5},
  [174] = {.lex_state = 58, .reserved_word_set_id = 5},
  [175] = {.lex_state = 58, .reserved_word_set_id = 1},
  [176] = {.lex_state = 58, .reserved_word_set_id = 5},
  [177] = {.lex_state = 58, .reserved_word_set_id = 5},
  [178] = {.lex_state = 58, .reserved_word_set_id = 5},
  [179] = {.lex_state = 58, .reserved_word_set_id = 5},
  [180] = {.lex_state = 58, .reserved_word_set_id = 5},
  [181] = {.lex_state = 58, .reserved_word_set_id = 5},
  [182] = {.lex_state = 58, .reserved_word_set_id = 5},
  [183] = {.lex_state = 58, .reserved_word_set_id = 5},
  [184] = {.lex_state = 58, .reserved_word_set_id = 5},
  [185] = {.lex_state = 58, .reserved_word_set_id = 5},
  [186] = {.lex_state = 58, .reserved_word_set_id = 1},
  [187] = {.lex_state = 58, .reserved_word_set_id = 1},
  [188] = {.lex_state = 58, .reserved_word_set_id = 5},
  [189] = {.lex_state = 58, .reserved_word_set_id = 1},
  [190] = {.lex_state = 58, .reserved_word_set_id = 5},
  [191] = {.lex_state = 58, .reserved_word_set_id = 5},
  [192] = {.lex_state = 58, .reserved_word_set_id = 5},
  [193] = {.lex_state = 58, .reserved_word_set_id = 5},
  [194] = {.lex_state = 58, .reserved_word_set_id = 5},
  [195] = {.lex_state = 58, .reserved_word_set_id = 1},
  [196] = {.lex_state = 58, .reserved_word_set_id = 5},
  [197] = {.lex_state = 58, .reserved_word_set_id = 5},
  [198] = {.lex_state = 58, .reserved_word_set_id = 5},
  [199] = {.lex_state = 58, .reserved_word_set_id = 5},
  [200] = {.lex_state = 58, .reserved_word_set_id = 5},
  [201] = {.lex_state = 58, .reserved_word_set_id = 1},
  [202] = {.lex_state = 58, .reserved_word_set_id = 1},
  [203] = {.lex_state = 58, .reserved_word_set_id = 5},
  [204] = {.lex_state = 58, .reserved_word_set_id = 5},
  [205] = {.lex_state = 58, .reserved_word_set_id = 5},
  [206] = {.lex_state = 58, .reserved_word_set_id = 1},
  [207] = {.lex_state = 58, .reserved_word_set_id = 5},
  [208] = {.lex_state = 58, .reserved_word_set_id = 5},
  [209] = {.lex_state = 58, .reserved_word_set_id = 5},
  [210] = {.lex_state = 58, .reserved_word_set_id = 1},
  [211] = {.lex_state = 58, .reserved_word_set_id = 1},
  [212] = {.lex_state = 58, .reserved_word_set_id = 1},
  [213] = {.lex_state = 58, .reserved_word_set_id = 5},
  [214] = {.lex_state = 58, .reserved_word_set_id = 1},
  [215] = {.lex_state = 58, .reserved_word_set_id = 5},
  [216] = {.lex_state = 58, .reserved_word_set_id = 5},
  [217] = {.lex_state = 58, .reserved_word_set_id = 5},
  [218] = {.lex_state = 58, .reserved_word_set_id = 5},
  [219] = {.lex_state = 58, .reserved_word_set_id = 5},
  [220] = {.lex_state = 58, .reserved_word_set_id = 5},
  [221] = {.lex_state = 58, .reserved_word_set_id = 5},
  [222] = {.lex_state = 58, .reserved_word_set_id = 5},
  [223] = {.lex_state = 58, .reserved_word_set_id = 5},
  [224] = {.lex_state = 58, .reserved_word_set_id = 5},
  [225] = {.lex_state = 58, .reserved_word_set_id = 5},
  [226] = {.lex_state = 58, .reserved_word_set_id = 5},
  [227] = {.lex_state = 58, .reserved_word_set_id = 1},
  [228] = {.lex_state = 58, .reserved_word_set_id = 1},
  [229] = {.lex_state = 58, .reserved_word_set_id = 1},
  [230] = {.lex_state = 58, .reserved_word_set_id = 1},
  [231] = {.lex_state = 58, .reserved_word_set_id = 1},
  [232] = {.lex_state = 58, .reserved_word_set_id = 1},
  [233] = {.lex_state = 58, .reserved_word_set_id = 1},
  [234] = {.lex_state = 58, .reserved_word_set_id = 1},
  [235] = {.lex_state = 58, .reserved_word_set_id = 1},
  [236] = {.lex_state = 3, .reserved_word_set_id = 1},
  [237] = {.lex_state = 56},
  [238] = {.lex_state = 56},
  [239] = {.lex_state = 56},
  [240] = {.lex_state = 56},
  [241] = {.lex_state = 58, .reserved_word_set_id = 1},
  [242] = {.lex_state = 56},
  [243] = {.lex_state = 56},
  [244] = {.lex_state = 56},
  [245] = {.lex_state = 56},
  [246] = {.lex_state = 56},
  [247] = {.lex_state = 56},
  [248] = {.lex_state = 56},
  [249] = {.lex_state = 56},
  [250] = {.lex_state = 56},
  [251] = {.lex_state = 56},
  [252] = {.lex_state = 56},
  [253] = {.lex_state = 56},
  [254] = {.lex_state = 56},
  [255] = {.lex_state = 56},
  [256] = {.lex_state = 56},
  [257] = {.lex_state = 56},
  [258] = {.lex_state = 56},
  [259] = {.lex_state = 56},
  [260] = {.lex_state = 56},
  [261] = {.lex_state = 56},
  [262] = {.lex_state = 56},
  [263] = {.lex_state = 56},
  [264] = {.lex_state = 56},
  [265] = {.lex_state = 56},
  [266] = {.lex_state = 56},
  [267] = {.lex_state = 56},
  [268] = {.lex_state = 56},
  [269] = {.lex_state = 56},
  [270] = {.lex_state = 56},
  [271] = {.lex_state = 4},
  [272] = {.lex_state = 56},
  [273] = {.lex_state = 56},
  [274] = {.lex_state = 56},
  [275] = {.lex_state = 56},
  [276] = {.lex_state = 56},
  [277] = {.lex_state = 56},
  [278] = {.lex_state = 56},
  [279] = {.lex_state = 56},
  [280] = {.lex_state = 56},
  [281] = {.lex_state = 4},
  [282] = {.lex_state = 56},
  [283] = {.lex_state = 56},
  [284] = {.lex_state = 56},
  [285] = {.lex_state = 56},
  [286] = {.lex_state = 56},
  [287] = {.lex_state = 56},
  [288] = {.lex_state = 56},
  [289] = {.lex_state = 56},
  [290] = {.lex_state = 4},
  [291] = {.lex_state = 4},
  [292] = {.lex_state = 4},
  [293] = {.lex_state = 4},
  [294] = {.lex_state = 4},
  [295] = {.lex_state = 4},
  [296] = {.lex_state = 4},
  [297] = {.lex_state = 4},
  [298] = {.lex_state = 4},
  [299] = {.lex_state = 4},
  [300] = {.lex_state = 4},
  [301] = {.lex_state = 4},
  [302] = {.lex_state = 4},
  [303] = {.lex_state = 4},
  [304] = {.lex_state = 4},
  [305] = {.lex_state = 4},
  [306] = {.lex_state = 4},
  [307] = {.lex_state = 4},
  [308] = {.lex_state = 4},
  [309] = {.lex_state = 4},
  [310] = {.lex_state = 4},
  [311] = {.lex_state = 4},
  [312] = {.lex_state = 4},
  [313] = {.lex_state = 4},
  [314] = {.lex_state = 4},
  [315] = {.lex_state = 4},
  [316] = {.lex_state = 4},
  [317] = {.lex_state = 4},
  [318] = {.lex_state = 4},
  [319] = {.lex_state = 4},
  [320] = {.lex_state = 4},
  [321] = {.lex_state = 4},
  [322] = {.lex_state = 4},
  [323] = {.lex_state = 4},
  [324] = {.lex_state = 4},
  [325] = {.lex_state = 4},
  [326] = {.lex_state = 4},
  [327] = {.lex_state = 4},
  [328] = {.lex_state = 4},
  [329] = {.lex_state = 4},
  [330] = {.lex_state = 4},
  [331] = {.lex_state = 4},
  [332] = {.lex_state = 4},
  [333] = {.lex_state = 4},
  [334] = {.lex_state = 4},
  [335] = {.lex_state = 4},
  [336] = {.lex_state = 4},
  [337] = {.lex_state = 4},
  [338] = {.lex_state = 4},
  [339] = {.lex_state = 5},
  [340] = {.lex_state = 5},
  [341] = {.lex_state = 5},
  [342] = {.lex_state = 5},
  [343] = {.lex_state = 5},
  [344] = {.lex_state = 5},
  [345] = {.lex_state = 5},
  [346] = {.lex_state = 5},
  [347] = {.lex_state = 5},
  [348] = {.lex_state = 5},
  [349] = {.lex_state = 5},
  [350] = {.lex_state = 5},
  [351] = {.lex_state = 55, .reserved_word_set_id = 6},
  [352] = {.lex_state = 5},
  [353] = {.lex_state = 5},
  [354] = {.lex_state = 5},
  [355] = {.lex_state = 5},
  [356] = {.lex_state = 5},
  [357] = {.lex_state = 5},
  [358] = {.lex_state = 5},
  [359] = {.lex_state = 5},
  [360] = {.lex_state = 5},
  [361] = {.lex_state = 5},
  [362] = {.lex_state = 5},
  [363] = {.lex_state = 5},
  [364] = {.lex_state = 55, .reserved_word_set_id = 6},
  [365] = {.lex_state = 5},
  [366] = {.lex_state = 5},
  [367] = {.lex_state = 5},
  [368] = {.lex_state = 56, .reserved_word_set_id = 1},
  [369] = {.lex_state = 5},
  [370] = {.lex_state = 5},
  [371] = {.lex_state = 5},
  [372] = {.lex_state = 5},
  [373] = {.lex_state = 5},
  [374] = {.lex_state = 5},
  [375] = {.lex_state = 5},
  [376] = {.lex_state = 5},
  [377] = {.lex_state = 5},
  [378] = {.lex_state = 5},
  [379] = {.lex_state = 5},
  [380] = {.lex_state = 5},
  [381] = {.lex_state = 5},
  [382] = {.lex_state = 5},
  [383] = {.lex_state = 5},
  [384] = {.lex_state = 5},
  [385] = {.lex_state = 5},
  [386] = {.lex_state = 5},
  [387] = {.lex_state = 5},
  [388] = {.lex_state = 5},
  [389] = {.lex_state = 5},
  [390] = {.lex_state = 6},
  [391] = {.lex_state = 6},
  [392] = {.lex_state = 55, .reserved_word_set_id = 6},
  [393] = {.lex_state = 55, .reserved_word_set_id = 6},
  [394] = {.lex_state = 6},
  [395] = {.lex_state = 6},
  [396] = {.lex_state = 6},
  [397] = {.lex_state = 6},
  [398] = {.lex_state = 6},
  [399] = {.lex_state = 6},
  [400] = {.lex_state = 1},
  [401] = {.lex_state = 6},
  [402] = {.lex_state = 6},
  [403] = {.lex_state = 6},
  [404] = {.lex_state = 6},
  [405] = {.lex_state = 6},
  [406] = {.lex_state = 6},
  [407] = {.lex_state = 6},
  [408] = {.lex_state = 6},
  [409] = {.lex_state = 6},
  [410] = {.lex_state = 6},
  [411] = {.lex_state = 6},
  [412] = {.lex_state = 6},
  [413] = {.lex_state = 6},
  [414] = {.lex_state = 6},
  [415] = {.lex_state = 6},
  [416] = {.lex_state = 6},
  [417] = {.lex_state = 6},
  [418] = {.lex_state = 6},
  [419] = {.lex_state = 6},
  [420] = {.lex_state = 6},
  [421] = {.lex_state = 6},
  [422] = {.lex_state = 6},
  [423] = {.lex_state = 6},
  [424] = {.lex_state = 6},
  [425] = {.lex_state = 6},
  [426] = {.lex_state = 6},
  [427] = {.lex_state = 6},
  [428] = {.lex_state = 6},
  [429] = {.lex_state = 6},
  [430] = {.lex_state = 6},
  [431] = {.lex_state = 1},
  [432] = {.lex_state = 6},
  [433] = {.lex_state = 6},
  [434] = {.lex_state = 6},
  [435] = {.lex_state = 6},
  [436] = {.lex_state = 6},
  [437] = {.lex_state = 1},
  [438] = {.lex_state = 6},
  [439] = {.lex_state = 0, .reserved_word_set_id = 5},
  [440] = {.lex_state = 1},
  [441] = {.lex_state = 6},
  [442] = {.lex_state = 6},
  [443] = {.lex_state = 6},
  [444] = {.lex_state = 1},
  [445] = {.lex_state = 1},
  [446] = {.lex_state = 1},
  [447] = {.lex_state = 1},
  [448] = {.lex_state = 1},
  [449] = {.lex_state = 0, .reserved_word_set_id = 5},
  [450] = {.lex_state = 1},
  [451] = {.lex_state = 3},
  [452] = {.lex_state = 0, .reserved_word_set_id = 5},
  [453] = {.lex_state = 3, .reserved_word_set_id = 1},
  [454] = {.lex_state = 55, .reserved_word_set_id = 1},
  [455] = {.lex_state = 1},
  [456] = {.lex_state = 1},
  [457] = {.lex_state = 1},
  [458] = {.lex_state = 3},
  [459] = {.lex_state = 55, .reserved_word_set_id = 1},
  [460] = {.lex_state = 55, .reserved_word_set_id = 1},
  [461] = {.lex_state = 1},
  [462] = {.lex_state = 1},
  [463] = {.lex_state = 1},
  [464] = {.lex_state = 1},
  [465] = {.lex_state = 1},
  [466] = {.lex_state = 1},
  [467] = {.lex_state = 1},
  [468] = {.lex_state = 1},
  [469] = {.lex_state = 1},
  [470] = {.lex_state = 1},
  [471] = {.lex_state = 1},
  [472] = {.lex_state = 1},
  [473] = {.lex_state = 1},
  [474] = {.lex_state = 1},
  [475] = {.lex_state = 1},
  [476] = {.lex_state = 1},
  [477] = {.lex_state = 1},
  [478] = {.lex_state = 1},
  [479] = {.lex_state = 1},
  [480] = {.lex_state = 1},
  [481] = {.lex_state = 1},
  [482] = {.lex_state = 1},
  [483] = {.lex_state = 1},
  [484] = {.lex_state = 1},
  [485] = {.lex_state = 1},
  [486] = {.lex_state = 1},
  [487] = {.lex_state = 1},
  [488] = {.lex_state = 6},
  [489] = {.lex_state = 1},
  [490] = {.lex_state = 1},
  [491] = {.lex_state = 1},
  [492] = {.lex_state = 55, .reserved_word_set_id = 1},
  [493] = {.lex_state = 55, .reserved_word_set_id = 1},
  [494] = {.lex_state = 1},
  [495] = {.lex_state = 1},
  [496] = {.lex_state = 55, .reserved_word_set_id = 1},
  [497] = {.lex_state = 1},
  [498] = {.lex_state = 0, .reserved_word_set_id = 5},
  [499] = {.lex_state = 3},
  [500] = {.lex_state = 55, .reserved_word_set_id = 1},
  [501] = {.lex_state = 3},
  [502] = {.lex_state = 55, .reserved_word_set_id = 1},
  [503] = {.lex_state = 1},
  [504] = {.lex_state = 1},
  [505] = {.lex_state = 3},
  [506] = {.lex_state = 3},
  [507] = {.lex_state = 3},
  [508] = {.lex_state = 3},
  [509] = {.lex_state = 1},
  [510] = {.lex_state = 0, .reserved_word_set_id = 5},
  [511] = {.lex_state = 0, .reserved_word_set_id = 5},
  [512] = {.lex_state = 1},
  [513] = {.lex_state = 1},
  [514] = {.lex_state = 1},
  [515] = {.lex_state = 6},
  [516] = {.lex_state = 6},
  [517] = {.lex_state = 3},
  [518] = {.lex_state = 6},
  [519] = {.lex_state = 0, .reserved_word_set_id = 5},
  [520] = {.lex_state = 0, .reserved_word_set_id = 5},
  [521] = {.lex_state = 0, .reserved_word_set_id = 5},
  [522] = {.lex_state = 6},
  [523] = {.lex_state = 0, .reserved_word_set_id = 5},
  [524] = {.lex_state = 6},
  [525] = {.lex_state = 6},
  [526] = {.lex_state = 0, .reserved_word_set_id = 5},
  [527] = {.lex_state = 0, .reserved_word_set_id = 5},
  [528] = {.lex_state = 6},
  [529] = {.lex_state = 55, .reserved_word_set_id = 1},
  [530] = {.lex_state = 6},
  [531] = {.lex_state = 6},
  [532] = {.lex_state = 6},
  [533] = {.lex_state = 3},
  [534] = {.lex_state = 6},
  [535] = {.lex_state = 0, .reserved_word_set_id = 1},
  [536] = {.lex_state = 6},
  [537] = {.lex_state = 6},
  [538] = {.lex_state = 6},
  [539] = {.lex_state = 6},
  [540] = {.lex_state = 0, .reserved_word_set_id = 5},
  [541] = {.lex_state = 0, .reserved_word_set_id = 5},
  [542] = {.lex_state = 6},
  [543] = {.lex_state = 0, .reserved_word_set_id = 1},
  [544] = {.lex_state = 3},
  [545] = {.lex_state = 6},
  [546] = {.lex_state = 3},
  [547] = {.lex_state = 3},
  [548] = {.lex_state = 3},
  [549] = {.lex_state = 3},
  [550] = {.lex_state = 3},
  [551] = {.lex_state = 3},
  [552] = {.lex_state = 3},
  [553] = {.lex_state = 3},
  [554] = {.lex_state = 3},
  [555] = {.lex_state = 3},
  [556] = {.lex_state = 3},
  [557] = {.lex_state = 3},
  [558] = {.lex_state = 3},
  [559] = {.lex_state = 3},
  [560] = {.lex_state = 3},
  [561] = {.lex_state = 3},
  [562] = {.lex_state = 3},
  [563] = {.lex_state = 3},
  [564] = {.lex_state = 3},
  [565] = {.lex_state = 3},
  [566] = {.lex_state = 3},
  [567] = {.lex_state = 3},
  [568] = {.lex_state = 3},
  [569] = {.lex_state = 0, .reserved_word_set_id = 5},
  [570] = {.lex_state = 3},
  [571] = {.lex_state = 3},
  [572] = {.lex_state = 3},
  [573] = {.lex_state = 3},
  [574] = {.lex_state = 3},
  [575] = {.lex_state = 3},
  [576] = {.lex_state = 3},
  [577] = {.lex_state = 3},
  [578] = {.lex_state = 3},
  [579] = {.lex_state = 3},
  [580] = {.lex_state = 3},
  [581] = {.lex_state = 3},
  [582] = {.lex_state = 3},
  [583] = {.lex_state = 0, .reserved_word_set_id = 5},
  [584] = {.lex_state = 0, .reserved_word_set_id = 5},
  [585] = {.lex_state = 0, .reserved_word_set_id = 5},
  [586] = {.lex_state = 6},
  [587] = {.lex_state = 6},
  [588] = {.lex_state = 6},
  [589] = {.lex_state = 6},
  [590] = {.lex_state = 0, .reserved_word_set_id = 5},
  [591] = {.lex_state = 0, .reserved_word_set_id = 1},
  [592] = {.lex_state = 0, .reserved_word_set_id = 5},
  [593] = {.lex_state = 6},
  [594] = {.lex_state = 0, .reserved_word_set_id = 5},
  [595] = {.lex_state = 0, .reserved_word_set_id = 5},
  [596] = {.lex_state = 0, .reserved_word_set_id = 5},
  [597] = {.lex_state = 6},
  [598] = {.lex_state = 3},
  [599] = {.lex_state = 6},
  [600] = {.lex_state = 6},
  [601] = {.lex_state = 0, .reserved_word_set_id = 5},
  [602] = {.lex_state = 6},
  [603] = {.lex_state = 0, .reserved_word_set_id = 5},
  [604] = {.lex_state = 6},
  [605] = {.lex_state = 6},
  [606] = {.lex_state = 0, .reserved_word_set_id = 1},
  [607] = {.lex_state = 0, .reserved_word_set_id = 5},
  [608] = {.lex_state = 6},
  [609] = {.lex_state = 6},
  [610] = {.lex_state = 0, .reserved_word_set_id = 5},
  [611] = {.lex_state = 6},
  [612] = {.lex_state = 0, .reserved_word_set_id = 1},
  [613] = {.lex_state = 6},
  [614] = {.lex_state = 0, .reserved_word_set_id = 1},
  [615] = {.lex_state = 6},
  [616] = {.lex_state = 0, .reserved_word_set_id = 5},
  [617] = {.lex_state = 6},
  [618] = {.lex_state = 6},
  [619] = {.lex_state = 0, .reserved_word_set_id = 5},
  [620] = {.lex_state = 6},
  [621] = {.lex_state = 0, .reserved_word_set_id = 5},
  [622] = {.lex_state = 6},
  [623] = {.lex_state = 0, .reserved_word_set_id = 5},
  [624] = {.lex_state = 6},
  [625] = {.lex_state = 6},
  [626] = {.lex_state = 0, .reserved_word_set_id = 1},
  [627] = {.lex_state = 6},
  [628] = {.lex_state = 6},
  [629] = {.lex_state = 0, .reserved_word_set_id = 1},
  [630] = {.lex_state = 6},
  [631] = {.lex_state = 6},
  [632] = {.lex_state = 6},
  [633] = {.lex_state = 6},
  [634] = {.lex_state = 6},
  [635] = {.lex_state = 0, .reserved_word_set_id = 1},
  [636] = {.lex_state = 6},
  [637] = {.lex_state = 6},
  [638] = {.lex_state = 6},
  [639] = {.lex_state = 6},
  [640] = {.lex_state = 0, .reserved_word_set_id = 5},
  [641] = {.lex_state = 0, .reserved_word_set_id = 1},
  [642] = {.lex_state = 0, .reserved_word_set_id = 5},
  [643] = {.lex_state = 6},
  [644] = {.lex_state = 0, .reserved_word_set_id = 1},
  [645] = {.lex_state = 0, .reserved_word_set_id = 1},
  [646] = {.lex_state = 6},
  [647] = {.lex_state = 0, .reserved_word_set_id = 5},
  [648] = {.lex_state = 6},
  [649] = {.lex_state = 6},
  [650] = {.lex_state = 6},
  [651] = {.lex_state = 6},
  [652] = {.lex_state = 6},
  [653] = {.lex_state = 0, .reserved_word_set_id = 5},
  [654] = {.lex_state = 6},
  [655] = {.lex_state = 6},
  [656] = {.lex_state = 0, .reserved_word_set_id = 5},
  [657] = {.lex_state = 0, .reserved_word_set_id = 5},
  [658] = {.lex_state = 0, .reserved_word_set_id = 5},
  [659] = {.lex_state = 0, .reserved_word_set_id = 5},
  [660] = {.lex_state = 0, .reserved_word_set_id = 5},
  [661] = {.lex_state = 0, .reserved_word_set_id = 1},
  [662] = {.lex_state = 6},
  [663] = {.lex_state = 6},
  [664] = {.lex_state = 6},
  [665] = {.lex_state = 6},
  [666] = {.lex_state = 0, .reserved_word_set_id = 1},
  [667] = {.lex_state = 6},
  [668] = {.lex_state = 0, .reserved_word_set_id = 1},
  [669] = {.lex_state = 6},
  [670] = {.lex_state = 6},
  [671] = {.lex_state = 0, .reserved_word_set_id = 5},
  [672] = {.lex_state = 6},
  [673] = {.lex_state = 6},
  [674] = {.lex_state = 0, .reserved_word_set_id = 1},
  [675] = {.lex_state = 0, .reserved_word_set_id = 1},
  [676] = {.lex_state = 6},
  [677] = {.lex_state = 6},
  [678] = {.lex_state = 6},
  [679] = {.lex_state = 6},
  [680] = {.lex_state = 0, .reserved_word_set_id = 1},
  [681] = {.lex_state = 6},
  [682] = {.lex_state = 0, .reserved_word_set_id = 5},
  [683] = {.lex_state = 0, .reserved_word_set_id = 5},
  [684] = {.lex_state = 6},
  [685] = {.lex_state = 0, .reserved_word_set_id = 1},
  [686] = {.lex_state = 0, .reserved_word_set_id = 5},
  [687] = {.lex_state = 6},
  [688] = {.lex_state = 0, .reserved_word_set_id = 1},
  [689] = {.lex_state = 6},
  [690] = {.lex_state = 6},
  [691] = {.lex_state = 6},
  [692] = {.lex_state = 0, .reserved_word_set_id = 1},
  [693] = {.lex_state = 6},
  [694] = {.lex_state = 0, .reserved_word_set_id = 5},
  [695] = {.lex_state = 0, .reserved_word_set_id = 5},
  [696] = {.lex_state = 6},
  [697] = {.lex_state = 0, .reserved_word_set_id = 5},
  [698] = {.lex_state = 0, .reserved_word_set_id = 5},
  [699] = {.lex_state = 0, .reserved_word_set_id = 5},
  [700] = {.lex_state = 6},
  [701] = {.lex_state = 0, .reserved_word_set_id = 5},
  [702] = {.lex_state = 0, .reserved_word_set_id = 5},
  [703] = {.lex_state = 0, .reserved_word_set_id = 5},
  [704] = {.lex_state = 0, .reserved_word_set_id = 1},
  [705] = {.lex_state = 0, .reserved_word_set_id = 5},
  [706] = {.lex_state = 0, .reserved_word_set_id = 1},
  [707] = {.lex_state = 0, .reserved_word_set_id = 1},
  [708] = {.lex_state = 0, .reserved_word_set_id = 5},
  [709] = {.lex_state = 0, .reserved_word_set_id = 5},
  [710] = {.lex_state = 0, .reserved_word_set_id = 1},
  [711] = {.lex_state = 0, .reserved_word_set_id = 5},
  [712] = {.lex_state = 0, .reserved_word_set_id = 1},
  [713] = {.lex_state = 0, .reserved_word_set_id = 5},
  [714] = {.lex_state = 0, .reserved_word_set_id = 1},
  [715] = {.lex_state = 0, .reserved_word_set_id = 1},
  [716] = {.lex_state = 0, .reserved_word_set_id = 5},
  [717] = {.lex_state = 0, .reserved_word_set_id = 1},
  [718] = {.lex_state = 0, .reserved_word_set_id = 5},
  [719] = {.lex_state = 0, .reserved_word_set_id = 5},
  [720] = {.lex_state = 0, .reserved_word_set_id = 5},
  [721] = {.lex_state = 0, .reserved_word_set_id = 5},
  [722] = {.lex_state = 0, .reserved_word_set_id = 5},
  [723] = {.lex_state = 0, .reserved_word_set_id = 5},
  [724] = {.lex_state = 0, .reserved_word_set_id = 1},
  [725] = {.lex_state = 0, .reserved_word_set_id = 1},
  [726] = {.lex_state = 0, .reserved_word_set_id = 5},
  [727] = {.lex_state = 6},
  [728] = {.lex_state = 0, .reserved_word_set_id = 5},
  [729] = {.lex_state = 0, .reserved_word_set_id = 5},
  [730] = {.lex_state = 0, .reserved_word_set_id = 5},
  [731] = {.lex_state = 0, .reserved_word_set_id = 5},
  [732] = {.lex_state = 0, .reserved_word_set_id = 5},
  [733] = {.lex_state = 0, .reserved_word_set_id = 5},
  [734] = {.lex_state = 0, .reserved_word_set_id = 5},
  [735] = {.lex_state = 0, .reserved_word_set_id = 5},
  [736] = {.lex_state = 0, .reserved_word_set_id = 5},
  [737] = {.lex_state = 0, .reserved_word_set_id = 5},
  [738] = {.lex_state = 0, .reserved_word_set_id = 1},
  [739] = {.lex_state = 0, .reserved_word_set_id = 5},
  [740] = {.lex_state = 0, .reserved_word_set_id = 5},
  [741] = {.lex_state = 0, .reserved_word_set_id = 5},
  [742] = {.lex_state = 0, .reserved_word_set_id = 1},
  [743] = {.lex_state = 0, .reserved_word_set_id = 1},
  [744] = {.lex_state = 0, .reserved_word_set_id = 1},
  [745] = {.lex_state = 0, .reserved_word_set_id = 5},
  [746] = {.lex_state = 0, .reserved_word_set_id = 1},
  [747] = {.lex_state = 0, .reserved_word_set_id = 1},
  [748] = {.lex_state = 0, .reserved_word_set_id = 5},
  [749] = {.lex_state = 0, .reserved_word_set_id = 5},
  [750] = {.lex_state = 0, .reserved_word_set_id = 5},
  [751] = {.lex_state = 0, .reserved_word_set_id = 5},
  [752] = {.lex_state = 0, .reserved_word_set_id = 1},
  [753] = {.lex_state = 0, .reserved_word_set_id = 1},
  [754] = {.lex_state = 0, .reserved_word_set_id = 1},
  [755] = {.lex_state = 0, .reserved_word_set_id = 5},
  [756] = {.lex_state = 0, .reserved_word_set_id = 5},
  [757] = {.lex_state = 0, .reserved_word_set_id = 5},
  [758] = {.lex_state = 0, .reserved_word_set_id = 1},
  [759] = {.lex_state = 0, .reserved_word_set_id = 5},
  [760] = {.lex_state = 0, .reserved_word_set_id = 1},
  [761] = {.lex_state = 0, .reserved_word_set_id = 5},
  [762] = {.lex_state = 0, .reserved_word_set_id = 5},
  [763] = {.lex_state = 0, .reserved_word_set_id = 5},
  [764] = {.lex_state = 0, .reserved_word_set_id = 5},
  [765] = {.lex_state = 0, .reserved_word_set_id = 5},
  [766] = {.lex_state = 0, .reserved_word_set_id = 1},
  [767] = {.lex_state = 0, .reserved_word_set_id = 5},
  [768] = {.lex_state = 0, .reserved_word_set_id = 5},
  [769] = {.lex_state = 0, .reserved_word_set_id = 5},
  [770] = {.lex_state = 0, .reserved_word_set_id = 5},
  [771] = {.lex_state = 0, .reserved_word_set_id = 5},
  [772] = {.lex_state = 0, .reserved_word_set_id = 5},
  [773] = {.lex_state = 0, .reserved_word_set_id = 5},
  [774] = {.lex_state = 0, .reserved_word_set_id = 5},
  [775] = {.lex_state = 0, .reserved_word_set_id = 5},
  [776] = {.lex_state = 0, .reserved_word_set_id = 1},
  [777] = {.lex_state = 0, .reserved_word_set_id = 1},
  [778] = {.lex_state = 0, .reserved_word_set_id = 1},
  [779] = {.lex_state = 0, .reserved_word_set_id = 1},
  [780] = {.lex_state = 0, .reserved_word_set_id = 1},
  [781] = {.lex_state = 0, .reserved_word_set_id = 1},
  [782] = {.lex_state = 0, .reserved_word_set_id = 1},
  [783] = {.lex_state = 55, .reserved_word_set_id = 6},
  [784] = {.lex_state = 55, .reserved_word_set_id = 6},
  [785] = {.lex_state = 55, .reserved_word_set_id = 6},
  [786] = {.lex_state = 55, .reserved_word_set_id = 6},
  [787] = {.lex_state = 55, .reserved_word_set_id = 6},
  [788] = {.lex_state = 55, .reserved_word_set_id = 6},
  [789] = {.lex_state = 0, .reserved_word_set_id = 5},
  [790] = {.lex_state = 0, .reserved_word_set_id = 5},
  [791] = {.lex_state = 0, .reserved_word_set_id = 5},
  [792] = {.lex_state = 0, .reserved_word_set_id = 5},
  [793] = {.lex_state = 55, .reserved_word_set_id = 5},
  [794] = {.lex_state = 55, .reserved_word_set_id = 5},
  [795] = {.lex_state = 55, .reserved_word_set_id = 5},
  [796] = {.lex_state = 55, .reserved_word_set_id = 5},
  [797] = {.lex_state = 56},
  [798] = {.lex_state = 0},
  [799] = {.lex_state = 56},
  [800] = {.lex_state = 0},
  [801] = {.lex_state = 55},
  [802] = {.lex_state = 0, .reserved_word_set_id = 5},
  [803] = {.lex_state = 55},
  [804] = {.lex_state = 55},
  [805] = {.lex_state = 55},
  [806] = {.lex_state = 55},
  [807] = {.lex_state = 55},
  [808] = {.lex_state = 55},
  [809] = {.lex_state = 55},
  [810] = {.lex_state = 55},
  [811] = {.lex_state = 55},
  [812] = {.lex_state = 55},
  [813] = {.lex_state = 55},
  [814] = {.lex_state = 55},
  [815] = {.lex_state = 55},
  [816] = {.lex_state = 55},
  [817] = {.lex_state = 55},
  [818] = {.lex_state = 55},
  [819] = {.lex_state = 55},
  [820] = {.lex_state = 0, .reserved_word_set_id = 5},
  [821] = {.lex_state = 55},
  [822] = {.lex_state = 0},
  [823] = {.lex_state = 0},
  [824] = {.lex_state = 55},
  [825] = {.lex_state = 0},
  [826] = {.lex_state = 55},
  [827] = {.lex_state = 55},
  [828] = {.lex_state = 0, .reserved_word_set_id = 1},
  [829] = {.lex_state = 55},
  [830] = {.lex_state = 55},
  [831] = {.lex_state = 55},
  [832] = {.lex_state = 55},
  [833] = {.lex_state = 55},
  [834] = {.lex_state = 55},
  [835] = {.lex_state = 0},
  [836] = {.lex_state = 55},
  [837] = {.lex_state = 55},
  [838] = {.lex_state = 0},
  [839] = {.lex_state = 0, .reserved_word_set_id = 5},
  [840] = {.lex_state = 0, .reserved_word_set_id = 1},
  [841] = {.lex_state = 0, .reserved_word_set_id = 5},
  [842] = {.lex_state = 0, .reserved_word_set_id = 5},
  [843] = {.lex_state = 3, .reserved_word_set_id = 1},
  [844] = {.lex_state = 3},
  [845] = {.lex_state = 56},
  [846] = {.lex_state = 3, .reserved_word_set_id = 1},
  [847] = {.lex_state = 0},
  [848] = {.lex_state = 55},
  [849] = {.lex_state = 3, .reserved_word_set_id = 1},
  [850] = {.lex_state = 3, .reserved_word_set_id = 1},
  [851] = {.lex_state = 3, .reserved_word_set_id = 1},
  [852] = {.lex_state = 55},
  [853] = {.lex_state = 3, .reserved_word_set_id = 1},
  [854] = {.lex_state = 3, .reserved_word_set_id = 1},
  [855] = {.lex_state = 3},
  [856] = {.lex_state = 56},
  [857] = {.lex_state = 3, .reserved_word_set_id = 1},
  [858] = {.lex_state = 0},
  [859] = {.lex_state = 0},
  [860] = {.lex_state = 0},
  [861] = {.lex_state = 0},
  [862] = {.lex_state = 0},
  [863] = {.lex_state = 0},
  [864] = {.lex_state = 0},
  [865] = {.lex_state = 3, .reserved_word_set_id = 1},
  [866] = {.lex_state = 0},
  [867] = {.lex_state = 0},
  [868] = {.lex_state = 0},
  [869] = {.lex_state = 0},
  [870] = {.lex_state = 0},
  [871] = {.lex_state = 0},
  [872] = {.lex_state = 0},
  [873] = {.lex_state = 0},
  [874] = {.lex_state = 0},
  [875] = {.lex_state = 0},
  [876] = {.lex_state = 0},
  [877] = {.lex_state = 0},
  [878] = {.lex_state = 0},
  [879] = {.lex_state = 0},
  [880] = {.lex_state = 0},
  [881] = {.lex_state = 0},
  [882] = {.lex_state = 0},
  [883] = {.lex_state = 0},
  [884] = {.lex_state = 0},
  [885] = {.lex_state = 0},
  [886] = {.lex_state = 0},
  [887] = {.lex_state = 0},
  [888] = {.lex_state = 0},
  [889] = {.lex_state = 55},
  [890] = {.lex_state = 0, .reserved_word_set_id = 1},
  [891] = {.lex_state = 0, .reserved_word_set_id = 1},
  [892] = {.lex_state = 0},
  [893] = {.lex_state = 55},
  [894] = {.lex_state = 55},
  [895] = {.lex_state = 0},
  [896] = {.lex_state = 55},
  [897] = {.lex_state = 55},
  [898] = {.lex_state = 56},
  [899] = {.lex_state = 0, .reserved_word_set_id = 1},
  [900] = {.lex_state = 56},
  [901] = {.lex_state = 0},
  [902] = {.lex_state = 0, .reserved_word_set_id = 1},
  [903] = {.lex_state = 0, .reserved_word_set_id = 1},
  [904] = {.lex_state = 0, .reserved_word_set_id = 1},
  [905] = {.lex_state = 55},
  [906] = {.lex_state = 0, .reserved_word_set_id = 1},
  [907] = {.lex_state = 0, .reserved_word_set_id = 1},
  [908] = {.lex_state = 55},
  [909] = {.lex_state = 0, .reserved_word_set_id = 1},
  [910] = {.lex_state = 55},
  [911] = {.lex_state = 55, .reserved_word_set_id = 7},
  [912] = {.lex_state = 0, .reserved_word_set_id = 1},
  [913] = {.lex_state = 55, .reserved_word_set_id = 7},
  [914] = {.lex_state = 55},
  [915] = {.lex_state = 55},
  [916] = {.lex_state = 55},
  [917] = {.lex_state = 55},
  [918] = {.lex_state = 55},
  [919] = {.lex_state = 55},
  [920] = {.lex_state = 55},
  [921] = {.lex_state = 55},
  [922] = {.lex_state = 55},
  [923] = {.lex_state = 55},
  [924] = {.lex_state = 55},
  [925] = {.lex_state = 55},
  [926] = {.lex_state = 55},
  [927] = {.lex_state = 55},
  [928] = {.lex_state = 55},
  [929] = {.lex_state = 55},
  [930] = {.lex_state = 55},
  [931] = {.lex_state = 55},
  [932] = {.lex_state = 55},
  [933] = {.lex_state = 0},
  [934] = {.lex_state = 55},
  [935] = {.lex_state = 55},
  [936] = {.lex_state = 55},
  [937] = {.lex_state = 55},
  [938] = {.lex_state = 55},
  [939] = {.lex_state = 55},
  [940] = {.lex_state = 0},
  [941] = {.lex_state = 55},
  [942] = {.lex_state = 55},
  [943] = {.lex_state = 0},
  [944] = {.lex_state = 55},
  [945] = {.lex_state = 55},
  [946] = {.lex_state = 55},
  [947] = {.lex_state = 55},
  [948] = {.lex_state = 55},
  [949] = {.lex_state = 55},
  [950] = {.lex_state = 55},
  [951] = {.lex_state = 55},
  [952] = {.lex_state = 55},
  [953] = {.lex_state = 55},
  [954] = {.lex_state = 55},
  [955] = {.lex_state = 55},
  [956] = {.lex_state = 55},
  [957] = {.lex_state = 55},
  [958] = {.lex_state = 55},
  [959] = {.lex_state = 0},
  [960] = {.lex_state = 0},
  [961] = {.lex_state = 0},
  [962] = {.lex_state = 55},
  [963] = {.lex_state = 55},
  [964] = {.lex_state = 55},
  [965] = {.lex_state = 55},
  [966] = {.lex_state = 55},
  [967] = {.lex_state = 55},
  [968] = {.lex_state = 55},
  [969] = {.lex_state = 55},
  [970] = {.lex_state = 0},
  [971] = {.lex_state = 55},
  [972] = {.lex_state = 55},
  [973] = {.lex_state = 0},
  [974] = {.lex_state = 55},
  [975] = {.lex_state = 55},
  [976] = {.lex_state = 55},
  [977] = {.lex_state = 0},
  [978] = {.lex_state = 55},
  [979] = {.lex_state = 55},
  [980] = {.lex_state = 55},
  [981] = {.lex_state = 55},
  [982] = {.lex_state = 55},
  [983] = {.lex_state = 3},
  [984] = {.lex_state = 55},
  [985] = {.lex_state = 55},
  [986] = {.lex_state = 55},
  [987] = {.lex_state = 55},
  [988] = {.lex_state = 55},
  [989] = {.lex_state = 55},
  [990] = {.lex_state = 55},
  [991] = {.lex_state = 55},
  [992] = {.lex_state = 0},
  [993] = {.lex_state = 55},
  [994] = {.lex_state = 55},
  [995] = {.lex_state = 55},
  [996] = {.lex_state = 55},
  [997] = {.lex_state = 55},
  [998] = {.lex_state = 55},
  [999] = {.lex_state = 0},
  [1000] = {.lex_state = 55},
  [1001] = {.lex_state = 55},
  [1002] = {.lex_state = 0},
  [1003] = {.lex_state = 0},
  [1004] = {.lex_state = 55},
  [1005] = {.lex_state = 55},
  [1006] = {.lex_state = 0},
  [1007] = {.lex_state = 0},
  [1008] = {.lex_state = 55},
  [1009] = {.lex_state = 55},
  [1010] = {.lex_state = 55},
  [1011] = {.lex_state = 0},
  [1012] = {.lex_state = 55},
  [1013] = {.lex_state = 0},
  [1014] = {.lex_state = 55},
  [1015] = {.lex_state = 55},
  [1016] = {.lex_state = 55},
  [1017] = {.lex_state = 0},
  [1018] = {.lex_state = 0},
  [1019] = {.lex_state = 3},
  [1020] = {.lex_state = 55},
  [1021] = {.lex_state = 55},
  [1022] = {.lex_state = 55},
  [1023] = {.lex_state = 55},
  [1024] = {.lex_state = 55},
  [1025] = {.lex_state = 55},
  [1026] = {.lex_state = 55},
  [1027] = {.lex_state = 55},
  [1028] = {.lex_state = 55},
  [1029] = {.lex_state = 55},
  [1030] = {.lex_state = 55},
  [1031] = {.lex_state = 55},
  [1032] = {.lex_state = 55},
  [1033] = {.lex_state = 55},
  [1034] = {.lex_state = 0},
  [1035] = {.lex_state = 55},
  [1036] = {.lex_state = 55},
  [1037] = {.lex_state = 0},
  [1038] = {.lex_state = 55},
  [1039] = {.lex_state = 0},
  [1040] = {.lex_state = 55},
  [1041] = {.lex_state = 55},
  [1042] = {.lex_state = 55},
  [1043] = {.lex_state = 55},
  [1044] = {.lex_state = 55},
  [1045] = {.lex_state = 55},
  [1046] = {.lex_state = 55},
  [1047] = {.lex_state = 55},
  [1048] = {.lex_state = 55},
  [1049] = {.lex_state = 55},
  [1050] = {.lex_state = 55},
  [1051] = {.lex_state = 55},
  [1052] = {.lex_state = 55},
  [1053] = {.lex_state = 0},
  [1054] = {.lex_state = 55},
  [1055] = {.lex_state = 55},
  [1056] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1057] = {.lex_state = 55},
  [1058] = {.lex_state = 0},
  [1059] = {.lex_state = 55},
  [1060] = {.lex_state = 55},
  [1061] = {.lex_state = 55},
  [1062] = {.lex_state = 55},
  [1063] = {.lex_state = 55},
  [1064] = {.lex_state = 0},
  [1065] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1066] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1067] = {.lex_state = 55},
  [1068] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1069] = {.lex_state = 55},
  [1070] = {.lex_state = 55},
  [1071] = {.lex_state = 55},
  [1072] = {.lex_state = 0},
  [1073] = {.lex_state = 55},
  [1074] = {.lex_state = 55},
  [1075] = {.lex_state = 0},
  [1076] = {.lex_state = 55},
  [1077] = {.lex_state = 0},
  [1078] = {.lex_state = 55},
  [1079] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1080] = {.lex_state = 2},
  [1081] = {.lex_state = 0},
  [1082] = {.lex_state = 0},
  [1083] = {.lex_state = 2},
  [1084] = {.lex_state = 2},
  [1085] = {.lex_state = 2},
  [1086] = {.lex_state = 55},
  [1087] = {.lex_state = 0},
  [1088] = {.lex_state = 0},
  [1089] = {.lex_state = 55},
  [1090] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1091] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1092] = {.lex_state = 0},
  [1093] = {.lex_state = 55},
  [1094] = {.lex_state = 55},
  [1095] = {.lex_state = 55},
  [1096] = {.lex_state = 55},
  [1097] = {.lex_state = 2},
  [1098] = {.lex_state = 0},
  [1099] = {.lex_state = 0},
  [1100] = {.lex_state = 55},
  [1101] = {.lex_state = 2},
  [1102] = {.lex_state = 55},
  [1103] = {.lex_state = 0},
  [1104] = {.lex_state = 0},
  [1105] = {.lex_state = 55},
  [1106] = {.lex_state = 55},
  [1107] = {.lex_state = 55},
  [1108] = {.lex_state = 55},
  [1109] = {.lex_state = 55},
  [1110] = {.lex_state = 55},
  [1111] = {.lex_state = 0},
  [1112] = {.lex_state = 55},
  [1113] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1114] = {.lex_state = 2},
  [1115] = {.lex_state = 55},
  [1116] = {.lex_state = 55},
  [1117] = {.lex_state = 0},
  [1118] = {.lex_state = 0},
  [1119] = {.lex_state = 55},
  [1120] = {.lex_state = 2},
  [1121] = {.lex_state = 58},
  [1122] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1123] = {.lex_state = 2},
  [1124] = {.lex_state = 58},
  [1125] = {.lex_state = 55},
  [1126] = {.lex_state = 0},
  [1127] = {.lex_state = 0},
  [1128] = {.lex_state = 55},
  [1129] = {.lex_state = 0},
  [1130] = {.lex_state = 0},
  [1131] = {.lex_state = 55},
  [1132] = {.lex_state = 55},
  [1133] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1134] = {.lex_state = 55},
  [1135] = {.lex_state = 0},
  [1136] = {.lex_state = 2},
  [1137] = {.lex_state = 0},
  [1138] = {.lex_state = 0},
  [1139] = {.lex_state = 55},
  [1140] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1141] = {.lex_state = 55},
  [1142] = {.lex_state = 2},
  [1143] = {.lex_state = 0},
  [1144] = {.lex_state = 0},
  [1145] = {.lex_state = 55},
  [1146] = {.lex_state = 58},
  [1147] = {.lex_state = 2},
  [1148] = {.lex_state = 55},
  [1149] = {.lex_state = 0},
  [1150] = {.lex_state = 2},
  [1151] = {.lex_state = 2},
  [1152] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1153] = {.lex_state = 0},
  [1154] = {.lex_state = 0},
  [1155] = {.lex_state = 55},
  [1156] = {.lex_state = 58},
  [1157] = {.lex_state = 55},
  [1158] = {.lex_state = 0},
  [1159] = {.lex_state = 55},
  [1160] = {.lex_state = 55},
  [1161] = {.lex_state = 2},
  [1162] = {.lex_state = 0},
  [1163] = {.lex_state = 0},
  [1164] = {.lex_state = 0},
  [1165] = {.lex_state = 0},
  [1166] = {.lex_state = 0},
  [1167] = {.lex_state = 0},
  [1168] = {.lex_state = 0},
  [1169] = {.lex_state = 0},
  [1170] = {.lex_state = 0},
  [1171] = {.lex_state = 55},
  [1172] = {.lex_state = 0},
  [1173] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1174] = {.lex_state = 0},
  [1175] = {.lex_state = 0},
  [1176] = {.lex_state = 0},
  [1177] = {.lex_state = 0},
  [1178] = {.lex_state = 0},
  [1179] = {.lex_state = 0},
  [1180] = {.lex_state = 0},
  [1181] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1182] = {.lex_state = 0},
  [1183] = {.lex_state = 55},
  [1184] = {.lex_state = 0},
  [1185] = {.lex_state = 55},
  [1186] = {.lex_state = 55},
  [1187] = {.lex_state = 0},
  [1188] = {.lex_state = 0},
  [1189] = {.lex_state = 0},
  [1190] = {.lex_state = 0},
  [1191] = {.lex_state = 0},
  [1192] = {.lex_state = 0},
  [1193] = {.lex_state = 0},
  [1194] = {.lex_state = 0},
  [1195] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1196] = {.lex_state = 0},
  [1197] = {.lex_state = 0},
  [1198] = {.lex_state = 0},
  [1199] = {.lex_state = 0},
  [1200] = {.lex_state = 0},
  [1201] = {.lex_state = 0},
  [1202] = {.lex_state = 55},
  [1203] = {.lex_state = 0},
  [1204] = {.lex_state = 0},
  [1205] = {.lex_state = 55},
  [1206] = {.lex_state = 0},
  [1207] = {.lex_state = 0},
  [1208] = {.lex_state = 55},
  [1209] = {.lex_state = 0},
  [1210] = {.lex_state = 0},
  [1211] = {.lex_state = 0},
  [1212] = {.lex_state = 0},
  [1213] = {.lex_state = 0},
  [1214] = {.lex_state = 0},
  [1215] = {.lex_state = 0},
  [1216] = {.lex_state = 0},
  [1217] = {.lex_state = 0},
  [1218] = {.lex_state = 0},
  [1219] = {.lex_state = 55},
  [1220] = {.lex_state = 0},
  [1221] = {.lex_state = 0},
  [1222] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1223] = {.lex_state = 55},
  [1224] = {.lex_state = 0},
  [1225] = {.lex_state = 0},
  [1226] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1227] = {.lex_state = 55},
  [1228] = {.lex_state = 0},
  [1229] = {.lex_state = 0},
  [1230] = {.lex_state = 0},
  [1231] = {.lex_state = 0},
  [1232] = {.lex_state = 0},
  [1233] = {.lex_state = 0},
  [1234] = {.lex_state = 55},
  [1235] = {.lex_state = 0},
  [1236] = {.lex_state = 55},
  [1237] = {.lex_state = 0},
  [1238] = {.lex_state = 0},
  [1239] = {.lex_state = 0},
  [1240] = {.lex_state = 0},
  [1241] = {.lex_state = 0},
  [1242] = {.lex_state = 0},
  [1243] = {.lex_state = 0},
  [1244] = {.lex_state = 0},
  [1245] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1246] = {.lex_state = 0},
  [1247] = {.lex_state = 0},
  [1248] = {.lex_state = 0},
  [1249] = {.lex_state = 0},
  [1250] = {.lex_state = 0},
  [1251] = {.lex_state = 0},
  [1252] = {.lex_state = 55},
  [1253] = {.lex_state = 0},
  [1254] = {.lex_state = 0},
  [1255] = {.lex_state = 0},
  [1256] = {.lex_state = 55},
  [1257] = {.lex_state = 0},
  [1258] = {.lex_state = 0},
  [1259] = {.lex_state = 0},
  [1260] = {.lex_state = 0},
  [1261] = {.lex_state = 55},
  [1262] = {.lex_state = 0},
  [1263] = {.lex_state = 0},
  [1264] = {.lex_state = 0},
  [1265] = {.lex_state = 0},
  [1266] = {.lex_state = 0},
  [1267] = {.lex_state = 0},
  [1268] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1269] = {.lex_state = 0},
  [1270] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1271] = {.lex_state = 0},
  [1272] = {.lex_state = 0},
  [1273] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1274] = {.lex_state = 0},
  [1275] = {.lex_state = 0},
  [1276] = {.lex_state = 0},
  [1277] = {.lex_state = 0},
  [1278] = {.lex_state = 0},
  [1279] = {.lex_state = 0},
  [1280] = {.lex_state = 0},
  [1281] = {.lex_state = 0},
  [1282] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1283] = {.lex_state = 0},
  [1284] = {.lex_state = 0},
  [1285] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1286] = {.lex_state = 0},
  [1287] = {.lex_state = 0},
  [1288] = {.lex_state = 0},
  [1289] = {.lex_state = 0},
  [1290] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1291] = {.lex_state = 0},
  [1292] = {.lex_state = 0},
  [1293] = {.lex_state = 0},
  [1294] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1295] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1296] = {.lex_state = 0},
  [1297] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1298] = {.lex_state = 0},
  [1299] = {.lex_state = 0},
  [1300] = {.lex_state = 0},
  [1301] = {.lex_state = 0},
  [1302] = {.lex_state = 0},
  [1303] = {.lex_state = 0},
  [1304] = {.lex_state = 0},
  [1305] = {.lex_state = 0},
  [1306] = {.lex_state = 0},
  [1307] = {.lex_state = 0},
  [1308] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1309] = {.lex_state = 0},
  [1310] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1311] = {.lex_state = 0},
  [1312] = {.lex_state = 0},
  [1313] = {.lex_state = 0},
  [1314] = {.lex_state = 0},
  [1315] = {.lex_state = 0},
  [1316] = {.lex_state = 58},
  [1317] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1318] = {.lex_state = 0},
  [1319] = {.lex_state = 0},
  [1320] = {.lex_state = 0},
  [1321] = {.lex_state = 0},
  [1322] = {.lex_state = 0},
  [1323] = {.lex_state = 0},
  [1324] = {.lex_state = 0},
  [1325] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1326] = {.lex_state = 0},
  [1327] = {.lex_state = 0},
  [1328] = {.lex_state = 0},
  [1329] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1330] = {.lex_state = 0},
  [1331] = {.lex_state = 0},
  [1332] = {.lex_state = 0},
  [1333] = {.lex_state = 0},
  [1334] = {.lex_state = 0},
  [1335] = {.lex_state = 0},
  [1336] = {.lex_state = 0},
  [1337] = {.lex_state = 0},
  [1338] = {.lex_state = 0},
  [1339] = {.lex_state = 0},
  [1340] = {.lex_state = 0},
  [1341] = {.lex_state = 0},
  [1342] = {.lex_state = 0},
  [1343] = {.lex_state = 0},
  [1344] = {.lex_state = 0},
  [1345] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1346] = {.lex_state = 0},
  [1347] = {.lex_state = 0},
  [1348] = {.lex_state = 0},
  [1349] = {.lex_state = 0},
  [1350] = {.lex_state = 0},
  [1351] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1352] = {.lex_state = 0},
  [1353] = {.lex_state = 0},
  [1354] = {.lex_state = 0},
  [1355] = {.lex_state = 0},
  [1356] = {.lex_state = 0},
  [1357] = {.lex_state = 0},
  [1358] = {.lex_state = 0},
  [1359] = {.lex_state = 0},
  [1360] = {.lex_state = 0},
  [1361] = {.lex_state = 0},
  [1362] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1363] = {.lex_state = 0},
  [1364] = {.lex_state = 0},
  [1365] = {.lex_state = 0},
  [1366] = {.lex_state = 0},
  [1367] = {.lex_state = 0},
  [1368] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1369] = {.lex_state = 0},
  [1370] = {.lex_state = 0},
  [1371] = {.lex_state = 0},
  [1372] = {.lex_state = 0},
  [1373] = {.lex_state = 0},
  [1374] = {.lex_state = 0},
  [1375] = {.lex_state = 140},
  [1376] = {.lex_state = 0},
  [1377] = {.lex_state = 0},
  [1378] = {.lex_state = 0},
  [1379] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1380] = {.lex_state = 0},
  [1381] = {.lex_state = 0},
  [1382] = {.lex_state = 0},
  [1383] = {.lex_state = 0},
  [1384] = {.lex_state = 0},
  [1385] = {.lex_state = 0},
  [1386] = {.lex_state = 0},
  [1387] = {.lex_state = 0},
  [1388] = {.lex_state = 0},
  [1389] = {.lex_state = 0},
  [1390] = {.lex_state = 140},
  [1391] = {.lex_state = 0},
  [1392] = {.lex_state = 0},
  [1393] = {.lex_state = 0},
  [1394] = {.lex_state = 0},
  [1395] = {.lex_state = 0},
  [1396] = {.lex_state = 0},
  [1397] = {.lex_state = 0},
  [1398] = {.lex_state = 0},
  [1399] = {.lex_state = 0},
  [1400] = {.lex_state = 0},
  [1401] = {.lex_state = 140},
  [1402] = {.lex_state = 0},
  [1403] = {.lex_state = 0},
  [1404] = {.lex_state = 0},
  [1405] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1406] = {.lex_state = 0},
  [1407] = {.lex_state = 140},
  [1408] = {.lex_state = 0},
  [1409] = {.lex_state = 0},
  [1410] = {.lex_state = 0},
  [1411] = {.lex_state = 140},
  [1412] = {.lex_state = 0},
  [1413] = {.lex_state = 0},
  [1414] = {.lex_state = 0},
  [1415] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1416] = {.lex_state = 0},
  [1417] = {.lex_state = 0},
  [1418] = {.lex_state = 0},
  [1419] = {.lex_state = 0},
  [1420] = {.lex_state = 0},
  [1421] = {.lex_state = 140},
  [1422] = {.lex_state = 0},
  [1423] = {.lex_state = 0},
  [1424] = {.lex_state = 0},
  [1425] = {.lex_state = 0},
  [1426] = {.lex_state = 0},
  [1427] = {.lex_state = 0},
  [1428] = {.lex_state = 0},
  [1429] = {.lex_state = 0},
  [1430] = {.lex_state = 140},
  [1431] = {.lex_state = 0},
  [1432] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1433] = {.lex_state = 0},
  [1434] = {.lex_state = 0},
  [1435] = {.lex_state = 0},
  [1436] = {.lex_state = 0},
  [1437] = {.lex_state = 0},
  [1438] = {.lex_state = 0, .reserved_word_set_id = 1},
  [1439] = {.lex_state = 0},
  [1440] = {.lex_state = 0},
  [1441] = {.lex_state = 0},
};

static const TSSymbol ts_reserved_words[8][MAX_RESERVED_WORD_SET_SIZE] = {
  [1] = {
    anon_sym_package,
    anon_sym_import,
    anon_sym_const,
    anon_sym_var,
    anon_sym_func,
    anon_sym_type,
    anon_sym_struct,
    anon_sym_interface,
    anon_sym_map,
    anon_sym_chan,
    anon_sym_fallthrough,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_goto,
    anon_sym_return,
    anon_sym_go,
    anon_sym_defer,
    anon_sym_if,
    anon_sym_else,
    anon_sym_for,
    anon_sym_range,
    anon_sym_switch,
    anon_sym_case,
    anon_sym_default,
    anon_sym_select,
  },
  [2] = {
    anon_sym_package,
    anon_sym_import,
    anon_sym_else,
    anon_sym_range,
  },
  [3] = {
    anon_sym_package,
    anon_sym_import,
    anon_sym_else,
    anon_sym_range,
    anon_sym_case,
    anon_sym_default,
  },
  [4] = {
    anon_sym_package,
    anon_sym_import,
    anon_sym_const,
    anon_sym_var,
    anon_sym_type,
    anon_sym_fallthrough,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_goto,
    anon_sym_return,
    anon_sym_go,
    anon_sym_defer,
    anon_sym_if,
    anon_sym_else,
    anon_sym_for,
    anon_sym_switch,
    anon_sym_case,
    anon_sym_default,
    anon_sym_select,
  },
  [5] = {
    anon_sym_package,
    anon_sym_import,
    anon_sym_const,
    anon_sym_var,
    anon_sym_type,
    anon_sym_fallthrough,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_goto,
    anon_sym_return,
    anon_sym_go,
    anon_sym_defer,
    anon_sym_if,
    anon_sym_else,
    anon_sym_for,
    anon_sym_range,
    anon_sym_switch,
    anon_sym_case,
    anon_sym_default,
    anon_sym_select,
  },
  [6] = {
    anon_sym_package,
    anon_sym_import,
    anon_sym_const,
    anon_sym_var,
    anon_sym_type,
    anon_sym_fallthrough,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_goto,
    anon_sym_return,
    anon_sym_go,
    anon_sym_defer,
    anon_sym_if,
    anon_sym_else,
    anon_sym_for,
    anon_sym_range,
    anon_sym_switch,
    anon_sym_select,
  },
  [7] = {
    anon_sym_package,
    anon_sym_import,
    anon_sym_const,
    anon_sym_var,
    anon_sym_func,
    anon_sym_type,
    anon_sym_struct,
    anon_sym_interface,
    anon_sym_map,
    anon_sym_chan,
    anon_sym_fallthrough,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_goto,
    anon_sym_return,
    anon_sym_go,
    anon_sym_defer,
    anon_sym_if,
    anon_sym_else,
    anon_sym_for,
    anon_sym_range,
    anon_sym_switch,
    anon_sym_select,
  },
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_NULL] = ACTIONS(1),
    [anon_sym_package] = ACTIONS(1),
    [anon_sym_import] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [sym_blank_identifier] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_const] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_var] = ACTIONS(1),
    [anon_sym_func] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1),
    [anon_sym_type] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_struct] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_interface] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_map] = ACTIONS(1),
    [anon_sym_chan] = ACTIONS(1),
    [anon_sym_LT_DASH] = ACTIONS(1),
    [anon_sym_COLON_EQ] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_LT_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_AMP_CARET_EQ] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_fallthrough] = ACTIONS(1),
    [anon_sym_break] = ACTIONS(1),
    [anon_sym_continue] = ACTIONS(1),
    [anon_sym_goto] = ACTIONS(1),
    [anon_sym_return] = ACTIONS(1),
    [anon_sym_go] = ACTIONS(1),
    [anon_sym_defer] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_range] = ACTIONS(1),
    [anon_sym_switch] = ACTIONS(1),
    [anon_sym_case] = ACTIONS(1),
    [anon_sym_default] = ACTIONS(1),
    [anon_sym_select] = ACTIONS(1),
    [anon_sym_new] = ACTIONS(1),
    [anon_sym_make] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_AMP_CARET] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_BQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE2] = ACTIONS(1),
    [sym_escape_sequence] = ACTIONS(1),
    [sym_int_literal] = ACTIONS(1),
    [sym_float_literal] = ACTIONS(1),
    [sym_imaginary_literal] = ACTIONS(1),
    [sym_rune_literal] = ACTIONS(1),
    [sym_nil] = ACTIONS(1),
    [sym_true] = ACTIONS(1),
    [sym_false] = ACTIONS(1),
    [sym_iota] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(1)] = {
    [sym_source_file] = STATE(1419),
    [sym_package_clause] = STATE(1100),
    [sym_import_declaration] = STATE(1100),
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_function_declaration] = STATE(1100),
    [sym_method_declaration] = STATE(1100),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym__statement] = STATE(1252),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [aux_sym_source_file_repeat1] = STATE(2),
    [ts_builtin_sym_end] = ACTIONS(5),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_package] = ACTIONS(11),
    [anon_sym_import] = ACTIONS(13),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(21),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(2)] = {
    [sym_package_clause] = STATE(1069),
    [sym_import_declaration] = STATE(1069),
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_function_declaration] = STATE(1069),
    [sym_method_declaration] = STATE(1069),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym__statement] = STATE(1252),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [aux_sym_source_file_repeat1] = STATE(3),
    [ts_builtin_sym_end] = ACTIONS(77),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_package] = ACTIONS(11),
    [anon_sym_import] = ACTIONS(13),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(21),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(3)] = {
    [sym_package_clause] = STATE(1252),
    [sym_import_declaration] = STATE(1252),
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_function_declaration] = STATE(1252),
    [sym_method_declaration] = STATE(1252),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym__statement] = STATE(1252),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [aux_sym_source_file_repeat1] = STATE(3),
    [ts_builtin_sym_end] = ACTIONS(79),
    [sym_identifier] = ACTIONS(81),
    [anon_sym_SEMI] = ACTIONS(84),
    [anon_sym_package] = ACTIONS(87),
    [anon_sym_import] = ACTIONS(90),
    [anon_sym_LPAREN] = ACTIONS(93),
    [anon_sym_const] = ACTIONS(96),
    [anon_sym_var] = ACTIONS(99),
    [anon_sym_func] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(105),
    [anon_sym_type] = ACTIONS(108),
    [anon_sym_STAR] = ACTIONS(111),
    [anon_sym_struct] = ACTIONS(114),
    [anon_sym_TILDE] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(120),
    [anon_sym_interface] = ACTIONS(123),
    [anon_sym_map] = ACTIONS(126),
    [anon_sym_chan] = ACTIONS(129),
    [anon_sym_LT_DASH] = ACTIONS(132),
    [anon_sym_fallthrough] = ACTIONS(135),
    [anon_sym_break] = ACTIONS(138),
    [anon_sym_continue] = ACTIONS(141),
    [anon_sym_goto] = ACTIONS(144),
    [anon_sym_return] = ACTIONS(147),
    [anon_sym_go] = ACTIONS(150),
    [anon_sym_defer] = ACTIONS(153),
    [anon_sym_if] = ACTIONS(156),
    [anon_sym_for] = ACTIONS(159),
    [anon_sym_switch] = ACTIONS(162),
    [anon_sym_select] = ACTIONS(165),
    [anon_sym_new] = ACTIONS(168),
    [anon_sym_make] = ACTIONS(168),
    [anon_sym_PLUS] = ACTIONS(171),
    [anon_sym_DASH] = ACTIONS(171),
    [anon_sym_BANG] = ACTIONS(171),
    [anon_sym_CARET] = ACTIONS(171),
    [anon_sym_AMP] = ACTIONS(171),
    [anon_sym_BQUOTE] = ACTIONS(174),
    [anon_sym_DQUOTE] = ACTIONS(177),
    [sym_int_literal] = ACTIONS(180),
    [sym_float_literal] = ACTIONS(180),
    [sym_imaginary_literal] = ACTIONS(183),
    [sym_rune_literal] = ACTIONS(183),
    [sym_nil] = ACTIONS(180),
    [sym_true] = ACTIONS(180),
    [sym_false] = ACTIONS(180),
    [sym_iota] = ACTIONS(180),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(4)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1225),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(190),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_case] = ACTIONS(192),
    [anon_sym_default] = ACTIONS(192),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(5)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1200),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(194),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_case] = ACTIONS(196),
    [anon_sym_default] = ACTIONS(196),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(6)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1174),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(198),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_case] = ACTIONS(200),
    [anon_sym_default] = ACTIONS(200),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(7)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1182),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(202),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_case] = ACTIONS(204),
    [anon_sym_default] = ACTIONS(204),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(8)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1180),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(206),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_case] = ACTIONS(208),
    [anon_sym_default] = ACTIONS(208),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(9)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym__statement] = STATE(987),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1162),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(210),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_case] = ACTIONS(212),
    [anon_sym_default] = ACTIONS(212),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(10)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym__statement] = STATE(987),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1215),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(214),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_case] = ACTIONS(216),
    [anon_sym_default] = ACTIONS(216),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(11)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1398),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(218),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(12)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym__statement] = STATE(941),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(220),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_case] = ACTIONS(222),
    [anon_sym_default] = ACTIONS(222),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(13)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1365),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(224),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(14)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1355),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(226),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(15)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1386),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(228),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(16)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1417),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(230),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(17)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1427),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(232),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(18)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym_statement_list] = STATE(1403),
    [sym__statement] = STATE(916),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_empty_labeled_statement] = STATE(1247),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(186),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_RBRACE] = ACTIONS(234),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(19)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym__statement] = STATE(941),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(20)] = {
    [sym__declaration] = STATE(990),
    [sym_const_declaration] = STATE(990),
    [sym_var_declaration] = STATE(990),
    [sym_type_declaration] = STATE(990),
    [sym_expression_list] = STATE(838),
    [sym_parenthesized_type] = STATE(1340),
    [sym__simple_type] = STATE(1340),
    [sym_generic_type] = STATE(1149),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1149),
    [sym_implicit_length_array_type] = STATE(1300),
    [sym_slice_type] = STATE(1149),
    [sym_struct_type] = STATE(1149),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1154),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(990),
    [sym__statement] = STATE(987),
    [sym_empty_statement] = STATE(990),
    [sym__simple_statement] = STATE(990),
    [sym_expression_statement] = STATE(939),
    [sym_send_statement] = STATE(939),
    [sym_inc_statement] = STATE(939),
    [sym_dec_statement] = STATE(939),
    [sym_assignment_statement] = STATE(939),
    [sym_short_var_declaration] = STATE(939),
    [sym_labeled_statement] = STATE(990),
    [sym_fallthrough_statement] = STATE(990),
    [sym_break_statement] = STATE(990),
    [sym_continue_statement] = STATE(990),
    [sym_goto_statement] = STATE(990),
    [sym_return_statement] = STATE(990),
    [sym_go_statement] = STATE(990),
    [sym_defer_statement] = STATE(990),
    [sym_if_statement] = STATE(990),
    [sym_for_statement] = STATE(990),
    [sym_expression_switch_statement] = STATE(990),
    [sym_type_switch_statement] = STATE(990),
    [sym_select_statement] = STATE(990),
    [sym__expression] = STATE(240),
    [sym_parenthesized_expression] = STATE(255),
    [sym_call_expression] = STATE(255),
    [sym_selector_expression] = STATE(255),
    [sym_index_expression] = STATE(255),
    [sym_slice_expression] = STATE(255),
    [sym_type_assertion_expression] = STATE(255),
    [sym_type_conversion_expression] = STATE(255),
    [sym_type_instantiation_expression] = STATE(255),
    [sym_composite_literal] = STATE(255),
    [sym_func_literal] = STATE(255),
    [sym_unary_expression] = STATE(255),
    [sym_binary_expression] = STATE(255),
    [sym_qualified_type] = STATE(1053),
    [sym_raw_string_literal] = STATE(255),
    [sym_interpreted_string_literal] = STATE(255),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_const] = ACTIONS(17),
    [anon_sym_var] = ACTIONS(19),
    [anon_sym_func] = ACTIONS(188),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_type] = ACTIONS(25),
    [anon_sym_STAR] = ACTIONS(27),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(41),
    [anon_sym_fallthrough] = ACTIONS(43),
    [anon_sym_break] = ACTIONS(45),
    [anon_sym_continue] = ACTIONS(47),
    [anon_sym_goto] = ACTIONS(49),
    [anon_sym_return] = ACTIONS(51),
    [anon_sym_go] = ACTIONS(53),
    [anon_sym_defer] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_for] = ACTIONS(59),
    [anon_sym_switch] = ACTIONS(61),
    [anon_sym_select] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_make] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_CARET] = ACTIONS(67),
    [anon_sym_AMP] = ACTIONS(67),
    [anon_sym_BQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_int_literal] = ACTIONS(73),
    [sym_float_literal] = ACTIONS(73),
    [sym_imaginary_literal] = ACTIONS(75),
    [sym_rune_literal] = ACTIONS(75),
    [sym_nil] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_iota] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(21)] = {
    [sym_expression_list] = STATE(835),
    [sym_parenthesized_type] = STATE(1338),
    [sym__simple_type] = STATE(1338),
    [sym_generic_type] = STATE(1098),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1098),
    [sym_implicit_length_array_type] = STATE(1292),
    [sym_slice_type] = STATE(1098),
    [sym_struct_type] = STATE(1098),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1099),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym_block] = STATE(965),
    [sym__simple_statement] = STATE(1358),
    [sym_expression_statement] = STATE(1302),
    [sym_send_statement] = STATE(1302),
    [sym_inc_statement] = STATE(1302),
    [sym_dec_statement] = STATE(1302),
    [sym_assignment_statement] = STATE(1302),
    [sym_short_var_declaration] = STATE(1302),
    [sym_for_clause] = STATE(1304),
    [sym_range_clause] = STATE(1304),
    [sym__expression] = STATE(281),
    [sym_parenthesized_expression] = STATE(331),
    [sym_call_expression] = STATE(331),
    [sym_selector_expression] = STATE(331),
    [sym_index_expression] = STATE(331),
    [sym_slice_expression] = STATE(331),
    [sym_type_assertion_expression] = STATE(331),
    [sym_type_conversion_expression] = STATE(331),
    [sym_type_instantiation_expression] = STATE(331),
    [sym_composite_literal] = STATE(331),
    [sym_func_literal] = STATE(331),
    [sym_unary_expression] = STATE(331),
    [sym_binary_expression] = STATE(331),
    [sym_qualified_type] = STATE(1039),
    [sym_raw_string_literal] = STATE(331),
    [sym_interpreted_string_literal] = STATE(331),
    [sym_identifier] = ACTIONS(236),
    [anon_sym_SEMI] = ACTIONS(238),
    [anon_sym_LPAREN] = ACTIONS(240),
    [anon_sym_func] = ACTIONS(242),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(244),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(33),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(246),
    [anon_sym_range] = ACTIONS(248),
    [anon_sym_new] = ACTIONS(250),
    [anon_sym_make] = ACTIONS(250),
    [anon_sym_PLUS] = ACTIONS(252),
    [anon_sym_DASH] = ACTIONS(252),
    [anon_sym_BANG] = ACTIONS(252),
    [anon_sym_CARET] = ACTIONS(252),
    [anon_sym_AMP] = ACTIONS(252),
    [anon_sym_BQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [sym_int_literal] = ACTIONS(258),
    [sym_float_literal] = ACTIONS(258),
    [sym_imaginary_literal] = ACTIONS(260),
    [sym_rune_literal] = ACTIONS(260),
    [sym_nil] = ACTIONS(258),
    [sym_true] = ACTIONS(258),
    [sym_false] = ACTIONS(258),
    [sym_iota] = ACTIONS(258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(22)] = {
    [sym_expression_list] = STATE(825),
    [sym_parenthesized_type] = STATE(1338),
    [sym__simple_type] = STATE(1338),
    [sym_generic_type] = STATE(1098),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1098),
    [sym_implicit_length_array_type] = STATE(1292),
    [sym_slice_type] = STATE(1098),
    [sym_struct_type] = STATE(1098),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1099),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym__simple_statement] = STATE(1384),
    [sym_expression_statement] = STATE(1302),
    [sym_send_statement] = STATE(1302),
    [sym_inc_statement] = STATE(1302),
    [sym_dec_statement] = STATE(1302),
    [sym_assignment_statement] = STATE(1302),
    [sym_short_var_declaration] = STATE(1302),
    [sym__type_switch_header] = STATE(1348),
    [sym__expression] = STATE(290),
    [sym_parenthesized_expression] = STATE(331),
    [sym_call_expression] = STATE(331),
    [sym_selector_expression] = STATE(331),
    [sym_index_expression] = STATE(331),
    [sym_slice_expression] = STATE(331),
    [sym_type_assertion_expression] = STATE(331),
    [sym_type_conversion_expression] = STATE(331),
    [sym_type_instantiation_expression] = STATE(331),
    [sym_composite_literal] = STATE(331),
    [sym_func_literal] = STATE(331),
    [sym_unary_expression] = STATE(331),
    [sym_binary_expression] = STATE(331),
    [sym_qualified_type] = STATE(1039),
    [sym_raw_string_literal] = STATE(331),
    [sym_interpreted_string_literal] = STATE(331),
    [sym_identifier] = ACTIONS(236),
    [anon_sym_LPAREN] = ACTIONS(240),
    [anon_sym_func] = ACTIONS(242),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(244),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(262),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(246),
    [anon_sym_new] = ACTIONS(250),
    [anon_sym_make] = ACTIONS(250),
    [anon_sym_PLUS] = ACTIONS(252),
    [anon_sym_DASH] = ACTIONS(252),
    [anon_sym_BANG] = ACTIONS(252),
    [anon_sym_CARET] = ACTIONS(252),
    [anon_sym_AMP] = ACTIONS(252),
    [anon_sym_BQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [sym_int_literal] = ACTIONS(258),
    [sym_float_literal] = ACTIONS(258),
    [sym_imaginary_literal] = ACTIONS(260),
    [sym_rune_literal] = ACTIONS(260),
    [sym_nil] = ACTIONS(258),
    [sym_true] = ACTIONS(258),
    [sym_false] = ACTIONS(258),
    [sym_iota] = ACTIONS(258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(23)] = {
    [sym_expression_list] = STATE(823),
    [sym_parenthesized_type] = STATE(1338),
    [sym__simple_type] = STATE(1338),
    [sym_generic_type] = STATE(1098),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1098),
    [sym_implicit_length_array_type] = STATE(1292),
    [sym_slice_type] = STATE(1098),
    [sym_struct_type] = STATE(1098),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1099),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym__simple_statement] = STATE(1359),
    [sym_expression_statement] = STATE(1302),
    [sym_send_statement] = STATE(1302),
    [sym_inc_statement] = STATE(1302),
    [sym_dec_statement] = STATE(1302),
    [sym_assignment_statement] = STATE(1302),
    [sym_short_var_declaration] = STATE(1302),
    [sym__expression] = STATE(296),
    [sym_parenthesized_expression] = STATE(331),
    [sym_call_expression] = STATE(331),
    [sym_selector_expression] = STATE(331),
    [sym_index_expression] = STATE(331),
    [sym_slice_expression] = STATE(331),
    [sym_type_assertion_expression] = STATE(331),
    [sym_type_conversion_expression] = STATE(331),
    [sym_type_instantiation_expression] = STATE(331),
    [sym_composite_literal] = STATE(331),
    [sym_func_literal] = STATE(331),
    [sym_unary_expression] = STATE(331),
    [sym_binary_expression] = STATE(331),
    [sym_qualified_type] = STATE(1039),
    [sym_raw_string_literal] = STATE(331),
    [sym_interpreted_string_literal] = STATE(331),
    [sym_identifier] = ACTIONS(236),
    [anon_sym_LPAREN] = ACTIONS(240),
    [anon_sym_func] = ACTIONS(242),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(244),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(264),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(246),
    [anon_sym_new] = ACTIONS(250),
    [anon_sym_make] = ACTIONS(250),
    [anon_sym_PLUS] = ACTIONS(252),
    [anon_sym_DASH] = ACTIONS(252),
    [anon_sym_BANG] = ACTIONS(252),
    [anon_sym_CARET] = ACTIONS(252),
    [anon_sym_AMP] = ACTIONS(252),
    [anon_sym_BQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [sym_int_literal] = ACTIONS(258),
    [sym_float_literal] = ACTIONS(258),
    [sym_imaginary_literal] = ACTIONS(260),
    [sym_rune_literal] = ACTIONS(260),
    [sym_nil] = ACTIONS(258),
    [sym_true] = ACTIONS(258),
    [sym_false] = ACTIONS(258),
    [sym_iota] = ACTIONS(258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(24)] = {
    [sym_expression_list] = STATE(823),
    [sym_parenthesized_type] = STATE(1338),
    [sym__simple_type] = STATE(1338),
    [sym_generic_type] = STATE(1098),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1098),
    [sym_implicit_length_array_type] = STATE(1292),
    [sym_slice_type] = STATE(1098),
    [sym_struct_type] = STATE(1098),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1099),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym__simple_statement] = STATE(1370),
    [sym_expression_statement] = STATE(1302),
    [sym_send_statement] = STATE(1302),
    [sym_inc_statement] = STATE(1302),
    [sym_dec_statement] = STATE(1302),
    [sym_assignment_statement] = STATE(1302),
    [sym_short_var_declaration] = STATE(1302),
    [sym__expression] = STATE(296),
    [sym_parenthesized_expression] = STATE(331),
    [sym_call_expression] = STATE(331),
    [sym_selector_expression] = STATE(331),
    [sym_index_expression] = STATE(331),
    [sym_slice_expression] = STATE(331),
    [sym_type_assertion_expression] = STATE(331),
    [sym_type_conversion_expression] = STATE(331),
    [sym_type_instantiation_expression] = STATE(331),
    [sym_composite_literal] = STATE(331),
    [sym_func_literal] = STATE(331),
    [sym_unary_expression] = STATE(331),
    [sym_binary_expression] = STATE(331),
    [sym_qualified_type] = STATE(1039),
    [sym_raw_string_literal] = STATE(331),
    [sym_interpreted_string_literal] = STATE(331),
    [sym_identifier] = ACTIONS(236),
    [anon_sym_LPAREN] = ACTIONS(240),
    [anon_sym_func] = ACTIONS(242),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(244),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(266),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(246),
    [anon_sym_new] = ACTIONS(250),
    [anon_sym_make] = ACTIONS(250),
    [anon_sym_PLUS] = ACTIONS(252),
    [anon_sym_DASH] = ACTIONS(252),
    [anon_sym_BANG] = ACTIONS(252),
    [anon_sym_CARET] = ACTIONS(252),
    [anon_sym_AMP] = ACTIONS(252),
    [anon_sym_BQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [sym_int_literal] = ACTIONS(258),
    [sym_float_literal] = ACTIONS(258),
    [sym_imaginary_literal] = ACTIONS(260),
    [sym_rune_literal] = ACTIONS(260),
    [sym_nil] = ACTIONS(258),
    [sym_true] = ACTIONS(258),
    [sym_false] = ACTIONS(258),
    [sym_iota] = ACTIONS(258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(25)] = {
    [sym_expression_list] = STATE(823),
    [sym_parenthesized_type] = STATE(1338),
    [sym__simple_type] = STATE(1338),
    [sym_generic_type] = STATE(1098),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1098),
    [sym_implicit_length_array_type] = STATE(1292),
    [sym_slice_type] = STATE(1098),
    [sym_struct_type] = STATE(1098),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1099),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym__simple_statement] = STATE(1373),
    [sym_expression_statement] = STATE(1302),
    [sym_send_statement] = STATE(1302),
    [sym_inc_statement] = STATE(1302),
    [sym_dec_statement] = STATE(1302),
    [sym_assignment_statement] = STATE(1302),
    [sym_short_var_declaration] = STATE(1302),
    [sym__expression] = STATE(296),
    [sym_parenthesized_expression] = STATE(331),
    [sym_call_expression] = STATE(331),
    [sym_selector_expression] = STATE(331),
    [sym_index_expression] = STATE(331),
    [sym_slice_expression] = STATE(331),
    [sym_type_assertion_expression] = STATE(331),
    [sym_type_conversion_expression] = STATE(331),
    [sym_type_instantiation_expression] = STATE(331),
    [sym_composite_literal] = STATE(331),
    [sym_func_literal] = STATE(331),
    [sym_unary_expression] = STATE(331),
    [sym_binary_expression] = STATE(331),
    [sym_qualified_type] = STATE(1039),
    [sym_raw_string_literal] = STATE(331),
    [sym_interpreted_string_literal] = STATE(331),
    [sym_identifier] = ACTIONS(236),
    [anon_sym_LPAREN] = ACTIONS(240),
    [anon_sym_func] = ACTIONS(242),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(244),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(268),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(246),
    [anon_sym_new] = ACTIONS(250),
    [anon_sym_make] = ACTIONS(250),
    [anon_sym_PLUS] = ACTIONS(252),
    [anon_sym_DASH] = ACTIONS(252),
    [anon_sym_BANG] = ACTIONS(252),
    [anon_sym_CARET] = ACTIONS(252),
    [anon_sym_AMP] = ACTIONS(252),
    [anon_sym_BQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [sym_int_literal] = ACTIONS(258),
    [sym_float_literal] = ACTIONS(258),
    [sym_imaginary_literal] = ACTIONS(260),
    [sym_rune_literal] = ACTIONS(260),
    [sym_nil] = ACTIONS(258),
    [sym_true] = ACTIONS(258),
    [sym_false] = ACTIONS(258),
    [sym_iota] = ACTIONS(258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(26)] = {
    [sym_expression_list] = STATE(823),
    [sym_parenthesized_type] = STATE(1338),
    [sym__simple_type] = STATE(1338),
    [sym_generic_type] = STATE(1098),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1098),
    [sym_implicit_length_array_type] = STATE(1292),
    [sym_slice_type] = STATE(1098),
    [sym_struct_type] = STATE(1098),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1099),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym__simple_statement] = STATE(1372),
    [sym_expression_statement] = STATE(1302),
    [sym_send_statement] = STATE(1302),
    [sym_inc_statement] = STATE(1302),
    [sym_dec_statement] = STATE(1302),
    [sym_assignment_statement] = STATE(1302),
    [sym_short_var_declaration] = STATE(1302),
    [sym__expression] = STATE(296),
    [sym_parenthesized_expression] = STATE(331),
    [sym_call_expression] = STATE(331),
    [sym_selector_expression] = STATE(331),
    [sym_index_expression] = STATE(331),
    [sym_slice_expression] = STATE(331),
    [sym_type_assertion_expression] = STATE(331),
    [sym_type_conversion_expression] = STATE(331),
    [sym_type_instantiation_expression] = STATE(331),
    [sym_composite_literal] = STATE(331),
    [sym_func_literal] = STATE(331),
    [sym_unary_expression] = STATE(331),
    [sym_binary_expression] = STATE(331),
    [sym_qualified_type] = STATE(1039),
    [sym_raw_string_literal] = STATE(331),
    [sym_interpreted_string_literal] = STATE(331),
    [sym_identifier] = ACTIONS(236),
    [anon_sym_LPAREN] = ACTIONS(240),
    [anon_sym_func] = ACTIONS(242),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(244),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_LBRACE] = ACTIONS(270),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(246),
    [anon_sym_new] = ACTIONS(250),
    [anon_sym_make] = ACTIONS(250),
    [anon_sym_PLUS] = ACTIONS(252),
    [anon_sym_DASH] = ACTIONS(252),
    [anon_sym_BANG] = ACTIONS(252),
    [anon_sym_CARET] = ACTIONS(252),
    [anon_sym_AMP] = ACTIONS(252),
    [anon_sym_BQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [sym_int_literal] = ACTIONS(258),
    [sym_float_literal] = ACTIONS(258),
    [sym_imaginary_literal] = ACTIONS(260),
    [sym_rune_literal] = ACTIONS(260),
    [sym_nil] = ACTIONS(258),
    [sym_true] = ACTIONS(258),
    [sym_false] = ACTIONS(258),
    [sym_iota] = ACTIONS(258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(27)] = {
    [sym_expression_list] = STATE(822),
    [sym_parenthesized_type] = STATE(1338),
    [sym__simple_type] = STATE(1338),
    [sym_generic_type] = STATE(1098),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1098),
    [sym_implicit_length_array_type] = STATE(1292),
    [sym_slice_type] = STATE(1098),
    [sym_struct_type] = STATE(1098),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1099),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym__simple_statement] = STATE(1385),
    [sym_expression_statement] = STATE(1302),
    [sym_send_statement] = STATE(1302),
    [sym_inc_statement] = STATE(1302),
    [sym_dec_statement] = STATE(1302),
    [sym_assignment_statement] = STATE(1302),
    [sym_short_var_declaration] = STATE(1302),
    [sym__expression] = STATE(271),
    [sym_parenthesized_expression] = STATE(331),
    [sym_call_expression] = STATE(331),
    [sym_selector_expression] = STATE(331),
    [sym_index_expression] = STATE(331),
    [sym_slice_expression] = STATE(331),
    [sym_type_assertion_expression] = STATE(331),
    [sym_type_conversion_expression] = STATE(331),
    [sym_type_instantiation_expression] = STATE(331),
    [sym_composite_literal] = STATE(331),
    [sym_func_literal] = STATE(331),
    [sym_unary_expression] = STATE(331),
    [sym_binary_expression] = STATE(331),
    [sym_qualified_type] = STATE(1039),
    [sym_raw_string_literal] = STATE(331),
    [sym_interpreted_string_literal] = STATE(331),
    [sym_identifier] = ACTIONS(236),
    [anon_sym_LPAREN] = ACTIONS(240),
    [anon_sym_func] = ACTIONS(242),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_STAR] = ACTIONS(244),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(31),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(246),
    [anon_sym_new] = ACTIONS(250),
    [anon_sym_make] = ACTIONS(250),
    [anon_sym_PLUS] = ACTIONS(252),
    [anon_sym_DASH] = ACTIONS(252),
    [anon_sym_BANG] = ACTIONS(252),
    [anon_sym_CARET] = ACTIONS(252),
    [anon_sym_AMP] = ACTIONS(252),
    [anon_sym_BQUOTE] = ACTIONS(254),
    [anon_sym_DQUOTE] = ACTIONS(256),
    [sym_int_literal] = ACTIONS(258),
    [sym_float_literal] = ACTIONS(258),
    [sym_imaginary_literal] = ACTIONS(260),
    [sym_rune_literal] = ACTIONS(260),
    [sym_nil] = ACTIONS(258),
    [sym_true] = ACTIONS(258),
    [sym_false] = ACTIONS(258),
    [sym_iota] = ACTIONS(258),
    [sym_comment] = ACTIONS(3),
  },
  [STATE(28)] = {
    [sym_expression_list] = STATE(1030),
    [sym_parenthesized_type] = STATE(1337),
    [sym__simple_type] = STATE(1337),
    [sym_generic_type] = STATE(1081),
    [sym_pointer_type] = STATE(885),
    [sym_array_type] = STATE(1081),
    [sym_implicit_length_array_type] = STATE(1322),
    [sym_slice_type] = STATE(1081),
    [sym_struct_type] = STATE(1081),
    [sym_negated_type] = STATE(847),
    [sym_interface_type] = STATE(885),
    [sym_map_type] = STATE(1082),
    [sym_channel_type] = STATE(885),
    [sym_function_type] = STATE(885),
    [sym__expression] = STATE(400),
    [sym_parenthesized_expression] = STATE(497),
    [sym_call_expression] = STATE(497),
    [sym_selector_expression] = STATE(497),
    [sym_index_expression] = STATE(497),
    [sym_slice_expression] = STATE(497),
    [sym_type_assertion_expression] = STATE(497),
    [sym_type_conversion_expression] = STATE(497),
    [sym_type_instantiation_expression] = STATE(497),
    [sym_composite_literal] = STATE(497),
    [sym_func_literal] = STATE(497),
    [sym_unary_expression] = STATE(497),
    [sym_binary_expression] = STATE(497),
    [sym_qualified_type] = STATE(1034),
    [sym_raw_string_literal] = STATE(497),
    [sym_interpreted_string_literal] = STATE(497),
    [sym_identifier] = ACTIONS(272),
    [aux_sym_source_file_token1] = ACTIONS(274),
    [anon_sym_SEMI] = ACTIONS(276),
    [anon_sym_NULL] = ACTIONS(276),
    [anon_sym_LPAREN] = ACTIONS(278),
    [anon_sym_func] = ACTIONS(280),
    [anon_sym_LBRACK] = ACTIONS(282),
    [anon_sym_STAR] = ACTIONS(284),
    [anon_sym_struct] = ACTIONS(29),
    [anon_sym_TILDE] = ACTIONS(286),
    [anon_sym_RBRACE] = ACTIONS(276),
    [anon_sym_interface] = ACTIONS(35),
    [anon_sym_map] = ACTIONS(37),
    [anon_sym_chan] = ACTIONS(39),
    [anon_sym_LT_DASH] = ACTIONS(288),
    [anon_sym_case] = ACTIONS(276),
    [anon_sym_default] = ACTIONS(276),
    [anon_sym_new] = ACTIONS(290),
    [anon_sym_make] = ACTIONS(290),
    [anon_sym_PLUS] = ACTIONS(292),
    [anon_sym_DASH] = ACTIONS(292),
    [anon_sym_BANG] = ACTIONS(292),
    [anon_sym_CARET] = ACTIONS(292),
    [anon_sym_AMP] = ACTIONS(292),
    [anon_sym_BQUOTE] = ACTIONS(294),
    [anon_sym_DQUOTE] = ACTIONS(296),
    [sym_int_literal] = ACTIONS(298),
    [sym_float_literal] = ACTIONS(298),
    [sym_imaginary_literal] = ACTIONS(298),
    [sym_rune_literal] = ACTIONS(298),
    [sym_nil] = ACTIONS(298),
    [sym_true] = ACTIONS(298),
    [sym_false] = ACTIONS(298),
    [sym_iota] = ACTIONS(298),
    [sym_comment] = ACTIONS(300),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 33,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(306), 1,
      anon_sym_COMMA,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(314), 1,
      anon_sym_RBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1087), 1,
      sym_literal_element,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1250), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [131] = 33,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(330), 1,
      anon_sym_COMMA,
    ACTIONS(332), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1092), 1,
      sym_literal_element,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1178), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [262] = 33,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(334), 1,
      anon_sym_COMMA,
    ACTIONS(336), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1104), 1,
      sym_literal_element,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1189), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [393] = 33,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(338), 1,
      anon_sym_COMMA,
    ACTIONS(340), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1158), 1,
      sym_literal_element,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1259), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [524] = 33,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(342), 1,
      anon_sym_COMMA,
    ACTIONS(344), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1127), 1,
      sym_literal_element,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1210), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [655] = 33,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(346), 1,
      anon_sym_COMMA,
    ACTIONS(348), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1144), 1,
      sym_literal_element,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1232), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [786] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(350), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [914] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(352), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [1042] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(354), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [1170] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(356), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [1298] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(358), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [1426] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(360), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [1554] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(362), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [1682] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(364), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [1810] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(366), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [1938] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(368), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [2066] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(370), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [2194] = 32,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(372), 1,
      anon_sym_RBRACE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [2322] = 31,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1177), 1,
      sym_literal_element,
    STATE(1276), 1,
      sym_keyed_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [2447] = 30,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(542), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1175), 1,
      sym_literal_value,
    STATE(1280), 1,
      sym_literal_element,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [2569] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(374), 1,
      sym_identifier,
    ACTIONS(376), 1,
      anon_sym_LPAREN,
    ACTIONS(378), 1,
      anon_sym_func,
    ACTIONS(380), 1,
      anon_sym_STAR,
    ACTIONS(382), 1,
      anon_sym_LT_DASH,
    ACTIONS(388), 1,
      anon_sym_BQUOTE,
    ACTIONS(390), 1,
      anon_sym_DQUOTE,
    STATE(451), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1058), 1,
      sym_qualified_type,
    STATE(1138), 1,
      sym_map_type,
    STATE(1281), 1,
      sym_implicit_length_array_type,
    STATE(1303), 1,
      sym_expression_list,
    ACTIONS(384), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(394), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1269), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1371), 2,
      sym_send_statement,
      sym_receive_statement,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1137), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(386), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(392), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(544), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [2689] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(396), 1,
      anon_sym_RBRACK,
    ACTIONS(398), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(689), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [2808] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(400), 1,
      sym_identifier,
    ACTIONS(402), 1,
      anon_sym_STAR,
    ACTIONS(404), 1,
      anon_sym_COLON,
    STATE(631), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1169), 1,
      sym_type_elem,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(970), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [2927] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(408), 1,
      anon_sym_RPAREN,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(525), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1228), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [3046] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(416), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [3165] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(400), 1,
      sym_identifier,
    ACTIONS(402), 1,
      anon_sym_STAR,
    ACTIONS(418), 1,
      anon_sym_COLON,
    STATE(600), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1169), 1,
      sym_type_elem,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(970), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [3284] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(420), 1,
      anon_sym_range,
    STATE(522), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1287), 1,
      sym_expression_list,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [3403] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(420), 1,
      anon_sym_range,
    STATE(522), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1291), 1,
      sym_expression_list,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [3522] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(422), 1,
      anon_sym_RPAREN,
    STATE(518), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1263), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [3641] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(400), 1,
      sym_identifier,
    ACTIONS(402), 1,
      anon_sym_STAR,
    ACTIONS(424), 1,
      anon_sym_RBRACK,
    STATE(673), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1255), 1,
      sym_type_elem,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(970), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [3760] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(400), 1,
      sym_identifier,
    ACTIONS(402), 1,
      anon_sym_STAR,
    ACTIONS(426), 1,
      anon_sym_COLON,
    STATE(637), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1169), 1,
      sym_type_elem,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(970), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [3879] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(428), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [3998] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(430), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [4117] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(432), 1,
      anon_sym_RPAREN,
    STATE(524), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1258), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [4236] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(434), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [4355] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(436), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [4474] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(442), 1,
      anon_sym_LBRACE,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(532), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    STATE(1354), 1,
      sym_expression_list,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [4593] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(448), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [4712] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(450), 1,
      anon_sym_RPAREN,
    STATE(528), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1191), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [4831] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(400), 1,
      sym_identifier,
    ACTIONS(402), 1,
      anon_sym_STAR,
    ACTIONS(452), 1,
      anon_sym_RBRACK,
    STATE(684), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1169), 1,
      sym_type_elem,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(970), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [4950] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(454), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [5069] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(456), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [5188] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(458), 1,
      anon_sym_RPAREN,
    STATE(530), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1214), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [5307] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(460), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [5426] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(462), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [5545] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(464), 1,
      anon_sym_RPAREN,
    STATE(531), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1235), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [5664] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(466), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [5783] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(424), 1,
      anon_sym_RBRACK,
    ACTIONS(468), 1,
      sym_identifier,
    STATE(673), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1211), 1,
      sym_type_parameter_declaration,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [5902] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(470), 1,
      anon_sym_RPAREN,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [6021] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(400), 1,
      sym_identifier,
    ACTIONS(402), 1,
      anon_sym_STAR,
    ACTIONS(472), 1,
      anon_sym_COLON,
    STATE(605), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1169), 1,
      sym_type_elem,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(970), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [6140] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(400), 1,
      sym_identifier,
    ACTIONS(402), 1,
      anon_sym_STAR,
    ACTIONS(474), 1,
      anon_sym_COLON,
    STATE(618), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1169), 1,
      sym_type_elem,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(970), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [6259] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(400), 1,
      sym_identifier,
    ACTIONS(402), 1,
      anon_sym_STAR,
    ACTIONS(476), 1,
      anon_sym_COLON,
    STATE(625), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1169), 1,
      sym_type_elem,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(970), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [6378] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(478), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [6494] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(480), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [6610] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(396), 1,
      anon_sym_RBRACK,
    STATE(689), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [6726] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(452), 1,
      anon_sym_RBRACK,
    STATE(684), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [6842] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(482), 1,
      anon_sym_RBRACK,
    STATE(622), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [6958] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(484), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [7074] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(486), 1,
      anon_sym_SEMI,
    STATE(696), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [7190] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(488), 1,
      anon_sym_SEMI,
    STATE(681), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [7306] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(593), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1287), 1,
      sym_expression_list,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [7422] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(490), 1,
      anon_sym_RBRACK,
    STATE(652), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [7538] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(593), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1291), 1,
      sym_expression_list,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [7654] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(942), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [7770] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(506), 1,
      anon_sym_RBRACK,
    STATE(627), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [7886] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(508), 1,
      anon_sym_RBRACK,
    STATE(608), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [8002] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(510), 1,
      anon_sym_RBRACK,
    STATE(613), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [8118] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(534), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1314), 1,
      sym_variadic_argument,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [8234] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(512), 1,
      anon_sym_RBRACK,
    STATE(628), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [8350] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(514), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [8466] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(980), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [8582] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(516), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [8698] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(518), 1,
      anon_sym_RBRACK,
    STATE(611), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [8814] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(520), 1,
      anon_sym_RBRACK,
    STATE(665), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [8930] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1031), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [9046] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(522), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [9162] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(946), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [9278] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(524), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [9394] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(981), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [9510] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(522), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1287), 1,
      sym_expression_list,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [9626] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(526), 1,
      anon_sym_RBRACK,
    STATE(634), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [9742] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(528), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [9858] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(982), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [9974] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(530), 1,
      anon_sym_RBRACK,
    STATE(599), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [10090] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(993), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [10206] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(532), 1,
      anon_sym_RBRACK,
    STATE(679), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [10322] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(991), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [10438] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(534), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [10554] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(536), 1,
      anon_sym_RBRACK,
    STATE(638), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [10670] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(538), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [10786] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(936), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [10902] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(540), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [11018] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(542), 1,
      anon_sym_RBRACK,
    STATE(639), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [11134] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(544), 1,
      anon_sym_RPAREN,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [11250] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(522), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1291), 1,
      sym_expression_list,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [11366] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(522), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    STATE(1416), 1,
      sym_expression_list,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [11482] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1004), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [11598] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(424), 1,
      anon_sym_RBRACK,
    STATE(673), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [11714] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(597), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1291), 1,
      sym_expression_list,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [11830] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(400), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1023), 1,
      sym_expression_list,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [11946] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(546), 1,
      anon_sym_RBRACK,
    STATE(620), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [12062] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(548), 1,
      anon_sym_RBRACK,
    STATE(632), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [12178] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(456), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [12291] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(655), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [12404] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(457), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [12517] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [12630] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(488), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [12743] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(390), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [12856] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(536), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [12969] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(602), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [13082] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(537), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [13195] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(538), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [13308] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(539), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [13421] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [13534] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(390), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [13647] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(441), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [13760] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(442), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [13873] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(443), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [13986] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(438), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [14099] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(448), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [14212] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(440), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [14325] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(444), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [14438] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(445), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [14551] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(446), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [14664] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(447), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [14777] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(236), 1,
      sym_identifier,
    ACTIONS(240), 1,
      anon_sym_LPAREN,
    ACTIONS(242), 1,
      anon_sym_func,
    ACTIONS(244), 1,
      anon_sym_STAR,
    ACTIONS(246), 1,
      anon_sym_LT_DASH,
    ACTIONS(254), 1,
      anon_sym_BQUOTE,
    ACTIONS(256), 1,
      anon_sym_DQUOTE,
    STATE(295), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1039), 1,
      sym_qualified_type,
    STATE(1099), 1,
      sym_map_type,
    STATE(1292), 1,
      sym_implicit_length_array_type,
    ACTIONS(250), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(260), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1338), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1098), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(252), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(258), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(331), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [14890] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(236), 1,
      sym_identifier,
    ACTIONS(240), 1,
      anon_sym_LPAREN,
    ACTIONS(242), 1,
      anon_sym_func,
    ACTIONS(244), 1,
      anon_sym_STAR,
    ACTIONS(246), 1,
      anon_sym_LT_DASH,
    ACTIONS(254), 1,
      anon_sym_BQUOTE,
    ACTIONS(256), 1,
      anon_sym_DQUOTE,
    STATE(293), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1039), 1,
      sym_qualified_type,
    STATE(1099), 1,
      sym_map_type,
    STATE(1292), 1,
      sym_implicit_length_array_type,
    ACTIONS(250), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(260), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1338), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1098), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(252), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(258), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(331), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [15003] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(236), 1,
      sym_identifier,
    ACTIONS(240), 1,
      anon_sym_LPAREN,
    ACTIONS(242), 1,
      anon_sym_func,
    ACTIONS(244), 1,
      anon_sym_STAR,
    ACTIONS(246), 1,
      anon_sym_LT_DASH,
    ACTIONS(254), 1,
      anon_sym_BQUOTE,
    ACTIONS(256), 1,
      anon_sym_DQUOTE,
    STATE(291), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1039), 1,
      sym_qualified_type,
    STATE(1099), 1,
      sym_map_type,
    STATE(1292), 1,
      sym_implicit_length_array_type,
    ACTIONS(250), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(260), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1338), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1098), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(252), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(258), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(331), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [15116] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(236), 1,
      sym_identifier,
    ACTIONS(240), 1,
      anon_sym_LPAREN,
    ACTIONS(242), 1,
      anon_sym_func,
    ACTIONS(244), 1,
      anon_sym_STAR,
    ACTIONS(246), 1,
      anon_sym_LT_DASH,
    ACTIONS(254), 1,
      anon_sym_BQUOTE,
    ACTIONS(256), 1,
      anon_sym_DQUOTE,
    STATE(297), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1039), 1,
      sym_qualified_type,
    STATE(1099), 1,
      sym_map_type,
    STATE(1292), 1,
      sym_implicit_length_array_type,
    ACTIONS(250), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(260), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1338), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1098), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(252), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(258), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(331), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [15229] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(236), 1,
      sym_identifier,
    ACTIONS(240), 1,
      anon_sym_LPAREN,
    ACTIONS(242), 1,
      anon_sym_func,
    ACTIONS(244), 1,
      anon_sym_STAR,
    ACTIONS(246), 1,
      anon_sym_LT_DASH,
    ACTIONS(254), 1,
      anon_sym_BQUOTE,
    ACTIONS(256), 1,
      anon_sym_DQUOTE,
    STATE(298), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1039), 1,
      sym_qualified_type,
    STATE(1099), 1,
      sym_map_type,
    STATE(1292), 1,
      sym_implicit_length_array_type,
    ACTIONS(250), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(260), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1338), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1098), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(252), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(258), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(331), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [15342] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(236), 1,
      sym_identifier,
    ACTIONS(240), 1,
      anon_sym_LPAREN,
    ACTIONS(242), 1,
      anon_sym_func,
    ACTIONS(244), 1,
      anon_sym_STAR,
    ACTIONS(246), 1,
      anon_sym_LT_DASH,
    ACTIONS(254), 1,
      anon_sym_BQUOTE,
    ACTIONS(256), 1,
      anon_sym_DQUOTE,
    STATE(292), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1039), 1,
      sym_qualified_type,
    STATE(1099), 1,
      sym_map_type,
    STATE(1292), 1,
      sym_implicit_length_array_type,
    ACTIONS(250), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(260), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1338), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1098), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(252), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(258), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(331), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [15455] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(550), 1,
      sym_identifier,
    ACTIONS(552), 1,
      anon_sym_LPAREN,
    ACTIONS(554), 1,
      anon_sym_func,
    ACTIONS(556), 1,
      anon_sym_STAR,
    ACTIONS(558), 1,
      anon_sym_LT_DASH,
    ACTIONS(564), 1,
      anon_sym_BQUOTE,
    ACTIONS(566), 1,
      anon_sym_DQUOTE,
    STATE(343), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1037), 1,
      sym_qualified_type,
    STATE(1118), 1,
      sym_map_type,
    STATE(1321), 1,
      sym_implicit_length_array_type,
    ACTIONS(560), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(570), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1339), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1117), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(562), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(568), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(348), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [15568] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(550), 1,
      sym_identifier,
    ACTIONS(552), 1,
      anon_sym_LPAREN,
    ACTIONS(554), 1,
      anon_sym_func,
    ACTIONS(556), 1,
      anon_sym_STAR,
    ACTIONS(558), 1,
      anon_sym_LT_DASH,
    ACTIONS(564), 1,
      anon_sym_BQUOTE,
    ACTIONS(566), 1,
      anon_sym_DQUOTE,
    STATE(344), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1037), 1,
      sym_qualified_type,
    STATE(1118), 1,
      sym_map_type,
    STATE(1321), 1,
      sym_implicit_length_array_type,
    ACTIONS(560), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(570), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1339), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1117), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(562), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(568), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(348), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [15681] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(572), 1,
      anon_sym_STAR,
    STATE(687), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1193), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [15794] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(550), 1,
      sym_identifier,
    ACTIONS(552), 1,
      anon_sym_LPAREN,
    ACTIONS(554), 1,
      anon_sym_func,
    ACTIONS(556), 1,
      anon_sym_STAR,
    ACTIONS(558), 1,
      anon_sym_LT_DASH,
    ACTIONS(564), 1,
      anon_sym_BQUOTE,
    ACTIONS(566), 1,
      anon_sym_DQUOTE,
    STATE(346), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1037), 1,
      sym_qualified_type,
    STATE(1118), 1,
      sym_map_type,
    STATE(1321), 1,
      sym_implicit_length_array_type,
    ACTIONS(560), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(570), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1339), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1117), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(562), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(568), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(348), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [15907] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(550), 1,
      sym_identifier,
    ACTIONS(552), 1,
      anon_sym_LPAREN,
    ACTIONS(554), 1,
      anon_sym_func,
    ACTIONS(556), 1,
      anon_sym_STAR,
    ACTIONS(558), 1,
      anon_sym_LT_DASH,
    ACTIONS(564), 1,
      anon_sym_BQUOTE,
    ACTIONS(566), 1,
      anon_sym_DQUOTE,
    STATE(342), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1037), 1,
      sym_qualified_type,
    STATE(1118), 1,
      sym_map_type,
    STATE(1321), 1,
      sym_implicit_length_array_type,
    ACTIONS(560), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(570), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1339), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1117), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(562), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(568), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(348), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [16020] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(550), 1,
      sym_identifier,
    ACTIONS(552), 1,
      anon_sym_LPAREN,
    ACTIONS(554), 1,
      anon_sym_func,
    ACTIONS(556), 1,
      anon_sym_STAR,
    ACTIONS(558), 1,
      anon_sym_LT_DASH,
    ACTIONS(564), 1,
      anon_sym_BQUOTE,
    ACTIONS(566), 1,
      anon_sym_DQUOTE,
    STATE(345), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1037), 1,
      sym_qualified_type,
    STATE(1118), 1,
      sym_map_type,
    STATE(1321), 1,
      sym_implicit_length_array_type,
    ACTIONS(560), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(570), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1339), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1117), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(562), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(568), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(348), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [16133] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(374), 1,
      sym_identifier,
    ACTIONS(376), 1,
      anon_sym_LPAREN,
    ACTIONS(378), 1,
      anon_sym_func,
    ACTIONS(380), 1,
      anon_sym_STAR,
    ACTIONS(382), 1,
      anon_sym_LT_DASH,
    ACTIONS(388), 1,
      anon_sym_BQUOTE,
    ACTIONS(390), 1,
      anon_sym_DQUOTE,
    STATE(499), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1058), 1,
      sym_qualified_type,
    STATE(1138), 1,
      sym_map_type,
    STATE(1281), 1,
      sym_implicit_length_array_type,
    ACTIONS(384), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(394), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1269), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1137), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(386), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(392), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(544), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [16246] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(374), 1,
      sym_identifier,
    ACTIONS(376), 1,
      anon_sym_LPAREN,
    ACTIONS(378), 1,
      anon_sym_func,
    ACTIONS(380), 1,
      anon_sym_STAR,
    ACTIONS(382), 1,
      anon_sym_LT_DASH,
    ACTIONS(388), 1,
      anon_sym_BQUOTE,
    ACTIONS(390), 1,
      anon_sym_DQUOTE,
    STATE(501), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1058), 1,
      sym_qualified_type,
    STATE(1138), 1,
      sym_map_type,
    STATE(1281), 1,
      sym_implicit_length_array_type,
    ACTIONS(384), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(394), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1269), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1137), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(386), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(392), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(544), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [16359] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(374), 1,
      sym_identifier,
    ACTIONS(376), 1,
      anon_sym_LPAREN,
    ACTIONS(378), 1,
      anon_sym_func,
    ACTIONS(380), 1,
      anon_sym_STAR,
    ACTIONS(382), 1,
      anon_sym_LT_DASH,
    ACTIONS(388), 1,
      anon_sym_BQUOTE,
    ACTIONS(390), 1,
      anon_sym_DQUOTE,
    STATE(505), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1058), 1,
      sym_qualified_type,
    STATE(1138), 1,
      sym_map_type,
    STATE(1281), 1,
      sym_implicit_length_array_type,
    ACTIONS(384), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(394), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1269), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1137), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(386), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(392), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(544), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [16472] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(374), 1,
      sym_identifier,
    ACTIONS(376), 1,
      anon_sym_LPAREN,
    ACTIONS(378), 1,
      anon_sym_func,
    ACTIONS(380), 1,
      anon_sym_STAR,
    ACTIONS(382), 1,
      anon_sym_LT_DASH,
    ACTIONS(388), 1,
      anon_sym_BQUOTE,
    ACTIONS(390), 1,
      anon_sym_DQUOTE,
    STATE(506), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1058), 1,
      sym_qualified_type,
    STATE(1138), 1,
      sym_map_type,
    STATE(1281), 1,
      sym_implicit_length_array_type,
    ACTIONS(384), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(394), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1269), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1137), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(386), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(392), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(544), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [16585] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(374), 1,
      sym_identifier,
    ACTIONS(376), 1,
      anon_sym_LPAREN,
    ACTIONS(378), 1,
      anon_sym_func,
    ACTIONS(380), 1,
      anon_sym_STAR,
    ACTIONS(382), 1,
      anon_sym_LT_DASH,
    ACTIONS(388), 1,
      anon_sym_BQUOTE,
    ACTIONS(390), 1,
      anon_sym_DQUOTE,
    STATE(507), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1058), 1,
      sym_qualified_type,
    STATE(1138), 1,
      sym_map_type,
    STATE(1281), 1,
      sym_implicit_length_array_type,
    ACTIONS(384), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(394), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1269), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1137), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(386), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(392), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(544), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [16698] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(374), 1,
      sym_identifier,
    ACTIONS(376), 1,
      anon_sym_LPAREN,
    ACTIONS(378), 1,
      anon_sym_func,
    ACTIONS(380), 1,
      anon_sym_STAR,
    ACTIONS(382), 1,
      anon_sym_LT_DASH,
    ACTIONS(388), 1,
      anon_sym_BQUOTE,
    ACTIONS(390), 1,
      anon_sym_DQUOTE,
    STATE(508), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1058), 1,
      sym_qualified_type,
    STATE(1138), 1,
      sym_map_type,
    STATE(1281), 1,
      sym_implicit_length_array_type,
    ACTIONS(384), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(394), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1269), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1137), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(386), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(392), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(544), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [16811] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(572), 1,
      anon_sym_STAR,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(973), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [16924] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(574), 1,
      anon_sym_chan,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [17037] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(437), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [17150] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(669), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [17263] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(27), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(41), 1,
      anon_sym_LT_DASH,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(188), 1,
      anon_sym_func,
    ACTIONS(576), 1,
      sym_identifier,
    STATE(246), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1053), 1,
      sym_qualified_type,
    STATE(1154), 1,
      sym_map_type,
    STATE(1300), 1,
      sym_implicit_length_array_type,
    ACTIONS(65), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(75), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1340), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1149), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(67), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(73), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(255), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [17376] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [17489] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(390), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [17602] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(586), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [17715] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(587), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [17828] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(588), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [17941] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(589), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [18054] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(973), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [18167] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(574), 1,
      anon_sym_chan,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [18280] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(609), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [18393] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(643), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [18506] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(550), 1,
      sym_identifier,
    ACTIONS(552), 1,
      anon_sym_LPAREN,
    ACTIONS(554), 1,
      anon_sym_func,
    ACTIONS(556), 1,
      anon_sym_STAR,
    ACTIONS(558), 1,
      anon_sym_LT_DASH,
    ACTIONS(564), 1,
      anon_sym_BQUOTE,
    ACTIONS(566), 1,
      anon_sym_DQUOTE,
    STATE(340), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1037), 1,
      sym_qualified_type,
    STATE(1118), 1,
      sym_map_type,
    STATE(1321), 1,
      sym_implicit_length_array_type,
    ACTIONS(560), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(570), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1339), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1117), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(562), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(568), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(348), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [18619] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(27), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(41), 1,
      anon_sym_LT_DASH,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(188), 1,
      anon_sym_func,
    ACTIONS(576), 1,
      sym_identifier,
    STATE(244), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1053), 1,
      sym_qualified_type,
    STATE(1154), 1,
      sym_map_type,
    STATE(1300), 1,
      sym_implicit_length_array_type,
    ACTIONS(65), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(75), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1340), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1149), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(67), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(73), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(255), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [18732] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(672), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [18845] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(27), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(41), 1,
      anon_sym_LT_DASH,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(188), 1,
      anon_sym_func,
    ACTIONS(576), 1,
      sym_identifier,
    STATE(243), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1053), 1,
      sym_qualified_type,
    STATE(1154), 1,
      sym_map_type,
    STATE(1300), 1,
      sym_implicit_length_array_type,
    ACTIONS(65), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(75), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1340), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1149), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(67), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(73), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(255), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [18958] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(448), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(973), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [19071] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    STATE(455), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [19184] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(272), 1,
      sym_identifier,
    ACTIONS(280), 1,
      anon_sym_func,
    ACTIONS(492), 1,
      anon_sym_LPAREN,
    ACTIONS(494), 1,
      anon_sym_STAR,
    ACTIONS(496), 1,
      anon_sym_LT_DASH,
    ACTIONS(500), 1,
      anon_sym_BQUOTE,
    ACTIONS(502), 1,
      anon_sym_DQUOTE,
    ACTIONS(574), 1,
      anon_sym_chan,
    STATE(448), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1034), 1,
      sym_qualified_type,
    STATE(1082), 1,
      sym_map_type,
    STATE(1322), 1,
      sym_implicit_length_array_type,
    ACTIONS(290), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(504), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1337), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1081), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(498), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(298), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(497), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [19297] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(27), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(41), 1,
      anon_sym_LT_DASH,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(188), 1,
      anon_sym_func,
    ACTIONS(576), 1,
      sym_identifier,
    STATE(242), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1053), 1,
      sym_qualified_type,
    STATE(1154), 1,
      sym_map_type,
    STATE(1300), 1,
      sym_implicit_length_array_type,
    ACTIONS(65), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(75), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1340), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1149), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(67), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(73), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(255), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [19410] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(615), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [19523] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(27), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(41), 1,
      anon_sym_LT_DASH,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(188), 1,
      anon_sym_func,
    ACTIONS(576), 1,
      sym_identifier,
    STATE(245), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1053), 1,
      sym_qualified_type,
    STATE(1154), 1,
      sym_map_type,
    STATE(1300), 1,
      sym_implicit_length_array_type,
    ACTIONS(65), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(75), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1340), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1149), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(67), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(73), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(255), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [19636] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(27), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(41), 1,
      anon_sym_LT_DASH,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(188), 1,
      anon_sym_func,
    ACTIONS(576), 1,
      sym_identifier,
    STATE(247), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1053), 1,
      sym_qualified_type,
    STATE(1154), 1,
      sym_map_type,
    STATE(1300), 1,
      sym_implicit_length_array_type,
    ACTIONS(65), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(75), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1340), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1149), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(67), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(73), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(255), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [19749] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(236), 1,
      sym_identifier,
    ACTIONS(240), 1,
      anon_sym_LPAREN,
    ACTIONS(242), 1,
      anon_sym_func,
    ACTIONS(244), 1,
      anon_sym_STAR,
    ACTIONS(246), 1,
      anon_sym_LT_DASH,
    ACTIONS(254), 1,
      anon_sym_BQUOTE,
    ACTIONS(256), 1,
      anon_sym_DQUOTE,
    STATE(295), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1039), 1,
      sym_qualified_type,
    STATE(1099), 1,
      sym_map_type,
    STATE(1292), 1,
      sym_implicit_length_array_type,
    ACTIONS(250), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(260), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(973), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1098), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(252), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(258), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(331), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [19862] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(236), 1,
      sym_identifier,
    ACTIONS(240), 1,
      anon_sym_LPAREN,
    ACTIONS(242), 1,
      anon_sym_func,
    ACTIONS(244), 1,
      anon_sym_STAR,
    ACTIONS(246), 1,
      anon_sym_LT_DASH,
    ACTIONS(254), 1,
      anon_sym_BQUOTE,
    ACTIONS(256), 1,
      anon_sym_DQUOTE,
    ACTIONS(574), 1,
      anon_sym_chan,
    STATE(295), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1039), 1,
      sym_qualified_type,
    STATE(1099), 1,
      sym_map_type,
    STATE(1292), 1,
      sym_implicit_length_array_type,
    ACTIONS(250), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(260), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1338), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1098), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(252), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(258), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(331), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [19975] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(27), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(41), 1,
      anon_sym_LT_DASH,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(188), 1,
      anon_sym_func,
    ACTIONS(574), 1,
      anon_sym_chan,
    ACTIONS(576), 1,
      sym_identifier,
    STATE(246), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1053), 1,
      sym_qualified_type,
    STATE(1154), 1,
      sym_map_type,
    STATE(1300), 1,
      sym_implicit_length_array_type,
    ACTIONS(65), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(75), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1340), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1149), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(67), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(73), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(255), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [20088] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(727), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [20201] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(693), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [20314] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(27), 1,
      anon_sym_STAR,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(41), 1,
      anon_sym_LT_DASH,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(188), 1,
      anon_sym_func,
    ACTIONS(576), 1,
      sym_identifier,
    STATE(246), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1053), 1,
      sym_qualified_type,
    STATE(1154), 1,
      sym_map_type,
    STATE(1300), 1,
      sym_implicit_length_array_type,
    ACTIONS(65), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(75), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(973), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1149), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(67), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(73), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(255), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [20427] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(973), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [20540] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    ACTIONS(574), 1,
      anon_sym_chan,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [20653] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(646), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [20766] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(633), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [20879] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(670), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [20992] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(438), 1,
      sym_identifier,
    ACTIONS(440), 1,
      anon_sym_STAR,
    ACTIONS(444), 1,
      anon_sym_LT_DASH,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(973), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(446), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [21105] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(550), 1,
      sym_identifier,
    ACTIONS(552), 1,
      anon_sym_LPAREN,
    ACTIONS(554), 1,
      anon_sym_func,
    ACTIONS(556), 1,
      anon_sym_STAR,
    ACTIONS(558), 1,
      anon_sym_LT_DASH,
    ACTIONS(564), 1,
      anon_sym_BQUOTE,
    ACTIONS(566), 1,
      anon_sym_DQUOTE,
    ACTIONS(574), 1,
      anon_sym_chan,
    STATE(343), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1037), 1,
      sym_qualified_type,
    STATE(1118), 1,
      sym_map_type,
    STATE(1321), 1,
      sym_implicit_length_array_type,
    ACTIONS(560), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(570), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1339), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1117), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(562), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(568), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(348), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [21218] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(550), 1,
      sym_identifier,
    ACTIONS(552), 1,
      anon_sym_LPAREN,
    ACTIONS(554), 1,
      anon_sym_func,
    ACTIONS(556), 1,
      anon_sym_STAR,
    ACTIONS(558), 1,
      anon_sym_LT_DASH,
    ACTIONS(564), 1,
      anon_sym_BQUOTE,
    ACTIONS(566), 1,
      anon_sym_DQUOTE,
    STATE(343), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1037), 1,
      sym_qualified_type,
    STATE(1118), 1,
      sym_map_type,
    STATE(1321), 1,
      sym_implicit_length_array_type,
    ACTIONS(560), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(570), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(973), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1117), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(562), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(568), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(348), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [21331] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(374), 1,
      sym_identifier,
    ACTIONS(376), 1,
      anon_sym_LPAREN,
    ACTIONS(378), 1,
      anon_sym_func,
    ACTIONS(380), 1,
      anon_sym_STAR,
    ACTIONS(382), 1,
      anon_sym_LT_DASH,
    ACTIONS(388), 1,
      anon_sym_BQUOTE,
    ACTIONS(390), 1,
      anon_sym_DQUOTE,
    ACTIONS(574), 1,
      anon_sym_chan,
    STATE(499), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1058), 1,
      sym_qualified_type,
    STATE(1138), 1,
      sym_map_type,
    STATE(1281), 1,
      sym_implicit_length_array_type,
    ACTIONS(384), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(394), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1269), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1137), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(386), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(392), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(544), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [21444] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(400), 1,
      sym_identifier,
    ACTIONS(402), 1,
      anon_sym_STAR,
    STATE(391), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(973), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [21557] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(374), 1,
      sym_identifier,
    ACTIONS(376), 1,
      anon_sym_LPAREN,
    ACTIONS(378), 1,
      anon_sym_func,
    ACTIONS(380), 1,
      anon_sym_STAR,
    ACTIONS(382), 1,
      anon_sym_LT_DASH,
    ACTIONS(388), 1,
      anon_sym_BQUOTE,
    ACTIONS(390), 1,
      anon_sym_DQUOTE,
    STATE(499), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1058), 1,
      sym_qualified_type,
    STATE(1138), 1,
      sym_map_type,
    STATE(1281), 1,
      sym_implicit_length_array_type,
    ACTIONS(384), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(394), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(973), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1137), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(386), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(392), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(544), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [21670] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(572), 1,
      anon_sym_STAR,
    STATE(667), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1193), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [21783] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(604), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [21896] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(648), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [22009] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(654), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [22122] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(572), 1,
      anon_sym_STAR,
    STATE(700), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1193), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [22235] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(617), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [22348] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(649), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [22461] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(650), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [22574] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(572), 1,
      anon_sym_STAR,
    STATE(651), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1193), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [22687] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(624), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [22800] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(662), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [22913] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(663), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [23026] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(572), 1,
      anon_sym_STAR,
    STATE(664), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1193), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [23139] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(630), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [23252] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(676), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [23365] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(677), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [23478] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    ACTIONS(572), 1,
      anon_sym_STAR,
    STATE(678), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1193), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [23591] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    ACTIONS(406), 1,
      sym_identifier,
    ACTIONS(410), 1,
      anon_sym_STAR,
    ACTIONS(412), 1,
      anon_sym_LT_DASH,
    STATE(636), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(414), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [23704] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(690), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [23817] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(302), 1,
      sym_identifier,
    ACTIONS(304), 1,
      anon_sym_LPAREN,
    ACTIONS(308), 1,
      anon_sym_func,
    ACTIONS(310), 1,
      anon_sym_STAR,
    ACTIONS(316), 1,
      anon_sym_LT_DASH,
    ACTIONS(322), 1,
      anon_sym_BQUOTE,
    ACTIONS(324), 1,
      anon_sym_DQUOTE,
    STATE(691), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(867), 1,
      sym_qualified_type,
    STATE(895), 1,
      sym_map_type,
    STATE(1323), 1,
      sym_implicit_length_array_type,
    ACTIONS(318), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(328), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1336), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(892), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(320), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(326), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(435), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [23930] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(550), 1,
      sym_identifier,
    ACTIONS(552), 1,
      anon_sym_LPAREN,
    ACTIONS(554), 1,
      anon_sym_func,
    ACTIONS(556), 1,
      anon_sym_STAR,
    ACTIONS(558), 1,
      anon_sym_LT_DASH,
    ACTIONS(564), 1,
      anon_sym_BQUOTE,
    ACTIONS(566), 1,
      anon_sym_DQUOTE,
    STATE(341), 1,
      sym__expression,
    STATE(847), 1,
      sym_negated_type,
    STATE(1037), 1,
      sym_qualified_type,
    STATE(1118), 1,
      sym_map_type,
    STATE(1321), 1,
      sym_implicit_length_array_type,
    ACTIONS(560), 2,
      anon_sym_new,
      anon_sym_make,
    ACTIONS(570), 2,
      sym_imaginary_literal,
      sym_rune_literal,
    STATE(1339), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 4,
      sym_pointer_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    STATE(1117), 4,
      sym_generic_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
    ACTIONS(562), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
    ACTIONS(568), 6,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
    STATE(348), 14,
      sym_parenthesized_expression,
      sym_call_expression,
      sym_selector_expression,
      sym_index_expression,
      sym_slice_expression,
      sym_type_assertion_expression,
      sym_type_conversion_expression,
      sym_type_instantiation_expression,
      sym_composite_literal,
      sym_func_literal,
      sym_unary_expression,
      sym_binary_expression,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [24043] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(580), 1,
      anon_sym_DOT,
    ACTIONS(583), 1,
      anon_sym_LPAREN,
    ACTIONS(587), 1,
      anon_sym_COMMA,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(591), 1,
      anon_sym_LBRACK,
    ACTIONS(597), 1,
      anon_sym_STAR,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    STATE(398), 1,
      sym_literal_value,
    STATE(543), 1,
      aux_sym_var_spec_repeat1,
    STATE(869), 1,
      sym_type_arguments,
    STATE(879), 1,
      sym_map_type,
    STATE(1328), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    ACTIONS(600), 5,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    ACTIONS(595), 14,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [24145] = 10,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(595), 1,
      aux_sym_source_file_token1,
    ACTIONS(608), 1,
      anon_sym_DOT,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(614), 1,
      anon_sym_LBRACK,
    ACTIONS(618), 1,
      anon_sym_LBRACE,
    ACTIONS(620), 1,
      anon_sym_COLON,
    STATE(285), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(600), 41,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [24216] = 10,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(595), 1,
      aux_sym_source_file_token1,
    ACTIONS(608), 1,
      anon_sym_DOT,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(614), 1,
      anon_sym_LBRACK,
    ACTIONS(618), 1,
      anon_sym_LBRACE,
    ACTIONS(622), 1,
      anon_sym_COLON,
    STATE(285), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(600), 41,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [24287] = 9,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(595), 1,
      aux_sym_source_file_token1,
    ACTIONS(608), 1,
      anon_sym_DOT,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(614), 1,
      anon_sym_LBRACK,
    ACTIONS(618), 1,
      anon_sym_LBRACE,
    STATE(285), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(600), 41,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [24355] = 19,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(624), 1,
      aux_sym_source_file_token1,
    ACTIONS(628), 1,
      anon_sym_DOT,
    ACTIONS(630), 1,
      anon_sym_LPAREN,
    ACTIONS(632), 1,
      anon_sym_COMMA,
    ACTIONS(636), 1,
      anon_sym_LBRACK,
    ACTIONS(642), 1,
      anon_sym_LT_DASH,
    ACTIONS(644), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(646), 1,
      anon_sym_DASH_DASH,
    ACTIONS(650), 1,
      anon_sym_AMP_AMP,
    ACTIONS(652), 1,
      anon_sym_PIPE_PIPE,
    STATE(288), 1,
      sym_argument_list,
    STATE(798), 1,
      aux_sym_expression_list_repeat1,
    STATE(1313), 1,
      sym_type_arguments,
    ACTIONS(640), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(626), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
    ACTIONS(648), 6,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(638), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(634), 13,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [24443] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(79), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
      sym_imaginary_literal,
      sym_rune_literal,
    ACTIONS(654), 30,
      anon_sym_package,
      anon_sym_import,
      anon_sym_const,
      anon_sym_var,
      anon_sym_func,
      anon_sym_type,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_fallthrough,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_goto,
      anon_sym_return,
      anon_sym_go,
      anon_sym_defer,
      anon_sym_if,
      anon_sym_for,
      anon_sym_switch,
      anon_sym_select,
      anon_sym_new,
      anon_sym_make,
      sym_identifier,
      sym_int_literal,
      sym_float_literal,
      sym_nil,
      sym_true,
      sym_false,
      sym_iota,
  [24498] = 10,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(628), 1,
      anon_sym_DOT,
    ACTIONS(630), 1,
      anon_sym_LPAREN,
    ACTIONS(636), 1,
      anon_sym_LBRACK,
    ACTIONS(656), 1,
      aux_sym_source_file_token1,
    STATE(288), 1,
      sym_argument_list,
    STATE(1313), 1,
      sym_type_arguments,
    ACTIONS(640), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(638), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(658), 30,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [24567] = 9,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(628), 1,
      anon_sym_DOT,
    ACTIONS(630), 1,
      anon_sym_LPAREN,
    ACTIONS(636), 1,
      anon_sym_LBRACK,
    ACTIONS(656), 1,
      aux_sym_source_file_token1,
    STATE(288), 1,
      sym_argument_list,
    STATE(1313), 1,
      sym_type_arguments,
    ACTIONS(638), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(658), 34,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [24634] = 8,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(628), 1,
      anon_sym_DOT,
    ACTIONS(630), 1,
      anon_sym_LPAREN,
    ACTIONS(636), 1,
      anon_sym_LBRACK,
    ACTIONS(656), 1,
      aux_sym_source_file_token1,
    STATE(288), 1,
      sym_argument_list,
    STATE(1313), 1,
      sym_type_arguments,
    ACTIONS(658), 41,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [24699] = 11,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(628), 1,
      anon_sym_DOT,
    ACTIONS(630), 1,
      anon_sym_LPAREN,
    ACTIONS(636), 1,
      anon_sym_LBRACK,
    ACTIONS(656), 1,
      aux_sym_source_file_token1,
    STATE(288), 1,
      sym_argument_list,
    STATE(1313), 1,
      sym_type_arguments,
    ACTIONS(640), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(648), 6,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(638), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(658), 24,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [24770] = 8,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(628), 1,
      anon_sym_DOT,
    ACTIONS(630), 1,
      anon_sym_LPAREN,
    ACTIONS(636), 1,
      anon_sym_LBRACK,
    ACTIONS(660), 1,
      aux_sym_source_file_token1,
    STATE(288), 1,
      sym_argument_list,
    STATE(1313), 1,
      sym_type_arguments,
    ACTIONS(662), 41,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [24835] = 12,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(628), 1,
      anon_sym_DOT,
    ACTIONS(630), 1,
      anon_sym_LPAREN,
    ACTIONS(636), 1,
      anon_sym_LBRACK,
    ACTIONS(650), 1,
      anon_sym_AMP_AMP,
    ACTIONS(656), 1,
      aux_sym_source_file_token1,
    STATE(288), 1,
      sym_argument_list,
    STATE(1313), 1,
      sym_type_arguments,
    ACTIONS(640), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(648), 6,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(638), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(658), 23,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PIPE_PIPE,
  [24908] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(595), 1,
      aux_sym_source_file_token1,
    ACTIONS(664), 1,
      anon_sym_LPAREN,
    STATE(288), 1,
      sym_special_argument_list,
    ACTIONS(600), 43,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [24966] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(666), 1,
      aux_sym_source_file_token1,
    ACTIONS(668), 45,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25020] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(670), 1,
      aux_sym_source_file_token1,
    ACTIONS(672), 45,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25074] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(674), 1,
      aux_sym_source_file_token1,
    ACTIONS(676), 45,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25128] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(678), 1,
      aux_sym_source_file_token1,
    ACTIONS(680), 45,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_else,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25182] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(682), 1,
      aux_sym_source_file_token1,
    ACTIONS(684), 45,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_else,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25236] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(686), 1,
      aux_sym_source_file_token1,
    ACTIONS(688), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25289] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(595), 1,
      aux_sym_source_file_token1,
    ACTIONS(600), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25342] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(690), 1,
      aux_sym_source_file_token1,
    ACTIONS(692), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25395] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(694), 1,
      aux_sym_source_file_token1,
    ACTIONS(696), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25448] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(698), 1,
      aux_sym_source_file_token1,
    ACTIONS(700), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25501] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(702), 1,
      aux_sym_source_file_token1,
    ACTIONS(704), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25554] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(706), 1,
      aux_sym_source_file_token1,
    ACTIONS(708), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25607] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(710), 1,
      aux_sym_source_file_token1,
    ACTIONS(712), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25660] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(714), 1,
      aux_sym_source_file_token1,
    ACTIONS(716), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25713] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(718), 1,
      aux_sym_source_file_token1,
    ACTIONS(720), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25766] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(722), 1,
      aux_sym_source_file_token1,
    ACTIONS(724), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25819] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(726), 1,
      aux_sym_source_file_token1,
    ACTIONS(728), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25872] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(730), 1,
      aux_sym_source_file_token1,
    ACTIONS(732), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25925] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(734), 1,
      aux_sym_source_file_token1,
    ACTIONS(736), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [25978] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(738), 1,
      aux_sym_source_file_token1,
    ACTIONS(740), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26031] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(742), 1,
      aux_sym_source_file_token1,
    ACTIONS(744), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26084] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(746), 1,
      aux_sym_source_file_token1,
    ACTIONS(748), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26137] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(33), 1,
      anon_sym_LBRACE,
    ACTIONS(624), 1,
      anon_sym_SEMI,
    ACTIONS(634), 1,
      anon_sym_EQ,
    ACTIONS(750), 1,
      anon_sym_DOT,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    ACTIONS(754), 1,
      anon_sym_COMMA,
    ACTIONS(756), 1,
      anon_sym_LBRACK,
    ACTIONS(762), 1,
      anon_sym_LT_DASH,
    ACTIONS(766), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(768), 1,
      anon_sym_DASH_DASH,
    ACTIONS(774), 1,
      anon_sym_AMP_AMP,
    ACTIONS(776), 1,
      anon_sym_PIPE_PIPE,
    STATE(312), 1,
      sym_argument_list,
    STATE(798), 1,
      aux_sym_expression_list_repeat1,
    STATE(910), 1,
      sym_block,
    STATE(1274), 1,
      sym_type_arguments,
    ACTIONS(772), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(760), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(770), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(758), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(764), 12,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [26228] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(778), 1,
      aux_sym_source_file_token1,
    ACTIONS(780), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26281] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(782), 1,
      aux_sym_source_file_token1,
    ACTIONS(784), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26334] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(786), 1,
      aux_sym_source_file_token1,
    ACTIONS(788), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26387] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(790), 1,
      aux_sym_source_file_token1,
    ACTIONS(792), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26440] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(794), 1,
      aux_sym_source_file_token1,
    ACTIONS(796), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26493] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(798), 1,
      aux_sym_source_file_token1,
    ACTIONS(800), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26546] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(802), 1,
      aux_sym_source_file_token1,
    ACTIONS(804), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26599] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(806), 1,
      aux_sym_source_file_token1,
    ACTIONS(808), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26652] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(810), 1,
      aux_sym_source_file_token1,
    ACTIONS(812), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26705] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(33), 1,
      anon_sym_LBRACE,
    ACTIONS(624), 1,
      anon_sym_SEMI,
    ACTIONS(634), 1,
      anon_sym_EQ,
    ACTIONS(750), 1,
      anon_sym_DOT,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    ACTIONS(754), 1,
      anon_sym_COMMA,
    ACTIONS(756), 1,
      anon_sym_LBRACK,
    ACTIONS(762), 1,
      anon_sym_LT_DASH,
    ACTIONS(766), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(768), 1,
      anon_sym_DASH_DASH,
    ACTIONS(774), 1,
      anon_sym_AMP_AMP,
    ACTIONS(776), 1,
      anon_sym_PIPE_PIPE,
    STATE(312), 1,
      sym_argument_list,
    STATE(798), 1,
      aux_sym_expression_list_repeat1,
    STATE(979), 1,
      sym_block,
    STATE(1274), 1,
      sym_type_arguments,
    ACTIONS(772), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(760), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(770), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(758), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(764), 12,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [26796] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(814), 1,
      aux_sym_source_file_token1,
    ACTIONS(816), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26849] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(818), 1,
      aux_sym_source_file_token1,
    ACTIONS(820), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26902] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(822), 1,
      aux_sym_source_file_token1,
    ACTIONS(824), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [26955] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(826), 1,
      aux_sym_source_file_token1,
    ACTIONS(828), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27008] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(830), 1,
      aux_sym_source_file_token1,
    ACTIONS(832), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27061] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(834), 1,
      aux_sym_source_file_token1,
    ACTIONS(836), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27114] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(838), 1,
      aux_sym_source_file_token1,
    ACTIONS(840), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27167] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(842), 1,
      aux_sym_source_file_token1,
    ACTIONS(844), 44,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27220] = 21,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_SEMI,
    ACTIONS(634), 1,
      anon_sym_EQ,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    ACTIONS(754), 1,
      anon_sym_COMMA,
    ACTIONS(756), 1,
      anon_sym_LBRACK,
    ACTIONS(762), 1,
      anon_sym_LT_DASH,
    ACTIONS(766), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(768), 1,
      anon_sym_DASH_DASH,
    ACTIONS(774), 1,
      anon_sym_AMP_AMP,
    ACTIONS(776), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(846), 1,
      anon_sym_DOT,
    ACTIONS(848), 1,
      anon_sym_LBRACE,
    STATE(312), 1,
      sym_argument_list,
    STATE(798), 1,
      aux_sym_expression_list_repeat1,
    STATE(1274), 1,
      sym_type_arguments,
    ACTIONS(772), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(760), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(770), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(758), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(764), 12,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [27308] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(750), 1,
      anon_sym_DOT,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    ACTIONS(756), 1,
      anon_sym_LBRACK,
    STATE(312), 1,
      sym_argument_list,
    STATE(1274), 1,
      sym_type_arguments,
    ACTIONS(658), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(758), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27371] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 1,
      anon_sym_EQ,
    ACTIONS(750), 1,
      anon_sym_DOT,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    ACTIONS(756), 1,
      anon_sym_LBRACK,
    ACTIONS(774), 1,
      anon_sym_AMP_AMP,
    STATE(312), 1,
      sym_argument_list,
    STATE(1274), 1,
      sym_type_arguments,
    ACTIONS(772), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(760), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(770), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(758), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_PIPE,
  [27442] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(750), 1,
      anon_sym_DOT,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    ACTIONS(756), 1,
      anon_sym_LBRACK,
    STATE(312), 1,
      sym_argument_list,
    STATE(1274), 1,
      sym_type_arguments,
    ACTIONS(658), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(656), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27503] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(580), 1,
      anon_sym_DOT,
    ACTIONS(850), 1,
      anon_sym_LPAREN,
    ACTIONS(853), 1,
      anon_sym_LBRACK,
    STATE(314), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(600), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27564] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(750), 1,
      anon_sym_DOT,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    ACTIONS(756), 1,
      anon_sym_LBRACK,
    STATE(312), 1,
      sym_argument_list,
    STATE(1274), 1,
      sym_type_arguments,
    ACTIONS(662), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(660), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27625] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(624), 1,
      anon_sym_LBRACE,
    ACTIONS(634), 1,
      anon_sym_EQ,
    ACTIONS(750), 1,
      anon_sym_DOT,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    ACTIONS(754), 1,
      anon_sym_COMMA,
    ACTIONS(756), 1,
      anon_sym_LBRACK,
    ACTIONS(766), 1,
      anon_sym_PLUS_PLUS,
    ACTIONS(768), 1,
      anon_sym_DASH_DASH,
    ACTIONS(774), 1,
      anon_sym_AMP_AMP,
    ACTIONS(776), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(857), 1,
      anon_sym_LT_DASH,
    STATE(312), 1,
      sym_argument_list,
    STATE(798), 1,
      aux_sym_expression_list_repeat1,
    STATE(1274), 1,
      sym_type_arguments,
    ACTIONS(772), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(760), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(770), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(758), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(764), 12,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [27710] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(750), 1,
      anon_sym_DOT,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    ACTIONS(756), 1,
      anon_sym_LBRACK,
    STATE(312), 1,
      sym_argument_list,
    STATE(1274), 1,
      sym_type_arguments,
    ACTIONS(658), 3,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(760), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(758), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27775] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 1,
      anon_sym_EQ,
    ACTIONS(750), 1,
      anon_sym_DOT,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    ACTIONS(756), 1,
      anon_sym_LBRACK,
    STATE(312), 1,
      sym_argument_list,
    STATE(1274), 1,
      sym_type_arguments,
    ACTIONS(772), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(760), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(770), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(758), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 20,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27844] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(859), 1,
      anon_sym_LPAREN,
    STATE(312), 1,
      sym_special_argument_list,
    ACTIONS(600), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 26,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27898] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(722), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27947] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(712), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(710), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [27996] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(716), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(714), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28045] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(718), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28094] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(728), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(726), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28143] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(680), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(678), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28192] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(732), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(730), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28241] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(780), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(778), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28290] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(788), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(786), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28339] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(792), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(790), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28388] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(796), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(794), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28437] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(800), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(798), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28486] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(840), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(838), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28535] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(804), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(802), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28584] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(828), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(826), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28633] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(808), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(806), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28682] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(812), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(810), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28731] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(816), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(814), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28780] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(824), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(822), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28829] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(820), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(818), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28878] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(832), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(830), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28927] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(836), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(834), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [28976] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(704), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(702), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29025] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(696), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(694), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29074] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(748), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(746), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29123] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(740), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(738), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29172] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(736), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(734), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29221] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(844), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(842), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29270] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(684), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(682), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29319] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(672), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(670), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29368] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(700), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(698), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29417] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(600), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29466] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(676), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(674), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29515] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(668), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(666), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29564] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(784), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(782), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29613] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(744), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(742), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29662] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(692), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(690), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29711] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(688), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(686), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29760] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(708), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(706), 27,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29809] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(580), 1,
      anon_sym_DOT,
    ACTIONS(850), 1,
      anon_sym_LPAREN,
    ACTIONS(853), 1,
      anon_sym_LBRACK,
    ACTIONS(861), 1,
      anon_sym_LBRACE,
    STATE(360), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(600), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 19,
      anon_sym_COMMA,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29868] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(863), 1,
      anon_sym_DOT,
    ACTIONS(865), 1,
      anon_sym_LPAREN,
    ACTIONS(869), 1,
      anon_sym_EQ,
    ACTIONS(871), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_AMP_AMP,
    ACTIONS(883), 1,
      anon_sym_PIPE_PIPE,
    STATE(358), 1,
      sym_argument_list,
    STATE(1271), 1,
      sym_type_arguments,
    ACTIONS(879), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(875), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(877), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(873), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(867), 13,
      anon_sym_COMMA,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [29936] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(863), 1,
      anon_sym_DOT,
    ACTIONS(865), 1,
      anon_sym_LPAREN,
    ACTIONS(871), 1,
      anon_sym_LBRACK,
    STATE(358), 1,
      sym_argument_list,
    STATE(1271), 1,
      sym_type_arguments,
    ACTIONS(658), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(873), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 19,
      anon_sym_COMMA,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [29994] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 1,
      anon_sym_EQ,
    ACTIONS(863), 1,
      anon_sym_DOT,
    ACTIONS(865), 1,
      anon_sym_LPAREN,
    ACTIONS(871), 1,
      anon_sym_LBRACK,
    STATE(358), 1,
      sym_argument_list,
    STATE(1271), 1,
      sym_type_arguments,
    ACTIONS(879), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(875), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(877), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(873), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 15,
      anon_sym_COMMA,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30058] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(863), 1,
      anon_sym_DOT,
    ACTIONS(865), 1,
      anon_sym_LPAREN,
    ACTIONS(871), 1,
      anon_sym_LBRACK,
    STATE(358), 1,
      sym_argument_list,
    STATE(1271), 1,
      sym_type_arguments,
    ACTIONS(662), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(660), 19,
      anon_sym_COMMA,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30114] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(863), 1,
      anon_sym_DOT,
    ACTIONS(865), 1,
      anon_sym_LPAREN,
    ACTIONS(871), 1,
      anon_sym_LBRACK,
    STATE(358), 1,
      sym_argument_list,
    STATE(1271), 1,
      sym_type_arguments,
    ACTIONS(658), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(656), 19,
      anon_sym_COMMA,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30170] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(658), 1,
      anon_sym_EQ,
    ACTIONS(863), 1,
      anon_sym_DOT,
    ACTIONS(865), 1,
      anon_sym_LPAREN,
    ACTIONS(871), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_AMP_AMP,
    STATE(358), 1,
      sym_argument_list,
    STATE(1271), 1,
      sym_type_arguments,
    ACTIONS(879), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(875), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(877), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(873), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 14,
      anon_sym_COMMA,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_PIPE,
  [30236] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(863), 1,
      anon_sym_DOT,
    ACTIONS(865), 1,
      anon_sym_LPAREN,
    ACTIONS(871), 1,
      anon_sym_LBRACK,
    STATE(358), 1,
      sym_argument_list,
    STATE(1271), 1,
      sym_type_arguments,
    ACTIONS(658), 3,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(875), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(873), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 19,
      anon_sym_COMMA,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30296] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(885), 1,
      anon_sym_LPAREN,
    STATE(358), 1,
      sym_special_argument_list,
    ACTIONS(600), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 21,
      anon_sym_DOT,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30345] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(600), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30389] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(676), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(674), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30433] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(732), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(730), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30477] = 18,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(887), 1,
      sym_identifier,
    ACTIONS(889), 1,
      aux_sym_source_file_token1,
    ACTIONS(893), 1,
      anon_sym_LPAREN,
    ACTIONS(895), 1,
      anon_sym_func,
    ACTIONS(897), 1,
      anon_sym_LBRACK,
    ACTIONS(899), 1,
      anon_sym_STAR,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(903), 1,
      anon_sym_TILDE,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(907), 1,
      anon_sym_map,
    ACTIONS(909), 1,
      anon_sym_chan,
    ACTIONS(911), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(803), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(811), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    ACTIONS(891), 11,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [30551] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(780), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(778), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30595] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(788), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(786), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30639] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(680), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(678), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30683] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(792), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(790), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30727] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(796), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(794), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30771] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(800), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(798), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30815] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(840), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(838), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30859] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(668), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(666), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30903] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(828), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(826), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30947] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(716), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(714), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [30991] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(824), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(822), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31035] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(718), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31079] = 18,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(889), 1,
      aux_sym_source_file_token1,
    ACTIONS(893), 1,
      anon_sym_LPAREN,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(903), 1,
      anon_sym_TILDE,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(917), 1,
      anon_sym_LBRACK,
    ACTIONS(919), 1,
      anon_sym_STAR,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(925), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(811), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
    ACTIONS(891), 11,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [31153] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(832), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(830), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31197] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(804), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(802), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31241] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(820), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(818), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31285] = 25,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(903), 1,
      anon_sym_TILDE,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(919), 1,
      anon_sym_STAR,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(925), 1,
      anon_sym_LT_DASH,
    ACTIONS(927), 1,
      aux_sym_source_file_token1,
    ACTIONS(931), 1,
      anon_sym_DOT,
    ACTIONS(933), 1,
      anon_sym_LPAREN,
    ACTIONS(935), 1,
      anon_sym_COMMA,
    ACTIONS(937), 1,
      anon_sym_LBRACK,
    ACTIONS(939), 1,
      anon_sym_BQUOTE,
    ACTIONS(941), 1,
      anon_sym_DQUOTE,
    STATE(626), 1,
      aux_sym_field_declaration_repeat1,
    STATE(816), 1,
      sym_map_type,
    STATE(821), 1,
      sym_type_arguments,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(894), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1107), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
    ACTIONS(929), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [31373] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(736), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(734), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31417] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(844), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(842), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31461] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(684), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(682), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31505] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(836), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(834), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31549] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(700), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(698), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31593] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(784), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(782), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31637] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(704), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(702), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31681] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(672), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(670), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31725] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(712), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(710), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31769] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(722), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31813] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(740), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(738), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31857] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(728), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(726), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31901] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(708), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(706), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31945] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(812), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(810), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [31989] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(748), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(746), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32033] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(696), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(694), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32077] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(744), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(742), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32121] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(816), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(814), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32165] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(692), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(690), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32209] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(688), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(686), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32253] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(808), 14,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(806), 22,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32297] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(658), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(656), 22,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32349] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(662), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(660), 22,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32401] = 21,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(286), 1,
      anon_sym_TILDE,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(949), 1,
      aux_sym_source_file_token1,
    ACTIONS(953), 1,
      anon_sym_LPAREN,
    ACTIONS(955), 1,
      anon_sym_COMMA,
    ACTIONS(957), 1,
      anon_sym_EQ,
    ACTIONS(959), 1,
      anon_sym_LBRACK,
    ACTIONS(961), 1,
      anon_sym_STAR,
    ACTIONS(963), 1,
      anon_sym_LT_DASH,
    STATE(787), 1,
      aux_sym_const_spec_repeat1,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1357), 2,
      sym_parenthesized_type,
      sym__simple_type,
    ACTIONS(951), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [32478] = 21,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(286), 1,
      anon_sym_TILDE,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(953), 1,
      anon_sym_LPAREN,
    ACTIONS(955), 1,
      anon_sym_COMMA,
    ACTIONS(959), 1,
      anon_sym_LBRACK,
    ACTIONS(961), 1,
      anon_sym_STAR,
    ACTIONS(963), 1,
      anon_sym_LT_DASH,
    ACTIONS(965), 1,
      aux_sym_source_file_token1,
    ACTIONS(969), 1,
      anon_sym_EQ,
    STATE(392), 1,
      aux_sym_const_spec_repeat1,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1378), 2,
      sym_parenthesized_type,
      sym__simple_type,
    ACTIONS(967), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [32555] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(971), 1,
      anon_sym_LPAREN,
    STATE(397), 1,
      sym_special_argument_list,
    ACTIONS(600), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 23,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32600] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(580), 1,
      anon_sym_DOT,
    ACTIONS(853), 1,
      anon_sym_LBRACK,
    STATE(398), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(850), 2,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
    ACTIONS(600), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32653] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(680), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(678), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32693] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(840), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(838), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32733] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(828), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(826), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32773] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(824), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(822), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32813] = 15,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(764), 1,
      aux_sym_source_file_token1,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(977), 1,
      anon_sym_COMMA,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    ACTIONS(987), 1,
      anon_sym_AMP_AMP,
    ACTIONS(989), 1,
      anon_sym_PIPE_PIPE,
    STATE(509), 1,
      sym_argument_list,
    STATE(889), 1,
      aux_sym_expression_list_repeat1,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(983), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(634), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
    ACTIONS(985), 6,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(981), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [32877] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(844), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(842), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32917] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(684), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(682), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32957] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(700), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(698), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [32997] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(784), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(782), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33037] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(722), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33077] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(740), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(738), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33117] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(744), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(742), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33157] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(692), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(690), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33197] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(688), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(686), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33237] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(708), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(706), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33277] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(712), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(710), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33317] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(716), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(714), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33357] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(718), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33397] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(728), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(726), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33437] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(732), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(730), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33477] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(780), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(778), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33517] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(788), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(786), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33557] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(792), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(790), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33597] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(796), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(794), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33637] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(800), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(798), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33677] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(804), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(802), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33717] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(808), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(806), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33757] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(812), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(810), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33797] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(816), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(814), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33837] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(820), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(818), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33877] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(832), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(830), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33917] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(836), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(834), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33957] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(704), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(702), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [33997] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(696), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(694), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34037] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(748), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(746), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34077] = 9,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(595), 1,
      aux_sym_source_file_token1,
    ACTIONS(608), 1,
      anon_sym_DOT,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(614), 1,
      anon_sym_LBRACK,
    ACTIONS(991), 1,
      anon_sym_LBRACE,
    STATE(512), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(600), 25,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34129] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(672), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(670), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34169] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(676), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(674), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34209] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(668), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(666), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34249] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(600), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34289] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(736), 8,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(734), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34329] = 13,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(867), 1,
      aux_sym_source_file_token1,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    ACTIONS(987), 1,
      anon_sym_AMP_AMP,
    ACTIONS(989), 1,
      anon_sym_PIPE_PIPE,
    STATE(509), 1,
      sym_argument_list,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(983), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(869), 6,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
    ACTIONS(985), 6,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(981), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [34388] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(658), 2,
      anon_sym_EQ,
      anon_sym_COLON,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PIPE_PIPE,
  [34451] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(877), 2,
      sym_parameter_list,
      sym__simple_type,
    ACTIONS(889), 7,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [34518] = 8,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(656), 1,
      aux_sym_source_file_token1,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    STATE(509), 1,
      sym_argument_list,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(658), 25,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34567] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(658), 5,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 14,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34620] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(658), 4,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 11,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34677] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(658), 2,
      anon_sym_EQ,
      anon_sym_COLON,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_COLON_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34738] = 9,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(656), 1,
      aux_sym_source_file_token1,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    STATE(509), 1,
      sym_argument_list,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(981), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(658), 18,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34789] = 10,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(656), 1,
      aux_sym_source_file_token1,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    STATE(509), 1,
      sym_argument_list,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(983), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(981), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(658), 14,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34842] = 11,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(656), 1,
      aux_sym_source_file_token1,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    STATE(509), 1,
      sym_argument_list,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(983), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(985), 6,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(981), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(658), 8,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [34897] = 12,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(656), 1,
      aux_sym_source_file_token1,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    ACTIONS(987), 1,
      anon_sym_AMP_AMP,
    STATE(509), 1,
      sym_argument_list,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(983), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(985), 6,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(658), 7,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PIPE_PIPE,
    ACTIONS(981), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [34954] = 8,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(660), 1,
      aux_sym_source_file_token1,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    STATE(509), 1,
      sym_argument_list,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(662), 25,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_COMMA,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35003] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(877), 2,
      sym_parameter_list,
      sym__simple_type,
    ACTIONS(889), 7,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [35070] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(595), 1,
      aux_sym_source_file_token1,
    ACTIONS(1025), 1,
      anon_sym_LPAREN,
    STATE(509), 1,
      sym_special_argument_list,
    ACTIONS(600), 27,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35112] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(634), 1,
      anon_sym_EQ,
    ACTIONS(762), 1,
      anon_sym_LT_DASH,
    ACTIONS(764), 1,
      anon_sym_COLON_EQ,
    ACTIONS(1027), 1,
      anon_sym_DOT,
    ACTIONS(1029), 1,
      anon_sym_LPAREN,
    ACTIONS(1031), 1,
      anon_sym_COMMA,
    ACTIONS(1033), 1,
      anon_sym_LBRACK,
    ACTIONS(1037), 1,
      anon_sym_PIPE,
    ACTIONS(1039), 1,
      anon_sym_COLON,
    ACTIONS(1049), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1051), 1,
      anon_sym_PIPE_PIPE,
    STATE(547), 1,
      sym_argument_list,
    STATE(943), 1,
      aux_sym_expression_list_repeat1,
    STATE(1286), 1,
      sym_type_arguments,
    ACTIONS(1043), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1047), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1041), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1045), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1035), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [35184] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1053), 1,
      anon_sym_LBRACE,
    STATE(401), 1,
      sym_block,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(901), 2,
      sym_parameter_list,
      sym__simple_type,
    ACTIONS(889), 4,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_PIPE,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [35254] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1055), 1,
      anon_sym_DOT,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1059), 1,
      anon_sym_RPAREN,
    ACTIONS(1061), 1,
      anon_sym_COMMA,
    ACTIONS(1064), 1,
      anon_sym_LBRACK,
    ACTIONS(1066), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(612), 1,
      aux_sym_var_spec_repeat1,
    STATE(869), 1,
      sym_type_arguments,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1275), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [35330] = 20,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1074), 1,
      anon_sym_LPAREN,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1078), 1,
      anon_sym_LBRACK,
    ACTIONS(1080), 1,
      anon_sym_STAR,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1084), 1,
      anon_sym_TILDE,
    ACTIONS(1086), 1,
      anon_sym_LBRACE,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1094), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(1086), 1,
      sym_block,
    ACTIONS(1068), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1072), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(956), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [35402] = 13,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    ACTIONS(987), 1,
      anon_sym_AMP_AMP,
    ACTIONS(989), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1096), 1,
      aux_sym_source_file_token1,
    STATE(509), 1,
      sym_argument_list,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(983), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1098), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
    ACTIONS(985), 6,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(981), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [35460] = 13,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    ACTIONS(987), 1,
      anon_sym_AMP_AMP,
    ACTIONS(989), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1100), 1,
      aux_sym_source_file_token1,
    STATE(509), 1,
      sym_argument_list,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(983), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1102), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
    ACTIONS(985), 6,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(981), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [35518] = 13,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(973), 1,
      anon_sym_DOT,
    ACTIONS(975), 1,
      anon_sym_LPAREN,
    ACTIONS(979), 1,
      anon_sym_LBRACK,
    ACTIONS(987), 1,
      anon_sym_AMP_AMP,
    ACTIONS(989), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1104), 1,
      aux_sym_source_file_token1,
    STATE(509), 1,
      sym_argument_list,
    STATE(1284), 1,
      sym_type_arguments,
    ACTIONS(983), 4,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1106), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
    ACTIONS(985), 6,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(981), 7,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [35576] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(580), 1,
      anon_sym_DOT,
    ACTIONS(850), 1,
      anon_sym_LPAREN,
    ACTIONS(853), 1,
      anon_sym_LBRACK,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    STATE(548), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(600), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 17,
      anon_sym_COMMA,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35626] = 20,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1074), 1,
      anon_sym_LPAREN,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1078), 1,
      anon_sym_LBRACK,
    ACTIONS(1080), 1,
      anon_sym_STAR,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1084), 1,
      anon_sym_TILDE,
    ACTIONS(1086), 1,
      anon_sym_LBRACE,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1094), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(1139), 1,
      sym_block,
    ACTIONS(1110), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1112), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1010), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [35698] = 20,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1074), 1,
      anon_sym_LPAREN,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1078), 1,
      anon_sym_LBRACK,
    ACTIONS(1080), 1,
      anon_sym_STAR,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1084), 1,
      anon_sym_TILDE,
    ACTIONS(1086), 1,
      anon_sym_LBRACE,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1094), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(1067), 1,
      sym_block,
    ACTIONS(1114), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1116), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(931), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [35770] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(730), 1,
      aux_sym_source_file_token1,
    ACTIONS(732), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35807] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(686), 1,
      aux_sym_source_file_token1,
    ACTIONS(688), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35844] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(706), 1,
      aux_sym_source_file_token1,
    ACTIONS(708), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35881] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(710), 1,
      aux_sym_source_file_token1,
    ACTIONS(712), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35918] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(714), 1,
      aux_sym_source_file_token1,
    ACTIONS(716), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35955] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(718), 1,
      aux_sym_source_file_token1,
    ACTIONS(720), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [35992] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(726), 1,
      aux_sym_source_file_token1,
    ACTIONS(728), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36029] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(778), 1,
      aux_sym_source_file_token1,
    ACTIONS(780), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36066] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(786), 1,
      aux_sym_source_file_token1,
    ACTIONS(788), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36103] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(790), 1,
      aux_sym_source_file_token1,
    ACTIONS(792), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36140] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(794), 1,
      aux_sym_source_file_token1,
    ACTIONS(796), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36177] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(798), 1,
      aux_sym_source_file_token1,
    ACTIONS(800), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36214] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(802), 1,
      aux_sym_source_file_token1,
    ACTIONS(804), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36251] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(806), 1,
      aux_sym_source_file_token1,
    ACTIONS(808), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36288] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(810), 1,
      aux_sym_source_file_token1,
    ACTIONS(812), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36325] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(814), 1,
      aux_sym_source_file_token1,
    ACTIONS(816), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36362] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(818), 1,
      aux_sym_source_file_token1,
    ACTIONS(820), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36399] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(830), 1,
      aux_sym_source_file_token1,
    ACTIONS(832), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36436] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(834), 1,
      aux_sym_source_file_token1,
    ACTIONS(836), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36473] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(702), 1,
      aux_sym_source_file_token1,
    ACTIONS(704), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36510] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(694), 1,
      aux_sym_source_file_token1,
    ACTIONS(696), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36547] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(746), 1,
      aux_sym_source_file_token1,
    ACTIONS(748), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36584] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(822), 1,
      aux_sym_source_file_token1,
    ACTIONS(824), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36621] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(670), 1,
      aux_sym_source_file_token1,
    ACTIONS(672), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36658] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(674), 1,
      aux_sym_source_file_token1,
    ACTIONS(676), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36695] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(666), 1,
      aux_sym_source_file_token1,
    ACTIONS(668), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36732] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(734), 1,
      aux_sym_source_file_token1,
    ACTIONS(736), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36769] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(869), 2,
      anon_sym_EQ,
      anon_sym_COLON,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(867), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_COLON_EQ,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [36832] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(842), 1,
      aux_sym_source_file_token1,
    ACTIONS(844), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36869] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(698), 1,
      aux_sym_source_file_token1,
    ACTIONS(700), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36906] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(782), 1,
      aux_sym_source_file_token1,
    ACTIONS(784), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [36943] = 20,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(893), 1,
      anon_sym_LPAREN,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(903), 1,
      anon_sym_TILDE,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(917), 1,
      anon_sym_LBRACK,
    ACTIONS(919), 1,
      anon_sym_STAR,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(925), 1,
      anon_sym_LT_DASH,
    ACTIONS(1068), 1,
      aux_sym_source_file_token1,
    ACTIONS(1120), 1,
      anon_sym_LBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(1208), 1,
      sym_block,
    ACTIONS(1072), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1059), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [37014] = 20,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(893), 1,
      anon_sym_LPAREN,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(903), 1,
      anon_sym_TILDE,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(917), 1,
      anon_sym_LBRACK,
    ACTIONS(919), 1,
      anon_sym_STAR,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(925), 1,
      anon_sym_LT_DASH,
    ACTIONS(1110), 1,
      aux_sym_source_file_token1,
    ACTIONS(1120), 1,
      anon_sym_LBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(1234), 1,
      sym_block,
    ACTIONS(1112), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1048), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [37085] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(722), 1,
      aux_sym_source_file_token1,
    ACTIONS(724), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [37122] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(690), 1,
      aux_sym_source_file_token1,
    ACTIONS(692), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [37159] = 18,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1074), 1,
      anon_sym_LPAREN,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1078), 1,
      anon_sym_LBRACK,
    ACTIONS(1080), 1,
      anon_sym_STAR,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1084), 1,
      anon_sym_TILDE,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1094), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    ACTIONS(889), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(966), 2,
      sym_parameter_list,
      sym__simple_type,
    ACTIONS(891), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACE,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [37226] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(595), 1,
      aux_sym_source_file_token1,
    ACTIONS(600), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [37263] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1122), 1,
      sym_identifier,
    ACTIONS(1124), 1,
      anon_sym_RPAREN,
    ACTIONS(1126), 1,
      anon_sym_COMMA,
    ACTIONS(1128), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1242), 2,
      sym_parameter_declaration,
      sym_variadic_parameter_declaration,
    STATE(1298), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [37334] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1027), 1,
      anon_sym_DOT,
    ACTIONS(1029), 1,
      anon_sym_LPAREN,
    ACTIONS(1033), 1,
      anon_sym_LBRACK,
    STATE(547), 1,
      sym_argument_list,
    STATE(1286), 1,
      sym_type_arguments,
    ACTIONS(662), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(660), 17,
      anon_sym_COMMA,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [37381] = 18,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1074), 1,
      anon_sym_LPAREN,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1084), 1,
      anon_sym_TILDE,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1130), 1,
      sym_identifier,
    ACTIONS(1132), 1,
      anon_sym_func,
    ACTIONS(1134), 1,
      anon_sym_LBRACK,
    ACTIONS(1136), 1,
      anon_sym_STAR,
    ACTIONS(1138), 1,
      anon_sym_map,
    ACTIONS(1140), 1,
      anon_sym_chan,
    ACTIONS(1142), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    ACTIONS(889), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    STATE(918), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(966), 2,
      sym_parameter_list,
      sym__simple_type,
    ACTIONS(891), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACE,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [37448] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1027), 1,
      anon_sym_DOT,
    ACTIONS(1029), 1,
      anon_sym_LPAREN,
    ACTIONS(1033), 1,
      anon_sym_LBRACK,
    STATE(547), 1,
      sym_argument_list,
    STATE(1286), 1,
      sym_type_arguments,
    ACTIONS(658), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(656), 17,
      anon_sym_COMMA,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [37495] = 20,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(893), 1,
      anon_sym_LPAREN,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(903), 1,
      anon_sym_TILDE,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(917), 1,
      anon_sym_LBRACK,
    ACTIONS(919), 1,
      anon_sym_STAR,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(925), 1,
      anon_sym_LT_DASH,
    ACTIONS(1114), 1,
      aux_sym_source_file_token1,
    ACTIONS(1120), 1,
      anon_sym_LBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(1256), 1,
      sym_block,
    ACTIONS(1116), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1049), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [37566] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(678), 1,
      aux_sym_source_file_token1,
    ACTIONS(680), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [37603] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(682), 1,
      aux_sym_source_file_token1,
    ACTIONS(684), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [37640] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1027), 1,
      anon_sym_DOT,
    ACTIONS(1029), 1,
      anon_sym_LPAREN,
    ACTIONS(1033), 1,
      anon_sym_LBRACK,
    STATE(547), 1,
      sym_argument_list,
    STATE(1286), 1,
      sym_type_arguments,
    ACTIONS(1043), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(658), 5,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1035), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 12,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [37691] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1027), 1,
      anon_sym_DOT,
    ACTIONS(1029), 1,
      anon_sym_LPAREN,
    ACTIONS(1033), 1,
      anon_sym_LBRACK,
    ACTIONS(1037), 1,
      anon_sym_PIPE,
    STATE(547), 1,
      sym_argument_list,
    STATE(1286), 1,
      sym_type_arguments,
    ACTIONS(1043), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1041), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(658), 4,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1035), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 9,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [37746] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1027), 1,
      anon_sym_DOT,
    ACTIONS(1029), 1,
      anon_sym_LPAREN,
    ACTIONS(1033), 1,
      anon_sym_LBRACK,
    ACTIONS(1037), 1,
      anon_sym_PIPE,
    STATE(547), 1,
      sym_argument_list,
    STATE(1286), 1,
      sym_type_arguments,
    ACTIONS(658), 2,
      anon_sym_EQ,
      anon_sym_COLON,
    ACTIONS(1043), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1047), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1041), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1045), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(656), 5,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1035), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [37805] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1027), 1,
      anon_sym_DOT,
    ACTIONS(1029), 1,
      anon_sym_LPAREN,
    ACTIONS(1033), 1,
      anon_sym_LBRACK,
    ACTIONS(1037), 1,
      anon_sym_PIPE,
    ACTIONS(1049), 1,
      anon_sym_AMP_AMP,
    STATE(547), 1,
      sym_argument_list,
    STATE(1286), 1,
      sym_type_arguments,
    ACTIONS(658), 2,
      anon_sym_EQ,
      anon_sym_COLON,
    ACTIONS(1043), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1047), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1041), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(656), 4,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PIPE_PIPE,
    ACTIONS(1045), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1035), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [37866] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(838), 1,
      aux_sym_source_file_token1,
    ACTIONS(840), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [37903] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1122), 1,
      sym_identifier,
    ACTIONS(1128), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1144), 1,
      anon_sym_RPAREN,
    ACTIONS(1146), 1,
      anon_sym_COMMA,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1246), 2,
      sym_parameter_declaration,
      sym_variadic_parameter_declaration,
    STATE(1298), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [37974] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1122), 1,
      sym_identifier,
    ACTIONS(1128), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1148), 1,
      anon_sym_RPAREN,
    ACTIONS(1150), 1,
      anon_sym_COMMA,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1231), 2,
      sym_parameter_declaration,
      sym_variadic_parameter_declaration,
    STATE(1298), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [38045] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(826), 1,
      aux_sym_source_file_token1,
    ACTIONS(828), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [38082] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(742), 1,
      aux_sym_source_file_token1,
    ACTIONS(744), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [38119] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(738), 1,
      aux_sym_source_file_token1,
    ACTIONS(740), 28,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [38156] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(580), 1,
      anon_sym_DOT,
    ACTIONS(611), 1,
      anon_sym_PIPE,
    ACTIONS(853), 1,
      anon_sym_LBRACK,
    ACTIONS(1059), 1,
      anon_sym_COMMA,
    STATE(398), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(850), 2,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
    ACTIONS(600), 4,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 15,
      anon_sym_STAR,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [38208] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(608), 1,
      anon_sym_DOT,
    ACTIONS(850), 1,
      anon_sym_LPAREN,
    ACTIONS(853), 1,
      anon_sym_LBRACK,
    STATE(398), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(600), 5,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT_DOT_DOT,
      anon_sym_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [38256] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1152), 1,
      anon_sym_LPAREN,
    STATE(547), 1,
      sym_special_argument_list,
    ACTIONS(600), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 19,
      anon_sym_DOT,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [38296] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1154), 1,
      anon_sym_RPAREN,
    ACTIONS(1156), 1,
      anon_sym_COMMA,
    ACTIONS(1158), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1267), 1,
      aux_sym_argument_list_repeat1,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [38362] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1122), 1,
      sym_identifier,
    ACTIONS(1128), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1176), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1298), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1301), 2,
      sym_parameter_declaration,
      sym_variadic_parameter_declaration,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [38430] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1122), 1,
      sym_identifier,
    ACTIONS(1128), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1178), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1298), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1301), 2,
      sym_parameter_declaration,
      sym_variadic_parameter_declaration,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [38498] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1122), 1,
      sym_identifier,
    ACTIONS(1128), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1180), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1298), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1301), 2,
      sym_parameter_declaration,
      sym_variadic_parameter_declaration,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [38566] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1031), 1,
      anon_sym_COMMA,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(943), 1,
      aux_sym_expression_list_repeat1,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(764), 2,
      anon_sym_SEMI,
      anon_sym_COLON,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [38630] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1122), 1,
      sym_identifier,
    ACTIONS(1128), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1182), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1298), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1301), 2,
      sym_parameter_declaration,
      sym_variadic_parameter_declaration,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [38698] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1184), 1,
      anon_sym_RPAREN,
    ACTIONS(1186), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1170), 1,
      aux_sym_argument_list_repeat1,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [38764] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1188), 1,
      anon_sym_RPAREN,
    ACTIONS(1190), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1264), 1,
      aux_sym_argument_list_repeat1,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [38830] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1122), 1,
      sym_identifier,
    ACTIONS(1128), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1192), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1298), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1301), 2,
      sym_parameter_declaration,
      sym_variadic_parameter_declaration,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [38898] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1122), 1,
      sym_identifier,
    ACTIONS(1128), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1194), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1298), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1301), 2,
      sym_parameter_declaration,
      sym_variadic_parameter_declaration,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [38966] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1196), 1,
      anon_sym_RPAREN,
    ACTIONS(1198), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1198), 1,
      aux_sym_argument_list_repeat1,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [39032] = 18,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(893), 1,
      anon_sym_LPAREN,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(903), 1,
      anon_sym_TILDE,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(917), 1,
      anon_sym_LBRACK,
    ACTIONS(919), 1,
      anon_sym_STAR,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(925), 1,
      anon_sym_LT_DASH,
    ACTIONS(1200), 1,
      aux_sym_source_file_token1,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1134), 2,
      sym_parameter_list,
      sym__simple_type,
    ACTIONS(1202), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [39098] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1204), 1,
      anon_sym_RPAREN,
    ACTIONS(1206), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1221), 1,
      aux_sym_argument_list_repeat1,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [39164] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1208), 1,
      anon_sym_RPAREN,
    ACTIONS(1210), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1240), 1,
      aux_sym_argument_list_repeat1,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [39230] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(764), 1,
      anon_sym_COLON_EQ,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1031), 1,
      anon_sym_COMMA,
    ACTIONS(1212), 1,
      anon_sym_DOT,
    ACTIONS(1216), 1,
      anon_sym_LBRACE,
    ACTIONS(1218), 1,
      anon_sym_PIPE,
    ACTIONS(1228), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1230), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(943), 1,
      aux_sym_expression_list_repeat1,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1226), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1220), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1224), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [39296] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(832), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(830), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [39331] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1232), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [39392] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1122), 1,
      sym_identifier,
    ACTIONS(1128), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1298), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1301), 2,
      sym_parameter_declaration,
      sym_variadic_parameter_declaration,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [39457] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(658), 3,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 12,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [39506] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(658), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 9,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT_DOT_DOT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [39559] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(656), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [39614] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(943), 1,
      anon_sym_DOT,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(656), 4,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE_PIPE,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [39671] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(587), 1,
      anon_sym_COMMA,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1236), 1,
      anon_sym_EQ,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    STATE(802), 1,
      aux_sym_var_spec_repeat1,
    STATE(816), 1,
      sym_map_type,
    STATE(915), 1,
      sym_parenthesized_type,
    STATE(917), 1,
      sym__simple_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [39740] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(587), 1,
      anon_sym_COMMA,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1246), 1,
      anon_sym_EQ,
    STATE(540), 1,
      aux_sym_var_spec_repeat1,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(914), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [39807] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1248), 3,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_COLON,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [39866] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(587), 1,
      anon_sym_COMMA,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(802), 1,
      aux_sym_var_spec_repeat1,
    STATE(879), 1,
      sym_map_type,
    STATE(1335), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [39933] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(600), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [39968] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(580), 1,
      anon_sym_DOT,
    ACTIONS(850), 1,
      anon_sym_LPAREN,
    ACTIONS(853), 1,
      anon_sym_LBRACK,
    STATE(398), 1,
      sym_literal_value,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(600), 5,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(595), 17,
      anon_sym_COMMA,
      anon_sym_STAR,
      anon_sym_LBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40013] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(680), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(678), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40048] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(840), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(838), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40083] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(828), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(826), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40118] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(824), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(822), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40153] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(736), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(734), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40188] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(844), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(842), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40223] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(684), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(682), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40258] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(700), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(698), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40293] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(784), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(782), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40328] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(724), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(722), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40363] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(740), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(738), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40398] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(744), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(742), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40433] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(692), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(690), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40468] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(688), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(686), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40503] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(708), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(706), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40538] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(712), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(710), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40573] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(716), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(714), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40608] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(720), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(718), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40643] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(728), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(726), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40678] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(732), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(730), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40713] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(780), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(778), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40748] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(788), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(786), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40783] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(792), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(790), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40818] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1250), 1,
      sym_identifier,
    ACTIONS(1252), 1,
      anon_sym_RBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(955), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1045), 2,
      sym_method_elem,
      sym_type_elem,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [40883] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(800), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(798), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40918] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(804), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(802), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40953] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(808), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(806), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [40988] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(812), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(810), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41023] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(816), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(814), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41058] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(820), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(818), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41093] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(836), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(834), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41128] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(704), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(702), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41163] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(696), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(694), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41198] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(748), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(746), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41233] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(672), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(670), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41268] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(676), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(674), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41303] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(668), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(666), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41338] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1250), 1,
      sym_identifier,
    ACTIONS(1254), 1,
      anon_sym_RBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(955), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1041), 2,
      sym_method_elem,
      sym_type_elem,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [41403] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1250), 1,
      sym_identifier,
    ACTIONS(1256), 1,
      anon_sym_RBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(955), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1145), 2,
      sym_method_elem,
      sym_type_elem,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [41468] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1250), 1,
      sym_identifier,
    ACTIONS(1258), 1,
      anon_sym_RBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(955), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1145), 2,
      sym_method_elem,
      sym_type_elem,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [41533] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(658), 3,
      anon_sym_PIPE,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 12,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41582] = 12,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1218), 1,
      anon_sym_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(658), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1220), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
    ACTIONS(656), 9,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_COLON_EQ,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [41635] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1218), 1,
      anon_sym_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1226), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1220), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1224), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(656), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_COLON_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [41690] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1218), 1,
      anon_sym_PIPE,
    ACTIONS(1228), 1,
      anon_sym_AMP_AMP,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1226), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1220), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(656), 4,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_COLON_EQ,
      anon_sym_PIPE_PIPE,
    ACTIONS(1224), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [41747] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1250), 1,
      sym_identifier,
    ACTIONS(1260), 1,
      anon_sym_RBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(955), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1145), 2,
      sym_method_elem,
      sym_type_elem,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [41812] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(587), 1,
      anon_sym_COMMA,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(543), 1,
      aux_sym_var_spec_repeat1,
    STATE(879), 1,
      sym_map_type,
    STATE(1328), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [41879] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1250), 1,
      sym_identifier,
    ACTIONS(1262), 1,
      anon_sym_RBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(955), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1063), 2,
      sym_method_elem,
      sym_type_elem,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [41944] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(764), 1,
      anon_sym_LBRACE,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1218), 1,
      anon_sym_PIPE,
    ACTIONS(1228), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1230), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1264), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1244), 1,
      aux_sym_expression_list_repeat1,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1226), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1220), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1224), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [42007] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1250), 1,
      sym_identifier,
    ACTIONS(1266), 1,
      anon_sym_RBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(955), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1145), 2,
      sym_method_elem,
      sym_type_elem,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [42072] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1250), 1,
      sym_identifier,
    ACTIONS(1268), 1,
      anon_sym_RBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(955), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1145), 2,
      sym_method_elem,
      sym_type_elem,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [42137] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1250), 1,
      sym_identifier,
    ACTIONS(1270), 1,
      anon_sym_RBRACE,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(955), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1145), 2,
      sym_method_elem,
      sym_type_elem,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [42202] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(764), 1,
      anon_sym_SEMI,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1031), 1,
      anon_sym_COMMA,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1272), 1,
      anon_sym_DOT,
    STATE(397), 1,
      sym_argument_list,
    STATE(943), 1,
      aux_sym_expression_list_repeat1,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [42265] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(796), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_COLON,
      anon_sym_AMP,
      anon_sym_SLASH,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(794), 20,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_LT_DASH,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [42300] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1274), 1,
      anon_sym_RBRACK,
    ACTIONS(1276), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [42360] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1278), 1,
      anon_sym_RBRACK,
    ACTIONS(1280), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [42420] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1282), 1,
      anon_sym_EQ,
    ACTIONS(1284), 1,
      anon_sym_LBRACK,
    STATE(686), 1,
      sym_type_parameter_list,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(924), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [42484] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1100), 2,
      anon_sym_SEMI,
      anon_sym_COLON,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [42542] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(33), 1,
      anon_sym_LBRACE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    STATE(289), 1,
      sym_block,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1153), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [42606] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1286), 1,
      anon_sym_RPAREN,
    ACTIONS(1288), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [42666] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1290), 1,
      anon_sym_RBRACK,
    ACTIONS(1292), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [42726] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(33), 1,
      anon_sym_LBRACE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1294), 1,
      sym_identifier,
    STATE(289), 1,
      sym_block,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1153), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [42790] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1296), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(1330), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [42854] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1298), 1,
      anon_sym_RBRACK,
    ACTIONS(1300), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [42914] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1302), 1,
      anon_sym_RPAREN,
    ACTIONS(1304), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [42974] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1306), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(1330), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [43038] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1308), 1,
      anon_sym_RBRACK,
    ACTIONS(1310), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [43098] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(587), 1,
      anon_sym_COMMA,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(802), 1,
      aux_sym_var_spec_repeat1,
    STATE(879), 1,
      sym_map_type,
    STATE(1279), 1,
      sym__simple_type,
    STATE(1288), 1,
      sym_parenthesized_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [43164] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1312), 1,
      anon_sym_RBRACK,
    ACTIONS(1314), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [43224] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1250), 1,
      sym_identifier,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(955), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1145), 2,
      sym_method_elem,
      sym_type_elem,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [43286] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1218), 1,
      anon_sym_PIPE,
    ACTIONS(1228), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1230), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(867), 2,
      anon_sym_COMMA,
      anon_sym_LBRACE,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1226), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1220), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1224), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [43344] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1316), 1,
      anon_sym_LBRACE,
    STATE(489), 1,
      sym_block,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1064), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [43408] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1318), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [43468] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1322), 1,
      anon_sym_RBRACK,
    ACTIONS(1324), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [43528] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1326), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(1330), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [43592] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1328), 1,
      anon_sym_RBRACK,
    ACTIONS(1330), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [43652] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1332), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(1330), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [43716] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1334), 1,
      anon_sym_RBRACK,
    ACTIONS(1336), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [43776] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1338), 1,
      anon_sym_LBRACE,
    STATE(327), 1,
      sym_block,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1103), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [43840] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1340), 1,
      anon_sym_RPAREN,
    ACTIONS(1342), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [43900] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1344), 1,
      anon_sym_RBRACK,
    ACTIONS(1346), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [43960] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1348), 1,
      anon_sym_COMMA,
    STATE(816), 1,
      sym_map_type,
    STATE(828), 1,
      aux_sym_field_declaration_repeat1,
    STATE(896), 1,
      sym_parenthesized_type,
    STATE(897), 1,
      sym__simple_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [44026] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1350), 1,
      anon_sym_RBRACK,
    ACTIONS(1352), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44086] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1354), 1,
      anon_sym_RBRACK,
    ACTIONS(1356), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44146] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1358), 1,
      anon_sym_LBRACE,
    STATE(370), 1,
      sym_block,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1126), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [44210] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1360), 1,
      anon_sym_RPAREN,
    ACTIONS(1362), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44270] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1364), 1,
      anon_sym_RBRACK,
    ACTIONS(1366), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44330] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1368), 1,
      anon_sym_RBRACK,
    ACTIONS(1370), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44390] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(33), 1,
      anon_sym_LBRACE,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1218), 1,
      anon_sym_PIPE,
    ACTIONS(1228), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1230), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(919), 1,
      sym_block,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1226), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1220), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1224), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44450] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1372), 1,
      anon_sym_RBRACK,
    ACTIONS(1374), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44510] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1376), 1,
      anon_sym_LBRACE,
    STATE(551), 1,
      sym_block,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1143), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [44574] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1378), 1,
      anon_sym_RPAREN,
    ACTIONS(1380), 1,
      anon_sym_COMMA,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44634] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1382), 1,
      anon_sym_RBRACK,
    ACTIONS(1384), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44694] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1386), 1,
      anon_sym_RBRACK,
    ACTIONS(1388), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44754] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1390), 1,
      anon_sym_RBRACK,
    ACTIONS(1392), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [44814] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1394), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(1330), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [44878] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(33), 1,
      anon_sym_LBRACE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1396), 1,
      sym_identifier,
    STATE(289), 1,
      sym_block,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1153), 2,
      sym_parameter_list,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [44942] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1398), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(1330), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [45006] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(867), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [45064] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1400), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [45125] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(1330), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [45186] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1218), 1,
      anon_sym_PIPE,
    ACTIONS(1228), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1230), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1402), 1,
      anon_sym_LBRACE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1226), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1220), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1224), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [45243] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1404), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [45304] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1406), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [45361] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1408), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [45418] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1410), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [45475] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1412), 1,
      anon_sym_RPAREN,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [45532] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1414), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [45589] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1416), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1188), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [45650] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1418), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [45707] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1420), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [45764] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1422), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [45825] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(1169), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [45886] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1424), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [45947] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1426), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1165), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [46008] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1428), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [46069] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1430), 1,
      anon_sym_type,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1414), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [46130] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1432), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [46187] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1434), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [46244] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1436), 1,
      anon_sym_RPAREN,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [46301] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1438), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [46358] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1440), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1209), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [46419] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1442), 1,
      anon_sym_RPAREN,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [46476] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1444), 1,
      anon_sym_type,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1414), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [46537] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1100), 1,
      anon_sym_LBRACE,
    ACTIONS(1218), 1,
      anon_sym_PIPE,
    ACTIONS(1228), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1230), 1,
      anon_sym_PIPE_PIPE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1226), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1220), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1224), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [46594] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1446), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [46651] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1448), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [46712] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1218), 1,
      anon_sym_PIPE,
    ACTIONS(1228), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1230), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1450), 1,
      anon_sym_LBRACE,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1222), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1226), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1220), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1224), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1214), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [46769] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1452), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [46826] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1454), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [46887] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1456), 1,
      anon_sym_type,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1414), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [46948] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1458), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [47005] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1460), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [47062] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1462), 1,
      anon_sym_RPAREN,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [47119] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1464), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [47176] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1466), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1230), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [47237] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1468), 1,
      anon_sym_SEMI,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [47294] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1470), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1241), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [47355] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1472), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [47416] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1474), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [47473] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1476), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [47534] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1478), 1,
      anon_sym_EQ,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(921), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [47595] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1480), 1,
      anon_sym_RPAREN,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [47652] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1482), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [47713] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1484), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [47770] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1486), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [47827] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1488), 1,
      anon_sym_RBRACK,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [47884] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1490), 1,
      anon_sym_type,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1343), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [47945] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1492), 1,
      anon_sym_COLON,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [48002] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1494), 1,
      anon_sym_RPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1254), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48063] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(1255), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48124] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1496), 1,
      anon_sym_SEMI,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [48181] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(1243), 1,
      sym_type_elem,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1075), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48242] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1498), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48303] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_RBRACK,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48364] = 15,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(993), 1,
      anon_sym_DOT,
    ACTIONS(1162), 1,
      anon_sym_PIPE,
    ACTIONS(1172), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1174), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1502), 1,
      anon_sym_RPAREN,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1166), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1170), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1164), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1168), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(1160), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [48421] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1399), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48479] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(871), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48537] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1429), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48595] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1506), 1,
      anon_sym_LBRACK,
    ACTIONS(1508), 1,
      anon_sym_STAR,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1512), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(963), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48653] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1216), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48711] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1506), 1,
      anon_sym_LBRACK,
    ACTIONS(1508), 1,
      anon_sym_STAR,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1512), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(967), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48769] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1163), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48827] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(813), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48885] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(871), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [48943] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1506), 1,
      anon_sym_LBRACK,
    ACTIONS(1508), 1,
      anon_sym_STAR,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1512), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(976), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49001] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(878), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49059] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1334), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49117] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1418), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49175] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1130), 1,
      sym_identifier,
    ACTIONS(1132), 1,
      anon_sym_func,
    ACTIONS(1138), 1,
      anon_sym_map,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1514), 1,
      anon_sym_LBRACK,
    ACTIONS(1516), 1,
      anon_sym_STAR,
    ACTIONS(1518), 1,
      anon_sym_chan,
    ACTIONS(1520), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(918), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(951), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49233] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1130), 1,
      sym_identifier,
    ACTIONS(1132), 1,
      anon_sym_func,
    ACTIONS(1138), 1,
      anon_sym_map,
    ACTIONS(1140), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1514), 1,
      anon_sym_LBRACK,
    ACTIONS(1516), 1,
      anon_sym_STAR,
    ACTIONS(1520), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(918), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1025), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49291] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1522), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(819), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49349] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1130), 1,
      sym_identifier,
    ACTIONS(1132), 1,
      anon_sym_func,
    ACTIONS(1138), 1,
      anon_sym_map,
    ACTIONS(1140), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1514), 1,
      anon_sym_LBRACK,
    ACTIONS(1516), 1,
      anon_sym_STAR,
    ACTIONS(1524), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(918), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(952), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49407] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(827), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49465] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1388), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49523] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(818), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49581] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(886), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49639] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    ACTIONS(1526), 1,
      anon_sym_chan,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(809), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49697] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1522), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(810), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49755] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1130), 1,
      sym_identifier,
    ACTIONS(1132), 1,
      anon_sym_func,
    ACTIONS(1138), 1,
      anon_sym_map,
    ACTIONS(1140), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1514), 1,
      anon_sym_LBRACK,
    ACTIONS(1516), 1,
      anon_sym_STAR,
    ACTIONS(1520), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(918), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(963), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49813] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1130), 1,
      sym_identifier,
    ACTIONS(1132), 1,
      anon_sym_func,
    ACTIONS(1138), 1,
      anon_sym_map,
    ACTIONS(1140), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1514), 1,
      anon_sym_LBRACK,
    ACTIONS(1516), 1,
      anon_sym_STAR,
    ACTIONS(1520), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(918), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(967), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49871] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(884), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [49929] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    ACTIONS(947), 1,
      anon_sym_LBRACK,
    ACTIONS(997), 1,
      anon_sym_PIPE,
    ACTIONS(1007), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1118), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1528), 1,
      anon_sym_DOT,
    STATE(397), 1,
      sym_argument_list,
    STATE(1299), 1,
      sym_type_arguments,
    ACTIONS(1001), 2,
      anon_sym_AMP,
      anon_sym_SLASH,
    ACTIONS(1005), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(999), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_CARET,
    ACTIONS(1003), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(995), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_CARET,
  [49983] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1530), 1,
      anon_sym_LT_DASH,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(888), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50041] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(887), 1,
      sym_identifier,
    ACTIONS(895), 1,
      anon_sym_func,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(907), 1,
      anon_sym_map,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1532), 1,
      anon_sym_LBRACK,
    ACTIONS(1534), 1,
      anon_sym_STAR,
    ACTIONS(1536), 1,
      anon_sym_chan,
    ACTIONS(1538), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(803), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(809), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50099] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(923), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50157] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(887), 1,
      sym_identifier,
    ACTIONS(895), 1,
      anon_sym_func,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(907), 1,
      anon_sym_map,
    ACTIONS(909), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1532), 1,
      anon_sym_LBRACK,
    ACTIONS(1534), 1,
      anon_sym_STAR,
    ACTIONS(1540), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(803), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(819), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50215] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(887), 1,
      sym_identifier,
    ACTIONS(895), 1,
      anon_sym_func,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(907), 1,
      anon_sym_map,
    ACTIONS(909), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1532), 1,
      anon_sym_LBRACK,
    ACTIONS(1534), 1,
      anon_sym_STAR,
    ACTIONS(1538), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(803), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(827), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50273] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(887), 1,
      sym_identifier,
    ACTIONS(895), 1,
      anon_sym_func,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(907), 1,
      anon_sym_map,
    ACTIONS(909), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1532), 1,
      anon_sym_LBRACK,
    ACTIONS(1534), 1,
      anon_sym_STAR,
    ACTIONS(1540), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(803), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(810), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50331] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(887), 1,
      sym_identifier,
    ACTIONS(895), 1,
      anon_sym_func,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(907), 1,
      anon_sym_map,
    ACTIONS(909), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1532), 1,
      anon_sym_LBRACK,
    ACTIONS(1534), 1,
      anon_sym_STAR,
    ACTIONS(1538), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(803), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(818), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50389] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(831), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50447] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(887), 1,
      sym_identifier,
    ACTIONS(895), 1,
      anon_sym_func,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(907), 1,
      anon_sym_map,
    ACTIONS(909), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1532), 1,
      anon_sym_LBRACK,
    ACTIONS(1534), 1,
      anon_sym_STAR,
    ACTIONS(1538), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(803), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(831), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50505] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(887), 1,
      sym_identifier,
    ACTIONS(895), 1,
      anon_sym_func,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(907), 1,
      anon_sym_map,
    ACTIONS(909), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1532), 1,
      anon_sym_LBRACK,
    ACTIONS(1534), 1,
      anon_sym_STAR,
    ACTIONS(1538), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(803), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(813), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50563] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1318), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50621] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1176), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50679] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(873), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50737] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(808), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50795] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1506), 1,
      anon_sym_LBRACK,
    ACTIONS(1508), 1,
      anon_sym_STAR,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1542), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(930), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50853] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1506), 1,
      anon_sym_LBRACK,
    ACTIONS(1508), 1,
      anon_sym_STAR,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1512), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(945), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50911] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1506), 1,
      anon_sym_LBRACK,
    ACTIONS(1508), 1,
      anon_sym_STAR,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1512), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(948), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [50969] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(887), 1,
      sym_identifier,
    ACTIONS(895), 1,
      anon_sym_func,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(907), 1,
      anon_sym_map,
    ACTIONS(909), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1532), 1,
      anon_sym_LBRACK,
    ACTIONS(1534), 1,
      anon_sym_STAR,
    ACTIONS(1538), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(803), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(808), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51027] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1051), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51085] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1440), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51143] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(872), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51201] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(884), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51259] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(913), 1,
      sym_identifier,
    ACTIONS(915), 1,
      anon_sym_func,
    ACTIONS(921), 1,
      anon_sym_map,
    ACTIONS(923), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1238), 1,
      anon_sym_LBRACK,
    ACTIONS(1240), 1,
      anon_sym_STAR,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1244), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(801), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(926), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51317] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(872), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51375] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1130), 1,
      sym_identifier,
    ACTIONS(1132), 1,
      anon_sym_func,
    ACTIONS(1138), 1,
      anon_sym_map,
    ACTIONS(1140), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1514), 1,
      anon_sym_LBRACK,
    ACTIONS(1516), 1,
      anon_sym_STAR,
    ACTIONS(1524), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(918), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(930), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51433] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1130), 1,
      sym_identifier,
    ACTIONS(1132), 1,
      anon_sym_func,
    ACTIONS(1138), 1,
      anon_sym_map,
    ACTIONS(1140), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1514), 1,
      anon_sym_LBRACK,
    ACTIONS(1516), 1,
      anon_sym_STAR,
    ACTIONS(1520), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(918), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(945), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51491] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1130), 1,
      sym_identifier,
    ACTIONS(1132), 1,
      anon_sym_func,
    ACTIONS(1138), 1,
      anon_sym_map,
    ACTIONS(1140), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1514), 1,
      anon_sym_LBRACK,
    ACTIONS(1516), 1,
      anon_sym_STAR,
    ACTIONS(1520), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(918), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(948), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51549] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(878), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51607] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1544), 1,
      anon_sym_chan,
    STATE(879), 1,
      sym_map_type,
    STATE(861), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51665] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(887), 1,
      sym_identifier,
    ACTIONS(895), 1,
      anon_sym_func,
    ACTIONS(901), 1,
      anon_sym_struct,
    ACTIONS(905), 1,
      anon_sym_interface,
    ACTIONS(907), 1,
      anon_sym_map,
    ACTIONS(909), 1,
      anon_sym_chan,
    ACTIONS(1234), 1,
      anon_sym_LPAREN,
    ACTIONS(1242), 1,
      anon_sym_TILDE,
    ACTIONS(1532), 1,
      anon_sym_LBRACK,
    ACTIONS(1534), 1,
      anon_sym_STAR,
    ACTIONS(1538), 1,
      anon_sym_LT_DASH,
    STATE(816), 1,
      sym_map_type,
    STATE(803), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(815), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(837), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51723] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1229), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51781] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1546), 1,
      anon_sym_LT_DASH,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(888), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51839] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1506), 1,
      anon_sym_LBRACK,
    ACTIONS(1508), 1,
      anon_sym_STAR,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1512), 1,
      anon_sym_LT_DASH,
    ACTIONS(1548), 1,
      anon_sym_chan,
    STATE(1027), 1,
      sym_map_type,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(951), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51897] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1260), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [51955] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1414), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52013] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1546), 1,
      anon_sym_LT_DASH,
    STATE(879), 1,
      sym_map_type,
    STATE(862), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52071] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1347), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52129] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(886), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52187] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1070), 1,
      sym_identifier,
    ACTIONS(1076), 1,
      anon_sym_func,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1090), 1,
      anon_sym_map,
    ACTIONS(1092), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1506), 1,
      anon_sym_LBRACK,
    ACTIONS(1508), 1,
      anon_sym_STAR,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1542), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(922), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(952), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52245] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1251), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52303] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1441), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52361] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1369), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52419] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(574), 1,
      anon_sym_chan,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1023), 1,
      anon_sym_LT_DASH,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(861), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52477] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1344), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52535] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1190), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52593] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1343), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52651] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(37), 1,
      anon_sym_map,
    ACTIONS(39), 1,
      anon_sym_chan,
    ACTIONS(1015), 1,
      sym_identifier,
    ACTIONS(1017), 1,
      anon_sym_func,
    ACTIONS(1019), 1,
      anon_sym_LBRACK,
    ACTIONS(1021), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    ACTIONS(1530), 1,
      anon_sym_LT_DASH,
    STATE(879), 1,
      sym_map_type,
    STATE(847), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(862), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52709] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1396), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52767] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1428), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52825] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1212), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52883] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1353), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52941] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1422), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [52999] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1233), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [53057] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(29), 1,
      anon_sym_struct,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(35), 1,
      anon_sym_interface,
    ACTIONS(578), 1,
      sym_identifier,
    ACTIONS(589), 1,
      anon_sym_func,
    ACTIONS(602), 1,
      anon_sym_map,
    ACTIONS(604), 1,
      anon_sym_chan,
    ACTIONS(606), 1,
      anon_sym_LT_DASH,
    ACTIONS(1011), 1,
      anon_sym_LBRACK,
    ACTIONS(1013), 1,
      anon_sym_STAR,
    ACTIONS(1057), 1,
      anon_sym_LPAREN,
    STATE(879), 1,
      sym_map_type,
    STATE(875), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(1349), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(885), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [53115] = 16,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1082), 1,
      anon_sym_struct,
    ACTIONS(1088), 1,
      anon_sym_interface,
    ACTIONS(1130), 1,
      sym_identifier,
    ACTIONS(1132), 1,
      anon_sym_func,
    ACTIONS(1138), 1,
      anon_sym_map,
    ACTIONS(1140), 1,
      anon_sym_chan,
    ACTIONS(1504), 1,
      anon_sym_LPAREN,
    ACTIONS(1510), 1,
      anon_sym_TILDE,
    ACTIONS(1514), 1,
      anon_sym_LBRACK,
    ACTIONS(1516), 1,
      anon_sym_STAR,
    ACTIONS(1520), 1,
      anon_sym_LT_DASH,
    STATE(1027), 1,
      sym_map_type,
    STATE(918), 2,
      sym_negated_type,
      sym_qualified_type,
    STATE(976), 2,
      sym_parenthesized_type,
      sym__simple_type,
    STATE(1020), 8,
      sym_generic_type,
      sym_pointer_type,
      sym_array_type,
      sym_slice_type,
      sym_struct_type,
      sym_interface_type,
      sym_channel_type,
      sym_function_type,
  [53173] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1552), 1,
      aux_sym_source_file_token1,
    ACTIONS(1550), 22,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_func,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_struct,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_interface,
      anon_sym_PIPE,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_LT_DASH,
      anon_sym_case,
      anon_sym_default,
      sym_identifier,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53204] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1556), 1,
      aux_sym_source_file_token1,
    ACTIONS(1554), 22,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_func,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_struct,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_interface,
      anon_sym_PIPE,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_LT_DASH,
      anon_sym_case,
      anon_sym_default,
      sym_identifier,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53235] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1560), 1,
      aux_sym_source_file_token1,
    ACTIONS(1558), 22,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_func,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_struct,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_interface,
      anon_sym_PIPE,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_LT_DASH,
      anon_sym_case,
      anon_sym_default,
      sym_identifier,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53266] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1564), 1,
      aux_sym_source_file_token1,
    ACTIONS(1562), 22,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_func,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_struct,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_interface,
      anon_sym_PIPE,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_LT_DASH,
      anon_sym_case,
      anon_sym_default,
      sym_identifier,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53297] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1568), 1,
      aux_sym_source_file_token1,
    ACTIONS(1570), 1,
      anon_sym_COMMA,
    STATE(787), 1,
      aux_sym_const_spec_repeat1,
    ACTIONS(1566), 17,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LPAREN,
      anon_sym_EQ,
      anon_sym_func,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_struct,
      anon_sym_TILDE,
      anon_sym_RBRACE,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_LT_DASH,
      anon_sym_case,
      anon_sym_default,
      sym_identifier,
  [53329] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1568), 1,
      aux_sym_source_file_token1,
    ACTIONS(1566), 18,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_func,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_struct,
      anon_sym_TILDE,
      anon_sym_RBRACE,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_LT_DASH,
      anon_sym_case,
      anon_sym_default,
      sym_identifier,
  [53356] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1550), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
    ACTIONS(1552), 12,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON,
  [53382] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1558), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
    ACTIONS(1560), 12,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON,
  [53408] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1562), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
    ACTIONS(1564), 12,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON,
  [53434] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1554), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
    ACTIONS(1556), 12,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      anon_sym_COLON,
  [53460] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1560), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1558), 14,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LPAREN,
      anon_sym_func,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_struct,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_LT_DASH,
      sym_identifier,
  [53484] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1556), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1554), 14,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LPAREN,
      anon_sym_func,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_struct,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_LT_DASH,
      sym_identifier,
  [53508] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1552), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1550), 14,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LPAREN,
      anon_sym_func,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_struct,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_LT_DASH,
      sym_identifier,
  [53532] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1564), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1562), 14,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LPAREN,
      anon_sym_func,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_struct,
      anon_sym_TILDE,
      anon_sym_LBRACE,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      anon_sym_LT_DASH,
      sym_identifier,
  [53556] = 6,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(931), 1,
      anon_sym_DOT,
    ACTIONS(1059), 1,
      aux_sym_source_file_token1,
    ACTIONS(1575), 1,
      anon_sym_LBRACK,
    STATE(821), 1,
      sym_type_arguments,
    ACTIONS(1573), 11,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53585] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(754), 1,
      anon_sym_COMMA,
    STATE(800), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1577), 13,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [53610] = 6,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(931), 1,
      anon_sym_DOT,
    ACTIONS(1059), 1,
      aux_sym_source_file_token1,
    ACTIONS(1579), 1,
      anon_sym_LBRACK,
    STATE(821), 1,
      sym_type_arguments,
    ACTIONS(1573), 11,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53639] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1582), 1,
      anon_sym_COMMA,
    STATE(800), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(867), 13,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [53664] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1575), 1,
      anon_sym_LBRACK,
    ACTIONS(1585), 1,
      aux_sym_source_file_token1,
    STATE(814), 1,
      sym_type_arguments,
    ACTIONS(1587), 11,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53690] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1593), 1,
      anon_sym_COMMA,
    STATE(802), 1,
      aux_sym_var_spec_repeat1,
    ACTIONS(1589), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
    ACTIONS(1591), 6,
      anon_sym_LPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LT_DASH,
  [53716] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1585), 1,
      aux_sym_source_file_token1,
    ACTIONS(1596), 1,
      anon_sym_LBRACK,
    STATE(814), 1,
      sym_type_arguments,
    ACTIONS(1587), 11,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53742] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1599), 1,
      aux_sym_source_file_token1,
    ACTIONS(1601), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53763] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1603), 1,
      aux_sym_source_file_token1,
    ACTIONS(1605), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53784] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1607), 1,
      aux_sym_source_file_token1,
    ACTIONS(1609), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53805] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1611), 1,
      aux_sym_source_file_token1,
    ACTIONS(1613), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53826] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1615), 1,
      aux_sym_source_file_token1,
    ACTIONS(1617), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53847] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1619), 1,
      aux_sym_source_file_token1,
    ACTIONS(1621), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53868] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1619), 1,
      aux_sym_source_file_token1,
    ACTIONS(1621), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53889] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1623), 1,
      aux_sym_source_file_token1,
    ACTIONS(1625), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53910] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1627), 1,
      aux_sym_source_file_token1,
    ACTIONS(1629), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53931] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1619), 1,
      aux_sym_source_file_token1,
    ACTIONS(1621), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53952] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1631), 1,
      aux_sym_source_file_token1,
    ACTIONS(1633), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53973] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1635), 1,
      aux_sym_source_file_token1,
    ACTIONS(1637), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [53994] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1639), 1,
      aux_sym_source_file_token1,
    ACTIONS(1641), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54015] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1643), 1,
      aux_sym_source_file_token1,
    ACTIONS(1645), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54036] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1647), 1,
      aux_sym_source_file_token1,
    ACTIONS(1649), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54057] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1651), 1,
      aux_sym_source_file_token1,
    ACTIONS(1653), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54078] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1655), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
    ACTIONS(1657), 7,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LT_DASH,
  [54099] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1659), 1,
      aux_sym_source_file_token1,
    ACTIONS(1661), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54120] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1665), 1,
      anon_sym_COLON_EQ,
    ACTIONS(1663), 12,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [54141] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1669), 1,
      anon_sym_COLON_EQ,
    ACTIONS(1667), 12,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [54162] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1671), 1,
      aux_sym_source_file_token1,
    ACTIONS(1673), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54183] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1675), 1,
      anon_sym_COLON_EQ,
    ACTIONS(1663), 12,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [54204] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1677), 1,
      aux_sym_source_file_token1,
    ACTIONS(1679), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54225] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1681), 1,
      aux_sym_source_file_token1,
    ACTIONS(1683), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54246] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1689), 1,
      anon_sym_COMMA,
    STATE(828), 1,
      aux_sym_field_declaration_repeat1,
    ACTIONS(1687), 5,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LT_DASH,
    ACTIONS(1685), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
  [54271] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1692), 1,
      aux_sym_source_file_token1,
    ACTIONS(1694), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54292] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1696), 1,
      aux_sym_source_file_token1,
    ACTIONS(1698), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54313] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1700), 1,
      aux_sym_source_file_token1,
    ACTIONS(1702), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54334] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1704), 1,
      aux_sym_source_file_token1,
    ACTIONS(1706), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54355] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1708), 1,
      aux_sym_source_file_token1,
    ACTIONS(1710), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54376] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1712), 1,
      aux_sym_source_file_token1,
    ACTIONS(1714), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54397] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1716), 1,
      anon_sym_EQ,
    ACTIONS(1718), 1,
      anon_sym_COLON_EQ,
    ACTIONS(1663), 11,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [54420] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1720), 1,
      aux_sym_source_file_token1,
    ACTIONS(1722), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54441] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1585), 1,
      aux_sym_source_file_token1,
    ACTIONS(1587), 12,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_case,
      anon_sym_default,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [54462] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1726), 1,
      anon_sym_COLON_EQ,
    ACTIONS(1724), 12,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_AMP_CARET_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
  [54483] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1728), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
    ACTIONS(1730), 6,
      anon_sym_LPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LT_DASH,
  [54503] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1732), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
    ACTIONS(1734), 6,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LT_DASH,
  [54523] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1736), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
    ACTIONS(1738), 6,
      anon_sym_LPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LT_DASH,
  [54543] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1740), 6,
      anon_sym_func,
      anon_sym_struct,
      anon_sym_interface,
      anon_sym_map,
      anon_sym_chan,
      sym_identifier,
    ACTIONS(1742), 6,
      anon_sym_LPAREN,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_TILDE,
      anon_sym_LT_DASH,
  [54563] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(1744), 1,
      sym_identifier,
    ACTIONS(1746), 1,
      anon_sym_DOT,
    ACTIONS(1748), 1,
      sym_blank_identifier,
    ACTIONS(1750), 1,
      anon_sym_LPAREN,
    STATE(1121), 1,
      sym_dot,
    STATE(1119), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
    STATE(1205), 2,
      sym_import_spec,
      sym_import_spec_list,
  [54596] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1752), 1,
      anon_sym_DOT,
    ACTIONS(1754), 1,
      anon_sym_LBRACK,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(1059), 8,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [54619] = 9,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(931), 1,
      anon_sym_DOT,
    ACTIONS(939), 1,
      anon_sym_BQUOTE,
    ACTIONS(941), 1,
      anon_sym_DQUOTE,
    ACTIONS(1575), 1,
      anon_sym_LBRACK,
    ACTIONS(1757), 1,
      aux_sym_source_file_token1,
    STATE(821), 1,
      sym_type_arguments,
    STATE(1093), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
    ACTIONS(1759), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [54650] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1746), 1,
      anon_sym_DOT,
    ACTIONS(1761), 1,
      sym_identifier,
    ACTIONS(1763), 1,
      sym_blank_identifier,
    ACTIONS(1765), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_BQUOTE,
    ACTIONS(1769), 1,
      anon_sym_DQUOTE,
    STATE(1156), 1,
      sym_dot,
    STATE(1128), 2,
      sym_import_spec,
      sym_import_spec_list,
    STATE(1155), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [54683] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1771), 1,
      anon_sym_LBRACK,
    STATE(874), 1,
      sym_type_arguments,
    ACTIONS(1585), 8,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [54703] = 8,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(939), 1,
      anon_sym_BQUOTE,
    ACTIONS(941), 1,
      anon_sym_DQUOTE,
    ACTIONS(1575), 1,
      anon_sym_LBRACK,
    ACTIONS(1774), 1,
      aux_sym_source_file_token1,
    STATE(814), 1,
      sym_type_arguments,
    STATE(1108), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
    ACTIONS(1776), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [54731] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(1744), 1,
      sym_identifier,
    ACTIONS(1746), 1,
      anon_sym_DOT,
    ACTIONS(1748), 1,
      sym_blank_identifier,
    ACTIONS(1778), 1,
      anon_sym_RPAREN,
    STATE(1096), 1,
      sym_import_spec,
    STATE(1121), 1,
      sym_dot,
    STATE(1119), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [54763] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(1744), 1,
      sym_identifier,
    ACTIONS(1746), 1,
      anon_sym_DOT,
    ACTIONS(1748), 1,
      sym_blank_identifier,
    ACTIONS(1780), 1,
      anon_sym_RPAREN,
    STATE(1042), 1,
      sym_import_spec,
    STATE(1121), 1,
      sym_dot,
    STATE(1119), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [54795] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(1744), 1,
      sym_identifier,
    ACTIONS(1746), 1,
      anon_sym_DOT,
    ACTIONS(1748), 1,
      sym_blank_identifier,
    ACTIONS(1782), 1,
      anon_sym_RPAREN,
    STATE(1057), 1,
      sym_import_spec,
    STATE(1121), 1,
      sym_dot,
    STATE(1119), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [54827] = 8,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(939), 1,
      anon_sym_BQUOTE,
    ACTIONS(941), 1,
      anon_sym_DQUOTE,
    ACTIONS(1575), 1,
      anon_sym_LBRACK,
    ACTIONS(1784), 1,
      aux_sym_source_file_token1,
    STATE(814), 1,
      sym_type_arguments,
    STATE(1094), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
    ACTIONS(1786), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [54855] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(1744), 1,
      sym_identifier,
    ACTIONS(1746), 1,
      anon_sym_DOT,
    ACTIONS(1748), 1,
      sym_blank_identifier,
    ACTIONS(1788), 1,
      anon_sym_RPAREN,
    STATE(1096), 1,
      sym_import_spec,
    STATE(1121), 1,
      sym_dot,
    STATE(1119), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [54887] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(1744), 1,
      sym_identifier,
    ACTIONS(1746), 1,
      anon_sym_DOT,
    ACTIONS(1748), 1,
      sym_blank_identifier,
    ACTIONS(1790), 1,
      anon_sym_RPAREN,
    STATE(1096), 1,
      sym_import_spec,
    STATE(1121), 1,
      sym_dot,
    STATE(1119), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [54919] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1752), 1,
      anon_sym_DOT,
    ACTIONS(1792), 1,
      anon_sym_LBRACK,
    STATE(869), 1,
      sym_type_arguments,
    ACTIONS(1059), 7,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [54941] = 8,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(893), 1,
      anon_sym_LPAREN,
    ACTIONS(931), 1,
      anon_sym_DOT,
    ACTIONS(1059), 1,
      aux_sym_source_file_token1,
    ACTIONS(1575), 1,
      anon_sym_LBRACK,
    STATE(529), 1,
      sym_parameter_list,
    STATE(821), 1,
      sym_type_arguments,
    ACTIONS(1573), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_PIPE,
  [54969] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(1744), 1,
      sym_identifier,
    ACTIONS(1746), 1,
      anon_sym_DOT,
    ACTIONS(1748), 1,
      sym_blank_identifier,
    ACTIONS(1794), 1,
      anon_sym_RPAREN,
    STATE(1096), 1,
      sym_import_spec,
    STATE(1121), 1,
      sym_dot,
    STATE(1119), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [55001] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1704), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55016] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1720), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55031] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1712), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55046] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1619), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55061] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1619), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55076] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1611), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55091] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1696), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55106] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(1744), 1,
      sym_identifier,
    ACTIONS(1746), 1,
      anon_sym_DOT,
    ACTIONS(1748), 1,
      sym_blank_identifier,
    STATE(1096), 1,
      sym_import_spec,
    STATE(1121), 1,
      sym_dot,
    STATE(1119), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [55135] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1607), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55150] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    ACTIONS(1771), 1,
      anon_sym_LBRACK,
    STATE(399), 1,
      sym_literal_value,
    STATE(874), 1,
      sym_type_arguments,
    ACTIONS(1585), 5,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_PIPE,
  [55173] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1692), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55188] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1659), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55203] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1627), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55218] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1647), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55233] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1700), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55248] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1635), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55263] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1631), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55278] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1792), 1,
      anon_sym_LBRACK,
    STATE(874), 1,
      sym_type_arguments,
    ACTIONS(1585), 7,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55297] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1603), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55312] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1623), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55327] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1619), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55342] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1639), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55357] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1643), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55372] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1599), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55387] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1671), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55402] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1677), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55417] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1615), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55432] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1585), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55447] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1681), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55462] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1708), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55477] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1651), 9,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LBRACE,
      anon_sym_PIPE,
      anon_sym_COLON,
  [55492] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(977), 1,
      anon_sym_COMMA,
    ACTIONS(1577), 1,
      aux_sym_source_file_token1,
    STATE(905), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1796), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [55512] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(1798), 1,
      sym_identifier,
    ACTIONS(1800), 1,
      anon_sym_STAR,
    ACTIONS(1802), 1,
      anon_sym_RBRACE,
    STATE(848), 1,
      sym_qualified_type,
    STATE(908), 1,
      sym_generic_type,
    STATE(1125), 1,
      sym_field_declaration,
    STATE(1332), 1,
      sym_negated_type,
  [55540] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(1798), 1,
      sym_identifier,
    ACTIONS(1800), 1,
      anon_sym_STAR,
    ACTIONS(1804), 1,
      anon_sym_RBRACE,
    STATE(848), 1,
      sym_qualified_type,
    STATE(908), 1,
      sym_generic_type,
    STATE(1125), 1,
      sym_field_declaration,
    STATE(1332), 1,
      sym_negated_type,
  [55568] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    STATE(399), 1,
      sym_literal_value,
    ACTIONS(1585), 6,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_PIPE,
  [55586] = 6,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(939), 1,
      anon_sym_BQUOTE,
    ACTIONS(941), 1,
      anon_sym_DQUOTE,
    ACTIONS(1784), 1,
      aux_sym_source_file_token1,
    STATE(1094), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
    ACTIONS(1786), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [55608] = 6,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(939), 1,
      anon_sym_BQUOTE,
    ACTIONS(941), 1,
      anon_sym_DQUOTE,
    ACTIONS(1806), 1,
      aux_sym_source_file_token1,
    STATE(1105), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
    ACTIONS(1808), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [55630] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    STATE(399), 1,
      sym_literal_value,
    ACTIONS(1639), 6,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_PIPE,
  [55648] = 7,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(939), 1,
      anon_sym_BQUOTE,
    ACTIONS(941), 1,
      anon_sym_DQUOTE,
    ACTIONS(1810), 1,
      aux_sym_source_file_token1,
    STATE(1109), 1,
      sym_raw_string_literal,
    STATE(1110), 1,
      sym_interpreted_string_literal,
    ACTIONS(1812), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [55672] = 7,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(939), 1,
      anon_sym_BQUOTE,
    ACTIONS(941), 1,
      anon_sym_DQUOTE,
    ACTIONS(1810), 1,
      aux_sym_source_file_token1,
    STATE(1112), 1,
      sym_raw_string_literal,
    STATE(1115), 1,
      sym_interpreted_string_literal,
    ACTIONS(1812), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [55696] = 6,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1814), 1,
      anon_sym_DOT,
    ACTIONS(1816), 1,
      anon_sym_LBRACK,
    STATE(934), 1,
      sym_type_arguments,
    ACTIONS(1059), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1573), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACE,
  [55718] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(1798), 1,
      sym_identifier,
    ACTIONS(1800), 1,
      anon_sym_STAR,
    ACTIONS(1819), 1,
      anon_sym_RBRACE,
    STATE(848), 1,
      sym_qualified_type,
    STATE(908), 1,
      sym_generic_type,
    STATE(1125), 1,
      sym_field_declaration,
    STATE(1332), 1,
      sym_negated_type,
  [55746] = 6,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1814), 1,
      anon_sym_DOT,
    ACTIONS(1821), 1,
      anon_sym_LBRACK,
    STATE(934), 1,
      sym_type_arguments,
    ACTIONS(1059), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1573), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACE,
  [55768] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1053), 1,
      anon_sym_LBRACE,
    STATE(408), 1,
      sym_block,
    ACTIONS(1623), 6,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_PIPE,
  [55786] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(1798), 1,
      sym_identifier,
    ACTIONS(1800), 1,
      anon_sym_STAR,
    ACTIONS(1823), 1,
      anon_sym_RBRACE,
    STATE(848), 1,
      sym_qualified_type,
    STATE(908), 1,
      sym_generic_type,
    STATE(1125), 1,
      sym_field_declaration,
    STATE(1332), 1,
      sym_negated_type,
  [55814] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(1798), 1,
      sym_identifier,
    ACTIONS(1800), 1,
      anon_sym_STAR,
    ACTIONS(1825), 1,
      anon_sym_RBRACE,
    STATE(848), 1,
      sym_qualified_type,
    STATE(908), 1,
      sym_generic_type,
    STATE(1125), 1,
      sym_field_declaration,
    STATE(1332), 1,
      sym_negated_type,
  [55842] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(1798), 1,
      sym_identifier,
    ACTIONS(1800), 1,
      anon_sym_STAR,
    ACTIONS(1827), 1,
      anon_sym_RBRACE,
    STATE(848), 1,
      sym_qualified_type,
    STATE(908), 1,
      sym_generic_type,
    STATE(1125), 1,
      sym_field_declaration,
    STATE(1332), 1,
      sym_negated_type,
  [55870] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(867), 1,
      aux_sym_source_file_token1,
    ACTIONS(1829), 1,
      anon_sym_COMMA,
    STATE(905), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(869), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [55890] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(1798), 1,
      sym_identifier,
    ACTIONS(1800), 1,
      anon_sym_STAR,
    ACTIONS(1832), 1,
      anon_sym_RBRACE,
    STATE(848), 1,
      sym_qualified_type,
    STATE(908), 1,
      sym_generic_type,
    STATE(1060), 1,
      sym_field_declaration,
    STATE(1332), 1,
      sym_negated_type,
  [55918] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(1798), 1,
      sym_identifier,
    ACTIONS(1800), 1,
      anon_sym_STAR,
    ACTIONS(1834), 1,
      anon_sym_RBRACE,
    STATE(848), 1,
      sym_qualified_type,
    STATE(908), 1,
      sym_generic_type,
    STATE(1062), 1,
      sym_field_declaration,
    STATE(1332), 1,
      sym_negated_type,
  [55946] = 6,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(939), 1,
      anon_sym_BQUOTE,
    ACTIONS(941), 1,
      anon_sym_DQUOTE,
    ACTIONS(1774), 1,
      aux_sym_source_file_token1,
    STATE(1108), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
    ACTIONS(1776), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [55968] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(1798), 1,
      sym_identifier,
    ACTIONS(1800), 1,
      anon_sym_STAR,
    ACTIONS(1836), 1,
      anon_sym_RBRACE,
    STATE(848), 1,
      sym_qualified_type,
    STATE(908), 1,
      sym_generic_type,
    STATE(1055), 1,
      sym_field_declaration,
    STATE(1332), 1,
      sym_negated_type,
  [55996] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1838), 1,
      aux_sym_source_file_token1,
    ACTIONS(1842), 1,
      anon_sym_else,
    ACTIONS(1840), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56013] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1844), 1,
      sym_identifier,
    ACTIONS(1846), 1,
      aux_sym_source_file_token1,
    ACTIONS(1848), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56030] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(1798), 1,
      sym_identifier,
    ACTIONS(1800), 1,
      anon_sym_STAR,
    STATE(848), 1,
      sym_qualified_type,
    STATE(908), 1,
      sym_generic_type,
    STATE(1125), 1,
      sym_field_declaration,
    STATE(1332), 1,
      sym_negated_type,
  [56055] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1850), 1,
      sym_identifier,
    ACTIONS(1852), 1,
      aux_sym_source_file_token1,
    ACTIONS(1854), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56072] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1856), 1,
      aux_sym_source_file_token1,
    ACTIONS(1860), 1,
      anon_sym_EQ,
    ACTIONS(1858), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56089] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1862), 1,
      aux_sym_source_file_token1,
    ACTIONS(1866), 1,
      anon_sym_EQ,
    ACTIONS(1864), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56106] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1868), 1,
      aux_sym_source_file_token1,
    STATE(920), 1,
      aux_sym_statement_list_repeat1,
    ACTIONS(1870), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
    ACTIONS(1872), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56125] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1862), 1,
      aux_sym_source_file_token1,
    ACTIONS(1874), 1,
      anon_sym_EQ,
    ACTIONS(1864), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56142] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1876), 1,
      anon_sym_LBRACK,
    STATE(938), 1,
      sym_type_arguments,
    ACTIONS(1585), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1587), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACE,
  [56161] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1879), 1,
      aux_sym_source_file_token1,
    ACTIONS(1883), 1,
      anon_sym_else,
    ACTIONS(1881), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56178] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1885), 1,
      aux_sym_source_file_token1,
    STATE(925), 1,
      aux_sym_statement_list_repeat1,
    ACTIONS(1887), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
    ACTIONS(212), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56197] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1889), 1,
      aux_sym_source_file_token1,
    ACTIONS(1891), 6,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56212] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1821), 1,
      anon_sym_LBRACK,
    STATE(938), 1,
      sym_type_arguments,
    ACTIONS(1585), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1587), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACE,
  [56231] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1893), 1,
      aux_sym_source_file_token1,
    ACTIONS(1895), 6,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56246] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1897), 1,
      aux_sym_source_file_token1,
    ACTIONS(1899), 6,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56261] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1901), 1,
      aux_sym_source_file_token1,
    STATE(925), 1,
      aux_sym_statement_list_repeat1,
    ACTIONS(1904), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
    ACTIONS(1907), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56280] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1909), 1,
      aux_sym_source_file_token1,
    ACTIONS(1911), 6,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56295] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1913), 1,
      aux_sym_source_file_token1,
    ACTIONS(1915), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56309] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1917), 1,
      aux_sym_source_file_token1,
    ACTIONS(1919), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56323] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1921), 1,
      aux_sym_source_file_token1,
    ACTIONS(1923), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56337] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1651), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1653), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56351] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1086), 1,
      anon_sym_LBRACE,
    STATE(1106), 1,
      sym_block,
    ACTIONS(1925), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1927), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [56369] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1929), 1,
      aux_sym_source_file_token1,
    ACTIONS(1931), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56383] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1933), 1,
      anon_sym_RBRACE,
    ACTIONS(1935), 1,
      anon_sym_case,
    ACTIONS(1937), 1,
      anon_sym_default,
    STATE(1006), 3,
      sym_expression_case,
      sym_default_case,
      aux_sym_expression_switch_statement_repeat1,
  [56401] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1659), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1661), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56415] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1939), 1,
      aux_sym_source_file_token1,
    ACTIONS(1941), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56429] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1943), 1,
      aux_sym_source_file_token1,
    ACTIONS(1945), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56443] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1947), 1,
      aux_sym_source_file_token1,
    ACTIONS(1949), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56457] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1631), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1633), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56471] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1951), 1,
      aux_sym_source_file_token1,
    ACTIONS(1953), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56485] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1955), 1,
      anon_sym_RBRACE,
    ACTIONS(1957), 1,
      anon_sym_case,
    ACTIONS(1960), 1,
      anon_sym_default,
    STATE(940), 3,
      sym_default_case,
      sym_type_case,
      aux_sym_type_switch_statement_repeat1,
  [56503] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1963), 1,
      aux_sym_source_file_token1,
    ACTIONS(1965), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56517] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1967), 1,
      aux_sym_source_file_token1,
    ACTIONS(1969), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56531] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1031), 1,
      anon_sym_COMMA,
    ACTIONS(1796), 1,
      anon_sym_COLON,
    STATE(960), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1577), 3,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [56549] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1599), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1601), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56563] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1681), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1683), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56577] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1971), 1,
      aux_sym_source_file_token1,
    ACTIONS(1973), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56591] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1720), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1722), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56605] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1647), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1649), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56619] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1975), 1,
      aux_sym_source_file_token1,
    ACTIONS(1977), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56633] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1712), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1714), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56647] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1619), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1621), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56661] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1619), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1621), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56675] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1611), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1613), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56689] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1979), 1,
      aux_sym_source_file_token1,
    ACTIONS(1981), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56703] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1983), 1,
      aux_sym_source_file_token1,
    ACTIONS(1987), 1,
      anon_sym_PIPE,
    STATE(995), 1,
      aux_sym_type_elem_repeat1,
    ACTIONS(1985), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [56721] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1086), 1,
      anon_sym_LBRACE,
    STATE(1160), 1,
      sym_block,
    ACTIONS(1989), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1991), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [56739] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1993), 1,
      aux_sym_source_file_token1,
    ACTIONS(1995), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56753] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1997), 1,
      aux_sym_source_file_token1,
    ACTIONS(1999), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56767] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1935), 1,
      anon_sym_case,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2001), 1,
      anon_sym_RBRACE,
    STATE(999), 3,
      sym_expression_case,
      sym_default_case,
      aux_sym_expression_switch_statement_repeat1,
  [56785] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(869), 1,
      anon_sym_COLON,
    ACTIONS(2003), 1,
      anon_sym_COMMA,
    STATE(960), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(867), 3,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [56803] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1935), 1,
      anon_sym_case,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2006), 1,
      anon_sym_RBRACE,
    STATE(1006), 3,
      sym_expression_case,
      sym_default_case,
      aux_sym_expression_switch_statement_repeat1,
  [56821] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2008), 1,
      aux_sym_source_file_token1,
    ACTIONS(2010), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56835] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1700), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1702), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56849] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1603), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1605), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56863] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2012), 1,
      aux_sym_source_file_token1,
    ACTIONS(2014), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56877] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1623), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1625), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56891] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1619), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1621), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56905] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1671), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1673), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56919] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2016), 1,
      aux_sym_source_file_token1,
    ACTIONS(2018), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56933] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2020), 1,
      anon_sym_LPAREN,
    ACTIONS(2022), 1,
      anon_sym_LBRACK,
    ACTIONS(2024), 1,
      anon_sym_PIPE,
    STATE(1072), 1,
      aux_sym_type_elem_repeat1,
    ACTIONS(1983), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [56953] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1708), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1710), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [56967] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2026), 1,
      aux_sym_source_file_token1,
    ACTIONS(2028), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [56981] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1647), 6,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_PIPE,
  [56993] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1643), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1645), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57007] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1677), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1679), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57021] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1615), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1617), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57035] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2030), 1,
      anon_sym_RBRACE,
    ACTIONS(2032), 1,
      anon_sym_case,
    STATE(1018), 3,
      sym_default_case,
      sym_communication_case,
      aux_sym_select_statement_repeat1,
  [57053] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1696), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1698), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57067] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2034), 1,
      aux_sym_source_file_token1,
    ACTIONS(2036), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57081] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2038), 1,
      aux_sym_source_file_token1,
    ACTIONS(2040), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57095] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2042), 1,
      aux_sym_source_file_token1,
    ACTIONS(2044), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57109] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2042), 1,
      aux_sym_source_file_token1,
    ACTIONS(2044), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57123] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1059), 1,
      anon_sym_LBRACE,
    ACTIONS(1752), 1,
      anon_sym_DOT,
    ACTIONS(1754), 1,
      anon_sym_LBRACK,
    ACTIONS(2046), 1,
      anon_sym_LPAREN,
    STATE(502), 1,
      sym_parameter_list,
    STATE(869), 1,
      sym_type_arguments,
  [57145] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1692), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1694), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57159] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1704), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1706), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57173] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2048), 1,
      aux_sym_source_file_token1,
    ACTIONS(2050), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57187] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2052), 1,
      aux_sym_source_file_token1,
    ACTIONS(1907), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57201] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1627), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1629), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57215] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2054), 1,
      aux_sym_source_file_token1,
    ACTIONS(2056), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57229] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2058), 1,
      aux_sym_source_file_token1,
    ACTIONS(2060), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57243] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2062), 1,
      aux_sym_source_file_token1,
    ACTIONS(2064), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57257] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1935), 1,
      anon_sym_case,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2066), 1,
      anon_sym_RBRACE,
    STATE(933), 3,
      sym_expression_case,
      sym_default_case,
      aux_sym_expression_switch_statement_repeat1,
  [57275] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2068), 1,
      aux_sym_source_file_token1,
    ACTIONS(2070), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57289] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2072), 1,
      aux_sym_source_file_token1,
    ACTIONS(2074), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57303] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1987), 1,
      anon_sym_PIPE,
    ACTIONS(2076), 1,
      aux_sym_source_file_token1,
    STATE(998), 1,
      aux_sym_type_elem_repeat1,
    ACTIONS(2078), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [57321] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2080), 1,
      aux_sym_source_file_token1,
    ACTIONS(2082), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57335] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2084), 1,
      aux_sym_source_file_token1,
    ACTIONS(2086), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57349] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2088), 1,
      aux_sym_source_file_token1,
    ACTIONS(2092), 1,
      anon_sym_PIPE,
    STATE(998), 1,
      aux_sym_type_elem_repeat1,
    ACTIONS(2090), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [57367] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1935), 1,
      anon_sym_case,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2095), 1,
      anon_sym_RBRACE,
    STATE(1006), 3,
      sym_expression_case,
      sym_default_case,
      aux_sym_expression_switch_statement_repeat1,
  [57385] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2097), 1,
      aux_sym_source_file_token1,
    ACTIONS(2099), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57399] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2101), 1,
      aux_sym_source_file_token1,
    ACTIONS(2103), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57413] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2105), 1,
      anon_sym_RBRACE,
    ACTIONS(2107), 1,
      anon_sym_case,
    STATE(1011), 3,
      sym_default_case,
      sym_type_case,
      aux_sym_type_switch_statement_repeat1,
  [57431] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1935), 1,
      anon_sym_case,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2109), 1,
      anon_sym_RBRACE,
    STATE(1013), 3,
      sym_expression_case,
      sym_default_case,
      aux_sym_expression_switch_statement_repeat1,
  [57449] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2111), 1,
      aux_sym_source_file_token1,
    ACTIONS(2113), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57463] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2115), 1,
      aux_sym_source_file_token1,
    ACTIONS(2117), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57477] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2119), 1,
      anon_sym_RBRACE,
    ACTIONS(2121), 1,
      anon_sym_case,
    ACTIONS(2124), 1,
      anon_sym_default,
    STATE(1006), 3,
      sym_expression_case,
      sym_default_case,
      aux_sym_expression_switch_statement_repeat1,
  [57495] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1935), 1,
      anon_sym_case,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2127), 1,
      anon_sym_RBRACE,
    STATE(961), 3,
      sym_expression_case,
      sym_default_case,
      aux_sym_expression_switch_statement_repeat1,
  [57513] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2129), 1,
      aux_sym_source_file_token1,
    ACTIONS(2131), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57527] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2133), 1,
      aux_sym_source_file_token1,
    ACTIONS(2135), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57541] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1086), 1,
      anon_sym_LBRACE,
    STATE(1102), 1,
      sym_block,
    ACTIONS(2137), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2139), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [57559] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2107), 1,
      anon_sym_case,
    ACTIONS(2141), 1,
      anon_sym_RBRACE,
    STATE(940), 3,
      sym_default_case,
      sym_type_case,
      aux_sym_type_switch_statement_repeat1,
  [57577] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2143), 1,
      aux_sym_source_file_token1,
    ACTIONS(2145), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57591] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1935), 1,
      anon_sym_case,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2147), 1,
      anon_sym_RBRACE,
    STATE(1006), 3,
      sym_expression_case,
      sym_default_case,
      aux_sym_expression_switch_statement_repeat1,
  [57609] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2149), 1,
      aux_sym_source_file_token1,
    ACTIONS(2151), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57623] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2153), 1,
      aux_sym_source_file_token1,
    ACTIONS(2155), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57637] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2157), 1,
      aux_sym_source_file_token1,
    ACTIONS(2159), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57651] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2161), 1,
      anon_sym_RBRACE,
    ACTIONS(2163), 1,
      anon_sym_case,
    ACTIONS(2166), 1,
      anon_sym_default,
    STATE(1017), 3,
      sym_default_case,
      sym_communication_case,
      aux_sym_select_statement_repeat1,
  [57669] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1937), 1,
      anon_sym_default,
    ACTIONS(2032), 1,
      anon_sym_case,
    ACTIONS(2169), 1,
      anon_sym_RBRACE,
    STATE(1017), 3,
      sym_default_case,
      sym_communication_case,
      aux_sym_select_statement_repeat1,
  [57687] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1059), 1,
      anon_sym_LBRACE,
    ACTIONS(1752), 1,
      anon_sym_DOT,
    ACTIONS(1754), 1,
      anon_sym_LBRACK,
    ACTIONS(2171), 1,
      anon_sym_LPAREN,
    STATE(460), 1,
      sym_parameter_list,
    STATE(869), 1,
      sym_type_arguments,
  [57709] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1585), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1587), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57723] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1607), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1609), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57737] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2173), 1,
      aux_sym_source_file_token1,
    ACTIONS(2175), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57751] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2177), 1,
      aux_sym_source_file_token1,
    ACTIONS(2179), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57765] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2181), 1,
      aux_sym_source_file_token1,
    ACTIONS(2183), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57779] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1635), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1637), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57793] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2185), 1,
      aux_sym_source_file_token1,
    ACTIONS(2187), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57807] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1639), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(1641), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_LBRACK,
      anon_sym_LBRACE,
  [57821] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2189), 1,
      aux_sym_source_file_token1,
    ACTIONS(2191), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57835] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2193), 1,
      aux_sym_source_file_token1,
    ACTIONS(2195), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57849] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2197), 1,
      aux_sym_source_file_token1,
    ACTIONS(2199), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57863] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2201), 1,
      aux_sym_source_file_token1,
    ACTIONS(2203), 5,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [57877] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2205), 1,
      aux_sym_source_file_token1,
    ACTIONS(2209), 1,
      anon_sym_RPAREN,
    STATE(1043), 1,
      aux_sym_type_declaration_repeat1,
    ACTIONS(2207), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [57894] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2211), 1,
      aux_sym_source_file_token1,
    ACTIONS(2215), 1,
      anon_sym_RPAREN,
    STATE(1047), 1,
      aux_sym_import_spec_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [57911] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1585), 1,
      anon_sym_LPAREN,
    ACTIONS(1771), 1,
      anon_sym_LBRACK,
    ACTIONS(2217), 1,
      anon_sym_LBRACE,
    STATE(483), 1,
      sym_literal_value,
    STATE(874), 1,
      sym_type_arguments,
  [57930] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2219), 1,
      aux_sym_source_file_token1,
    ACTIONS(2223), 1,
      anon_sym_RBRACE,
    STATE(1036), 1,
      aux_sym_interface_type_repeat1,
    ACTIONS(2221), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [57947] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2225), 1,
      aux_sym_source_file_token1,
    ACTIONS(2231), 1,
      anon_sym_RBRACE,
    STATE(1036), 1,
      aux_sym_interface_type_repeat1,
    ACTIONS(2228), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [57964] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(861), 1,
      anon_sym_LBRACE,
    ACTIONS(1585), 1,
      anon_sym_LPAREN,
    ACTIONS(1771), 1,
      anon_sym_LBRACK,
    STATE(362), 1,
      sym_literal_value,
    STATE(874), 1,
      sym_type_arguments,
  [57983] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2233), 1,
      aux_sym_source_file_token1,
    ACTIONS(2237), 1,
      anon_sym_RBRACE,
    STATE(1046), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(2235), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58000] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1585), 1,
      anon_sym_LPAREN,
    ACTIONS(1771), 1,
      anon_sym_LBRACK,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    STATE(318), 1,
      sym_literal_value,
    STATE(874), 1,
      sym_type_arguments,
  [58019] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2241), 1,
      aux_sym_source_file_token1,
    ACTIONS(2245), 1,
      anon_sym_RBRACE,
    STATE(1036), 1,
      aux_sym_interface_type_repeat1,
    ACTIONS(2243), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58036] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2247), 1,
      aux_sym_source_file_token1,
    ACTIONS(2251), 1,
      anon_sym_RBRACE,
    STATE(1035), 1,
      aux_sym_interface_type_repeat1,
    ACTIONS(2249), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58053] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2253), 1,
      aux_sym_source_file_token1,
    ACTIONS(2257), 1,
      anon_sym_RPAREN,
    STATE(1044), 1,
      aux_sym_import_spec_list_repeat1,
    ACTIONS(2255), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58070] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2259), 1,
      aux_sym_source_file_token1,
    ACTIONS(2265), 1,
      anon_sym_RPAREN,
    STATE(1043), 1,
      aux_sym_type_declaration_repeat1,
    ACTIONS(2262), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58087] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2267), 1,
      aux_sym_source_file_token1,
    ACTIONS(2271), 1,
      anon_sym_RPAREN,
    STATE(1047), 1,
      aux_sym_import_spec_list_repeat1,
    ACTIONS(2269), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58104] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2273), 1,
      aux_sym_source_file_token1,
    ACTIONS(2277), 1,
      anon_sym_RBRACE,
    STATE(1050), 1,
      aux_sym_interface_type_repeat1,
    ACTIONS(2275), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58121] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2279), 1,
      aux_sym_source_file_token1,
    ACTIONS(2285), 1,
      anon_sym_RBRACE,
    STATE(1046), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(2282), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58138] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2287), 1,
      aux_sym_source_file_token1,
    ACTIONS(2293), 1,
      anon_sym_RPAREN,
    STATE(1047), 1,
      aux_sym_import_spec_list_repeat1,
    ACTIONS(2290), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58155] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1120), 1,
      anon_sym_LBRACE,
    ACTIONS(2137), 1,
      aux_sym_source_file_token1,
    STATE(1185), 1,
      sym_block,
    ACTIONS(2139), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58172] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1120), 1,
      anon_sym_LBRACE,
    ACTIONS(1925), 1,
      aux_sym_source_file_token1,
    STATE(1186), 1,
      sym_block,
    ACTIONS(1927), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58189] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2295), 1,
      aux_sym_source_file_token1,
    ACTIONS(2299), 1,
      anon_sym_RBRACE,
    STATE(1036), 1,
      aux_sym_interface_type_repeat1,
    ACTIONS(2297), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58206] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2088), 1,
      aux_sym_source_file_token1,
    ACTIONS(2090), 4,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
      anon_sym_PIPE,
  [58219] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2301), 1,
      aux_sym_source_file_token1,
    ACTIONS(2305), 1,
      anon_sym_RPAREN,
    STATE(1032), 1,
      aux_sym_type_declaration_repeat1,
    ACTIONS(2303), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58236] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1585), 1,
      anon_sym_LPAREN,
    ACTIONS(1771), 1,
      anon_sym_LBRACK,
    ACTIONS(2307), 1,
      anon_sym_LBRACE,
    STATE(284), 1,
      sym_literal_value,
    STATE(874), 1,
      sym_type_arguments,
  [58255] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2309), 1,
      aux_sym_source_file_token1,
    ACTIONS(2313), 1,
      anon_sym_RBRACE,
    STATE(1046), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(2311), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58272] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2315), 1,
      aux_sym_source_file_token1,
    ACTIONS(2319), 1,
      anon_sym_RBRACE,
    STATE(1054), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(2317), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58289] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(31), 1,
      anon_sym_TILDE,
    ACTIONS(2321), 1,
      sym_identifier,
    STATE(852), 1,
      sym_qualified_type,
    STATE(893), 1,
      sym_generic_type,
    STATE(1332), 1,
      sym_negated_type,
  [58308] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2323), 1,
      aux_sym_source_file_token1,
    ACTIONS(2327), 1,
      anon_sym_RPAREN,
    STATE(1033), 1,
      aux_sym_import_spec_list_repeat1,
    ACTIONS(2325), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58325] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    ACTIONS(1585), 1,
      anon_sym_LPAREN,
    ACTIONS(1771), 1,
      anon_sym_LBRACK,
    STATE(549), 1,
      sym_literal_value,
    STATE(874), 1,
      sym_type_arguments,
  [58344] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(1120), 1,
      anon_sym_LBRACE,
    ACTIONS(1989), 1,
      aux_sym_source_file_token1,
    STATE(1236), 1,
      sym_block,
    ACTIONS(1991), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58361] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2329), 1,
      aux_sym_source_file_token1,
    ACTIONS(2333), 1,
      anon_sym_RBRACE,
    STATE(1038), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(2331), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58378] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2335), 1,
      aux_sym_source_file_token1,
    ACTIONS(2339), 1,
      anon_sym_RBRACE,
    STATE(1046), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(2337), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58395] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2341), 1,
      aux_sym_source_file_token1,
    ACTIONS(2345), 1,
      anon_sym_RBRACE,
    STATE(1061), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(2343), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58412] = 5,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2347), 1,
      aux_sym_source_file_token1,
    ACTIONS(2351), 1,
      anon_sym_RBRACE,
    STATE(1040), 1,
      aux_sym_interface_type_repeat1,
    ACTIONS(2349), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58429] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1316), 1,
      anon_sym_LBRACE,
    STATE(495), 1,
      sym_block,
    ACTIONS(1623), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [58443] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2353), 1,
      sym_identifier,
    ACTIONS(2356), 1,
      anon_sym_RPAREN,
    STATE(1065), 1,
      aux_sym_var_spec_list_repeat1,
    STATE(1227), 1,
      sym_var_spec,
  [58459] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2358), 1,
      sym_identifier,
    ACTIONS(2360), 1,
      anon_sym_RPAREN,
    STATE(1052), 2,
      sym_type_alias,
      sym_type_spec,
  [58473] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2362), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2364), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58485] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2366), 1,
      sym_identifier,
    ACTIONS(2368), 1,
      anon_sym_RPAREN,
    STATE(1140), 1,
      aux_sym_var_spec_list_repeat1,
    STATE(1227), 1,
      sym_var_spec,
  [58501] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2370), 1,
      ts_builtin_sym_end,
    ACTIONS(2372), 1,
      aux_sym_source_file_token1,
    ACTIONS(2374), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58515] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2376), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2378), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58527] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2380), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2382), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58539] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2024), 1,
      anon_sym_PIPE,
    STATE(1088), 1,
      aux_sym_type_elem_repeat1,
    ACTIONS(2076), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [58553] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2384), 1,
      aux_sym_source_file_token1,
    ACTIONS(2265), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
  [58565] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2386), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2388), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58577] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2024), 1,
      anon_sym_PIPE,
    STATE(1072), 1,
      aux_sym_type_elem_repeat1,
    ACTIONS(1983), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [58591] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2390), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2392), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58603] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2394), 1,
      anon_sym_COMMA,
    STATE(1077), 1,
      aux_sym_type_case_repeat1,
    ACTIONS(2397), 2,
      anon_sym_RBRACK,
      anon_sym_COLON,
  [58617] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(682), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(684), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58629] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2358), 1,
      sym_identifier,
    ACTIONS(2399), 1,
      anon_sym_RPAREN,
    STATE(1073), 2,
      sym_type_alias,
      sym_type_spec,
  [58643] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2403), 1,
      anon_sym_DQUOTE2,
    STATE(1085), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2401), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [58657] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2217), 1,
      anon_sym_LBRACE,
    STATE(483), 1,
      sym_literal_value,
    ACTIONS(1585), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [58671] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2217), 1,
      anon_sym_LBRACE,
    STATE(483), 1,
      sym_literal_value,
    ACTIONS(1639), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [58685] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2407), 1,
      anon_sym_DQUOTE2,
    STATE(1120), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2405), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [58699] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2411), 1,
      anon_sym_DQUOTE2,
    STATE(1083), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2409), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [58713] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2413), 1,
      anon_sym_DQUOTE2,
    STATE(1120), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2405), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [58727] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2415), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2417), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58739] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2419), 1,
      anon_sym_COMMA,
    ACTIONS(2421), 1,
      anon_sym_RBRACE,
    ACTIONS(2423), 1,
      anon_sym_COLON,
    STATE(1167), 1,
      aux_sym_literal_value_repeat1,
  [58755] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2425), 1,
      anon_sym_PIPE,
    STATE(1088), 1,
      aux_sym_type_elem_repeat1,
    ACTIONS(2088), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [58769] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2428), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2430), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58781] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2358), 1,
      sym_identifier,
    ACTIONS(2432), 1,
      anon_sym_RPAREN,
    STATE(1073), 2,
      sym_type_alias,
      sym_type_spec,
  [58795] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2366), 1,
      sym_identifier,
    ACTIONS(2434), 1,
      anon_sym_LPAREN,
    STATE(928), 2,
      sym_var_spec,
      sym_var_spec_list,
  [58809] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2423), 1,
      anon_sym_COLON,
    ACTIONS(2436), 1,
      anon_sym_COMMA,
    ACTIONS(2438), 1,
      anon_sym_RBRACE,
    STATE(1257), 1,
      aux_sym_literal_value_repeat1,
  [58825] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2440), 1,
      aux_sym_source_file_token1,
    ACTIONS(2442), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [58837] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2444), 1,
      aux_sym_source_file_token1,
    ACTIONS(2446), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [58849] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2448), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2450), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58861] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2452), 1,
      aux_sym_source_file_token1,
    ACTIONS(2293), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
  [58873] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2456), 1,
      anon_sym_DQUOTE2,
    STATE(1101), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2454), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [58887] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    STATE(318), 1,
      sym_literal_value,
    ACTIONS(1585), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [58901] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    STATE(318), 1,
      sym_literal_value,
    ACTIONS(1639), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [58915] = 4,
    ACTIONS(77), 1,
      ts_builtin_sym_end,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2372), 1,
      aux_sym_source_file_token1,
    ACTIONS(2374), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58929] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2458), 1,
      anon_sym_DQUOTE2,
    STATE(1120), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2405), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [58943] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2460), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2462), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [58955] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1338), 1,
      anon_sym_LBRACE,
    STATE(336), 1,
      sym_block,
    ACTIONS(1623), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [58969] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2423), 1,
      anon_sym_COLON,
    ACTIONS(2464), 1,
      anon_sym_COMMA,
    ACTIONS(2466), 1,
      anon_sym_RBRACE,
    STATE(1196), 1,
      aux_sym_literal_value_repeat1,
  [58985] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2468), 1,
      aux_sym_source_file_token1,
    ACTIONS(2470), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [58997] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2472), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2474), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59009] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2476), 1,
      aux_sym_source_file_token1,
    ACTIONS(2478), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [59021] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2480), 1,
      aux_sym_source_file_token1,
    ACTIONS(2482), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [59033] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2484), 1,
      aux_sym_source_file_token1,
    ACTIONS(2486), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [59045] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2484), 1,
      aux_sym_source_file_token1,
    ACTIONS(2486), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [59057] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2171), 1,
      anon_sym_LPAREN,
    ACTIONS(2488), 1,
      anon_sym_LBRACK,
    STATE(454), 1,
      sym_parameter_list,
    STATE(1309), 1,
      sym_type_parameter_list,
  [59073] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2484), 1,
      aux_sym_source_file_token1,
    ACTIONS(2486), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [59085] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2490), 1,
      sym_identifier,
    ACTIONS(2493), 1,
      anon_sym_RPAREN,
    STATE(1113), 1,
      aux_sym_const_declaration_repeat1,
    STATE(1219), 1,
      sym_const_spec,
  [59101] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2497), 1,
      anon_sym_DQUOTE2,
    STATE(1147), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2495), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [59115] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2484), 1,
      aux_sym_source_file_token1,
    ACTIONS(2486), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [59127] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2390), 1,
      aux_sym_source_file_token1,
    ACTIONS(2392), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
  [59139] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(861), 1,
      anon_sym_LBRACE,
    STATE(362), 1,
      sym_literal_value,
    ACTIONS(1585), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [59153] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(861), 1,
      anon_sym_LBRACE,
    STATE(362), 1,
      sym_literal_value,
    ACTIONS(1639), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [59167] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2499), 1,
      aux_sym_source_file_token1,
    ACTIONS(2501), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
  [59179] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2506), 1,
      anon_sym_DQUOTE2,
    STATE(1120), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2503), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [59193] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    STATE(1157), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [59207] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2508), 1,
      sym_identifier,
    ACTIONS(2510), 1,
      anon_sym_RPAREN,
    STATE(1133), 1,
      aux_sym_const_declaration_repeat1,
    STATE(1219), 1,
      sym_const_spec,
  [59223] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2512), 1,
      anon_sym_DQUOTE2,
    STATE(1120), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2405), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [59237] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(69), 1,
      anon_sym_BQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    STATE(1116), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [59251] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2514), 1,
      aux_sym_source_file_token1,
    ACTIONS(2285), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [59263] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1358), 1,
      anon_sym_LBRACE,
    STATE(387), 1,
      sym_block,
    ACTIONS(1623), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [59277] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2423), 1,
      anon_sym_COLON,
    ACTIONS(2516), 1,
      anon_sym_COMMA,
    ACTIONS(2518), 1,
      anon_sym_RBRACE,
    STATE(1218), 1,
      aux_sym_literal_value_repeat1,
  [59293] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2520), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2522), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59305] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 1,
      anon_sym_LPAREN,
    ACTIONS(2488), 1,
      anon_sym_LBRACK,
    STATE(492), 1,
      sym_parameter_list,
    STATE(1312), 1,
      sym_type_parameter_list,
  [59321] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(33), 1,
      anon_sym_LBRACE,
    ACTIONS(2524), 1,
      anon_sym_if,
    STATE(927), 2,
      sym_block,
      sym_if_statement,
  [59335] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(674), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(676), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59347] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(666), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(668), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59359] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2508), 1,
      sym_identifier,
    ACTIONS(2526), 1,
      anon_sym_RPAREN,
    STATE(1113), 1,
      aux_sym_const_declaration_repeat1,
    STATE(1219), 1,
      sym_const_spec,
  [59375] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2528), 1,
      aux_sym_source_file_token1,
    ACTIONS(2530), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [59387] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(33), 1,
      anon_sym_LBRACE,
    ACTIONS(2524), 1,
      anon_sym_if,
    STATE(958), 2,
      sym_block,
      sym_if_statement,
  [59401] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2534), 1,
      anon_sym_DQUOTE2,
    STATE(1142), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2532), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [59415] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    STATE(549), 1,
      sym_literal_value,
    ACTIONS(1585), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [59429] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    STATE(549), 1,
      sym_literal_value,
    ACTIONS(1639), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [59443] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2536), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2538), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59455] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2366), 1,
      sym_identifier,
    ACTIONS(2540), 1,
      anon_sym_RPAREN,
    STATE(1065), 1,
      aux_sym_var_spec_list_repeat1,
    STATE(1227), 1,
      sym_var_spec,
  [59471] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2542), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2544), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59483] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2546), 1,
      anon_sym_DQUOTE2,
    STATE(1120), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2405), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [59497] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1376), 1,
      anon_sym_LBRACE,
    STATE(558), 1,
      sym_block,
    ACTIONS(1623), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [59511] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2423), 1,
      anon_sym_COLON,
    ACTIONS(2548), 1,
      anon_sym_COMMA,
    ACTIONS(2550), 1,
      anon_sym_RBRACE,
    STATE(1238), 1,
      aux_sym_literal_value_repeat1,
  [59527] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2552), 1,
      aux_sym_source_file_token1,
    ACTIONS(2231), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RBRACE,
  [59539] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1767), 1,
      anon_sym_BQUOTE,
    ACTIONS(1769), 1,
      anon_sym_DQUOTE,
    STATE(1076), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [59553] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_DQUOTE2,
    STATE(1120), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2405), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [59567] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(670), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(672), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59579] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2307), 1,
      anon_sym_LBRACE,
    STATE(284), 1,
      sym_literal_value,
    ACTIONS(1585), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [59593] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2558), 1,
      anon_sym_DQUOTE2,
    STATE(1151), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2556), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [59607] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2560), 1,
      anon_sym_DQUOTE2,
    STATE(1120), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2405), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [59621] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2358), 1,
      sym_identifier,
    ACTIONS(2562), 1,
      anon_sym_LPAREN,
    STATE(972), 2,
      sym_type_alias,
      sym_type_spec,
  [59635] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(33), 1,
      anon_sym_LBRACE,
    STATE(256), 1,
      sym_block,
    ACTIONS(1623), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [59649] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2307), 1,
      anon_sym_LBRACE,
    STATE(284), 1,
      sym_literal_value,
    ACTIONS(1639), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [59663] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2499), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2501), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59675] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1767), 1,
      anon_sym_BQUOTE,
    ACTIONS(1769), 1,
      anon_sym_DQUOTE,
    STATE(1070), 2,
      sym_raw_string_literal,
      sym_interpreted_string_literal,
  [59689] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2376), 1,
      aux_sym_source_file_token1,
    ACTIONS(2378), 3,
      anon_sym_SEMI,
      anon_sym_NULL,
      anon_sym_RPAREN,
  [59701] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2423), 1,
      anon_sym_COLON,
    ACTIONS(2564), 1,
      anon_sym_COMMA,
    ACTIONS(2566), 1,
      anon_sym_RBRACE,
    STATE(1199), 1,
      aux_sym_literal_value_repeat1,
  [59717] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(678), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(680), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59729] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2568), 2,
      ts_builtin_sym_end,
      aux_sym_source_file_token1,
    ACTIONS(2570), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59741] = 4,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2574), 1,
      anon_sym_DQUOTE2,
    STATE(1123), 1,
      aux_sym_interpreted_string_literal_repeat1,
    ACTIONS(2572), 2,
      aux_sym_interpreted_string_literal_token1,
      sym_escape_sequence,
  [59755] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(214), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [59764] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2088), 3,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_PIPE,
  [59773] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(480), 1,
      anon_sym_RPAREN,
    ACTIONS(2576), 1,
      anon_sym_COMMA,
    STATE(1266), 1,
      aux_sym_expression_list_repeat1,
  [59786] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2578), 1,
      anon_sym_RPAREN,
    ACTIONS(2580), 1,
      anon_sym_COMMA,
    STATE(1249), 1,
      aux_sym_expression_list_repeat1,
  [59799] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1326), 1,
      anon_sym_RBRACK,
    ACTIONS(2582), 1,
      anon_sym_COMMA,
    STATE(1224), 1,
      aux_sym_type_arguments_repeat1,
  [59812] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(364), 1,
      anon_sym_RBRACE,
    ACTIONS(2584), 1,
      anon_sym_COMMA,
    STATE(1172), 1,
      aux_sym_literal_value_repeat1,
  [59825] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1498), 1,
      anon_sym_RBRACK,
    ACTIONS(2586), 1,
      anon_sym_COMMA,
    STATE(1077), 1,
      aux_sym_type_case_repeat1,
  [59838] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2588), 1,
      anon_sym_COMMA,
    ACTIONS(2590), 1,
      anon_sym_RBRACK,
    STATE(1253), 1,
      aux_sym_type_arguments_repeat1,
  [59851] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(434), 1,
      anon_sym_RPAREN,
    ACTIONS(2592), 1,
      anon_sym_COMMA,
    STATE(1206), 1,
      aux_sym_argument_list_repeat1,
  [59864] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2448), 1,
      aux_sym_source_file_token1,
    ACTIONS(2450), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [59875] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2594), 1,
      anon_sym_COMMA,
    ACTIONS(2597), 1,
      anon_sym_RBRACE,
    STATE(1172), 1,
      aux_sym_literal_value_repeat1,
  [59888] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2599), 1,
      sym_identifier,
    ACTIONS(2601), 1,
      anon_sym_RBRACK,
    STATE(1277), 1,
      sym_type_parameter_declaration,
  [59901] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2603), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [59910] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1248), 3,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_COLON,
  [59919] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2397), 3,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_COLON,
  [59928] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2423), 1,
      anon_sym_COLON,
    ACTIONS(2597), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [59939] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2436), 1,
      anon_sym_COMMA,
    ACTIONS(2438), 1,
      anon_sym_RBRACE,
    STATE(1257), 1,
      aux_sym_literal_value_repeat1,
  [59952] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1176), 1,
      anon_sym_RPAREN,
    ACTIONS(2605), 1,
      anon_sym_COMMA,
    STATE(1207), 1,
      aux_sym_parameter_list_repeat1,
  [59965] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2607), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [59974] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2599), 1,
      sym_identifier,
    ACTIONS(2609), 1,
      anon_sym_RBRACK,
    STATE(1277), 1,
      sym_type_parameter_declaration,
  [59987] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2611), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [59996] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2428), 1,
      aux_sym_source_file_token1,
    ACTIONS(2430), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60007] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2613), 1,
      anon_sym_COMMA,
    ACTIONS(2616), 1,
      anon_sym_RBRACK,
    STATE(1184), 1,
      aux_sym_type_parameter_list_repeat1,
  [60020] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2460), 1,
      aux_sym_source_file_token1,
    ACTIONS(2462), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60031] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2472), 1,
      aux_sym_source_file_token1,
    ACTIONS(2474), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60042] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(516), 1,
      anon_sym_RPAREN,
    ACTIONS(2618), 1,
      anon_sym_COMMA,
    STATE(1266), 1,
      aux_sym_expression_list_repeat1,
  [60055] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2620), 1,
      anon_sym_RPAREN,
    ACTIONS(2622), 1,
      anon_sym_COMMA,
    STATE(1194), 1,
      aux_sym_expression_list_repeat1,
  [60068] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2464), 1,
      anon_sym_COMMA,
    ACTIONS(2466), 1,
      anon_sym_RBRACE,
    STATE(1196), 1,
      aux_sym_literal_value_repeat1,
  [60081] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2624), 1,
      anon_sym_COMMA,
    ACTIONS(2626), 1,
      anon_sym_RBRACK,
    STATE(1197), 1,
      aux_sym_type_case_repeat1,
  [60094] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1196), 1,
      anon_sym_RPAREN,
    ACTIONS(1198), 1,
      anon_sym_COMMA,
    STATE(1198), 1,
      aux_sym_argument_list_repeat1,
  [60107] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1296), 1,
      anon_sym_RBRACK,
    ACTIONS(2628), 1,
      anon_sym_COMMA,
    STATE(1224), 1,
      aux_sym_type_arguments_repeat1,
  [60120] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2020), 1,
      anon_sym_LPAREN,
    ACTIONS(2022), 1,
      anon_sym_LBRACK,
    ACTIONS(2630), 1,
      anon_sym_RPAREN,
  [60133] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(478), 1,
      anon_sym_RPAREN,
    ACTIONS(2632), 1,
      anon_sym_COMMA,
    STATE(1266), 1,
      aux_sym_expression_list_repeat1,
  [60146] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2358), 1,
      sym_identifier,
    STATE(1073), 2,
      sym_type_alias,
      sym_type_spec,
  [60157] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(356), 1,
      anon_sym_RBRACE,
    ACTIONS(2634), 1,
      anon_sym_COMMA,
    STATE(1172), 1,
      aux_sym_literal_value_repeat1,
  [60170] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1424), 1,
      anon_sym_RBRACK,
    ACTIONS(2636), 1,
      anon_sym_COMMA,
    STATE(1077), 1,
      aux_sym_type_case_repeat1,
  [60183] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(454), 1,
      anon_sym_RPAREN,
    ACTIONS(2638), 1,
      anon_sym_COMMA,
    STATE(1206), 1,
      aux_sym_argument_list_repeat1,
  [60196] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(350), 1,
      anon_sym_RBRACE,
    ACTIONS(2640), 1,
      anon_sym_COMMA,
    STATE(1172), 1,
      aux_sym_literal_value_repeat1,
  [60209] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2642), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [60218] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1180), 1,
      anon_sym_RPAREN,
    ACTIONS(2644), 1,
      anon_sym_COMMA,
    STATE(1207), 1,
      aux_sym_parameter_list_repeat1,
  [60231] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2542), 1,
      aux_sym_source_file_token1,
    ACTIONS(2544), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60242] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2646), 1,
      anon_sym_COMMA,
    ACTIONS(2648), 1,
      anon_sym_COLON,
    STATE(1077), 1,
      aux_sym_type_case_repeat1,
  [60255] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1472), 1,
      anon_sym_RBRACK,
    ACTIONS(2650), 1,
      anon_sym_COMMA,
    STATE(1077), 1,
      aux_sym_type_case_repeat1,
  [60268] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2520), 1,
      aux_sym_source_file_token1,
    ACTIONS(2522), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60279] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1232), 1,
      anon_sym_RPAREN,
    ACTIONS(2652), 1,
      anon_sym_COMMA,
    STATE(1206), 1,
      aux_sym_argument_list_repeat1,
  [60292] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2655), 1,
      anon_sym_RPAREN,
    ACTIONS(2657), 1,
      anon_sym_COMMA,
    STATE(1207), 1,
      aux_sym_parameter_list_repeat1,
  [60305] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2415), 1,
      aux_sym_source_file_token1,
    ACTIONS(2417), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60316] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2660), 1,
      anon_sym_RPAREN,
    ACTIONS(2662), 1,
      anon_sym_COMMA,
    STATE(1217), 1,
      aux_sym_expression_list_repeat1,
  [60329] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2516), 1,
      anon_sym_COMMA,
    ACTIONS(2518), 1,
      anon_sym_RBRACE,
    STATE(1218), 1,
      aux_sym_literal_value_repeat1,
  [60342] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2664), 1,
      anon_sym_COMMA,
    ACTIONS(2666), 1,
      anon_sym_RBRACK,
    STATE(1213), 1,
      aux_sym_type_parameter_list_repeat1,
  [60355] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2668), 1,
      anon_sym_COMMA,
    ACTIONS(2670), 1,
      anon_sym_RBRACK,
    STATE(1220), 1,
      aux_sym_type_case_repeat1,
  [60368] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2601), 1,
      anon_sym_RBRACK,
    ACTIONS(2672), 1,
      anon_sym_COMMA,
    STATE(1184), 1,
      aux_sym_type_parameter_list_repeat1,
  [60381] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1204), 1,
      anon_sym_RPAREN,
    ACTIONS(1206), 1,
      anon_sym_COMMA,
    STATE(1221), 1,
      aux_sym_argument_list_repeat1,
  [60394] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2674), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [60403] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2676), 1,
      anon_sym_COMMA,
    ACTIONS(2678), 1,
      anon_sym_RBRACK,
    STATE(1262), 1,
      aux_sym_type_case_repeat1,
  [60416] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(524), 1,
      anon_sym_RPAREN,
    ACTIONS(2680), 1,
      anon_sym_COMMA,
    STATE(1266), 1,
      aux_sym_expression_list_repeat1,
  [60429] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(360), 1,
      anon_sym_RBRACE,
    ACTIONS(2682), 1,
      anon_sym_COMMA,
    STATE(1172), 1,
      aux_sym_literal_value_repeat1,
  [60442] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2684), 1,
      aux_sym_source_file_token1,
    ACTIONS(2686), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60453] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1400), 1,
      anon_sym_RBRACK,
    ACTIONS(2688), 1,
      anon_sym_COMMA,
    STATE(1077), 1,
      aux_sym_type_case_repeat1,
  [60466] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(460), 1,
      anon_sym_RPAREN,
    ACTIONS(2690), 1,
      anon_sym_COMMA,
    STATE(1206), 1,
      aux_sym_argument_list_repeat1,
  [60479] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(2692), 1,
      sym_identifier,
    STATE(606), 1,
      sym_parameter_list,
  [60492] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2386), 1,
      aux_sym_source_file_token1,
    ACTIONS(2388), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60503] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2694), 1,
      anon_sym_COMMA,
    ACTIONS(2697), 1,
      anon_sym_RBRACK,
    STATE(1224), 1,
      aux_sym_type_arguments_repeat1,
  [60516] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2699), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [60525] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2508), 1,
      sym_identifier,
    ACTIONS(2701), 1,
      anon_sym_LPAREN,
    STATE(1001), 1,
      sym_const_spec,
  [60538] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2703), 1,
      aux_sym_source_file_token1,
    ACTIONS(2705), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60549] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1188), 1,
      anon_sym_RPAREN,
    ACTIONS(1190), 1,
      anon_sym_COMMA,
    STATE(1264), 1,
      aux_sym_argument_list_repeat1,
  [60562] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2646), 1,
      anon_sym_COMMA,
    ACTIONS(2707), 1,
      anon_sym_COLON,
    STATE(1203), 1,
      aux_sym_type_case_repeat1,
  [60575] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2709), 1,
      anon_sym_RPAREN,
    ACTIONS(2711), 1,
      anon_sym_COMMA,
    STATE(1237), 1,
      aux_sym_expression_list_repeat1,
  [60588] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2713), 1,
      anon_sym_RPAREN,
    ACTIONS(2715), 1,
      anon_sym_COMMA,
    STATE(1265), 1,
      aux_sym_parameter_list_repeat1,
  [60601] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2548), 1,
      anon_sym_COMMA,
    ACTIONS(2550), 1,
      anon_sym_RBRACE,
    STATE(1238), 1,
      aux_sym_literal_value_repeat1,
  [60614] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2717), 1,
      anon_sym_COMMA,
    ACTIONS(2719), 1,
      anon_sym_RBRACK,
    STATE(1239), 1,
      aux_sym_type_case_repeat1,
  [60627] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2536), 1,
      aux_sym_source_file_token1,
    ACTIONS(2538), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60638] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1208), 1,
      anon_sym_RPAREN,
    ACTIONS(1210), 1,
      anon_sym_COMMA,
    STATE(1240), 1,
      aux_sym_argument_list_repeat1,
  [60651] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2568), 1,
      aux_sym_source_file_token1,
    ACTIONS(2570), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60662] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(538), 1,
      anon_sym_RPAREN,
    ACTIONS(2721), 1,
      anon_sym_COMMA,
    STATE(1266), 1,
      aux_sym_expression_list_repeat1,
  [60675] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(372), 1,
      anon_sym_RBRACE,
    ACTIONS(2723), 1,
      anon_sym_COMMA,
    STATE(1172), 1,
      aux_sym_literal_value_repeat1,
  [60688] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1476), 1,
      anon_sym_RBRACK,
    ACTIONS(2725), 1,
      anon_sym_COMMA,
    STATE(1077), 1,
      aux_sym_type_case_repeat1,
  [60701] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(466), 1,
      anon_sym_RPAREN,
    ACTIONS(2727), 1,
      anon_sym_COMMA,
    STATE(1206), 1,
      aux_sym_argument_list_repeat1,
  [60714] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2729), 1,
      anon_sym_RPAREN,
    ACTIONS(2731), 1,
      anon_sym_COMMA,
    STATE(1164), 1,
      aux_sym_expression_list_repeat1,
  [60727] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2733), 1,
      anon_sym_RPAREN,
    ACTIONS(2735), 1,
      anon_sym_COMMA,
    STATE(1201), 1,
      aux_sym_parameter_list_repeat1,
  [60740] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2737), 1,
      anon_sym_COMMA,
    ACTIONS(2739), 1,
      anon_sym_RBRACK,
    STATE(1166), 1,
      aux_sym_type_arguments_repeat1,
  [60753] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1264), 1,
      anon_sym_COMMA,
    ACTIONS(1577), 1,
      anon_sym_LBRACE,
    STATE(1248), 1,
      aux_sym_expression_list_repeat1,
  [60766] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    ACTIONS(2741), 1,
      sym_identifier,
    STATE(641), 1,
      sym_parameter_list,
  [60779] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2743), 1,
      anon_sym_RPAREN,
    ACTIONS(2745), 1,
      anon_sym_COMMA,
    STATE(1179), 1,
      aux_sym_parameter_list_repeat1,
  [60792] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2747), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [60801] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(867), 1,
      anon_sym_LBRACE,
    ACTIONS(2749), 1,
      anon_sym_COMMA,
    STATE(1248), 1,
      aux_sym_expression_list_repeat1,
  [60814] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(540), 1,
      anon_sym_RPAREN,
    ACTIONS(2752), 1,
      anon_sym_COMMA,
    STATE(1266), 1,
      aux_sym_expression_list_repeat1,
  [60827] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2419), 1,
      anon_sym_COMMA,
    ACTIONS(2421), 1,
      anon_sym_RBRACE,
    STATE(1167), 1,
      aux_sym_literal_value_repeat1,
  [60840] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2754), 1,
      anon_sym_COMMA,
    ACTIONS(2756), 1,
      anon_sym_RBRACK,
    STATE(1168), 1,
      aux_sym_type_case_repeat1,
  [60853] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2372), 1,
      aux_sym_source_file_token1,
    ACTIONS(2374), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60864] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1398), 1,
      anon_sym_RBRACK,
    ACTIONS(2758), 1,
      anon_sym_COMMA,
    STATE(1224), 1,
      aux_sym_type_arguments_repeat1,
  [60877] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2760), 1,
      anon_sym_RPAREN,
    ACTIONS(2762), 1,
      anon_sym_COMMA,
    STATE(1187), 1,
      aux_sym_expression_list_repeat1,
  [60890] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2764), 1,
      anon_sym_COMMA,
    ACTIONS(2766), 1,
      anon_sym_RBRACK,
    STATE(1192), 1,
      aux_sym_type_arguments_repeat1,
  [60903] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2362), 1,
      aux_sym_source_file_token1,
    ACTIONS(2364), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60914] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(354), 1,
      anon_sym_RBRACE,
    ACTIONS(2768), 1,
      anon_sym_COMMA,
    STATE(1172), 1,
      aux_sym_literal_value_repeat1,
  [60927] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1184), 1,
      anon_sym_RPAREN,
    ACTIONS(1186), 1,
      anon_sym_COMMA,
    STATE(1170), 1,
      aux_sym_argument_list_repeat1,
  [60940] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2564), 1,
      anon_sym_COMMA,
    ACTIONS(2566), 1,
      anon_sym_RBRACE,
    STATE(1199), 1,
      aux_sym_literal_value_repeat1,
  [60953] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2770), 1,
      anon_sym_COMMA,
    ACTIONS(2772), 1,
      anon_sym_RBRACK,
    STATE(1204), 1,
      aux_sym_type_case_repeat1,
  [60966] = 3,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2380), 1,
      aux_sym_source_file_token1,
    ACTIONS(2382), 2,
      anon_sym_SEMI,
      anon_sym_NULL,
  [60977] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1500), 1,
      anon_sym_RBRACK,
    ACTIONS(2774), 1,
      anon_sym_COMMA,
    STATE(1077), 1,
      aux_sym_type_case_repeat1,
  [60990] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1154), 1,
      anon_sym_RPAREN,
    ACTIONS(1156), 1,
      anon_sym_COMMA,
    STATE(1267), 1,
      aux_sym_argument_list_repeat1,
  [61003] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(448), 1,
      anon_sym_RPAREN,
    ACTIONS(2776), 1,
      anon_sym_COMMA,
    STATE(1206), 1,
      aux_sym_argument_list_repeat1,
  [61016] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1192), 1,
      anon_sym_RPAREN,
    ACTIONS(2778), 1,
      anon_sym_COMMA,
    STATE(1207), 1,
      aux_sym_parameter_list_repeat1,
  [61029] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(867), 1,
      anon_sym_RPAREN,
    ACTIONS(2780), 1,
      anon_sym_COMMA,
    STATE(1266), 1,
      aux_sym_expression_list_repeat1,
  [61042] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(428), 1,
      anon_sym_RPAREN,
    ACTIONS(2783), 1,
      anon_sym_COMMA,
    STATE(1206), 1,
      aux_sym_argument_list_repeat1,
  [61055] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2785), 1,
      sym_identifier,
    ACTIONS(2787), 1,
      anon_sym_LPAREN,
  [61065] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2789), 1,
      anon_sym_LPAREN,
    ACTIONS(2791), 1,
      anon_sym_LBRACK,
  [61075] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2793), 1,
      sym_identifier,
    ACTIONS(2795), 1,
      anon_sym_LPAREN,
  [61085] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(865), 1,
      anon_sym_LPAREN,
    STATE(385), 1,
      sym_argument_list,
  [61095] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 1,
      anon_sym_LPAREN,
    STATE(364), 1,
      sym_parameter_list,
  [61105] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2785), 1,
      sym_identifier,
    ACTIONS(2797), 1,
      anon_sym_LPAREN,
  [61115] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(752), 1,
      anon_sym_LPAREN,
    STATE(335), 1,
      sym_argument_list,
  [61125] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2799), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61133] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2597), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [61141] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2616), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [61149] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    STATE(439), 1,
      sym_parameter_list,
  [61159] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2801), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61167] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2803), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [61175] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    STATE(549), 1,
      sym_literal_value,
  [61185] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2805), 1,
      sym_identifier,
    ACTIONS(2807), 1,
      anon_sym_LPAREN,
  [61195] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2171), 1,
      anon_sym_LPAREN,
    STATE(500), 1,
      sym_parameter_list,
  [61205] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2809), 1,
      anon_sym_LPAREN,
    STATE(513), 1,
      sym_argument_list,
  [61215] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2811), 1,
      sym_identifier,
    ACTIONS(2813), 1,
      anon_sym_LPAREN,
  [61225] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1029), 1,
      anon_sym_LPAREN,
    STATE(557), 1,
      sym_argument_list,
  [61235] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1967), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [61243] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2801), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61251] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    STATE(452), 1,
      sym_parameter_list,
  [61261] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2815), 1,
      sym_identifier,
    ACTIONS(2817), 1,
      anon_sym_LPAREN,
  [61271] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1971), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [61279] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    STATE(318), 1,
      sym_literal_value,
  [61289] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    STATE(449), 1,
      sym_parameter_list,
  [61299] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2493), 2,
      anon_sym_RPAREN,
      sym_identifier,
  [61307] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2815), 1,
      sym_identifier,
    ACTIONS(2819), 1,
      anon_sym_LPAREN,
  [61317] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2821), 1,
      anon_sym_LBRACE,
    STATE(806), 1,
      sym_field_declaration_list,
  [61327] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2356), 2,
      anon_sym_RPAREN,
      sym_identifier,
  [61335] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2823), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61343] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(945), 1,
      anon_sym_LPAREN,
    STATE(407), 1,
      sym_argument_list,
  [61353] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2307), 1,
      anon_sym_LBRACE,
    STATE(284), 1,
      sym_literal_value,
  [61363] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2655), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61371] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1951), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [61379] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2825), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [61387] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(33), 1,
      anon_sym_LBRACE,
    STATE(979), 1,
      sym_block,
  [61397] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2827), 1,
      anon_sym_LBRACE,
    STATE(1021), 1,
      sym_field_declaration_list,
  [61407] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    STATE(603), 1,
      sym_parameter_list,
  [61417] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2149), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [61425] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2599), 1,
      sym_identifier,
    STATE(1211), 1,
      sym_type_parameter_declaration,
  [61435] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2171), 1,
      anon_sym_LPAREN,
    STATE(459), 1,
      sym_parameter_list,
  [61445] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2829), 1,
      sym_identifier,
    ACTIONS(2831), 1,
      anon_sym_LPAREN,
  [61455] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2153), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [61463] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 1,
      anon_sym_LPAREN,
    STATE(493), 1,
      sym_parameter_list,
  [61473] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2833), 1,
      anon_sym_LPAREN,
    STATE(269), 1,
      sym_argument_list,
  [61483] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1232), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61491] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    STATE(616), 1,
      sym_parameter_list,
  [61501] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2835), 2,
      anon_sym_BQUOTE,
      anon_sym_DQUOTE,
  [61509] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2599), 1,
      sym_identifier,
    STATE(1277), 1,
      sym_type_parameter_declaration,
  [61519] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2837), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61527] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 1,
      anon_sym_LPAREN,
    STATE(351), 1,
      sym_parameter_list,
  [61537] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    STATE(623), 1,
      sym_parameter_list,
  [61547] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(861), 1,
      anon_sym_LBRACE,
    STATE(362), 1,
      sym_literal_value,
  [61557] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2217), 1,
      anon_sym_LBRACE,
    STATE(483), 1,
      sym_literal_value,
  [61567] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(312), 1,
      anon_sym_LBRACE,
    STATE(399), 1,
      sym_literal_value,
  [61577] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    STATE(629), 1,
      sym_parameter_list,
  [61587] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2815), 1,
      sym_identifier,
    ACTIONS(2839), 1,
      anon_sym_LPAREN,
  [61597] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2171), 1,
      anon_sym_LPAREN,
    STATE(496), 1,
      sym_parameter_list,
  [61607] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2841), 1,
      anon_sym_LBRACE,
    STATE(866), 1,
      sym_field_declaration_list,
  [61617] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2843), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [61625] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2815), 1,
      sym_identifier,
    ACTIONS(2845), 1,
      anon_sym_LPAREN,
  [61635] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2697), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [61643] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1009), 1,
      anon_sym_LPAREN,
    STATE(635), 1,
      sym_parameter_list,
  [61653] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2847), 1,
      anon_sym_LBRACK,
    STATE(814), 1,
      sym_type_arguments,
  [61663] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2849), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61671] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2851), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61679] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2853), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [61687] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2020), 1,
      anon_sym_LPAREN,
    ACTIONS(2022), 1,
      anon_sym_LBRACK,
  [61697] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2855), 1,
      anon_sym_LPAREN,
    ACTIONS(2857), 1,
      anon_sym_LBRACK,
  [61707] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2859), 1,
      anon_sym_LPAREN,
    ACTIONS(2861), 1,
      anon_sym_LBRACK,
  [61717] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2863), 1,
      anon_sym_LPAREN,
    ACTIONS(2865), 1,
      anon_sym_LBRACK,
  [61727] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2867), 1,
      anon_sym_LPAREN,
    ACTIONS(2869), 1,
      anon_sym_LBRACK,
  [61737] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2871), 1,
      anon_sym_RPAREN,
  [61744] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2873), 1,
      anon_sym_RPAREN,
  [61751] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2875), 1,
      anon_sym_RPAREN,
  [61758] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2877), 1,
      anon_sym_RBRACK,
  [61765] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2879), 1,
      sym_identifier,
  [61772] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2881), 1,
      anon_sym_RPAREN,
  [61779] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2883), 1,
      anon_sym_RBRACK,
  [61786] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2885), 1,
      anon_sym_LBRACE,
  [61793] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2887), 1,
      anon_sym_RPAREN,
  [61800] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2733), 1,
      anon_sym_RPAREN,
  [61807] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2889), 1,
      sym_identifier,
  [61814] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2891), 1,
      anon_sym_BQUOTE,
  [61821] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2893), 1,
      anon_sym_RPAREN,
  [61828] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2895), 1,
      anon_sym_COLON_EQ,
  [61835] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2897), 1,
      anon_sym_RBRACE,
  [61842] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2466), 1,
      anon_sym_RBRACE,
  [61849] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2899), 1,
      anon_sym_EQ,
  [61856] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2901), 1,
      anon_sym_SEMI,
  [61863] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2903), 1,
      anon_sym_LBRACE,
  [61870] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2518), 1,
      anon_sym_RBRACE,
  [61877] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2905), 1,
      anon_sym_LBRACE,
  [61884] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2907), 1,
      sym_identifier,
  [61891] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2909), 1,
      anon_sym_LBRACE,
  [61898] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2911), 1,
      anon_sym_LBRACK,
  [61905] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2913), 1,
      anon_sym_RBRACE,
  [61912] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2915), 1,
      anon_sym_BQUOTE,
  [61919] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2421), 1,
      anon_sym_RBRACE,
  [61926] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2917), 1,
      sym_identifier,
  [61933] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2919), 1,
      anon_sym_RBRACK,
  [61940] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2921), 1,
      anon_sym_LBRACE,
  [61947] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2923), 1,
      anon_sym_COLON,
  [61954] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2925), 1,
      anon_sym_LBRACE,
  [61961] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2927), 1,
      anon_sym_LBRACE,
  [61968] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2929), 1,
      anon_sym_LBRACE,
  [61975] = 2,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2931), 1,
      aux_sym_raw_string_literal_token1,
  [61982] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2743), 1,
      anon_sym_RPAREN,
  [61989] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2933), 1,
      anon_sym_chan,
  [61996] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2935), 1,
      anon_sym_EQ,
  [62003] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2937), 1,
      sym_identifier,
  [62010] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2939), 1,
      anon_sym_BQUOTE,
  [62017] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2941), 1,
      anon_sym_BQUOTE,
  [62024] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2943), 1,
      anon_sym_chan,
  [62031] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2945), 1,
      anon_sym_chan,
  [62038] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2947), 1,
      anon_sym_SEMI,
  [62045] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2949), 1,
      anon_sym_SEMI,
  [62052] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2951), 1,
      anon_sym_RBRACE,
  [62059] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2953), 1,
      anon_sym_LBRACE,
  [62066] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2955), 1,
      anon_sym_RPAREN,
  [62073] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2957), 1,
      anon_sym_LBRACE,
  [62080] = 2,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2959), 1,
      aux_sym_raw_string_literal_token1,
  [62087] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_LBRACE,
  [62094] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2963), 1,
      anon_sym_BQUOTE,
  [62101] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2965), 1,
      anon_sym_BQUOTE,
  [62108] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2967), 1,
      anon_sym_COLON,
  [62115] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2969), 1,
      anon_sym_RPAREN,
  [62122] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2971), 1,
      anon_sym_RPAREN,
  [62129] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2973), 1,
      anon_sym_RBRACK,
  [62136] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2975), 1,
      anon_sym_RBRACE,
  [62143] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2977), 1,
      anon_sym_RBRACK,
  [62150] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2979), 1,
      anon_sym_RPAREN,
  [62157] = 2,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2981), 1,
      aux_sym_raw_string_literal_token1,
  [62164] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2983), 1,
      anon_sym_chan,
  [62171] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2985), 1,
      anon_sym_RBRACE,
  [62178] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2566), 1,
      anon_sym_RBRACE,
  [62185] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2987), 1,
      sym_identifier,
  [62192] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2989), 1,
      anon_sym_RPAREN,
  [62199] = 2,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2991), 1,
      aux_sym_raw_string_literal_token1,
  [62206] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2993), 1,
      anon_sym_RPAREN,
  [62213] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2995), 1,
      anon_sym_RPAREN,
  [62220] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2997), 1,
      anon_sym_RPAREN,
  [62227] = 2,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(2999), 1,
      aux_sym_raw_string_literal_token1,
  [62234] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3001), 1,
      anon_sym_BQUOTE,
  [62241] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3003), 1,
      anon_sym_LBRACE,
  [62248] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3005), 1,
      anon_sym_RPAREN,
  [62255] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3007), 1,
      sym_identifier,
  [62262] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3009), 1,
      anon_sym_COLON,
  [62269] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3011), 1,
      anon_sym_RBRACE,
  [62276] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3013), 1,
      anon_sym_RPAREN,
  [62283] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3015), 1,
      ts_builtin_sym_end,
  [62290] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2713), 1,
      anon_sym_RPAREN,
  [62297] = 2,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(3017), 1,
      aux_sym_raw_string_literal_token1,
  [62304] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3019), 1,
      anon_sym_RBRACK,
  [62311] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3021), 1,
      anon_sym_chan,
  [62318] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2550), 1,
      anon_sym_RBRACE,
  [62325] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3023), 1,
      anon_sym_LBRACE,
  [62332] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3025), 1,
      anon_sym_chan,
  [62339] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3027), 1,
      anon_sym_RBRACE,
  [62346] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3029), 1,
      anon_sym_RBRACK,
  [62353] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2630), 1,
      anon_sym_RPAREN,
  [62360] = 2,
    ACTIONS(300), 1,
      sym_comment,
    ACTIONS(3031), 1,
      aux_sym_raw_string_literal_token1,
  [62367] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3033), 1,
      anon_sym_LBRACK,
  [62374] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3035), 1,
      sym_identifier,
  [62381] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3037), 1,
      anon_sym_LBRACK,
  [62388] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2438), 1,
      anon_sym_RBRACE,
  [62395] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3039), 1,
      anon_sym_LBRACK,
  [62402] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_RPAREN,
  [62409] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3043), 1,
      anon_sym_LBRACK,
  [62416] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3045), 1,
      sym_identifier,
  [62423] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_LBRACK,
  [62430] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3049), 1,
      anon_sym_LBRACE,
  [62437] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3051), 1,
      anon_sym_RPAREN,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(29)] = 0,
  [SMALL_STATE(30)] = 131,
  [SMALL_STATE(31)] = 262,
  [SMALL_STATE(32)] = 393,
  [SMALL_STATE(33)] = 524,
  [SMALL_STATE(34)] = 655,
  [SMALL_STATE(35)] = 786,
  [SMALL_STATE(36)] = 914,
  [SMALL_STATE(37)] = 1042,
  [SMALL_STATE(38)] = 1170,
  [SMALL_STATE(39)] = 1298,
  [SMALL_STATE(40)] = 1426,
  [SMALL_STATE(41)] = 1554,
  [SMALL_STATE(42)] = 1682,
  [SMALL_STATE(43)] = 1810,
  [SMALL_STATE(44)] = 1938,
  [SMALL_STATE(45)] = 2066,
  [SMALL_STATE(46)] = 2194,
  [SMALL_STATE(47)] = 2322,
  [SMALL_STATE(48)] = 2447,
  [SMALL_STATE(49)] = 2569,
  [SMALL_STATE(50)] = 2689,
  [SMALL_STATE(51)] = 2808,
  [SMALL_STATE(52)] = 2927,
  [SMALL_STATE(53)] = 3046,
  [SMALL_STATE(54)] = 3165,
  [SMALL_STATE(55)] = 3284,
  [SMALL_STATE(56)] = 3403,
  [SMALL_STATE(57)] = 3522,
  [SMALL_STATE(58)] = 3641,
  [SMALL_STATE(59)] = 3760,
  [SMALL_STATE(60)] = 3879,
  [SMALL_STATE(61)] = 3998,
  [SMALL_STATE(62)] = 4117,
  [SMALL_STATE(63)] = 4236,
  [SMALL_STATE(64)] = 4355,
  [SMALL_STATE(65)] = 4474,
  [SMALL_STATE(66)] = 4593,
  [SMALL_STATE(67)] = 4712,
  [SMALL_STATE(68)] = 4831,
  [SMALL_STATE(69)] = 4950,
  [SMALL_STATE(70)] = 5069,
  [SMALL_STATE(71)] = 5188,
  [SMALL_STATE(72)] = 5307,
  [SMALL_STATE(73)] = 5426,
  [SMALL_STATE(74)] = 5545,
  [SMALL_STATE(75)] = 5664,
  [SMALL_STATE(76)] = 5783,
  [SMALL_STATE(77)] = 5902,
  [SMALL_STATE(78)] = 6021,
  [SMALL_STATE(79)] = 6140,
  [SMALL_STATE(80)] = 6259,
  [SMALL_STATE(81)] = 6378,
  [SMALL_STATE(82)] = 6494,
  [SMALL_STATE(83)] = 6610,
  [SMALL_STATE(84)] = 6726,
  [SMALL_STATE(85)] = 6842,
  [SMALL_STATE(86)] = 6958,
  [SMALL_STATE(87)] = 7074,
  [SMALL_STATE(88)] = 7190,
  [SMALL_STATE(89)] = 7306,
  [SMALL_STATE(90)] = 7422,
  [SMALL_STATE(91)] = 7538,
  [SMALL_STATE(92)] = 7654,
  [SMALL_STATE(93)] = 7770,
  [SMALL_STATE(94)] = 7886,
  [SMALL_STATE(95)] = 8002,
  [SMALL_STATE(96)] = 8118,
  [SMALL_STATE(97)] = 8234,
  [SMALL_STATE(98)] = 8350,
  [SMALL_STATE(99)] = 8466,
  [SMALL_STATE(100)] = 8582,
  [SMALL_STATE(101)] = 8698,
  [SMALL_STATE(102)] = 8814,
  [SMALL_STATE(103)] = 8930,
  [SMALL_STATE(104)] = 9046,
  [SMALL_STATE(105)] = 9162,
  [SMALL_STATE(106)] = 9278,
  [SMALL_STATE(107)] = 9394,
  [SMALL_STATE(108)] = 9510,
  [SMALL_STATE(109)] = 9626,
  [SMALL_STATE(110)] = 9742,
  [SMALL_STATE(111)] = 9858,
  [SMALL_STATE(112)] = 9974,
  [SMALL_STATE(113)] = 10090,
  [SMALL_STATE(114)] = 10206,
  [SMALL_STATE(115)] = 10322,
  [SMALL_STATE(116)] = 10438,
  [SMALL_STATE(117)] = 10554,
  [SMALL_STATE(118)] = 10670,
  [SMALL_STATE(119)] = 10786,
  [SMALL_STATE(120)] = 10902,
  [SMALL_STATE(121)] = 11018,
  [SMALL_STATE(122)] = 11134,
  [SMALL_STATE(123)] = 11250,
  [SMALL_STATE(124)] = 11366,
  [SMALL_STATE(125)] = 11482,
  [SMALL_STATE(126)] = 11598,
  [SMALL_STATE(127)] = 11714,
  [SMALL_STATE(128)] = 11830,
  [SMALL_STATE(129)] = 11946,
  [SMALL_STATE(130)] = 12062,
  [SMALL_STATE(131)] = 12178,
  [SMALL_STATE(132)] = 12291,
  [SMALL_STATE(133)] = 12404,
  [SMALL_STATE(134)] = 12517,
  [SMALL_STATE(135)] = 12630,
  [SMALL_STATE(136)] = 12743,
  [SMALL_STATE(137)] = 12856,
  [SMALL_STATE(138)] = 12969,
  [SMALL_STATE(139)] = 13082,
  [SMALL_STATE(140)] = 13195,
  [SMALL_STATE(141)] = 13308,
  [SMALL_STATE(142)] = 13421,
  [SMALL_STATE(143)] = 13534,
  [SMALL_STATE(144)] = 13647,
  [SMALL_STATE(145)] = 13760,
  [SMALL_STATE(146)] = 13873,
  [SMALL_STATE(147)] = 13986,
  [SMALL_STATE(148)] = 14099,
  [SMALL_STATE(149)] = 14212,
  [SMALL_STATE(150)] = 14325,
  [SMALL_STATE(151)] = 14438,
  [SMALL_STATE(152)] = 14551,
  [SMALL_STATE(153)] = 14664,
  [SMALL_STATE(154)] = 14777,
  [SMALL_STATE(155)] = 14890,
  [SMALL_STATE(156)] = 15003,
  [SMALL_STATE(157)] = 15116,
  [SMALL_STATE(158)] = 15229,
  [SMALL_STATE(159)] = 15342,
  [SMALL_STATE(160)] = 15455,
  [SMALL_STATE(161)] = 15568,
  [SMALL_STATE(162)] = 15681,
  [SMALL_STATE(163)] = 15794,
  [SMALL_STATE(164)] = 15907,
  [SMALL_STATE(165)] = 16020,
  [SMALL_STATE(166)] = 16133,
  [SMALL_STATE(167)] = 16246,
  [SMALL_STATE(168)] = 16359,
  [SMALL_STATE(169)] = 16472,
  [SMALL_STATE(170)] = 16585,
  [SMALL_STATE(171)] = 16698,
  [SMALL_STATE(172)] = 16811,
  [SMALL_STATE(173)] = 16924,
  [SMALL_STATE(174)] = 17037,
  [SMALL_STATE(175)] = 17150,
  [SMALL_STATE(176)] = 17263,
  [SMALL_STATE(177)] = 17376,
  [SMALL_STATE(178)] = 17489,
  [SMALL_STATE(179)] = 17602,
  [SMALL_STATE(180)] = 17715,
  [SMALL_STATE(181)] = 17828,
  [SMALL_STATE(182)] = 17941,
  [SMALL_STATE(183)] = 18054,
  [SMALL_STATE(184)] = 18167,
  [SMALL_STATE(185)] = 18280,
  [SMALL_STATE(186)] = 18393,
  [SMALL_STATE(187)] = 18506,
  [SMALL_STATE(188)] = 18619,
  [SMALL_STATE(189)] = 18732,
  [SMALL_STATE(190)] = 18845,
  [SMALL_STATE(191)] = 18958,
  [SMALL_STATE(192)] = 19071,
  [SMALL_STATE(193)] = 19184,
  [SMALL_STATE(194)] = 19297,
  [SMALL_STATE(195)] = 19410,
  [SMALL_STATE(196)] = 19523,
  [SMALL_STATE(197)] = 19636,
  [SMALL_STATE(198)] = 19749,
  [SMALL_STATE(199)] = 19862,
  [SMALL_STATE(200)] = 19975,
  [SMALL_STATE(201)] = 20088,
  [SMALL_STATE(202)] = 20201,
  [SMALL_STATE(203)] = 20314,
  [SMALL_STATE(204)] = 20427,
  [SMALL_STATE(205)] = 20540,
  [SMALL_STATE(206)] = 20653,
  [SMALL_STATE(207)] = 20766,
  [SMALL_STATE(208)] = 20879,
  [SMALL_STATE(209)] = 20992,
  [SMALL_STATE(210)] = 21105,
  [SMALL_STATE(211)] = 21218,
  [SMALL_STATE(212)] = 21331,
  [SMALL_STATE(213)] = 21444,
  [SMALL_STATE(214)] = 21557,
  [SMALL_STATE(215)] = 21670,
  [SMALL_STATE(216)] = 21783,
  [SMALL_STATE(217)] = 21896,
  [SMALL_STATE(218)] = 22009,
  [SMALL_STATE(219)] = 22122,
  [SMALL_STATE(220)] = 22235,
  [SMALL_STATE(221)] = 22348,
  [SMALL_STATE(222)] = 22461,
  [SMALL_STATE(223)] = 22574,
  [SMALL_STATE(224)] = 22687,
  [SMALL_STATE(225)] = 22800,
  [SMALL_STATE(226)] = 22913,
  [SMALL_STATE(227)] = 23026,
  [SMALL_STATE(228)] = 23139,
  [SMALL_STATE(229)] = 23252,
  [SMALL_STATE(230)] = 23365,
  [SMALL_STATE(231)] = 23478,
  [SMALL_STATE(232)] = 23591,
  [SMALL_STATE(233)] = 23704,
  [SMALL_STATE(234)] = 23817,
  [SMALL_STATE(235)] = 23930,
  [SMALL_STATE(236)] = 24043,
  [SMALL_STATE(237)] = 24145,
  [SMALL_STATE(238)] = 24216,
  [SMALL_STATE(239)] = 24287,
  [SMALL_STATE(240)] = 24355,
  [SMALL_STATE(241)] = 24443,
  [SMALL_STATE(242)] = 24498,
  [SMALL_STATE(243)] = 24567,
  [SMALL_STATE(244)] = 24634,
  [SMALL_STATE(245)] = 24699,
  [SMALL_STATE(246)] = 24770,
  [SMALL_STATE(247)] = 24835,
  [SMALL_STATE(248)] = 24908,
  [SMALL_STATE(249)] = 24966,
  [SMALL_STATE(250)] = 25020,
  [SMALL_STATE(251)] = 25074,
  [SMALL_STATE(252)] = 25128,
  [SMALL_STATE(253)] = 25182,
  [SMALL_STATE(254)] = 25236,
  [SMALL_STATE(255)] = 25289,
  [SMALL_STATE(256)] = 25342,
  [SMALL_STATE(257)] = 25395,
  [SMALL_STATE(258)] = 25448,
  [SMALL_STATE(259)] = 25501,
  [SMALL_STATE(260)] = 25554,
  [SMALL_STATE(261)] = 25607,
  [SMALL_STATE(262)] = 25660,
  [SMALL_STATE(263)] = 25713,
  [SMALL_STATE(264)] = 25766,
  [SMALL_STATE(265)] = 25819,
  [SMALL_STATE(266)] = 25872,
  [SMALL_STATE(267)] = 25925,
  [SMALL_STATE(268)] = 25978,
  [SMALL_STATE(269)] = 26031,
  [SMALL_STATE(270)] = 26084,
  [SMALL_STATE(271)] = 26137,
  [SMALL_STATE(272)] = 26228,
  [SMALL_STATE(273)] = 26281,
  [SMALL_STATE(274)] = 26334,
  [SMALL_STATE(275)] = 26387,
  [SMALL_STATE(276)] = 26440,
  [SMALL_STATE(277)] = 26493,
  [SMALL_STATE(278)] = 26546,
  [SMALL_STATE(279)] = 26599,
  [SMALL_STATE(280)] = 26652,
  [SMALL_STATE(281)] = 26705,
  [SMALL_STATE(282)] = 26796,
  [SMALL_STATE(283)] = 26849,
  [SMALL_STATE(284)] = 26902,
  [SMALL_STATE(285)] = 26955,
  [SMALL_STATE(286)] = 27008,
  [SMALL_STATE(287)] = 27061,
  [SMALL_STATE(288)] = 27114,
  [SMALL_STATE(289)] = 27167,
  [SMALL_STATE(290)] = 27220,
  [SMALL_STATE(291)] = 27308,
  [SMALL_STATE(292)] = 27371,
  [SMALL_STATE(293)] = 27442,
  [SMALL_STATE(294)] = 27503,
  [SMALL_STATE(295)] = 27564,
  [SMALL_STATE(296)] = 27625,
  [SMALL_STATE(297)] = 27710,
  [SMALL_STATE(298)] = 27775,
  [SMALL_STATE(299)] = 27844,
  [SMALL_STATE(300)] = 27898,
  [SMALL_STATE(301)] = 27947,
  [SMALL_STATE(302)] = 27996,
  [SMALL_STATE(303)] = 28045,
  [SMALL_STATE(304)] = 28094,
  [SMALL_STATE(305)] = 28143,
  [SMALL_STATE(306)] = 28192,
  [SMALL_STATE(307)] = 28241,
  [SMALL_STATE(308)] = 28290,
  [SMALL_STATE(309)] = 28339,
  [SMALL_STATE(310)] = 28388,
  [SMALL_STATE(311)] = 28437,
  [SMALL_STATE(312)] = 28486,
  [SMALL_STATE(313)] = 28535,
  [SMALL_STATE(314)] = 28584,
  [SMALL_STATE(315)] = 28633,
  [SMALL_STATE(316)] = 28682,
  [SMALL_STATE(317)] = 28731,
  [SMALL_STATE(318)] = 28780,
  [SMALL_STATE(319)] = 28829,
  [SMALL_STATE(320)] = 28878,
  [SMALL_STATE(321)] = 28927,
  [SMALL_STATE(322)] = 28976,
  [SMALL_STATE(323)] = 29025,
  [SMALL_STATE(324)] = 29074,
  [SMALL_STATE(325)] = 29123,
  [SMALL_STATE(326)] = 29172,
  [SMALL_STATE(327)] = 29221,
  [SMALL_STATE(328)] = 29270,
  [SMALL_STATE(329)] = 29319,
  [SMALL_STATE(330)] = 29368,
  [SMALL_STATE(331)] = 29417,
  [SMALL_STATE(332)] = 29466,
  [SMALL_STATE(333)] = 29515,
  [SMALL_STATE(334)] = 29564,
  [SMALL_STATE(335)] = 29613,
  [SMALL_STATE(336)] = 29662,
  [SMALL_STATE(337)] = 29711,
  [SMALL_STATE(338)] = 29760,
  [SMALL_STATE(339)] = 29809,
  [SMALL_STATE(340)] = 29868,
  [SMALL_STATE(341)] = 29936,
  [SMALL_STATE(342)] = 29994,
  [SMALL_STATE(343)] = 30058,
  [SMALL_STATE(344)] = 30114,
  [SMALL_STATE(345)] = 30170,
  [SMALL_STATE(346)] = 30236,
  [SMALL_STATE(347)] = 30296,
  [SMALL_STATE(348)] = 30345,
  [SMALL_STATE(349)] = 30389,
  [SMALL_STATE(350)] = 30433,
  [SMALL_STATE(351)] = 30477,
  [SMALL_STATE(352)] = 30551,
  [SMALL_STATE(353)] = 30595,
  [SMALL_STATE(354)] = 30639,
  [SMALL_STATE(355)] = 30683,
  [SMALL_STATE(356)] = 30727,
  [SMALL_STATE(357)] = 30771,
  [SMALL_STATE(358)] = 30815,
  [SMALL_STATE(359)] = 30859,
  [SMALL_STATE(360)] = 30903,
  [SMALL_STATE(361)] = 30947,
  [SMALL_STATE(362)] = 30991,
  [SMALL_STATE(363)] = 31035,
  [SMALL_STATE(364)] = 31079,
  [SMALL_STATE(365)] = 31153,
  [SMALL_STATE(366)] = 31197,
  [SMALL_STATE(367)] = 31241,
  [SMALL_STATE(368)] = 31285,
  [SMALL_STATE(369)] = 31373,
  [SMALL_STATE(370)] = 31417,
  [SMALL_STATE(371)] = 31461,
  [SMALL_STATE(372)] = 31505,
  [SMALL_STATE(373)] = 31549,
  [SMALL_STATE(374)] = 31593,
  [SMALL_STATE(375)] = 31637,
  [SMALL_STATE(376)] = 31681,
  [SMALL_STATE(377)] = 31725,
  [SMALL_STATE(378)] = 31769,
  [SMALL_STATE(379)] = 31813,
  [SMALL_STATE(380)] = 31857,
  [SMALL_STATE(381)] = 31901,
  [SMALL_STATE(382)] = 31945,
  [SMALL_STATE(383)] = 31989,
  [SMALL_STATE(384)] = 32033,
  [SMALL_STATE(385)] = 32077,
  [SMALL_STATE(386)] = 32121,
  [SMALL_STATE(387)] = 32165,
  [SMALL_STATE(388)] = 32209,
  [SMALL_STATE(389)] = 32253,
  [SMALL_STATE(390)] = 32297,
  [SMALL_STATE(391)] = 32349,
  [SMALL_STATE(392)] = 32401,
  [SMALL_STATE(393)] = 32478,
  [SMALL_STATE(394)] = 32555,
  [SMALL_STATE(395)] = 32600,
  [SMALL_STATE(396)] = 32653,
  [SMALL_STATE(397)] = 32693,
  [SMALL_STATE(398)] = 32733,
  [SMALL_STATE(399)] = 32773,
  [SMALL_STATE(400)] = 32813,
  [SMALL_STATE(401)] = 32877,
  [SMALL_STATE(402)] = 32917,
  [SMALL_STATE(403)] = 32957,
  [SMALL_STATE(404)] = 32997,
  [SMALL_STATE(405)] = 33037,
  [SMALL_STATE(406)] = 33077,
  [SMALL_STATE(407)] = 33117,
  [SMALL_STATE(408)] = 33157,
  [SMALL_STATE(409)] = 33197,
  [SMALL_STATE(410)] = 33237,
  [SMALL_STATE(411)] = 33277,
  [SMALL_STATE(412)] = 33317,
  [SMALL_STATE(413)] = 33357,
  [SMALL_STATE(414)] = 33397,
  [SMALL_STATE(415)] = 33437,
  [SMALL_STATE(416)] = 33477,
  [SMALL_STATE(417)] = 33517,
  [SMALL_STATE(418)] = 33557,
  [SMALL_STATE(419)] = 33597,
  [SMALL_STATE(420)] = 33637,
  [SMALL_STATE(421)] = 33677,
  [SMALL_STATE(422)] = 33717,
  [SMALL_STATE(423)] = 33757,
  [SMALL_STATE(424)] = 33797,
  [SMALL_STATE(425)] = 33837,
  [SMALL_STATE(426)] = 33877,
  [SMALL_STATE(427)] = 33917,
  [SMALL_STATE(428)] = 33957,
  [SMALL_STATE(429)] = 33997,
  [SMALL_STATE(430)] = 34037,
  [SMALL_STATE(431)] = 34077,
  [SMALL_STATE(432)] = 34129,
  [SMALL_STATE(433)] = 34169,
  [SMALL_STATE(434)] = 34209,
  [SMALL_STATE(435)] = 34249,
  [SMALL_STATE(436)] = 34289,
  [SMALL_STATE(437)] = 34329,
  [SMALL_STATE(438)] = 34388,
  [SMALL_STATE(439)] = 34451,
  [SMALL_STATE(440)] = 34518,
  [SMALL_STATE(441)] = 34567,
  [SMALL_STATE(442)] = 34620,
  [SMALL_STATE(443)] = 34677,
  [SMALL_STATE(444)] = 34738,
  [SMALL_STATE(445)] = 34789,
  [SMALL_STATE(446)] = 34842,
  [SMALL_STATE(447)] = 34897,
  [SMALL_STATE(448)] = 34954,
  [SMALL_STATE(449)] = 35003,
  [SMALL_STATE(450)] = 35070,
  [SMALL_STATE(451)] = 35112,
  [SMALL_STATE(452)] = 35184,
  [SMALL_STATE(453)] = 35254,
  [SMALL_STATE(454)] = 35330,
  [SMALL_STATE(455)] = 35402,
  [SMALL_STATE(456)] = 35460,
  [SMALL_STATE(457)] = 35518,
  [SMALL_STATE(458)] = 35576,
  [SMALL_STATE(459)] = 35626,
  [SMALL_STATE(460)] = 35698,
  [SMALL_STATE(461)] = 35770,
  [SMALL_STATE(462)] = 35807,
  [SMALL_STATE(463)] = 35844,
  [SMALL_STATE(464)] = 35881,
  [SMALL_STATE(465)] = 35918,
  [SMALL_STATE(466)] = 35955,
  [SMALL_STATE(467)] = 35992,
  [SMALL_STATE(468)] = 36029,
  [SMALL_STATE(469)] = 36066,
  [SMALL_STATE(470)] = 36103,
  [SMALL_STATE(471)] = 36140,
  [SMALL_STATE(472)] = 36177,
  [SMALL_STATE(473)] = 36214,
  [SMALL_STATE(474)] = 36251,
  [SMALL_STATE(475)] = 36288,
  [SMALL_STATE(476)] = 36325,
  [SMALL_STATE(477)] = 36362,
  [SMALL_STATE(478)] = 36399,
  [SMALL_STATE(479)] = 36436,
  [SMALL_STATE(480)] = 36473,
  [SMALL_STATE(481)] = 36510,
  [SMALL_STATE(482)] = 36547,
  [SMALL_STATE(483)] = 36584,
  [SMALL_STATE(484)] = 36621,
  [SMALL_STATE(485)] = 36658,
  [SMALL_STATE(486)] = 36695,
  [SMALL_STATE(487)] = 36732,
  [SMALL_STATE(488)] = 36769,
  [SMALL_STATE(489)] = 36832,
  [SMALL_STATE(490)] = 36869,
  [SMALL_STATE(491)] = 36906,
  [SMALL_STATE(492)] = 36943,
  [SMALL_STATE(493)] = 37014,
  [SMALL_STATE(494)] = 37085,
  [SMALL_STATE(495)] = 37122,
  [SMALL_STATE(496)] = 37159,
  [SMALL_STATE(497)] = 37226,
  [SMALL_STATE(498)] = 37263,
  [SMALL_STATE(499)] = 37334,
  [SMALL_STATE(500)] = 37381,
  [SMALL_STATE(501)] = 37448,
  [SMALL_STATE(502)] = 37495,
  [SMALL_STATE(503)] = 37566,
  [SMALL_STATE(504)] = 37603,
  [SMALL_STATE(505)] = 37640,
  [SMALL_STATE(506)] = 37691,
  [SMALL_STATE(507)] = 37746,
  [SMALL_STATE(508)] = 37805,
  [SMALL_STATE(509)] = 37866,
  [SMALL_STATE(510)] = 37903,
  [SMALL_STATE(511)] = 37974,
  [SMALL_STATE(512)] = 38045,
  [SMALL_STATE(513)] = 38082,
  [SMALL_STATE(514)] = 38119,
  [SMALL_STATE(515)] = 38156,
  [SMALL_STATE(516)] = 38208,
  [SMALL_STATE(517)] = 38256,
  [SMALL_STATE(518)] = 38296,
  [SMALL_STATE(519)] = 38362,
  [SMALL_STATE(520)] = 38430,
  [SMALL_STATE(521)] = 38498,
  [SMALL_STATE(522)] = 38566,
  [SMALL_STATE(523)] = 38630,
  [SMALL_STATE(524)] = 38698,
  [SMALL_STATE(525)] = 38764,
  [SMALL_STATE(526)] = 38830,
  [SMALL_STATE(527)] = 38898,
  [SMALL_STATE(528)] = 38966,
  [SMALL_STATE(529)] = 39032,
  [SMALL_STATE(530)] = 39098,
  [SMALL_STATE(531)] = 39164,
  [SMALL_STATE(532)] = 39230,
  [SMALL_STATE(533)] = 39296,
  [SMALL_STATE(534)] = 39331,
  [SMALL_STATE(535)] = 39392,
  [SMALL_STATE(536)] = 39457,
  [SMALL_STATE(537)] = 39506,
  [SMALL_STATE(538)] = 39559,
  [SMALL_STATE(539)] = 39614,
  [SMALL_STATE(540)] = 39671,
  [SMALL_STATE(541)] = 39740,
  [SMALL_STATE(542)] = 39807,
  [SMALL_STATE(543)] = 39866,
  [SMALL_STATE(544)] = 39933,
  [SMALL_STATE(545)] = 39968,
  [SMALL_STATE(546)] = 40013,
  [SMALL_STATE(547)] = 40048,
  [SMALL_STATE(548)] = 40083,
  [SMALL_STATE(549)] = 40118,
  [SMALL_STATE(550)] = 40153,
  [SMALL_STATE(551)] = 40188,
  [SMALL_STATE(552)] = 40223,
  [SMALL_STATE(553)] = 40258,
  [SMALL_STATE(554)] = 40293,
  [SMALL_STATE(555)] = 40328,
  [SMALL_STATE(556)] = 40363,
  [SMALL_STATE(557)] = 40398,
  [SMALL_STATE(558)] = 40433,
  [SMALL_STATE(559)] = 40468,
  [SMALL_STATE(560)] = 40503,
  [SMALL_STATE(561)] = 40538,
  [SMALL_STATE(562)] = 40573,
  [SMALL_STATE(563)] = 40608,
  [SMALL_STATE(564)] = 40643,
  [SMALL_STATE(565)] = 40678,
  [SMALL_STATE(566)] = 40713,
  [SMALL_STATE(567)] = 40748,
  [SMALL_STATE(568)] = 40783,
  [SMALL_STATE(569)] = 40818,
  [SMALL_STATE(570)] = 40883,
  [SMALL_STATE(571)] = 40918,
  [SMALL_STATE(572)] = 40953,
  [SMALL_STATE(573)] = 40988,
  [SMALL_STATE(574)] = 41023,
  [SMALL_STATE(575)] = 41058,
  [SMALL_STATE(576)] = 41093,
  [SMALL_STATE(577)] = 41128,
  [SMALL_STATE(578)] = 41163,
  [SMALL_STATE(579)] = 41198,
  [SMALL_STATE(580)] = 41233,
  [SMALL_STATE(581)] = 41268,
  [SMALL_STATE(582)] = 41303,
  [SMALL_STATE(583)] = 41338,
  [SMALL_STATE(584)] = 41403,
  [SMALL_STATE(585)] = 41468,
  [SMALL_STATE(586)] = 41533,
  [SMALL_STATE(587)] = 41582,
  [SMALL_STATE(588)] = 41635,
  [SMALL_STATE(589)] = 41690,
  [SMALL_STATE(590)] = 41747,
  [SMALL_STATE(591)] = 41812,
  [SMALL_STATE(592)] = 41879,
  [SMALL_STATE(593)] = 41944,
  [SMALL_STATE(594)] = 42007,
  [SMALL_STATE(595)] = 42072,
  [SMALL_STATE(596)] = 42137,
  [SMALL_STATE(597)] = 42202,
  [SMALL_STATE(598)] = 42265,
  [SMALL_STATE(599)] = 42300,
  [SMALL_STATE(600)] = 42360,
  [SMALL_STATE(601)] = 42420,
  [SMALL_STATE(602)] = 42484,
  [SMALL_STATE(603)] = 42542,
  [SMALL_STATE(604)] = 42606,
  [SMALL_STATE(605)] = 42666,
  [SMALL_STATE(606)] = 42726,
  [SMALL_STATE(607)] = 42790,
  [SMALL_STATE(608)] = 42854,
  [SMALL_STATE(609)] = 42914,
  [SMALL_STATE(610)] = 42974,
  [SMALL_STATE(611)] = 43038,
  [SMALL_STATE(612)] = 43098,
  [SMALL_STATE(613)] = 43164,
  [SMALL_STATE(614)] = 43224,
  [SMALL_STATE(615)] = 43286,
  [SMALL_STATE(616)] = 43344,
  [SMALL_STATE(617)] = 43408,
  [SMALL_STATE(618)] = 43468,
  [SMALL_STATE(619)] = 43528,
  [SMALL_STATE(620)] = 43592,
  [SMALL_STATE(621)] = 43652,
  [SMALL_STATE(622)] = 43716,
  [SMALL_STATE(623)] = 43776,
  [SMALL_STATE(624)] = 43840,
  [SMALL_STATE(625)] = 43900,
  [SMALL_STATE(626)] = 43960,
  [SMALL_STATE(627)] = 44026,
  [SMALL_STATE(628)] = 44086,
  [SMALL_STATE(629)] = 44146,
  [SMALL_STATE(630)] = 44210,
  [SMALL_STATE(631)] = 44270,
  [SMALL_STATE(632)] = 44330,
  [SMALL_STATE(633)] = 44390,
  [SMALL_STATE(634)] = 44450,
  [SMALL_STATE(635)] = 44510,
  [SMALL_STATE(636)] = 44574,
  [SMALL_STATE(637)] = 44634,
  [SMALL_STATE(638)] = 44694,
  [SMALL_STATE(639)] = 44754,
  [SMALL_STATE(640)] = 44814,
  [SMALL_STATE(641)] = 44878,
  [SMALL_STATE(642)] = 44942,
  [SMALL_STATE(643)] = 45006,
  [SMALL_STATE(644)] = 45064,
  [SMALL_STATE(645)] = 45125,
  [SMALL_STATE(646)] = 45186,
  [SMALL_STATE(647)] = 45243,
  [SMALL_STATE(648)] = 45304,
  [SMALL_STATE(649)] = 45361,
  [SMALL_STATE(650)] = 45418,
  [SMALL_STATE(651)] = 45475,
  [SMALL_STATE(652)] = 45532,
  [SMALL_STATE(653)] = 45589,
  [SMALL_STATE(654)] = 45650,
  [SMALL_STATE(655)] = 45707,
  [SMALL_STATE(656)] = 45764,
  [SMALL_STATE(657)] = 45825,
  [SMALL_STATE(658)] = 45886,
  [SMALL_STATE(659)] = 45947,
  [SMALL_STATE(660)] = 46008,
  [SMALL_STATE(661)] = 46069,
  [SMALL_STATE(662)] = 46130,
  [SMALL_STATE(663)] = 46187,
  [SMALL_STATE(664)] = 46244,
  [SMALL_STATE(665)] = 46301,
  [SMALL_STATE(666)] = 46358,
  [SMALL_STATE(667)] = 46419,
  [SMALL_STATE(668)] = 46476,
  [SMALL_STATE(669)] = 46537,
  [SMALL_STATE(670)] = 46594,
  [SMALL_STATE(671)] = 46651,
  [SMALL_STATE(672)] = 46712,
  [SMALL_STATE(673)] = 46769,
  [SMALL_STATE(674)] = 46826,
  [SMALL_STATE(675)] = 46887,
  [SMALL_STATE(676)] = 46948,
  [SMALL_STATE(677)] = 47005,
  [SMALL_STATE(678)] = 47062,
  [SMALL_STATE(679)] = 47119,
  [SMALL_STATE(680)] = 47176,
  [SMALL_STATE(681)] = 47237,
  [SMALL_STATE(682)] = 47294,
  [SMALL_STATE(683)] = 47355,
  [SMALL_STATE(684)] = 47416,
  [SMALL_STATE(685)] = 47473,
  [SMALL_STATE(686)] = 47534,
  [SMALL_STATE(687)] = 47595,
  [SMALL_STATE(688)] = 47652,
  [SMALL_STATE(689)] = 47713,
  [SMALL_STATE(690)] = 47770,
  [SMALL_STATE(691)] = 47827,
  [SMALL_STATE(692)] = 47884,
  [SMALL_STATE(693)] = 47945,
  [SMALL_STATE(694)] = 48002,
  [SMALL_STATE(695)] = 48063,
  [SMALL_STATE(696)] = 48124,
  [SMALL_STATE(697)] = 48181,
  [SMALL_STATE(698)] = 48242,
  [SMALL_STATE(699)] = 48303,
  [SMALL_STATE(700)] = 48364,
  [SMALL_STATE(701)] = 48421,
  [SMALL_STATE(702)] = 48479,
  [SMALL_STATE(703)] = 48537,
  [SMALL_STATE(704)] = 48595,
  [SMALL_STATE(705)] = 48653,
  [SMALL_STATE(706)] = 48711,
  [SMALL_STATE(707)] = 48769,
  [SMALL_STATE(708)] = 48827,
  [SMALL_STATE(709)] = 48885,
  [SMALL_STATE(710)] = 48943,
  [SMALL_STATE(711)] = 49001,
  [SMALL_STATE(712)] = 49059,
  [SMALL_STATE(713)] = 49117,
  [SMALL_STATE(714)] = 49175,
  [SMALL_STATE(715)] = 49233,
  [SMALL_STATE(716)] = 49291,
  [SMALL_STATE(717)] = 49349,
  [SMALL_STATE(718)] = 49407,
  [SMALL_STATE(719)] = 49465,
  [SMALL_STATE(720)] = 49523,
  [SMALL_STATE(721)] = 49581,
  [SMALL_STATE(722)] = 49639,
  [SMALL_STATE(723)] = 49697,
  [SMALL_STATE(724)] = 49755,
  [SMALL_STATE(725)] = 49813,
  [SMALL_STATE(726)] = 49871,
  [SMALL_STATE(727)] = 49929,
  [SMALL_STATE(728)] = 49983,
  [SMALL_STATE(729)] = 50041,
  [SMALL_STATE(730)] = 50099,
  [SMALL_STATE(731)] = 50157,
  [SMALL_STATE(732)] = 50215,
  [SMALL_STATE(733)] = 50273,
  [SMALL_STATE(734)] = 50331,
  [SMALL_STATE(735)] = 50389,
  [SMALL_STATE(736)] = 50447,
  [SMALL_STATE(737)] = 50505,
  [SMALL_STATE(738)] = 50563,
  [SMALL_STATE(739)] = 50621,
  [SMALL_STATE(740)] = 50679,
  [SMALL_STATE(741)] = 50737,
  [SMALL_STATE(742)] = 50795,
  [SMALL_STATE(743)] = 50853,
  [SMALL_STATE(744)] = 50911,
  [SMALL_STATE(745)] = 50969,
  [SMALL_STATE(746)] = 51027,
  [SMALL_STATE(747)] = 51085,
  [SMALL_STATE(748)] = 51143,
  [SMALL_STATE(749)] = 51201,
  [SMALL_STATE(750)] = 51259,
  [SMALL_STATE(751)] = 51317,
  [SMALL_STATE(752)] = 51375,
  [SMALL_STATE(753)] = 51433,
  [SMALL_STATE(754)] = 51491,
  [SMALL_STATE(755)] = 51549,
  [SMALL_STATE(756)] = 51607,
  [SMALL_STATE(757)] = 51665,
  [SMALL_STATE(758)] = 51723,
  [SMALL_STATE(759)] = 51781,
  [SMALL_STATE(760)] = 51839,
  [SMALL_STATE(761)] = 51897,
  [SMALL_STATE(762)] = 51955,
  [SMALL_STATE(763)] = 52013,
  [SMALL_STATE(764)] = 52071,
  [SMALL_STATE(765)] = 52129,
  [SMALL_STATE(766)] = 52187,
  [SMALL_STATE(767)] = 52245,
  [SMALL_STATE(768)] = 52303,
  [SMALL_STATE(769)] = 52361,
  [SMALL_STATE(770)] = 52419,
  [SMALL_STATE(771)] = 52477,
  [SMALL_STATE(772)] = 52535,
  [SMALL_STATE(773)] = 52593,
  [SMALL_STATE(774)] = 52651,
  [SMALL_STATE(775)] = 52709,
  [SMALL_STATE(776)] = 52767,
  [SMALL_STATE(777)] = 52825,
  [SMALL_STATE(778)] = 52883,
  [SMALL_STATE(779)] = 52941,
  [SMALL_STATE(780)] = 52999,
  [SMALL_STATE(781)] = 53057,
  [SMALL_STATE(782)] = 53115,
  [SMALL_STATE(783)] = 53173,
  [SMALL_STATE(784)] = 53204,
  [SMALL_STATE(785)] = 53235,
  [SMALL_STATE(786)] = 53266,
  [SMALL_STATE(787)] = 53297,
  [SMALL_STATE(788)] = 53329,
  [SMALL_STATE(789)] = 53356,
  [SMALL_STATE(790)] = 53382,
  [SMALL_STATE(791)] = 53408,
  [SMALL_STATE(792)] = 53434,
  [SMALL_STATE(793)] = 53460,
  [SMALL_STATE(794)] = 53484,
  [SMALL_STATE(795)] = 53508,
  [SMALL_STATE(796)] = 53532,
  [SMALL_STATE(797)] = 53556,
  [SMALL_STATE(798)] = 53585,
  [SMALL_STATE(799)] = 53610,
  [SMALL_STATE(800)] = 53639,
  [SMALL_STATE(801)] = 53664,
  [SMALL_STATE(802)] = 53690,
  [SMALL_STATE(803)] = 53716,
  [SMALL_STATE(804)] = 53742,
  [SMALL_STATE(805)] = 53763,
  [SMALL_STATE(806)] = 53784,
  [SMALL_STATE(807)] = 53805,
  [SMALL_STATE(808)] = 53826,
  [SMALL_STATE(809)] = 53847,
  [SMALL_STATE(810)] = 53868,
  [SMALL_STATE(811)] = 53889,
  [SMALL_STATE(812)] = 53910,
  [SMALL_STATE(813)] = 53931,
  [SMALL_STATE(814)] = 53952,
  [SMALL_STATE(815)] = 53973,
  [SMALL_STATE(816)] = 53994,
  [SMALL_STATE(817)] = 54015,
  [SMALL_STATE(818)] = 54036,
  [SMALL_STATE(819)] = 54057,
  [SMALL_STATE(820)] = 54078,
  [SMALL_STATE(821)] = 54099,
  [SMALL_STATE(822)] = 54120,
  [SMALL_STATE(823)] = 54141,
  [SMALL_STATE(824)] = 54162,
  [SMALL_STATE(825)] = 54183,
  [SMALL_STATE(826)] = 54204,
  [SMALL_STATE(827)] = 54225,
  [SMALL_STATE(828)] = 54246,
  [SMALL_STATE(829)] = 54271,
  [SMALL_STATE(830)] = 54292,
  [SMALL_STATE(831)] = 54313,
  [SMALL_STATE(832)] = 54334,
  [SMALL_STATE(833)] = 54355,
  [SMALL_STATE(834)] = 54376,
  [SMALL_STATE(835)] = 54397,
  [SMALL_STATE(836)] = 54420,
  [SMALL_STATE(837)] = 54441,
  [SMALL_STATE(838)] = 54462,
  [SMALL_STATE(839)] = 54483,
  [SMALL_STATE(840)] = 54503,
  [SMALL_STATE(841)] = 54523,
  [SMALL_STATE(842)] = 54543,
  [SMALL_STATE(843)] = 54563,
  [SMALL_STATE(844)] = 54596,
  [SMALL_STATE(845)] = 54619,
  [SMALL_STATE(846)] = 54650,
  [SMALL_STATE(847)] = 54683,
  [SMALL_STATE(848)] = 54703,
  [SMALL_STATE(849)] = 54731,
  [SMALL_STATE(850)] = 54763,
  [SMALL_STATE(851)] = 54795,
  [SMALL_STATE(852)] = 54827,
  [SMALL_STATE(853)] = 54855,
  [SMALL_STATE(854)] = 54887,
  [SMALL_STATE(855)] = 54919,
  [SMALL_STATE(856)] = 54941,
  [SMALL_STATE(857)] = 54969,
  [SMALL_STATE(858)] = 55001,
  [SMALL_STATE(859)] = 55016,
  [SMALL_STATE(860)] = 55031,
  [SMALL_STATE(861)] = 55046,
  [SMALL_STATE(862)] = 55061,
  [SMALL_STATE(863)] = 55076,
  [SMALL_STATE(864)] = 55091,
  [SMALL_STATE(865)] = 55106,
  [SMALL_STATE(866)] = 55135,
  [SMALL_STATE(867)] = 55150,
  [SMALL_STATE(868)] = 55173,
  [SMALL_STATE(869)] = 55188,
  [SMALL_STATE(870)] = 55203,
  [SMALL_STATE(871)] = 55218,
  [SMALL_STATE(872)] = 55233,
  [SMALL_STATE(873)] = 55248,
  [SMALL_STATE(874)] = 55263,
  [SMALL_STATE(875)] = 55278,
  [SMALL_STATE(876)] = 55297,
  [SMALL_STATE(877)] = 55312,
  [SMALL_STATE(878)] = 55327,
  [SMALL_STATE(879)] = 55342,
  [SMALL_STATE(880)] = 55357,
  [SMALL_STATE(881)] = 55372,
  [SMALL_STATE(882)] = 55387,
  [SMALL_STATE(883)] = 55402,
  [SMALL_STATE(884)] = 55417,
  [SMALL_STATE(885)] = 55432,
  [SMALL_STATE(886)] = 55447,
  [SMALL_STATE(887)] = 55462,
  [SMALL_STATE(888)] = 55477,
  [SMALL_STATE(889)] = 55492,
  [SMALL_STATE(890)] = 55512,
  [SMALL_STATE(891)] = 55540,
  [SMALL_STATE(892)] = 55568,
  [SMALL_STATE(893)] = 55586,
  [SMALL_STATE(894)] = 55608,
  [SMALL_STATE(895)] = 55630,
  [SMALL_STATE(896)] = 55648,
  [SMALL_STATE(897)] = 55672,
  [SMALL_STATE(898)] = 55696,
  [SMALL_STATE(899)] = 55718,
  [SMALL_STATE(900)] = 55746,
  [SMALL_STATE(901)] = 55768,
  [SMALL_STATE(902)] = 55786,
  [SMALL_STATE(903)] = 55814,
  [SMALL_STATE(904)] = 55842,
  [SMALL_STATE(905)] = 55870,
  [SMALL_STATE(906)] = 55890,
  [SMALL_STATE(907)] = 55918,
  [SMALL_STATE(908)] = 55946,
  [SMALL_STATE(909)] = 55968,
  [SMALL_STATE(910)] = 55996,
  [SMALL_STATE(911)] = 56013,
  [SMALL_STATE(912)] = 56030,
  [SMALL_STATE(913)] = 56055,
  [SMALL_STATE(914)] = 56072,
  [SMALL_STATE(915)] = 56089,
  [SMALL_STATE(916)] = 56106,
  [SMALL_STATE(917)] = 56125,
  [SMALL_STATE(918)] = 56142,
  [SMALL_STATE(919)] = 56161,
  [SMALL_STATE(920)] = 56178,
  [SMALL_STATE(921)] = 56197,
  [SMALL_STATE(922)] = 56212,
  [SMALL_STATE(923)] = 56231,
  [SMALL_STATE(924)] = 56246,
  [SMALL_STATE(925)] = 56261,
  [SMALL_STATE(926)] = 56280,
  [SMALL_STATE(927)] = 56295,
  [SMALL_STATE(928)] = 56309,
  [SMALL_STATE(929)] = 56323,
  [SMALL_STATE(930)] = 56337,
  [SMALL_STATE(931)] = 56351,
  [SMALL_STATE(932)] = 56369,
  [SMALL_STATE(933)] = 56383,
  [SMALL_STATE(934)] = 56401,
  [SMALL_STATE(935)] = 56415,
  [SMALL_STATE(936)] = 56429,
  [SMALL_STATE(937)] = 56443,
  [SMALL_STATE(938)] = 56457,
  [SMALL_STATE(939)] = 56471,
  [SMALL_STATE(940)] = 56485,
  [SMALL_STATE(941)] = 56503,
  [SMALL_STATE(942)] = 56517,
  [SMALL_STATE(943)] = 56531,
  [SMALL_STATE(944)] = 56549,
  [SMALL_STATE(945)] = 56563,
  [SMALL_STATE(946)] = 56577,
  [SMALL_STATE(947)] = 56591,
  [SMALL_STATE(948)] = 56605,
  [SMALL_STATE(949)] = 56619,
  [SMALL_STATE(950)] = 56633,
  [SMALL_STATE(951)] = 56647,
  [SMALL_STATE(952)] = 56661,
  [SMALL_STATE(953)] = 56675,
  [SMALL_STATE(954)] = 56689,
  [SMALL_STATE(955)] = 56703,
  [SMALL_STATE(956)] = 56721,
  [SMALL_STATE(957)] = 56739,
  [SMALL_STATE(958)] = 56753,
  [SMALL_STATE(959)] = 56767,
  [SMALL_STATE(960)] = 56785,
  [SMALL_STATE(961)] = 56803,
  [SMALL_STATE(962)] = 56821,
  [SMALL_STATE(963)] = 56835,
  [SMALL_STATE(964)] = 56849,
  [SMALL_STATE(965)] = 56863,
  [SMALL_STATE(966)] = 56877,
  [SMALL_STATE(967)] = 56891,
  [SMALL_STATE(968)] = 56905,
  [SMALL_STATE(969)] = 56919,
  [SMALL_STATE(970)] = 56933,
  [SMALL_STATE(971)] = 56953,
  [SMALL_STATE(972)] = 56967,
  [SMALL_STATE(973)] = 56981,
  [SMALL_STATE(974)] = 56993,
  [SMALL_STATE(975)] = 57007,
  [SMALL_STATE(976)] = 57021,
  [SMALL_STATE(977)] = 57035,
  [SMALL_STATE(978)] = 57053,
  [SMALL_STATE(979)] = 57067,
  [SMALL_STATE(980)] = 57081,
  [SMALL_STATE(981)] = 57095,
  [SMALL_STATE(982)] = 57109,
  [SMALL_STATE(983)] = 57123,
  [SMALL_STATE(984)] = 57145,
  [SMALL_STATE(985)] = 57159,
  [SMALL_STATE(986)] = 57173,
  [SMALL_STATE(987)] = 57187,
  [SMALL_STATE(988)] = 57201,
  [SMALL_STATE(989)] = 57215,
  [SMALL_STATE(990)] = 57229,
  [SMALL_STATE(991)] = 57243,
  [SMALL_STATE(992)] = 57257,
  [SMALL_STATE(993)] = 57275,
  [SMALL_STATE(994)] = 57289,
  [SMALL_STATE(995)] = 57303,
  [SMALL_STATE(996)] = 57321,
  [SMALL_STATE(997)] = 57335,
  [SMALL_STATE(998)] = 57349,
  [SMALL_STATE(999)] = 57367,
  [SMALL_STATE(1000)] = 57385,
  [SMALL_STATE(1001)] = 57399,
  [SMALL_STATE(1002)] = 57413,
  [SMALL_STATE(1003)] = 57431,
  [SMALL_STATE(1004)] = 57449,
  [SMALL_STATE(1005)] = 57463,
  [SMALL_STATE(1006)] = 57477,
  [SMALL_STATE(1007)] = 57495,
  [SMALL_STATE(1008)] = 57513,
  [SMALL_STATE(1009)] = 57527,
  [SMALL_STATE(1010)] = 57541,
  [SMALL_STATE(1011)] = 57559,
  [SMALL_STATE(1012)] = 57577,
  [SMALL_STATE(1013)] = 57591,
  [SMALL_STATE(1014)] = 57609,
  [SMALL_STATE(1015)] = 57623,
  [SMALL_STATE(1016)] = 57637,
  [SMALL_STATE(1017)] = 57651,
  [SMALL_STATE(1018)] = 57669,
  [SMALL_STATE(1019)] = 57687,
  [SMALL_STATE(1020)] = 57709,
  [SMALL_STATE(1021)] = 57723,
  [SMALL_STATE(1022)] = 57737,
  [SMALL_STATE(1023)] = 57751,
  [SMALL_STATE(1024)] = 57765,
  [SMALL_STATE(1025)] = 57779,
  [SMALL_STATE(1026)] = 57793,
  [SMALL_STATE(1027)] = 57807,
  [SMALL_STATE(1028)] = 57821,
  [SMALL_STATE(1029)] = 57835,
  [SMALL_STATE(1030)] = 57849,
  [SMALL_STATE(1031)] = 57863,
  [SMALL_STATE(1032)] = 57877,
  [SMALL_STATE(1033)] = 57894,
  [SMALL_STATE(1034)] = 57911,
  [SMALL_STATE(1035)] = 57930,
  [SMALL_STATE(1036)] = 57947,
  [SMALL_STATE(1037)] = 57964,
  [SMALL_STATE(1038)] = 57983,
  [SMALL_STATE(1039)] = 58000,
  [SMALL_STATE(1040)] = 58019,
  [SMALL_STATE(1041)] = 58036,
  [SMALL_STATE(1042)] = 58053,
  [SMALL_STATE(1043)] = 58070,
  [SMALL_STATE(1044)] = 58087,
  [SMALL_STATE(1045)] = 58104,
  [SMALL_STATE(1046)] = 58121,
  [SMALL_STATE(1047)] = 58138,
  [SMALL_STATE(1048)] = 58155,
  [SMALL_STATE(1049)] = 58172,
  [SMALL_STATE(1050)] = 58189,
  [SMALL_STATE(1051)] = 58206,
  [SMALL_STATE(1052)] = 58219,
  [SMALL_STATE(1053)] = 58236,
  [SMALL_STATE(1054)] = 58255,
  [SMALL_STATE(1055)] = 58272,
  [SMALL_STATE(1056)] = 58289,
  [SMALL_STATE(1057)] = 58308,
  [SMALL_STATE(1058)] = 58325,
  [SMALL_STATE(1059)] = 58344,
  [SMALL_STATE(1060)] = 58361,
  [SMALL_STATE(1061)] = 58378,
  [SMALL_STATE(1062)] = 58395,
  [SMALL_STATE(1063)] = 58412,
  [SMALL_STATE(1064)] = 58429,
  [SMALL_STATE(1065)] = 58443,
  [SMALL_STATE(1066)] = 58459,
  [SMALL_STATE(1067)] = 58473,
  [SMALL_STATE(1068)] = 58485,
  [SMALL_STATE(1069)] = 58501,
  [SMALL_STATE(1070)] = 58515,
  [SMALL_STATE(1071)] = 58527,
  [SMALL_STATE(1072)] = 58539,
  [SMALL_STATE(1073)] = 58553,
  [SMALL_STATE(1074)] = 58565,
  [SMALL_STATE(1075)] = 58577,
  [SMALL_STATE(1076)] = 58591,
  [SMALL_STATE(1077)] = 58603,
  [SMALL_STATE(1078)] = 58617,
  [SMALL_STATE(1079)] = 58629,
  [SMALL_STATE(1080)] = 58643,
  [SMALL_STATE(1081)] = 58657,
  [SMALL_STATE(1082)] = 58671,
  [SMALL_STATE(1083)] = 58685,
  [SMALL_STATE(1084)] = 58699,
  [SMALL_STATE(1085)] = 58713,
  [SMALL_STATE(1086)] = 58727,
  [SMALL_STATE(1087)] = 58739,
  [SMALL_STATE(1088)] = 58755,
  [SMALL_STATE(1089)] = 58769,
  [SMALL_STATE(1090)] = 58781,
  [SMALL_STATE(1091)] = 58795,
  [SMALL_STATE(1092)] = 58809,
  [SMALL_STATE(1093)] = 58825,
  [SMALL_STATE(1094)] = 58837,
  [SMALL_STATE(1095)] = 58849,
  [SMALL_STATE(1096)] = 58861,
  [SMALL_STATE(1097)] = 58873,
  [SMALL_STATE(1098)] = 58887,
  [SMALL_STATE(1099)] = 58901,
  [SMALL_STATE(1100)] = 58915,
  [SMALL_STATE(1101)] = 58929,
  [SMALL_STATE(1102)] = 58943,
  [SMALL_STATE(1103)] = 58955,
  [SMALL_STATE(1104)] = 58969,
  [SMALL_STATE(1105)] = 58985,
  [SMALL_STATE(1106)] = 58997,
  [SMALL_STATE(1107)] = 59009,
  [SMALL_STATE(1108)] = 59021,
  [SMALL_STATE(1109)] = 59033,
  [SMALL_STATE(1110)] = 59045,
  [SMALL_STATE(1111)] = 59057,
  [SMALL_STATE(1112)] = 59073,
  [SMALL_STATE(1113)] = 59085,
  [SMALL_STATE(1114)] = 59101,
  [SMALL_STATE(1115)] = 59115,
  [SMALL_STATE(1116)] = 59127,
  [SMALL_STATE(1117)] = 59139,
  [SMALL_STATE(1118)] = 59153,
  [SMALL_STATE(1119)] = 59167,
  [SMALL_STATE(1120)] = 59179,
  [SMALL_STATE(1121)] = 59193,
  [SMALL_STATE(1122)] = 59207,
  [SMALL_STATE(1123)] = 59223,
  [SMALL_STATE(1124)] = 59237,
  [SMALL_STATE(1125)] = 59251,
  [SMALL_STATE(1126)] = 59263,
  [SMALL_STATE(1127)] = 59277,
  [SMALL_STATE(1128)] = 59293,
  [SMALL_STATE(1129)] = 59305,
  [SMALL_STATE(1130)] = 59321,
  [SMALL_STATE(1131)] = 59335,
  [SMALL_STATE(1132)] = 59347,
  [SMALL_STATE(1133)] = 59359,
  [SMALL_STATE(1134)] = 59375,
  [SMALL_STATE(1135)] = 59387,
  [SMALL_STATE(1136)] = 59401,
  [SMALL_STATE(1137)] = 59415,
  [SMALL_STATE(1138)] = 59429,
  [SMALL_STATE(1139)] = 59443,
  [SMALL_STATE(1140)] = 59455,
  [SMALL_STATE(1141)] = 59471,
  [SMALL_STATE(1142)] = 59483,
  [SMALL_STATE(1143)] = 59497,
  [SMALL_STATE(1144)] = 59511,
  [SMALL_STATE(1145)] = 59527,
  [SMALL_STATE(1146)] = 59539,
  [SMALL_STATE(1147)] = 59553,
  [SMALL_STATE(1148)] = 59567,
  [SMALL_STATE(1149)] = 59579,
  [SMALL_STATE(1150)] = 59593,
  [SMALL_STATE(1151)] = 59607,
  [SMALL_STATE(1152)] = 59621,
  [SMALL_STATE(1153)] = 59635,
  [SMALL_STATE(1154)] = 59649,
  [SMALL_STATE(1155)] = 59663,
  [SMALL_STATE(1156)] = 59675,
  [SMALL_STATE(1157)] = 59689,
  [SMALL_STATE(1158)] = 59701,
  [SMALL_STATE(1159)] = 59717,
  [SMALL_STATE(1160)] = 59729,
  [SMALL_STATE(1161)] = 59741,
  [SMALL_STATE(1162)] = 59755,
  [SMALL_STATE(1163)] = 59764,
  [SMALL_STATE(1164)] = 59773,
  [SMALL_STATE(1165)] = 59786,
  [SMALL_STATE(1166)] = 59799,
  [SMALL_STATE(1167)] = 59812,
  [SMALL_STATE(1168)] = 59825,
  [SMALL_STATE(1169)] = 59838,
  [SMALL_STATE(1170)] = 59851,
  [SMALL_STATE(1171)] = 59864,
  [SMALL_STATE(1172)] = 59875,
  [SMALL_STATE(1173)] = 59888,
  [SMALL_STATE(1174)] = 59901,
  [SMALL_STATE(1175)] = 59910,
  [SMALL_STATE(1176)] = 59919,
  [SMALL_STATE(1177)] = 59928,
  [SMALL_STATE(1178)] = 59939,
  [SMALL_STATE(1179)] = 59952,
  [SMALL_STATE(1180)] = 59965,
  [SMALL_STATE(1181)] = 59974,
  [SMALL_STATE(1182)] = 59987,
  [SMALL_STATE(1183)] = 59996,
  [SMALL_STATE(1184)] = 60007,
  [SMALL_STATE(1185)] = 60020,
  [SMALL_STATE(1186)] = 60031,
  [SMALL_STATE(1187)] = 60042,
  [SMALL_STATE(1188)] = 60055,
  [SMALL_STATE(1189)] = 60068,
  [SMALL_STATE(1190)] = 60081,
  [SMALL_STATE(1191)] = 60094,
  [SMALL_STATE(1192)] = 60107,
  [SMALL_STATE(1193)] = 60120,
  [SMALL_STATE(1194)] = 60133,
  [SMALL_STATE(1195)] = 60146,
  [SMALL_STATE(1196)] = 60157,
  [SMALL_STATE(1197)] = 60170,
  [SMALL_STATE(1198)] = 60183,
  [SMALL_STATE(1199)] = 60196,
  [SMALL_STATE(1200)] = 60209,
  [SMALL_STATE(1201)] = 60218,
  [SMALL_STATE(1202)] = 60231,
  [SMALL_STATE(1203)] = 60242,
  [SMALL_STATE(1204)] = 60255,
  [SMALL_STATE(1205)] = 60268,
  [SMALL_STATE(1206)] = 60279,
  [SMALL_STATE(1207)] = 60292,
  [SMALL_STATE(1208)] = 60305,
  [SMALL_STATE(1209)] = 60316,
  [SMALL_STATE(1210)] = 60329,
  [SMALL_STATE(1211)] = 60342,
  [SMALL_STATE(1212)] = 60355,
  [SMALL_STATE(1213)] = 60368,
  [SMALL_STATE(1214)] = 60381,
  [SMALL_STATE(1215)] = 60394,
  [SMALL_STATE(1216)] = 60403,
  [SMALL_STATE(1217)] = 60416,
  [SMALL_STATE(1218)] = 60429,
  [SMALL_STATE(1219)] = 60442,
  [SMALL_STATE(1220)] = 60453,
  [SMALL_STATE(1221)] = 60466,
  [SMALL_STATE(1222)] = 60479,
  [SMALL_STATE(1223)] = 60492,
  [SMALL_STATE(1224)] = 60503,
  [SMALL_STATE(1225)] = 60516,
  [SMALL_STATE(1226)] = 60525,
  [SMALL_STATE(1227)] = 60538,
  [SMALL_STATE(1228)] = 60549,
  [SMALL_STATE(1229)] = 60562,
  [SMALL_STATE(1230)] = 60575,
  [SMALL_STATE(1231)] = 60588,
  [SMALL_STATE(1232)] = 60601,
  [SMALL_STATE(1233)] = 60614,
  [SMALL_STATE(1234)] = 60627,
  [SMALL_STATE(1235)] = 60638,
  [SMALL_STATE(1236)] = 60651,
  [SMALL_STATE(1237)] = 60662,
  [SMALL_STATE(1238)] = 60675,
  [SMALL_STATE(1239)] = 60688,
  [SMALL_STATE(1240)] = 60701,
  [SMALL_STATE(1241)] = 60714,
  [SMALL_STATE(1242)] = 60727,
  [SMALL_STATE(1243)] = 60740,
  [SMALL_STATE(1244)] = 60753,
  [SMALL_STATE(1245)] = 60766,
  [SMALL_STATE(1246)] = 60779,
  [SMALL_STATE(1247)] = 60792,
  [SMALL_STATE(1248)] = 60801,
  [SMALL_STATE(1249)] = 60814,
  [SMALL_STATE(1250)] = 60827,
  [SMALL_STATE(1251)] = 60840,
  [SMALL_STATE(1252)] = 60853,
  [SMALL_STATE(1253)] = 60864,
  [SMALL_STATE(1254)] = 60877,
  [SMALL_STATE(1255)] = 60890,
  [SMALL_STATE(1256)] = 60903,
  [SMALL_STATE(1257)] = 60914,
  [SMALL_STATE(1258)] = 60927,
  [SMALL_STATE(1259)] = 60940,
  [SMALL_STATE(1260)] = 60953,
  [SMALL_STATE(1261)] = 60966,
  [SMALL_STATE(1262)] = 60977,
  [SMALL_STATE(1263)] = 60990,
  [SMALL_STATE(1264)] = 61003,
  [SMALL_STATE(1265)] = 61016,
  [SMALL_STATE(1266)] = 61029,
  [SMALL_STATE(1267)] = 61042,
  [SMALL_STATE(1268)] = 61055,
  [SMALL_STATE(1269)] = 61065,
  [SMALL_STATE(1270)] = 61075,
  [SMALL_STATE(1271)] = 61085,
  [SMALL_STATE(1272)] = 61095,
  [SMALL_STATE(1273)] = 61105,
  [SMALL_STATE(1274)] = 61115,
  [SMALL_STATE(1275)] = 61125,
  [SMALL_STATE(1276)] = 61133,
  [SMALL_STATE(1277)] = 61141,
  [SMALL_STATE(1278)] = 61149,
  [SMALL_STATE(1279)] = 61159,
  [SMALL_STATE(1280)] = 61167,
  [SMALL_STATE(1281)] = 61175,
  [SMALL_STATE(1282)] = 61185,
  [SMALL_STATE(1283)] = 61195,
  [SMALL_STATE(1284)] = 61205,
  [SMALL_STATE(1285)] = 61215,
  [SMALL_STATE(1286)] = 61225,
  [SMALL_STATE(1287)] = 61235,
  [SMALL_STATE(1288)] = 61243,
  [SMALL_STATE(1289)] = 61251,
  [SMALL_STATE(1290)] = 61261,
  [SMALL_STATE(1291)] = 61271,
  [SMALL_STATE(1292)] = 61279,
  [SMALL_STATE(1293)] = 61289,
  [SMALL_STATE(1294)] = 61299,
  [SMALL_STATE(1295)] = 61307,
  [SMALL_STATE(1296)] = 61317,
  [SMALL_STATE(1297)] = 61327,
  [SMALL_STATE(1298)] = 61335,
  [SMALL_STATE(1299)] = 61343,
  [SMALL_STATE(1300)] = 61353,
  [SMALL_STATE(1301)] = 61363,
  [SMALL_STATE(1302)] = 61371,
  [SMALL_STATE(1303)] = 61379,
  [SMALL_STATE(1304)] = 61387,
  [SMALL_STATE(1305)] = 61397,
  [SMALL_STATE(1306)] = 61407,
  [SMALL_STATE(1307)] = 61417,
  [SMALL_STATE(1308)] = 61425,
  [SMALL_STATE(1309)] = 61435,
  [SMALL_STATE(1310)] = 61445,
  [SMALL_STATE(1311)] = 61455,
  [SMALL_STATE(1312)] = 61463,
  [SMALL_STATE(1313)] = 61473,
  [SMALL_STATE(1314)] = 61483,
  [SMALL_STATE(1315)] = 61491,
  [SMALL_STATE(1316)] = 61501,
  [SMALL_STATE(1317)] = 61509,
  [SMALL_STATE(1318)] = 61519,
  [SMALL_STATE(1319)] = 61527,
  [SMALL_STATE(1320)] = 61537,
  [SMALL_STATE(1321)] = 61547,
  [SMALL_STATE(1322)] = 61557,
  [SMALL_STATE(1323)] = 61567,
  [SMALL_STATE(1324)] = 61577,
  [SMALL_STATE(1325)] = 61587,
  [SMALL_STATE(1326)] = 61597,
  [SMALL_STATE(1327)] = 61607,
  [SMALL_STATE(1328)] = 61617,
  [SMALL_STATE(1329)] = 61625,
  [SMALL_STATE(1330)] = 61635,
  [SMALL_STATE(1331)] = 61643,
  [SMALL_STATE(1332)] = 61653,
  [SMALL_STATE(1333)] = 61663,
  [SMALL_STATE(1334)] = 61671,
  [SMALL_STATE(1335)] = 61679,
  [SMALL_STATE(1336)] = 61687,
  [SMALL_STATE(1337)] = 61697,
  [SMALL_STATE(1338)] = 61707,
  [SMALL_STATE(1339)] = 61717,
  [SMALL_STATE(1340)] = 61727,
  [SMALL_STATE(1341)] = 61737,
  [SMALL_STATE(1342)] = 61744,
  [SMALL_STATE(1343)] = 61751,
  [SMALL_STATE(1344)] = 61758,
  [SMALL_STATE(1345)] = 61765,
  [SMALL_STATE(1346)] = 61772,
  [SMALL_STATE(1347)] = 61779,
  [SMALL_STATE(1348)] = 61786,
  [SMALL_STATE(1349)] = 61793,
  [SMALL_STATE(1350)] = 61800,
  [SMALL_STATE(1351)] = 61807,
  [SMALL_STATE(1352)] = 61814,
  [SMALL_STATE(1353)] = 61821,
  [SMALL_STATE(1354)] = 61828,
  [SMALL_STATE(1355)] = 61835,
  [SMALL_STATE(1356)] = 61842,
  [SMALL_STATE(1357)] = 61849,
  [SMALL_STATE(1358)] = 61856,
  [SMALL_STATE(1359)] = 61863,
  [SMALL_STATE(1360)] = 61870,
  [SMALL_STATE(1361)] = 61877,
  [SMALL_STATE(1362)] = 61884,
  [SMALL_STATE(1363)] = 61891,
  [SMALL_STATE(1364)] = 61898,
  [SMALL_STATE(1365)] = 61905,
  [SMALL_STATE(1366)] = 61912,
  [SMALL_STATE(1367)] = 61919,
  [SMALL_STATE(1368)] = 61926,
  [SMALL_STATE(1369)] = 61933,
  [SMALL_STATE(1370)] = 61940,
  [SMALL_STATE(1371)] = 61947,
  [SMALL_STATE(1372)] = 61954,
  [SMALL_STATE(1373)] = 61961,
  [SMALL_STATE(1374)] = 61968,
  [SMALL_STATE(1375)] = 61975,
  [SMALL_STATE(1376)] = 61982,
  [SMALL_STATE(1377)] = 61989,
  [SMALL_STATE(1378)] = 61996,
  [SMALL_STATE(1379)] = 62003,
  [SMALL_STATE(1380)] = 62010,
  [SMALL_STATE(1381)] = 62017,
  [SMALL_STATE(1382)] = 62024,
  [SMALL_STATE(1383)] = 62031,
  [SMALL_STATE(1384)] = 62038,
  [SMALL_STATE(1385)] = 62045,
  [SMALL_STATE(1386)] = 62052,
  [SMALL_STATE(1387)] = 62059,
  [SMALL_STATE(1388)] = 62066,
  [SMALL_STATE(1389)] = 62073,
  [SMALL_STATE(1390)] = 62080,
  [SMALL_STATE(1391)] = 62087,
  [SMALL_STATE(1392)] = 62094,
  [SMALL_STATE(1393)] = 62101,
  [SMALL_STATE(1394)] = 62108,
  [SMALL_STATE(1395)] = 62115,
  [SMALL_STATE(1396)] = 62122,
  [SMALL_STATE(1397)] = 62129,
  [SMALL_STATE(1398)] = 62136,
  [SMALL_STATE(1399)] = 62143,
  [SMALL_STATE(1400)] = 62150,
  [SMALL_STATE(1401)] = 62157,
  [SMALL_STATE(1402)] = 62164,
  [SMALL_STATE(1403)] = 62171,
  [SMALL_STATE(1404)] = 62178,
  [SMALL_STATE(1405)] = 62185,
  [SMALL_STATE(1406)] = 62192,
  [SMALL_STATE(1407)] = 62199,
  [SMALL_STATE(1408)] = 62206,
  [SMALL_STATE(1409)] = 62213,
  [SMALL_STATE(1410)] = 62220,
  [SMALL_STATE(1411)] = 62227,
  [SMALL_STATE(1412)] = 62234,
  [SMALL_STATE(1413)] = 62241,
  [SMALL_STATE(1414)] = 62248,
  [SMALL_STATE(1415)] = 62255,
  [SMALL_STATE(1416)] = 62262,
  [SMALL_STATE(1417)] = 62269,
  [SMALL_STATE(1418)] = 62276,
  [SMALL_STATE(1419)] = 62283,
  [SMALL_STATE(1420)] = 62290,
  [SMALL_STATE(1421)] = 62297,
  [SMALL_STATE(1422)] = 62304,
  [SMALL_STATE(1423)] = 62311,
  [SMALL_STATE(1424)] = 62318,
  [SMALL_STATE(1425)] = 62325,
  [SMALL_STATE(1426)] = 62332,
  [SMALL_STATE(1427)] = 62339,
  [SMALL_STATE(1428)] = 62346,
  [SMALL_STATE(1429)] = 62353,
  [SMALL_STATE(1430)] = 62360,
  [SMALL_STATE(1431)] = 62367,
  [SMALL_STATE(1432)] = 62374,
  [SMALL_STATE(1433)] = 62381,
  [SMALL_STATE(1434)] = 62388,
  [SMALL_STATE(1435)] = 62395,
  [SMALL_STATE(1436)] = 62402,
  [SMALL_STATE(1437)] = 62409,
  [SMALL_STATE(1438)] = 62416,
  [SMALL_STATE(1439)] = 62423,
  [SMALL_STATE(1440)] = 62430,
  [SMALL_STATE(1441)] = 62437,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 0, 0, 0),
  [7] = {.entry = {.count = 1, .reusable = false}}, SHIFT(238),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(954),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1345),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(846),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1226),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1091),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1222),
  [23] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1152),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1327),
  [31] = {.entry = {.count = 1, .reusable = true}}, SHIFT(740),
  [33] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1363),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1364),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(728),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(986),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(911),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(913),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1379),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(28),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(133),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(192),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(27),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(21),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(22),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1391),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(248),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1407),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1084),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(255),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [77] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 1, 0, 0),
  [79] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0),
  [81] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(238),
  [84] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(954),
  [87] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1438),
  [90] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(843),
  [93] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(162),
  [96] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1226),
  [99] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1091),
  [102] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1245),
  [105] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(50),
  [108] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1152),
  [111] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(203),
  [114] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1327),
  [117] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(740),
  [120] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(11),
  [123] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1363),
  [126] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1364),
  [129] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(728),
  [132] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(200),
  [135] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(986),
  [138] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(911),
  [141] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(913),
  [144] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1379),
  [147] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(28),
  [150] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(133),
  [153] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(192),
  [156] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(27),
  [159] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(21),
  [162] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(22),
  [165] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1391),
  [168] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(248),
  [171] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(176),
  [174] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1407),
  [177] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1084),
  [180] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(255),
  [183] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(255),
  [186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(237),
  [188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1306),
  [190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_case, 3, 0, 7),
  [192] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_case, 3, 0, 7),
  [194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_default_case, 2, 0, 0),
  [196] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_default_case, 2, 0, 0),
  [198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_case, 3, 0, 41),
  [200] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_case, 3, 0, 41),
  [202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_case, 4, 0, 105),
  [204] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_case, 4, 0, 105),
  [206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_communication_case, 3, 0, 89),
  [208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_communication_case, 3, 0, 89),
  [210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_list, 2, 0, 0),
  [212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement_list, 2, 0, 0),
  [214] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_list, 3, 0, 0),
  [216] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement_list, 3, 0, 0),
  [218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_empty_labeled_statement, 2, 0, 26),
  [222] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_empty_labeled_statement, 2, 0, 26),
  [224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(546),
  [226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1159),
  [228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(503),
  [232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(294),
  [238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1320),
  [244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(206),
  [250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(299),
  [252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1411),
  [256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1161),
  [258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(331),
  [260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_clause, 3, 0, 57),
  [266] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_clause, 4, 0, 87),
  [268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_clause, 2, 0, 0),
  [270] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_clause, 3, 0, 58),
  [272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(431),
  [274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 1, 0, 0),
  [276] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 1, 0, 0),
  [278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(219),
  [280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1315),
  [282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(50),
  [284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(191),
  [286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(740),
  [288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(193),
  [290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(450),
  [292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(148),
  [294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1401),
  [296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1097),
  [298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(497),
  [300] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(395),
  [304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1367),
  [308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1289),
  [310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(394),
  [320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1390),
  [324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1080),
  [326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(435),
  [328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1434),
  [332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1356),
  [336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1404),
  [340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1360),
  [344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1424),
  [348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(554),
  [350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(469),
  [366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(567),
  [374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(458),
  [376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1331),
  [380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(517),
  [386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1430),
  [390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1150),
  [392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(544),
  [394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(721),
  [398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1397),
  [400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(515),
  [402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(516),
  [408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(189),
  [422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(718),
  [426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(473),
  [436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(545),
  [440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1007),
  [444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(765),
  [454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [468] = {.entry = {.count = 1, .reusable = false}}, SHIFT(236),
  [470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(576),
  [472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(732),
  [492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1401),
  [502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1097),
  [504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(743),
  [522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(424),
  [536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(564),
  [538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(573),
  [544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(339),
  [552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1324),
  [556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(347),
  [562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1136),
  [568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(348),
  [570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(774),
  [576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(239),
  [578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(855),
  [580] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(1351),
  [583] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__simple_type, 1, -1, 1), REDUCE(sym__expression, 1, 0, 0), SHIFT(703),
  [587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1405),
  [589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1278),
  [591] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__simple_type, 1, -1, 1), REDUCE(sym__expression, 1, 0, 0), SHIFT(68),
  [595] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0),
  [597] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(702),
  [600] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0),
  [602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1431),
  [604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(759),
  [606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1377),
  [608] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), SHIFT(1351),
  [611] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__simple_type, 1, -1, 1), REDUCE(sym__expression, 1, 0, 0),
  [614] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym__simple_type, 1, -1, 1), REDUCE(sym__expression, 1, 0, 0), SHIFT(657),
  [618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(30),
  [620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(12),
  [622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(19),
  [624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 1, 0, 0),
  [626] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_statement, 1, 0, 0),
  [628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1310),
  [630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(52),
  [632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(187),
  [634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_list, 1, 0, 0),
  [636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(54),
  [638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(188),
  [640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(190),
  [642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(131),
  [644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1014),
  [646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1015),
  [648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(194),
  [650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(196),
  [652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(197),
  [654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0),
  [656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, 0, 32),
  [658] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, 0, 32),
  [660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_expression, 2, 0, 6),
  [662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_expression, 2, 0, 6),
  [664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(659),
  [666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpreted_string_literal, 3, 0, 0),
  [668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpreted_string_literal, 3, 0, 0),
  [670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpreted_string_literal, 2, 0, 0),
  [672] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpreted_string_literal, 2, 0, 0),
  [674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_raw_string_literal, 3, 0, 0),
  [676] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_raw_string_literal, 3, 0, 0),
  [678] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 2, 0, 0),
  [680] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 2, 0, 0),
  [682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 3, 0, 0),
  [684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 3, 0, 0),
  [686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_special_argument_list, 3, 0, 0),
  [688] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_special_argument_list, 3, 0, 0),
  [690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_func_literal, 4, 0, 46),
  [692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_func_literal, 4, 0, 46),
  [694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_expression, 7, 0, 106),
  [696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_expression, 7, 0, 106),
  [698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_special_argument_list, 2, 0, 0),
  [700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_special_argument_list, 2, 0, 0),
  [702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_expression, 6, 0, 103),
  [704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_expression, 6, 0, 103),
  [706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_literal_value, 3, 0, 0),
  [708] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_literal_value, 3, 0, 0),
  [710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_conversion_expression, 4, -1, 61),
  [712] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_conversion_expression, 4, -1, 61),
  [714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_instantiation_expression, 4, -1, 19),
  [716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_instantiation_expression, 4, -1, 19),
  [718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 3, 0, 0),
  [720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 3, 0, 0),
  [722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selector_expression, 3, 0, 34),
  [724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_selector_expression, 3, 0, 34),
  [726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_expression, 4, 0, 62),
  [728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_expression, 4, 0, 62),
  [730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_index_expression, 4, 1, 63),
  [732] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_index_expression, 4, 1, 63),
  [734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 2, 0, 0),
  [740] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 2, 0, 0),
  [742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call_expression, 3, 0, 36),
  [744] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call_expression, 3, 0, 36),
  [746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_expression, 8, 0, 109),
  [748] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_expression, 8, 0, 109),
  [750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(155),
  [760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(156),
  [762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_list, 1, 0, 0),
  [766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1307),
  [768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1311),
  [770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(157),
  [774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_special_argument_list, 4, 0, 0),
  [780] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_special_argument_list, 4, 0, 0),
  [782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_literal_value, 2, 0, 0),
  [784] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_literal_value, 2, 0, 0),
  [786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_literal_value, 4, 0, 0),
  [788] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_literal_value, 4, 0, 0),
  [790] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_conversion_expression, 5, -1, 61),
  [792] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_conversion_expression, 5, -1, 61),
  [794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_instantiation_expression, 5, -1, 19),
  [796] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_instantiation_expression, 5, -1, 19),
  [798] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_assertion_expression, 5, 0, 91),
  [800] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_assertion_expression, 5, 0, 91),
  [802] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 4, 0, 0),
  [804] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 4, 0, 0),
  [806] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_expression, 5, 0, 92),
  [808] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_expression, 5, 0, 92),
  [810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_expression, 5, 0, 93),
  [812] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_expression, 5, 0, 93),
  [814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_special_argument_list, 5, 0, 0),
  [816] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_special_argument_list, 5, 0, 0),
  [818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_literal_value, 5, 0, 0),
  [820] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_literal_value, 5, 0, 0),
  [822] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_composite_literal, 2, 0, 13),
  [824] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_composite_literal, 2, 0, 13),
  [826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_composite_literal, 2, 0, 12),
  [828] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_composite_literal, 2, 0, 12),
  [830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_instantiation_expression, 6, -1, 19),
  [832] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_instantiation_expression, 6, -1, 19),
  [834] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 5, 0, 0),
  [836] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 5, 0, 0),
  [838] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call_expression, 2, 0, 10),
  [840] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call_expression, 2, 0, 10),
  [842] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_func_literal, 3, 0, 22),
  [844] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_func_literal, 3, 0, 22),
  [846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1268),
  [848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [850] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__simple_type, 1, -1, 1), REDUCE(sym__expression, 1, 0, 0),
  [853] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__simple_type, 1, -1, 1), REDUCE(sym__expression, 1, 0, 0), SHIFT(657),
  [857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1270),
  [865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [867] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0),
  [869] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0),
  [871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(161),
  [875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(235),
  [877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(163),
  [881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(666),
  [887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(799),
  [889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 2, 0, 5),
  [891] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 2, 0, 5),
  [893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(511),
  [895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1319),
  [897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(90),
  [899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(734),
  [901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1296),
  [903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(757),
  [905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1374),
  [907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1435),
  [909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(731),
  [911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1383),
  [913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(797),
  [915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1272),
  [917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(126),
  [919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(720),
  [921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1433),
  [923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(716),
  [925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1423),
  [927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 1, 0, 25),
  [929] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 1, 0, 25),
  [931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1432),
  [933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(775),
  [935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1415),
  [937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(58),
  [939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1407),
  [941] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1084),
  [943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1295),
  [945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_spec, 2, 0, 17),
  [951] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_spec, 2, 0, 17),
  [953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(703),
  [955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1368),
  [957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(113),
  [959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(84),
  [961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(702),
  [963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1377),
  [965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_spec, 1, 0, 4),
  [967] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_spec, 1, 0, 4),
  [969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(128),
  [971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1282),
  [975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(62),
  [977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(174),
  [979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(79),
  [981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(149),
  [983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(150),
  [985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(151),
  [987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(152),
  [989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(153),
  [991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(29),
  [993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(144),
  [999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [1001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(143),
  [1003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [1005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(145),
  [1007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [1009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [1011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [1013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [1015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(844),
  [1017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1293),
  [1019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [1021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(709),
  [1023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1382),
  [1025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(682),
  [1027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1285),
  [1029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [1031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [1033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [1035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [1037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(168),
  [1039] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_receive_statement, 1, 0, 60),
  [1041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [1043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(167),
  [1045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [1047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(169),
  [1049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [1051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [1053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [1055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1351),
  [1057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [1059] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_type, 1, -1, 1),
  [1061] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__simple_type, 1, -1, 1), SHIFT(1405),
  [1064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [1066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(712),
  [1068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 3, 0, 20),
  [1070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(900),
  [1072] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 3, 0, 20),
  [1074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(510),
  [1076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1326),
  [1078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(102),
  [1080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(744),
  [1082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1305),
  [1084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(715),
  [1086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(14),
  [1088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1389),
  [1090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1437),
  [1092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(742),
  [1094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1426),
  [1096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_defer_statement, 2, 0, 0),
  [1098] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_defer_statement, 2, 0, 0),
  [1100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_send_statement, 3, 0, 35),
  [1102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_send_statement, 3, 0, 35),
  [1104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_go_statement, 2, 0, 0),
  [1106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_go_statement, 2, 0, 0),
  [1108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(34),
  [1110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 4, 0, 42),
  [1112] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 4, 0, 42),
  [1114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 4, 0, 45),
  [1116] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 4, 0, 45),
  [1118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [1120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(11),
  [1122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(453),
  [1124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [1126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [1128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(738),
  [1130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(898),
  [1132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1283),
  [1134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(114),
  [1136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(754),
  [1138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1439),
  [1140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(752),
  [1142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1402),
  [1144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(793),
  [1146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1376),
  [1148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [1150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1420),
  [1152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [1154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [1156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [1158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1333),
  [1160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [1162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(137),
  [1164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [1166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(136),
  [1168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [1170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(139),
  [1172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [1174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [1176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(794),
  [1178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(795),
  [1180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [1182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(789),
  [1184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [1186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [1188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [1190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [1192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [1194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(783),
  [1196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [1198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [1200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_elem, 2, 0, 55),
  [1202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_elem, 2, 0, 55),
  [1204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [1206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [1208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(563),
  [1210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [1212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1329),
  [1214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [1216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(992),
  [1218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(179),
  [1220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [1222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(178),
  [1224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [1226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(180),
  [1228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [1230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [1232] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0),
  [1234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [1236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [1238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [1240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(720),
  [1242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(757),
  [1244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [1246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [1248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_literal_element, 1, 0, 0),
  [1250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(856),
  [1252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [1254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [1256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [1258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [1260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [1262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(950),
  [1264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [1266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(975),
  [1268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [1270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(883),
  [1272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1290),
  [1274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [1276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [1278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [1280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(95),
  [1282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [1284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [1286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [1288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1341),
  [1290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [1292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [1294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1019),
  [1296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(830),
  [1298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [1300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [1302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [1304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1409),
  [1306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [1308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [1310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [1312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [1314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [1316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [1318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [1320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1410),
  [1322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [1324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [1326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(978),
  [1328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [1330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [1332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(988),
  [1334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [1336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [1338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [1340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [1342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1436),
  [1344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [1346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [1348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1415),
  [1350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [1352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [1354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [1356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [1358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [1360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [1362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1342),
  [1364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [1366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [1368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [1370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [1372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [1374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [1376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(13),
  [1378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(561),
  [1380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1346),
  [1382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [1384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [1386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(572),
  [1388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [1390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [1392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [1394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [1396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(983),
  [1398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [1400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [1402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_clause, 2, 0, 29),
  [1404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [1406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [1408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [1410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [1412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [1414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(736),
  [1416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [1418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [1420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [1422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [1424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [1426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [1428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [1430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1395),
  [1432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [1434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [1436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [1438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(704),
  [1440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [1442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [1444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1408),
  [1446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [1448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [1450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_clause, 4, 0, 85),
  [1452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(735),
  [1454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [1456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1400),
  [1458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [1460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [1462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(550),
  [1464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [1466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [1468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [1470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [1472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [1474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(751),
  [1476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(598),
  [1478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(730),
  [1480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [1482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [1484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [1486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [1488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [1490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1406),
  [1492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_receive_statement, 3, 0, 33),
  [1494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [1496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [1498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(471),
  [1500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [1502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [1504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(719),
  [1506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [1508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(744),
  [1510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [1512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1426),
  [1514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(114),
  [1516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(754),
  [1518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(717),
  [1520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1402),
  [1522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(722),
  [1524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [1526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(723),
  [1528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1325),
  [1530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [1532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [1534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(734),
  [1536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(733),
  [1538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1383),
  [1540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [1542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(760),
  [1544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(763),
  [1546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [1548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(766),
  [1550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 5, 0, 0),
  [1552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 5, 0, 0),
  [1554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 4, 0, 0),
  [1556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 4, 0, 0),
  [1558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 2, 0, 0),
  [1560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 2, 0, 0),
  [1562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 3, 0, 0),
  [1564] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 3, 0, 0),
  [1566] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_const_spec_repeat1, 2, 0, 0),
  [1568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_const_spec_repeat1, 2, 0, 0),
  [1570] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_const_spec_repeat1, 2, 0, 0), SHIFT_REPEAT(1368),
  [1573] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_type, 1, -1, 1),
  [1575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(695),
  [1577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_list, 2, 0, 0),
  [1579] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__simple_type, 1, -1, 1), SHIFT(695),
  [1582] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(187),
  [1585] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_type, 1, 0, 0),
  [1587] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_type, 1, 0, 0),
  [1589] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_var_spec_repeat1, 2, 0, 40),
  [1591] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_var_spec_repeat1, 2, 0, 40),
  [1593] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_var_spec_repeat1, 2, 0, 40), SHIFT_REPEAT(1405),
  [1596] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__simple_type, 1, 0, 0), SHIFT(695),
  [1599] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_type, 3, 0, 0),
  [1601] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_type, 3, 0, 0),
  [1603] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 3, 0, 0),
  [1605] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 3, 0, 0),
  [1607] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_type, 2, 0, 0),
  [1609] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_type, 2, 0, 0),
  [1611] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_qualified_type, 3, 0, 31),
  [1613] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_qualified_type, 3, 0, 31),
  [1615] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map_type, 5, 0, 81),
  [1617] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map_type, 5, 0, 81),
  [1619] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_channel_type, 3, 0, 27),
  [1621] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_channel_type, 3, 0, 27),
  [1623] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 3, 0, 21),
  [1625] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 3, 0, 21),
  [1627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_arguments, 5, 2, 0),
  [1629] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_arguments, 5, 2, 0),
  [1631] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_type, 2, 1, 14),
  [1633] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_type, 2, 1, 14),
  [1635] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_negated_type, 2, 0, 0),
  [1637] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_negated_type, 2, 0, 0),
  [1639] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_type, 1, 3, 0),
  [1641] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_type, 1, 3, 0),
  [1643] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 4, 0, 0),
  [1645] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 4, 0, 0),
  [1647] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_type, 2, 0, 0),
  [1649] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_type, 2, 0, 0),
  [1651] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_channel_type, 2, 0, 7),
  [1653] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_channel_type, 2, 0, 7),
  [1655] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_var_spec_repeat1, 2, 0, 38),
  [1657] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_var_spec_repeat1, 2, 0, 38),
  [1659] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_type, 2, 1, 11),
  [1661] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_type, 2, 1, 11),
  [1663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [1665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [1667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [1669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [1671] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_type, 4, 0, 0),
  [1673] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_type, 4, 0, 0),
  [1675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [1677] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_type, 5, 0, 0),
  [1679] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_type, 5, 0, 0),
  [1681] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_type, 3, 0, 23),
  [1683] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_type, 3, 0, 23),
  [1685] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 40),
  [1687] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 40),
  [1689] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 40), SHIFT_REPEAT(1415),
  [1692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 5, 0, 0),
  [1694] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 5, 0, 0),
  [1696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_arguments, 4, 2, 0),
  [1698] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_arguments, 4, 2, 0),
  [1700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type, 4, 0, 48),
  [1702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type, 4, 0, 48),
  [1704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_type, 6, 0, 0),
  [1706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_type, 6, 0, 0),
  [1708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_arguments, 3, 2, 0),
  [1710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_arguments, 3, 2, 0),
  [1712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_type, 3, 0, 0),
  [1714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_type, 3, 0, 0),
  [1716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [1718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [1720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 2, 0, 0),
  [1722] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 2, 0, 0),
  [1724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [1726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [1728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameter_list, 4, 0, 0),
  [1730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_list, 4, 0, 0),
  [1732] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 77),
  [1734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_field_declaration_repeat1, 2, 0, 77),
  [1736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameter_list, 5, 0, 0),
  [1738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_list, 5, 0, 0),
  [1740] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameter_list, 3, 0, 0),
  [1742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_list, 3, 0, 0),
  [1744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1124),
  [1746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1316),
  [1748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1121),
  [1750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(851),
  [1752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1351),
  [1754] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__simple_type, 1, -1, 1), SHIFT(657),
  [1757] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 2, 0, 51),
  [1759] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 2, 0, 51),
  [1761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1146),
  [1763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1156),
  [1765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [1767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1375),
  [1769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1114),
  [1771] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__simple_type, 1, 0, 0), SHIFT(657),
  [1774] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 1, 0, 19),
  [1776] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 1, 0, 19),
  [1778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [1780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1071),
  [1782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1261),
  [1784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 2, 0, 41),
  [1786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 2, 0, 41),
  [1788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [1790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1074),
  [1792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(657),
  [1794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1223),
  [1796] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_list, 2, 0, 0),
  [1798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [1800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1056),
  [1802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(974),
  [1804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [1806] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 2, 0, 52),
  [1808] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 2, 0, 52),
  [1810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 79),
  [1812] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 3, 0, 79),
  [1814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1362),
  [1816] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__simple_type, 1, -1, 1), SHIFT(697),
  [1819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [1821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(697),
  [1823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [1825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(868),
  [1827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [1829] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(174),
  [1832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(947),
  [1834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [1836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(836),
  [1838] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 3, 0, 28),
  [1840] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 28),
  [1842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1135),
  [1844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1026),
  [1846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_statement, 1, 0, 0),
  [1848] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_statement, 1, 0, 0),
  [1850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1028),
  [1852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_statement, 1, 0, 0),
  [1854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_statement, 1, 0, 0),
  [1856] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_var_spec, 2, 0, 18),
  [1858] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_var_spec, 2, 0, 18),
  [1860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(115),
  [1862] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_var_spec, 3, 0, 39),
  [1864] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_var_spec, 3, 0, 39),
  [1866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(107),
  [1868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [1870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(9),
  [1872] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement_list, 1, 0, 0),
  [1874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(111),
  [1876] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__simple_type, 1, 0, 0), SHIFT(697),
  [1879] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 82),
  [1881] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, 0, 82),
  [1883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1130),
  [1885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [1887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(10),
  [1889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_spec, 3, 0, 50),
  [1891] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_spec, 3, 0, 50),
  [1893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_alias, 4, 0, 74),
  [1895] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_alias, 4, 0, 74),
  [1897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_spec, 2, 0, 24),
  [1899] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_spec, 2, 0, 24),
  [1901] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(20),
  [1904] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0), SHIFT_REPEAT(20),
  [1907] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0),
  [1909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_alias, 3, 0, 49),
  [1911] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_alias, 3, 0, 49),
  [1913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 7, 0, 104),
  [1915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 7, 0, 104),
  [1917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_var_declaration, 2, 0, 0),
  [1919] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_var_declaration, 2, 0, 0),
  [1921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_var_spec_list, 3, 0, 0),
  [1923] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_var_spec_list, 3, 0, 0),
  [1925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 5, 0, 72),
  [1927] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 5, 0, 72),
  [1929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_switch_statement, 6, 0, 101),
  [1931] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_switch_statement, 6, 0, 101),
  [1933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [1935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [1937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [1939] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_declaration, 5, 0, 0),
  [1941] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_declaration, 5, 0, 0),
  [1943] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_var_spec, 3, 0, 37),
  [1945] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_var_spec, 3, 0, 37),
  [1947] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_switch_statement, 5, 0, 59),
  [1949] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_switch_statement, 5, 0, 59),
  [1951] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_statement, 1, 0, 0),
  [1953] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_statement, 1, 0, 0),
  [1955] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_switch_statement_repeat1, 2, 0, 0),
  [1957] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_switch_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(758),
  [1960] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_switch_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1394),
  [1963] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_statement, 3, 0, 26),
  [1965] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_statement, 3, 0, 26),
  [1967] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_statement, 3, 0, 32),
  [1969] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment_statement, 3, 0, 32),
  [1971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_short_var_declaration, 3, 0, 33),
  [1973] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_short_var_declaration, 3, 0, 33),
  [1975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_switch_statement, 5, 0, 7),
  [1977] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_switch_statement, 5, 0, 7),
  [1979] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_empty_statement, 1, 0, 0),
  [1981] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_empty_statement, 1, 0, 0),
  [1983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_elem, 1, 0, 0),
  [1985] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_elem, 1, 0, 0),
  [1987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(746),
  [1989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 4, 0, 43),
  [1991] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 4, 0, 43),
  [1993] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_switch_statement, 5, 0, 88),
  [1995] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_switch_statement, 5, 0, 88),
  [1997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 83),
  [1999] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, 0, 83),
  [2001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [2003] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(135),
  [2006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1024),
  [2008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_switch_statement, 7, 0, 101),
  [2010] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_switch_statement, 7, 0, 101),
  [2012] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 2, 0, 9),
  [2014] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 2, 0, 9),
  [2016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_declaration, 4, 0, 0),
  [2018] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_declaration, 4, 0, 0),
  [2020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [2022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(761),
  [2024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [2026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_declaration, 2, 0, 0),
  [2028] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_declaration, 2, 0, 0),
  [2030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [2032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [2034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 3, 0, 30),
  [2036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 3, 0, 30),
  [2038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_spec, 5, 0, 94),
  [2040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_spec, 5, 0, 94),
  [2042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_var_spec, 5, 0, 95),
  [2044] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_var_spec, 5, 0, 95),
  [2046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [2048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fallthrough_statement, 1, 0, 0),
  [2050] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fallthrough_statement, 1, 0, 0),
  [2052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_statement_list_repeat1, 2, 0, 0),
  [2054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_switch_statement, 3, 0, 0),
  [2056] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_switch_statement, 3, 0, 0),
  [2058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statement, 1, 0, 0),
  [2060] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statement, 1, 0, 0),
  [2062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_var_spec, 4, 0, 64),
  [2064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_var_spec, 4, 0, 64),
  [2066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [2068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_spec, 4, 0, 65),
  [2070] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_spec, 4, 0, 65),
  [2072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_declaration, 6, 0, 0),
  [2074] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_declaration, 6, 0, 0),
  [2076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_elem, 2, 0, 0),
  [2078] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_elem, 2, 0, 0),
  [2080] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_declaration, 3, 0, 0),
  [2082] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_declaration, 3, 0, 0),
  [2084] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_declaration, 4, 0, 0),
  [2086] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_declaration, 4, 0, 0),
  [2088] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_elem_repeat1, 2, 0, 0),
  [2090] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_elem_repeat1, 2, 0, 0),
  [2092] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_elem_repeat1, 2, 0, 0), SHIFT_REPEAT(746),
  [2095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1005),
  [2097] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_var_spec_list, 2, 0, 0),
  [2099] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_var_spec_list, 2, 0, 0),
  [2101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_declaration, 2, 0, 0),
  [2103] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_declaration, 2, 0, 0),
  [2105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [2107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(758),
  [2109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1012),
  [2111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_var_spec, 4, 0, 66),
  [2113] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_var_spec, 4, 0, 66),
  [2115] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_switch_statement, 4, 0, 0),
  [2117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_switch_statement, 4, 0, 0),
  [2119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expression_switch_statement_repeat1, 2, 0, 0),
  [2121] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_switch_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(124),
  [2124] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_switch_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1394),
  [2127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [2129] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_switch_statement, 4, 0, 59),
  [2131] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_switch_statement, 4, 0, 59),
  [2133] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_select_statement, 3, 0, 0),
  [2135] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_select_statement, 3, 0, 0),
  [2137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 5, 0, 69),
  [2139] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 5, 0, 69),
  [2141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(937),
  [2143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_switch_statement, 4, 0, 7),
  [2145] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_switch_statement, 4, 0, 7),
  [2147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [2149] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inc_statement, 2, 0, 0),
  [2151] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_inc_statement, 2, 0, 0),
  [2153] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dec_statement, 2, 0, 0),
  [2155] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dec_statement, 2, 0, 0),
  [2157] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_select_statement, 4, 0, 0),
  [2159] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_select_statement, 4, 0, 0),
  [2161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_select_statement_repeat1, 2, 0, 0),
  [2163] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_select_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(49),
  [2166] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_select_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(1394),
  [2169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1016),
  [2171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(510),
  [2173] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_declaration, 3, 0, 0),
  [2175] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_declaration, 3, 0, 0),
  [2177] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_spec, 3, 0, 37),
  [2179] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_spec, 3, 0, 37),
  [2181] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_switch_statement, 6, 0, 88),
  [2183] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_switch_statement, 6, 0, 88),
  [2185] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_statement, 2, 0, 8),
  [2187] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_statement, 2, 0, 8),
  [2189] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_statement, 2, 0, 8),
  [2191] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_statement, 2, 0, 8),
  [2193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_goto_statement, 2, 0, 8),
  [2195] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_goto_statement, 2, 0, 8),
  [2197] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 2, 0, 0),
  [2199] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 2, 0, 0),
  [2201] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_spec, 4, 0, 64),
  [2203] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_spec, 4, 0, 64),
  [2205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1079),
  [2207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1079),
  [2209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(935),
  [2211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(849),
  [2213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(849),
  [2215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1223),
  [2217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(29),
  [2219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [2221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(585),
  [2223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(826),
  [2225] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_interface_type_repeat1, 2, 0, 0), SHIFT_REPEAT(614),
  [2228] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_interface_type_repeat1, 2, 0, 0), SHIFT_REPEAT(614),
  [2231] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_interface_type_repeat1, 2, 0, 0),
  [2233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(891),
  [2235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(891),
  [2237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(974),
  [2239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(31),
  [2241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [2243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(595),
  [2245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(975),
  [2247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(584),
  [2249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(584),
  [2251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(824),
  [2253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [2255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(854),
  [2257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1095),
  [2259] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1195),
  [2262] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_type_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1195),
  [2265] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_declaration_repeat1, 2, 0, 0),
  [2267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [2269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(853),
  [2271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1074),
  [2273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [2275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(596),
  [2277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(882),
  [2279] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(912),
  [2282] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_field_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(912),
  [2285] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_field_declaration_list_repeat1, 2, 0, 0),
  [2287] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_import_spec_list_repeat1, 2, 0, 0), SHIFT_REPEAT(865),
  [2290] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_import_spec_list_repeat1, 2, 0, 0), SHIFT_REPEAT(865),
  [2293] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_import_spec_list_repeat1, 2, 0, 0),
  [2295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(590),
  [2297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(590),
  [2299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(883),
  [2301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1090),
  [2303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1090),
  [2305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(969),
  [2307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [2309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(904),
  [2311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(904),
  [2313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(817),
  [2315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [2317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(902),
  [2319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(805),
  [2321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(845),
  [2323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [2325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(857),
  [2327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1171),
  [2329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(890),
  [2331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(890),
  [2333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(964),
  [2335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [2337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(903),
  [2339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(880),
  [2341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [2343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(899),
  [2345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(876),
  [2347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [2349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(594),
  [2351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(968),
  [2353] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_var_spec_list_repeat1, 2, 0, 0), SHIFT_REPEAT(541),
  [2356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_var_spec_list_repeat1, 2, 0, 0),
  [2358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(601),
  [2360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1022),
  [2362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 5, 0, 73),
  [2364] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 5, 0, 73),
  [2366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(541),
  [2368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1000),
  [2370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 2, 0, 0),
  [2372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [2374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(241),
  [2376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_spec, 2, 0, 15),
  [2378] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_spec, 2, 0, 15),
  [2380] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_spec_list, 2, 0, 0),
  [2382] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_spec_list, 2, 0, 0),
  [2384] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_declaration_repeat1, 2, 0, 0),
  [2386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_spec_list, 4, 0, 0),
  [2388] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_spec_list, 4, 0, 0),
  [2390] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_spec, 2, 0, 16),
  [2392] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_spec, 2, 0, 16),
  [2394] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_case_repeat1, 2, 0, 0), SHIFT_REPEAT(739),
  [2397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_case_repeat1, 2, 0, 0),
  [2399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [2401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [2403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [2405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [2407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [2409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1083),
  [2411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [2413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [2415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 4, 0, 44),
  [2417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 4, 0, 44),
  [2419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [2421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [2423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [2425] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_elem_repeat1, 2, 0, 0), SHIFT_REPEAT(707),
  [2428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_spec_list, 5, 0, 0),
  [2430] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_spec_list, 5, 0, 0),
  [2432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(935),
  [2434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [2436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [2438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [2440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 75),
  [2442] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 3, 0, 75),
  [2444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 76),
  [2446] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 3, 0, 76),
  [2448] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_spec_list, 3, 0, 0),
  [2450] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_spec_list, 3, 0, 0),
  [2452] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_import_spec_list_repeat1, 2, 0, 0),
  [2454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [2456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [2458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [2460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 6, 0, 97),
  [2462] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 6, 0, 97),
  [2464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [2466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [2468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 78),
  [2470] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 3, 0, 78),
  [2472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 6, 0, 98),
  [2474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_declaration, 6, 0, 98),
  [2476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 2, 0, 53),
  [2478] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 2, 0, 53),
  [2480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 2, 0, 54),
  [2482] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 2, 0, 54),
  [2484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 4, 0, 99),
  [2486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration, 4, 0, 99),
  [2488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1308),
  [2490] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_const_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(393),
  [2493] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_const_declaration_repeat1, 2, 0, 0),
  [2495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [2497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1148),
  [2499] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_spec, 1, 0, 3),
  [2501] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_spec, 1, 0, 3),
  [2503] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_interpreted_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(1120),
  [2506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_interpreted_string_literal_repeat1, 2, 0, 0),
  [2508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [2510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [2512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [2514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_field_declaration_list_repeat1, 2, 0, 0),
  [2516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [2518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [2520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_declaration, 2, 0, 0),
  [2522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_declaration, 2, 0, 0),
  [2524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [2526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(997),
  [2528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_elem, 3, 0, 80),
  [2530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_elem, 3, 0, 80),
  [2532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [2534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [2536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 5, 0, 70),
  [2538] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 5, 0, 70),
  [2540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [2542] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_clause, 2, 0, 2),
  [2544] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_clause, 2, 0, 2),
  [2546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [2548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [2550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(560),
  [2552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_interface_type_repeat1, 2, 0, 0),
  [2554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1132),
  [2556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1151),
  [2558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [2560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(582),
  [2562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1066),
  [2564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [2566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [2568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 5, 0, 71),
  [2570] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 5, 0, 71),
  [2572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1123),
  [2574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [2576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [2578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [2580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(120),
  [2582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(621),
  [2584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [2586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [2588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [2590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(887),
  [2592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [2594] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_literal_value_repeat1, 2, 0, 0), SHIFT_REPEAT(47),
  [2597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_literal_value_repeat1, 2, 0, 0),
  [2599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [2601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [2603] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_case, 4, 0, 41),
  [2605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [2607] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_communication_case, 4, 0, 89),
  [2609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [2611] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_case, 5, 0, 105),
  [2613] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_parameter_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1317),
  [2616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_parameter_list_repeat1, 2, 0, 0),
  [2618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [2622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [2624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [2626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [2628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [2630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(881),
  [2632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(98),
  [2634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [2636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [2638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [2640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [2642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_default_case, 3, 0, 0),
  [2644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [2646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(739),
  [2648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7),
  [2650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [2652] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(96),
  [2655] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameter_list_repeat1, 2, 0, 0),
  [2657] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameter_list_repeat1, 2, 0, 0), SHIFT_REPEAT(535),
  [2660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [2662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [2664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1173),
  [2666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [2668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(644),
  [2670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [2672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1181),
  [2674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_list, 4, 0, 0),
  [2676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [2678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [2680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(110),
  [2682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [2684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1294),
  [2686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1294),
  [2688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [2690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [2692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1111),
  [2694] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(645),
  [2697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_arguments_repeat1, 2, 0, 0),
  [2699] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_case, 4, 0, 7),
  [2701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1122),
  [2703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1297),
  [2705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1297),
  [2707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [2709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(559),
  [2711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [2713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(786),
  [2715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [2717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [2719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(562),
  [2721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [2723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [2725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [2727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [2729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [2731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [2733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(791),
  [2735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(521),
  [2737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(619),
  [2739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(971),
  [2741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1129),
  [2743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(796),
  [2745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [2747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_list, 1, 0, 0),
  [2749] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(195),
  [2752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [2754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [2756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [2758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(640),
  [2760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [2762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(100),
  [2764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(607),
  [2766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(833),
  [2768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [2770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(683),
  [2772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [2774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [2776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [2778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(527),
  [2780] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(186),
  [2783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [2785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [2787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(692),
  [2789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [2791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(780),
  [2793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [2795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [2797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [2799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 2, 0, 18),
  [2801] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 3, 0, 39),
  [2803] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyed_element, 3, 0, 90),
  [2805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [2807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(768),
  [2809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(62),
  [2811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(555),
  [2813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [2815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [2817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [2819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [2821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(909),
  [2823] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 1, 0, 19),
  [2825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [2827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [2829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [2831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(713),
  [2833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [2835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dot, 1, 0, 0),
  [2837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter_declaration, 2, 0, 41),
  [2839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(661),
  [2841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(907),
  [2843] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_declaration, 2, 0, 68),
  [2845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [2847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(695),
  [2849] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_argument, 2, 0, 0),
  [2851] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter_declaration, 3, 0, 67),
  [2853] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_declaration, 3, 0, 96),
  [2855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [2857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(767),
  [2859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [2861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [2863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [2865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(777),
  [2867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [2869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(705),
  [2871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [2873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [2875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [2877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(745),
  [2879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1141),
  [2881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [2883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [2885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1002),
  [2887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [2889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(863),
  [2891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [2893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [2895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [2897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1078),
  [2899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [2901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [2903] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_clause, 4, 0, 84),
  [2905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_switch_header, 9, 0, 110),
  [2907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(953),
  [2909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [2911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [2913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(552),
  [2915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(581),
  [2917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [2919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(749),
  [2921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_clause, 5, 0, 100),
  [2923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8),
  [2925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_clause, 4, 0, 86),
  [2927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_clause, 3, 0, 56),
  [2929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [2931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1393),
  [2933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [2935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [2937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [2939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [2941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [2943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(755),
  [2945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(737),
  [2947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [2949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [2951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [2953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_switch_header, 7, 0, 108),
  [2955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(944),
  [2957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [2959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1380),
  [2961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(977),
  [2963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [2965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1131),
  [2967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5),
  [2969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1361),
  [2971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [2973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(747),
  [2975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [2977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(726),
  [2979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1387),
  [2981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1381),
  [2983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(725),
  [2985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [2987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [2989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1425),
  [2991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [2993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1413),
  [2995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [2997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [2999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1352),
  [3001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [3003] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_switch_header, 7, 0, 107),
  [3005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [3007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [3009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4),
  [3011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(504),
  [3013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [3015] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [3017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1412),
  [3019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [3021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [3023] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_switch_header, 5, 0, 102),
  [3025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [3027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [3029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [3031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1366),
  [3033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [3035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(807),
  [3037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(764),
  [3039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [3041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [3043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [3045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1202),
  [3047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [3049] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_implicit_length_array_type, 4, 0, 47),
  [3051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
};

#ifdef __cplusplus
extern "C" {
#endif
#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_go(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .supertype_map_slices = ts_supertype_map_slices,
    .supertype_map_entries = ts_supertype_map_entries,
    .supertype_symbols = ts_supertype_symbols,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_identifier,
    .primary_state_ids = ts_primary_state_ids,
    .name = "go",
    .reserved_words = &ts_reserved_words[0][0],
    .max_reserved_word_set_size = 25,
    .metadata = {
      .major_version = 0,
      .minor_version = 25,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
